/*========================== begin_copyright_notice ============================

Copyright (C) 2020-2023 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

// This is an auto-generated file. Please do not edit!
// If changes are needed here please reach out to the codeowners, thanks.


#include "wa_def.h"

#define ACM_G11_GT_REV_ID_A0   SI_REV_ID(0,0)
#define ACM_G11_GT_REV_ID_B0   SI_REV_ID(4,4)
#define ACM_G11_GT_REV_ID_B1   SI_REV_ID(5,5)
#define ACM_G11_GT_REV_ID_C0   SI_REV_ID(8,8)