/*========================== begin_copyright_notice ============================

Copyright (C) 2020-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

#ifndef ZE_INFO_READER_H
#define ZE_INFO_READER_H

int zeinfo_reader_main(int argc, const char** argv);

#endif //_ZEINFOREADER_H_
