/*========================== begin_copyright_notice ============================

Copyright (C) 2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

syntax = "proto3";

package IGC_METRICS;

enum ParallelizationType
{
  SINGLE_THREAD = 0;
  SINGLE_EXECUTION_UNIT = 1;
  GROUP_OF_EXECUTION_UNITS = 2;
  SINGLE_DEVICE = 3;
}