/*========================== begin_copyright_notice ============================

Copyright (C) 2021-2022 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

// This is an auto-generated file. Please do not edit!
// If changes are needed here please reach out to the codeowners, thanks.


#include "wa_def.h"

#define PVC_GT_REV_ID_CT_A0P   SI_REV_ID(0,0)
#define PVC_GT_REV_ID_CTXT_A0   SI_REV_ID(3,3)
#define PVC_GT_REV_ID_CT_A0   SI_REV_ID(0,0)
#define PVC_GT_REV_ID_CTXT_B0   SI_REV_ID(5,5)
#define PVC_GT_REV_ID_CTXT_C0   SI_REV_ID(7,7)

