/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

#ifndef _IGA_VERSION_HPP_
#define _IGA_VERSION_HPP_
#define IGA_VERSION_PREFIX_STRING "1.1.0"

#endif // _IGA_VERSION_HPP_
