/*========================== begin_copyright_notice ============================

Copyright (C) 2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

syntax = "proto3";

import "Metrics/proto_schema/code_reference.proto";

package IGC_METRICS;

message MemAllocStats {

  message VarSharedLocalMem {
    string name = 1;
    int32 size = 2;
    optional CodeRef varLoc = 3;
  }

  int32 countBytesSharedLocalMemAlloc = 1;
  int32 countSendsSharedLocalMem = 2;

  int32 countBytesStaticLocaMemAlloc = 3;
  int32 countBytesL3CacheMemAlloc = 4;

  repeated VarSharedLocalMem variables = 5;
}