/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

#include "common/ged_types_internal.h"


const char* ged_signed_enum_value_t_str = "ged_signed_enum_value_t";
const char* ged_unsigned_enum_value_t_str = "ged_unsigned_enum_value_t";
