/*========================== begin_copyright_notice ============================

Copyright (C) 2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

syntax = "proto3";

import "Metrics/proto_schema/function.proto";
import "Metrics/proto_schema/device_enum.proto";

package IGC_METRICS;

message Program {
  optional string hash = 1;

  DeviceType device = 2;

  repeated Function functions = 3;
}