;=========================== begin_copyright_notice ============================
;
; Copyright (C) 2017-2021 Intel Corporation
;
; SPDX-License-Identifier: MIT
;
;============================ end_copyright_notice =============================

;
; !!! DO NOT EDIT THIS FILE !!!
;
; This file was automagically crafted by GED's model parser.
;

LIBRARY ged
EXPORTS
  _GED_GetModelVersionString
  _GED_GetReturnValueString
  _GED_GetReturnValuePad
  _GED_InitEmptyIns
  _GED_DecodeIns
  _GED_EncodeIns
  _GED_InsSize
  _GED_InsModified
  _GED_CountCompactEncodings
  _GED_GetCompactEncodings
  _GED_PrintFieldBitLocation
  _GED_QueryFieldBitLocation
  _GED_FieldSize
  _GED_GetOpcode
  _GED_GetMnemonic
  _GED_SetOpcode
  _GED_IsCompact
  _GED_GetModel
  _GED_Get___SrcImm
  _GED_Set___SrcImm
  _GED_GetNumOfSourceOperands
  _GED_GetHasDestinationOperand
  _GED_GetAccessMode
  _GED_SetAccessMode
  _GED_GetMaskCtrl
  _GED_SetMaskCtrl
  _GED_GetDepCtrl
  _GED_SetDepCtrl
  _GED_GetExecMaskOffsetCtrl
  _GED_SetExecMaskOffsetCtrl
  _GED_GetChannelOffset
  _GED_SetChannelOffset
  _GED_GetThreadCtrl
  _GED_SetThreadCtrl
  _GED_GetPredCtrl
  _GED_SetPredCtrl
  _GED_GetPredInv
  _GED_SetPredInv
  _GED_GetExecSize
  _GED_SetExecSize
  _GED_GetCondModifier
  _GED_SetCondModifier
  _GED_GetAccWrCtrl
  _GED_SetAccWrCtrl
  _GED_GetDebugCtrl
  _GED_SetDebugCtrl
  _GED_GetSaturate
  _GED_SetSaturate
  _GED_GetDstRegFile
  _GED_SetDstRegFile
  _GED_GetDstDataType
  _GED_SetDstDataType
  _GED_GetSrc0RegFile
  _GED_SetSrc0RegFile
  _GED_GetSrc0DataType
  _GED_SetSrc0DataType
  _GED_GetSrc1RegFile
  _GED_SetSrc1RegFile
  _GED_GetSrc1DataType
  _GED_SetSrc1DataType
  _GED_GetDstChanEn
  _GED_SetDstChanEn
  _GED_GetDstSubRegNum
  _GED_SetDstSubRegNum
  _GED_GetDstAddrImm
  _GED_SetDstAddrImm
  _GED_GetDstRegNum
  _GED_SetDstRegNum
  _GED_GetDstAddrSubRegNum
  _GED_SetDstAddrSubRegNum
  _GED_GetDstHorzStride
  _GED_SetDstHorzStride
  _GED_GetDstAddrMode
  _GED_SetDstAddrMode
  _GED_GetSrc0ChanSel
  _GED_SetSrc0ChanSel
  _GED_GetSrc0SubRegNum
  _GED_SetSrc0SubRegNum
  _GED_GetSrc0AddrImm
  _GED_SetSrc0AddrImm
  _GED_GetSrc0RegNum
  _GED_SetSrc0RegNum
  _GED_GetSrc0AddrSubRegNum
  _GED_SetSrc0AddrSubRegNum
  _GED_GetSrc0SrcMod
  _GED_SetSrc0SrcMod
  _GED_GetSrc0AddrMode
  _GED_SetSrc0AddrMode
  _GED_GetSrc0HorzStride
  _GED_SetSrc0HorzStride
  _GED_GetSrc0Width
  _GED_SetSrc0Width
  _GED_GetSrc0VertStride
  _GED_SetSrc0VertStride
  _GED_GetFlagSubRegNum
  _GED_SetFlagSubRegNum
  _GED_GetFlagRegNum
  _GED_SetFlagRegNum
  _GED_GetSrc1ChanSel
  _GED_SetSrc1ChanSel
  _GED_GetSrc1SubRegNum
  _GED_SetSrc1SubRegNum
  _GED_GetSrc1AddrImm
  _GED_SetSrc1AddrImm
  _GED_GetSrc1RegNum
  _GED_SetSrc1RegNum
  _GED_GetSrc1AddrSubRegNum
  _GED_SetSrc1AddrSubRegNum
  _GED_GetSrc1SrcMod
  _GED_SetSrc1SrcMod
  _GED_GetSrc1AddrMode
  _GED_SetSrc1AddrMode
  _GED_GetSrc1HorzStride
  _GED_SetSrc1HorzStride
  _GED_GetSrc1Width
  _GED_SetSrc1Width
  _GED_GetSrc1VertStride
  _GED_SetSrc1VertStride
  _GED_GetImm
  _GED_SetImm
  _GED_GetSrc2SrcMod
  _GED_SetSrc2SrcMod
  _GED_GetSrcDataType
  _GED_SetSrcDataType
  _GED_GetSrc0RepCtrl
  _GED_SetSrc0RepCtrl
  _GED_GetSrc1RepCtrl
  _GED_SetSrc1RepCtrl
  _GED_GetSrc2RepCtrl
  _GED_SetSrc2RepCtrl
  _GED_GetSrc2ChanSel
  _GED_SetSrc2ChanSel
  _GED_GetSrc2SubRegNum
  _GED_SetSrc2SubRegNum
  _GED_GetSrc2RegNum
  _GED_SetSrc2RegNum
  _GED_GetSrc2RegFile
  _GED_SetSrc2RegFile
  _GED_GetSrc2AddrMode
  _GED_SetSrc2AddrMode
  _GED_GetSrc2VertStride
  _GED_SetSrc2VertStride
  _GED_GetSFID
  _GED_SetSFID
  _GED_GetDescRegFile
  _GED_SetDescRegFile
  _GED_GetDescDataType
  _GED_SetDescDataType
  _GED_GetDescAddrSubRegNum
  _GED_SetDescAddrSubRegNum
  _GED_GetDescRegNum
  _GED_SetDescRegNum
  _GED_GetDescHorzStride
  _GED_SetDescHorzStride
  _GED_GetDescWidth
  _GED_SetDescWidth
  _GED_GetDescVertStride
  _GED_SetDescVertStride
  _GED_GetMsgDesc
  _GED_SetMsgDesc
  _GED_GetExMsgDescImm
  _GED_SetExMsgDescImm
  _GED_GetEOT
  _GED_SetEOT
  _GED_GetMathFC
  _GED_SetMathFC
  _GED_GetJIP
  _GED_SetJIP
  _GED_GetUIP
  _GED_SetUIP
  _GED_GetControlIndex
  _GED_SetControlIndex
  _GED_GetDataTypeIndex
  _GED_SetDataTypeIndex
  _GED_GetSubRegIndex
  _GED_SetSubRegIndex
  _GED_GetSrc0Index
  _GED_SetSrc0Index
  _GED_GetSrc1Index
  _GED_SetSrc1Index
  _GED_GetDescIndex
  _GED_SetDescIndex
  _GED_GetExDescRegFile
  _GED_SetExDescRegFile
  _GED_GetDstMathMacroExt
  _GED_SetDstMathMacroExt
  _GED_GetSrc0MathMacroExt
  _GED_SetSrc0MathMacroExt
  _GED_GetSrc1MathMacroExt
  _GED_SetSrc1MathMacroExt
  _GED_GetSrc2MathMacroExt
  _GED_SetSrc2MathMacroExt
  _GED_GetBranchCtrl
  _GED_SetBranchCtrl
  _GED_GetSourceIndex
  _GED_SetSourceIndex
  _GED_GetSrc2DataType
  _GED_SetSrc2DataType
  _GED_GetNoSrcDepSet
  _GED_SetNoSrcDepSet
  _GED_GetExFuncCtrl
  _GED_SetExFuncCtrl
  _GED_GetExMsgLength
  _GED_SetExMsgLength
  _GED_GetExDescAddrSubRegNum
  _GED_SetExDescAddrSubRegNum
  _GED_GetExDescRegNum
  _GED_SetExDescRegNum
  _GED_GetMsgDescCategory
  _GED_SetMsgDescCategory
  _GED_GetMsgDescScratchAddrOffset
  _GED_SetMsgDescScratchAddrOffset
  _GED_GetMsgDescScratchBlockSize
  _GED_SetMsgDescScratchBlockSize
  _GED_GetMsgDescScratchInvalidateAfterRead
  _GED_SetMsgDescScratchInvalidateAfterRead
  _GED_GetMsgDescScratchChannelMode
  _GED_SetMsgDescScratchChannelMode
  _GED_GetMsgDescScratchMessageType
  _GED_SetMsgDescScratchMessageType
  _GED_GetExecutionDataType
  _GED_SetExecutionDataType
  _GED_GetSrc0TernaryImm
  _GED_SetSrc0TernaryImm
  _GED_GetSrc2TernaryImm
  _GED_SetSrc2TernaryImm
  _GED_GetSrc2HorzStride
  _GED_SetSrc2HorzStride
  _GED_GetSWSB
  _GED_SetSWSB
  _GED_GetSrc1IsImm
  _GED_SetSrc1IsImm
  _GED_GetSrc0IsImm
  _GED_SetSrc0IsImm
  _GED_GetSrc0SubRegNumByte
  _GED_SetSrc0SubRegNumByte
  _GED_GetSyncFC
  _GED_SetSyncFC
  _GED_GetFusionCtrl
  _GED_SetFusionCtrl
  _GED_GetDataTypeIndexNoDep
  _GED_SetDataTypeIndexNoDep
  _GED_GetCompactedImm
  _GED_SetCompactedImm
  _GED_GetRepeatCount
  _GED_SetRepeatCount
  _GED_GetSystolicDepth
  _GED_SetSystolicDepth
  _GED_GetSrc2Precision
  _GED_SetSrc2Precision
  _GED_GetSrc2SubBytePrecision
  _GED_SetSrc2SubBytePrecision
  _GED_GetSrc1Precision
  _GED_SetSrc1Precision
  _GED_GetSrc1SubBytePrecision
  _GED_SetSrc1SubBytePrecision
  _GED_GetBfnFC
  _GED_SetBfnFC
  _GED_GetExBSO
  _GED_SetExBSO
  _GED_GetCPS
  _GED_SetCPS
  _GED_GetSrc1Length
  _GED_SetSrc1Length
  _GED_GetSrc2IsImm
  _GED_SetSrc2IsImm
  _GED_GetIndexedSrcAddrImm
  _GED_SetIndexedSrcAddrImm
  _GED_GetIndexedSrcAddrMode
  _GED_SetIndexedSrcAddrMode
  _GED_GetIndexedSrcAddrSubRegNum
  _GED_SetIndexedSrcAddrSubRegNum
  _GED_GetIndexedSrcChanSel
  _GED_SetIndexedSrcChanSel
  _GED_GetIndexedSrcDataType
  _GED_SetIndexedSrcDataType
  _GED_GetIndexedSrcHorzStride
  _GED_SetIndexedSrcHorzStride
  _GED_GetIndexedSrcIndex
  _GED_SetIndexedSrcIndex
  _GED_GetIndexedSrcIsImm
  _GED_SetIndexedSrcIsImm
  _GED_GetIndexedSrcMathMacroExt
  _GED_SetIndexedSrcMathMacroExt
  _GED_GetIndexedSrcRegFile
  _GED_SetIndexedSrcRegFile
  _GED_GetIndexedSrcRegNum
  _GED_SetIndexedSrcRegNum
  _GED_GetIndexedSrcRepCtrl
  _GED_SetIndexedSrcRepCtrl
  _GED_GetIndexedSrcSrcMod
  _GED_SetIndexedSrcSrcMod
  _GED_GetIndexedSrcSubRegNum
  _GED_SetIndexedSrcSubRegNum
  _GED_GetIndexedSrcSubRegNumByte
  _GED_SetIndexedSrcSubRegNumByte
  _GED_GetIndexedSrcVertStride
  _GED_SetIndexedSrcVertStride
  _GED_GetIndexedSrcWidth
  _GED_SetIndexedSrcWidth
  _GED_GetArchReg
  _GED_SetArchReg
  _GED_GetArchRegNum
  _GED_SetArchRegNum
  _GED_GetSwizzleX
  _GED_SetSwizzleX
  _GED_GetSwizzleY
  _GED_SetSwizzleY
  _GED_GetSwizzleZ
  _GED_SetSwizzleZ
  _GED_GetSwizzleW
  _GED_SetSwizzleW
  _GED_GetMessageLength
  _GED_SetMessageLength
  _GED_GetResponseLength
  _GED_SetResponseLength
  _GED_GetHeaderPresent
  _GED_SetHeaderPresent
  _GED_GetMessageTypeDP_SAMPLER
  _GED_SetMessageTypeDP_SAMPLER
  _GED_GetMessageTypeDP_RC
  _GED_SetMessageTypeDP_RC
  _GED_GetMessageTypeDP_CC
  _GED_SetMessageTypeDP_CC
  _GED_GetMessageTypeDP_DC0
  _GED_SetMessageTypeDP_DC0
  _GED_GetTypedSurfaceSlotGroup
  _GED_SetTypedSurfaceSlotGroup
  _GED_GetTypedAtomicSlotGroup
  _GED_SetTypedAtomicSlotGroup
  _GED_GetUntypedSurfaceSIMDMode
  _GED_SetUntypedSurfaceSIMDMode
  _GED_GetUntypedAtomicSIMDMode
  _GED_SetUntypedAtomicSIMDMode
  _GED_GetInvalidateAfterRead
  _GED_SetInvalidateAfterRead
  _GED_GetBlockSize
  _GED_SetBlockSize
  _GED_GetRedChannel
  _GED_SetRedChannel
  _GED_GetGreenChannel
  _GED_SetGreenChannel
  _GED_GetBlueChannel
  _GED_SetBlueChannel
  _GED_GetAlphaChannel
  _GED_SetAlphaChannel
  _GED_GetReturnDataControl
  _GED_SetReturnDataControl
  _GED_GetAtomicOperationType
  _GED_SetAtomicOperationType
  _GED_GetAtomicCounterOperationType
  _GED_SetAtomicCounterOperationType
  _GED_GetSubFuncID
  _GED_SetSubFuncID
  _GED_GetBindingTableIndex
  _GED_SetBindingTableIndex
  _GED_GetFuncControl
  _GED_SetFuncControl
  _GED_GetMessageTypeDP_DC1
  _GED_SetMessageTypeDP_DC1
  _GED_GetMessageTypeDP0Category
  _GED_SetMessageTypeDP0Category
  _GED_GetMessageTypeDP_DC0Legacy
  _GED_SetMessageTypeDP_DC0Legacy
  _GED_GetMessageTypeDP_DC0ScratchBlock
  _GED_SetMessageTypeDP_DC0ScratchBlock
  _GED_GetMessageTypeDP_DC2
  _GED_SetMessageTypeDP_DC2
  _GED_GetMessageTypeDP_DCRO
  _GED_SetMessageTypeDP_DCRO
  _GED_GetExMessageLength
  _GED_SetExMessageLength
  _GED_GetDPOpcode
  _GED_SetDPOpcode
  _GED_GetDPAddrSurfaceType
  _GED_SetDPAddrSurfaceType
  _GED_GetDPVectSize
  _GED_SetDPVectSize
  _GED_GetDPFlushType
  _GED_SetDPFlushType
  _GED_GetDPTranspose
  _GED_SetDPTranspose
  _GED_GetDPFlushRange
  _GED_SetDPFlushRange
  _GED_GetDPDataSize
  _GED_SetDPDataSize
  _GED_GetDPFenceScope
  _GED_SetDPFenceScope
  _GED_GetDPAddrSize
  _GED_SetDPAddrSize
  _GED_SetRawBits
  _GED_GetAccessModeString
  _GED_GetAccWrCtrlString
  _GED_GetAddrModeString
  _GED_GetArchRegString
  _GED_GetAtomicOperationTypeString
  _GED_GetBlockSizeString
  _GED_GetBranchCtrlString
  _GED_GetChannelMaskString
  _GED_GetChannelModeString
  _GED_GetChannelOffsetString
  _GED_GetCondModifierString
  _GED_GetDataTypeString
  _GED_GetDebugCtrlString
  _GED_GetDepCtrlString
  _GED_GetDpAddrSizeString
  _GED_GetDpAddrSurfaceTypeString
  _GED_GetDpDataSizeString
  _GED_GetDpFenceScopeString
  _GED_GetDpFlushTypeString
  _GED_GetDpOpcodeString
  _GED_GetDpTransposeString
  _GED_GetDpVectSizeString
  _GED_GetDstChanEnString
  _GED_GetEOTString
  _GED_GetExecMaskOffsetCtrlString
  _GED_GetExecutionDataTypeString
  _GED_GetFusionCtrlString
  _GED_GetHeaderPresentString
  _GED_GetMaskCtrlString
  _GED_GetMathFCString
  _GED_GetMathMacroExtString
  _GED_GetMessageTypeString
  _GED_GetNoSrcDepSetString
  _GED_GetOpcodeString
  _GED_GetPrecisionString
  _GED_GetPredCtrlString
  _GED_GetPredInvString
  _GED_GetRegFileString
  _GED_GetRepCtrlString
  _GED_GetReturnDataControlString
  _GED_GetSaturateString
  _GED_GetSFIDString
  _GED_GetSIMDModeString
  _GED_GetSlotGroupString
  _GED_GetSrcModString
  _GED_GetSubBytePrecisionString
  _GED_GetSubFuncIDString
  _GED_GetSwizzleString
  _GED_GetSyncFCString
  _GED_GetThreadCtrlString
  _GED_GetOperandWidth
  _GED_GetOperandNumericType
  _GED_GetNibCtrl
