/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

/* The error code name should be meaningful since it is part of error message */
_SPIRV_OP(Success,"")
_SPIRV_OP(InvalidTargetTriple, "Expects spir-unknown-unknown or spir64-unknown-unknown.")
_SPIRV_OP(InvalidAddressingModel, "Expects 0-2.")
_SPIRV_OP(InvalidMemoryModel, "Expects 0-3.")
_SPIRV_OP(InvalidFunctionControlMask,"")
_SPIRV_OP(InvalidBuiltinSetName, "Expects OpenCL12, OpenCL20.")
_SPIRV_OP(UnsupportedSPIRVOpcode, "")
