/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

/*
 * !!! DO NOT EDIT THIS FILE !!!
 *
 * This file was automagically crafted by GED's model parser.
 */


#ifndef GED_COLLECTORS_TABLES_H
#define GED_COLLECTORS_TABLES_H

#include "common/ged_interpreter_types.h"
#endif // GED_COLLECTORS_TABLES_H
