/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

/*
 * !!! DO NOT EDIT THIS FILE !!!
 *
 * This file was automagically crafted by GED's model parser.
 */


#ifndef GED_COMPACTION_TABLES_H
#define GED_COMPACTION_TABLES_H

#include "common/ged_compact_mapping_table.h"
extern const ged_compaction_table_entry_t CompactionTable0[32];
extern const ged_compaction_table_entry_t CompactionTable1[32];
extern const ged_compaction_table_entry_t CompactionTable2[32];
extern const ged_compaction_table_entry_t CompactionTable3[32];
extern const ged_compaction_table_entry_t CompactionTable4[32];
extern const ged_compaction_table_entry_t CompactionTable5[32];
extern const ged_compaction_table_entry_t CompactionTable6[4];
extern const ged_compaction_table_entry_t CompactionTable7[4];
extern const ged_compaction_table_entry_t CompactionTable8[32];
extern const ged_compaction_table_entry_t CompactionTable9[32];
extern const ged_compaction_table_entry_t CompactionTable10[32];
extern const ged_compaction_table_entry_t CompactionTable11[32];
extern const ged_compaction_table_entry_t CompactionTable12[32];
extern const ged_compaction_table_entry_t CompactionTable13[32];
extern const ged_compaction_table_entry_t CompactionTable14[32];
extern const ged_compaction_table_entry_t CompactionTable15[16];
extern const ged_compaction_table_entry_t CompactionTable16[16];
extern const ged_compaction_table_entry_t CompactionTable17[32];
extern const ged_compaction_table_entry_t CompactionTable18[32];
extern const ged_compaction_table_entry_t CompactionTable19[32];
extern const ged_compaction_table_entry_t CompactionTable20[16];
extern const ged_compaction_table_entry_t CompactionTable21[16];
extern const ged_compaction_table_entry_t CompactionTable22[32];
extern const ged_compaction_table_entry_t CompactionTable23[32];
extern const ged_compaction_table_entry_t CompactionTable24[32];
extern const ged_compaction_table_entry_t CompactionTable25[16];
extern const ged_compaction_table_entry_t CompactionTable26[8];
extern const ged_compaction_table_entry_t CompactionTable27[16];
extern const ged_compaction_table_entry_t CompactionTable28[16];
extern const ged_compaction_table_entry_t CompactionTable29[16];
extern const ged_compaction_table_entry_t CompactionTable30[16];
extern const ged_compaction_table_entry_t CompactionTable31[16];
extern const ged_compaction_table_entry_t CompactionTable32[32];
extern const ged_compaction_table_entry_t CompactionTable33[32];
extern const ged_compaction_table_entry_t CompactionTable34[32];
extern const ged_compaction_table_entry_t CompactionTable35[32];
extern const ged_compaction_table_entry_t CompactionTable36[16];
extern const ged_compaction_table_entry_t CompactionTable37[8];
extern const ged_compaction_table_entry_t CompactionTable38[16];
extern const ged_compaction_table_entry_t CompactionTable39[16];
extern const ged_compaction_table_entry_t CompactionTable40[16];
extern const ged_compaction_table_entry_t CompactionTable41[16];
extern const ged_compaction_table_entry_t CompactionTable42[16];
extern const ged_compaction_table_entry_t CompactionTable43[32];
extern const ged_compaction_table_entry_t CompactionTable44[32];
extern const ged_compaction_table_entry_t CompactionTable45[4];
extern const ged_compaction_table_entry_t CompactionTable46[4];
#endif // GED_COMPACTION_TABLES_H
