/*========================== begin_copyright_notice ============================

Copyright (C) 2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

syntax = "proto3";

package IGC_METRICS;

message LocalRegStats {

  int32 countRegAllocated = 1;
  int32 countRegTmpAllocated = 2;
  int32 countBytesRegTmpAllocated = 3;
  int32 countRegSpillsAllocated = 4;
  int32 countRegFillsAllocated = 5;
  int32 percentGRFUsage = 6;

  int32 EstimatedSpillTraffic = 7;
  int32 SpillFootprint = 8;

  int32 countPrvArray2GRFPromoted = 9;
  int32 countPrvArray2GRFNotPromoted = 10;
}