/*========================== begin_copyright_notice ============================

Copyright (C) 2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

syntax = "proto3";

import "Metrics/proto_schema/code_reference.proto";

package IGC_METRICS;

message FuncEmuCalls {
  enum Reason4FuncEmu {
    NO_HW_SUPPORT = 0;
    FP_MODEL_MODE = 1;
  }

  string name = 1;
  Reason4FuncEmu type = 2;
  optional int32 count = 3;
  int32 modelCost = 4;

  // All places where it was used
  repeated CodeRef funcCallLoc = 5;
}