/*========================== begin_copyright_notice ============================

Copyright (C) 2020-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

#ifndef IGCLLVM_ANALYSIS_TARGETLIBRARYINFO_H
#define IGCLLVM_ANALYSIS_TARGETLIBRARYINFO_H

#include "llvm/Config/llvm-config.h"
#include "llvm/Analysis/TargetLibraryInfo.h"

#if LLVM_VERSION_MAJOR >= 10
#define getTLI() getTLI(F)
#endif

#endif
