/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

// Sets igc32.dll name and file description

#include <winver.h>
#include "..\install\intcver.h"

#if defined(_WIN64)
    #define VER_ORIGINALNAME "igc64.dll"
#else
    #define VER_ORIGINALNAME "igc32.dll"
#endif
#define VER_FILETYPE             VFT_DLL
#define VER_FILESUBTYPE          VFT2_DRV_INSTALLABLE
#define VER_FILEOS               VOS_DOS_WINDOWS32
#define VER_FILEDESCRIPTION      "Intel Graphics Shader Compiler for Intel(R) Graphics Accelerator"
#define VER_INTERNALNAME         "IGC"

#include "..\inc\verinfo.ver"
