/*========================== begin_copyright_notice ============================

Copyright (C) 2020-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

#include "ZEInfoReader.h"

int main(int argc, const char** argv) {
    return zeinfo_reader_main(argc, argv);
}
