/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

/*
 * !!! DO NOT EDIT THIS FILE !!!
 *
 * This file was automagically crafted by GED's model parser.
 */


#ifndef GED_ENCODING_MASKS_TABLES_H
#define GED_ENCODING_MASKS_TABLES_H

#include "common/ged_ins_encoding_masks.h"
extern const ged_instruction_masks_entry_t EncodingMasksTable0[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable1[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable2[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable3[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable4[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable5[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable6[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable7[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable8[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable9[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable10[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable11[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable12[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable13[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable14[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable15[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable16[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable17[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable18[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable19[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable20[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable21[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable22[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable23[3];
extern const ged_instruction_masks_entry_t EncodingMasksTable24[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable25[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable26[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable27[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable28[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable29[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable30[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable31[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable32[7];
extern const ged_instruction_masks_entry_t EncodingMasksTable33[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable34[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable35[3];
extern const ged_instruction_masks_entry_t EncodingMasksTable36[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable37[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable38[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable39[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable40[7];
extern const ged_instruction_masks_entry_t EncodingMasksTable41[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable42[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable43[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable44[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable45[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable46[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable47[9];
extern const ged_instruction_masks_entry_t EncodingMasksTable48[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable49[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable50[16];
extern const ged_instruction_masks_entry_t EncodingMasksTable51[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable52[9];
extern const ged_instruction_masks_entry_t EncodingMasksTable53[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable54[7];
extern const ged_instruction_masks_entry_t EncodingMasksTable55[16];
extern const ged_instruction_masks_entry_t EncodingMasksTable56[16];
extern const ged_instruction_masks_entry_t EncodingMasksTable57[16];
extern const ged_instruction_masks_entry_t EncodingMasksTable58[16];
extern const ged_instruction_masks_entry_t EncodingMasksTable59[16];
extern const ged_instruction_masks_entry_t EncodingMasksTable60[12];
extern const ged_instruction_masks_entry_t EncodingMasksTable61[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable62[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable63[8];
extern const ged_instruction_masks_entry_t EncodingMasksTable64[7];
extern const ged_instruction_masks_entry_t EncodingMasksTable65[9];
extern const ged_instruction_masks_entry_t EncodingMasksTable66[10];
extern const ged_instruction_masks_entry_t EncodingMasksTable67[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable68[8];
extern const ged_instruction_masks_entry_t EncodingMasksTable69[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable70[7];
extern const ged_instruction_masks_entry_t EncodingMasksTable71[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable72[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable73[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable74[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable75[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable76[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable77[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable78[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable79[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable80[9];
extern const ged_instruction_masks_entry_t EncodingMasksTable81[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable82[8];
extern const ged_instruction_masks_entry_t EncodingMasksTable83[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable84[8];
extern const ged_instruction_masks_entry_t EncodingMasksTable85[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable86[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable87[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable88[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable89[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable90[8];
extern const ged_instruction_masks_entry_t EncodingMasksTable91[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable92[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable93[12];
extern const ged_instruction_masks_entry_t EncodingMasksTable94[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable95[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable96[16];
extern const ged_instruction_masks_entry_t EncodingMasksTable97[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable98[7];
extern const ged_instruction_masks_entry_t EncodingMasksTable99[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable100[8];
extern const ged_instruction_masks_entry_t EncodingMasksTable101[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable102[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable103[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable104[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable105[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable106[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable107[9];
extern const ged_instruction_masks_entry_t EncodingMasksTable108[3];
extern const ged_instruction_masks_entry_t EncodingMasksTable109[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable110[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable111[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable112[7];
extern const ged_instruction_masks_entry_t EncodingMasksTable113[7];
extern const ged_instruction_masks_entry_t EncodingMasksTable114[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable115[3];
extern const ged_instruction_masks_entry_t EncodingMasksTable116[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable117[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable118[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable119[3];
extern const ged_instruction_masks_entry_t EncodingMasksTable120[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable121[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable122[3];
extern const ged_instruction_masks_entry_t EncodingMasksTable123[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable124[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable125[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable126[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable127[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable128[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable129[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable130[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable131[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable132[128];
extern const ged_instruction_masks_entry_t EncodingMasksTable133[128];
extern const ged_instruction_masks_entry_t EncodingMasksTable134[128];
extern const ged_instruction_masks_entry_t EncodingMasksTable135[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable136[3];
extern const ged_instruction_masks_entry_t EncodingMasksTable137[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable138[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable139[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable140[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable141[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable142[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable143[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable144[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable145[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable146[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable147[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable148[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable149[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable150[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable151[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable152[128];
extern const ged_instruction_masks_entry_t EncodingMasksTable153[128];
extern const ged_instruction_masks_entry_t EncodingMasksTable154[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable155[128];
extern const ged_instruction_masks_entry_t EncodingMasksTable156[128];
extern const ged_instruction_masks_entry_t EncodingMasksTable157[9];
extern const ged_instruction_masks_entry_t EncodingMasksTable158[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable159[3];
extern const ged_instruction_masks_entry_t EncodingMasksTable160[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable161[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable162[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable163[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable164[16];
extern const ged_instruction_masks_entry_t EncodingMasksTable165[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable166[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable167[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable168[3];
extern const ged_instruction_masks_entry_t EncodingMasksTable169[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable170[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable171[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable172[3];
extern const ged_instruction_masks_entry_t EncodingMasksTable173[3];
#endif // GED_ENCODING_MASKS_TABLES_H
