/*========================== begin_copyright_notice ============================

Copyright (C) 2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

// Sets iga32.dll name and file description
#ifndef WDDM_VERSION_NUMBER
#define WDDM_VERSION_NUMBER 31
#endif

#ifndef LEGACY_D3DAPI_VERSION
#define LEGACY_D3DAPI_VERSION 0
#endif

#include <winver.h>
#include "..\..\..\install\intcver.h"
#if defined(_WIN64)
    #define VER_ORIGINALNAME "iga64.dll"
#else
    #define VER_ORIGINALNAME "iga32.dll"
#endif
#define VER_FILETYPE             VFT_DLL
#define VER_FILESUBTYPE          VFT2_DRV_INSTALLABLE
#define VER_FILEOS               VOS_DOS_WINDOWS32
#define VER_FILEDESCRIPTION      "Intel Graphics Shader Assembler for Intel(R) Graphics Accelerator"
#define VER_INTERNALNAME         "IGA"

#include "..\..\..\inc\verinfo.ver"
