/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

/*
 * !!! DO NOT EDIT THIS FILE !!!
 *
 * This file was automagically crafted by GED's model parser.
 */


#ifndef GED_MAPPING_FRAGMENTS_H
#define GED_MAPPING_FRAGMENTS_H

#include "common/ged_compact_mapping_table.h"
extern const ged_compact_mapping_fragment_t MappingFragmentsTable0[3];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable1[2];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable2[2];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable3[3];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable4[2];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable5[6];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable6[3];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable7[2];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable8[8];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable9[2];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable10[2];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable11[8];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable12[4];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable13[4];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable14[5];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable15[3];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable16[2];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable17[3];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable18[2];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable19[2];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable20[2];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable21[2];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable22[4];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable23[4];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable24[4];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable25[9];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable26[7];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable27[4];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable28[9];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable29[5];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable30[4];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable31[3];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable32[4];
extern const ged_compact_mapping_fragment_t MappingFragmentsTable33[10];
#endif // GED_MAPPING_FRAGMENTS_H
