/*========================== begin_copyright_notice ============================

Copyright (C) 2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

syntax = "proto3";

package IGC_METRICS;

message KernelArg {
  enum ArgumentType {
    EXPLICIT = 0;
    IMPLICIT = 1;
  }

  string name = 1;
  ArgumentType type = 2;
  int32 compileSize = 3;
}