/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

extern "C" int iga_main(int argc, const char **argv);

int main(int argc, const char **argv) { return iga_main(argc, argv); }
