/*========================== begin_copyright_notice ============================

Copyright (C) 2019-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

#pragma once

// reuse GFX enumeration types common to gen6 and gen7 
#include "usc.h"
#include "SurfaceFormats.h"
#include "ShaderTypes.h"

#include "usc_gen7_types.h"
