/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

#include "ged_basic_types.h"

const char* BoolTypeStr = "bool";

const char* InlineStr = "GED_INLINE";
