/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

/*
 * !!! DO NOT EDIT THIS FILE !!!
 *
 * This file was automagically crafted by GED's model parser.
 */


#ifndef GED_FRAGMENTS_H
#define GED_FRAGMENTS_H

#include "common/ged_ins_position_fragment.h"
extern const ged_ins_field_position_fragment_t FragmentsTable0[4];
extern const ged_ins_field_position_fragment_t FragmentsTable1[2];
extern const ged_ins_field_position_fragment_t FragmentsTable2[2];
extern const ged_ins_field_position_fragment_t FragmentsTable3[2];
extern const ged_ins_field_position_fragment_t FragmentsTable4[4];
extern const ged_ins_field_position_fragment_t FragmentsTable5[4];
extern const ged_ins_field_position_fragment_t FragmentsTable6[2];
extern const ged_ins_field_position_fragment_t FragmentsTable7[2];
extern const ged_ins_field_position_fragment_t FragmentsTable8[2];
extern const ged_ins_field_position_fragment_t FragmentsTable9[2];
extern const ged_ins_field_position_fragment_t FragmentsTable10[2];
extern const ged_ins_field_position_fragment_t FragmentsTable11[2];
extern const ged_ins_field_position_fragment_t FragmentsTable12[2];
extern const ged_ins_field_position_fragment_t FragmentsTable13[2];
extern const ged_ins_field_position_fragment_t FragmentsTable14[2];
extern const ged_ins_field_position_fragment_t FragmentsTable15[2];
extern const ged_ins_field_position_fragment_t FragmentsTable16[2];
extern const ged_ins_field_position_fragment_t FragmentsTable17[3];
extern const ged_ins_field_position_fragment_t FragmentsTable18[2];
extern const ged_ins_field_position_fragment_t FragmentsTable19[4];
extern const ged_ins_field_position_fragment_t FragmentsTable20[6];
extern const ged_ins_field_position_fragment_t FragmentsTable21[2];
extern const ged_ins_field_position_fragment_t FragmentsTable22[2];
extern const ged_ins_field_position_fragment_t FragmentsTable23[4];
extern const ged_ins_field_position_fragment_t FragmentsTable24[2];
extern const ged_ins_field_position_fragment_t FragmentsTable25[2];
extern const ged_ins_field_position_fragment_t FragmentsTable26[2];
extern const ged_ins_field_position_fragment_t FragmentsTable27[2];
extern const ged_ins_field_position_fragment_t FragmentsTable28[2];
extern const ged_ins_field_position_fragment_t FragmentsTable29[2];
extern const ged_ins_field_position_fragment_t FragmentsTable30[2];
extern const ged_ins_field_position_fragment_t FragmentsTable31[2];
extern const ged_ins_field_position_fragment_t FragmentsTable32[2];
extern const ged_ins_field_position_fragment_t FragmentsTable33[2];
extern const ged_ins_field_position_fragment_t FragmentsTable34[5];
extern const ged_ins_field_position_fragment_t FragmentsTable35[5];
extern const ged_ins_field_position_fragment_t FragmentsTable36[6];
extern const ged_ins_field_position_fragment_t FragmentsTable37[2];
extern const ged_ins_field_position_fragment_t FragmentsTable38[2];
extern const ged_ins_field_position_fragment_t FragmentsTable39[2];
extern const ged_ins_field_position_fragment_t FragmentsTable40[2];
extern const ged_ins_field_position_fragment_t FragmentsTable41[5];
extern const ged_ins_field_position_fragment_t FragmentsTable42[2];
extern const ged_ins_field_position_fragment_t FragmentsTable43[5];
extern const ged_ins_field_position_fragment_t FragmentsTable44[4];
extern const ged_ins_field_position_fragment_t FragmentsTable45[2];
extern const ged_ins_field_position_fragment_t FragmentsTable46[2];
extern const ged_ins_field_position_fragment_t FragmentsTable47[2];
extern const ged_ins_field_position_fragment_t FragmentsTable48[2];
extern const ged_ins_field_position_fragment_t FragmentsTable49[2];
extern const ged_ins_field_position_fragment_t FragmentsTable50[5];
extern const ged_ins_field_position_fragment_t FragmentsTable51[2];
extern const ged_ins_field_position_fragment_t FragmentsTable52[2];
extern const ged_ins_field_position_fragment_t FragmentsTable53[3];
#endif // GED_FRAGMENTS_H
