/*========================== begin_copyright_notice ============================

Copyright (C) 2021-2023 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

// This is an auto-generated file. Please do not edit!
// If changes are needed here please reach out to the codeowners, thanks.


#include "wa_def.h"

#define GT_12_71_REV_ID_A0   SI_REV_ID(0,0)

#define GT_12_71_REV_ID_B0   SI_REV_ID(4,4)

