/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

#ifndef _CFGSTRUCTURIZER_H_
#define _CFGSTRUCTURIZER_H_
namespace vISA {
class FlowGraph;
}
void doCFGStructurize(vISA::FlowGraph *FG);

#endif
