/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

_SPIRV_OP(BuiltInPosition, 0)
_SPIRV_OP(BuiltInPointSize, 1)
_SPIRV_OP(BuiltInClipDistance, 3)
_SPIRV_OP(BuiltInCullDistance, 4)
_SPIRV_OP(BuiltInVertexId, 5)
_SPIRV_OP(BuiltInInstanceId, 6)
_SPIRV_OP(BuiltInPrimitiveId, 7)
_SPIRV_OP(BuiltInInvocationId, 8)
_SPIRV_OP(BuiltInLayer, 9)
_SPIRV_OP(BuiltInViewportIndex, 10)
_SPIRV_OP(BuiltInTessLevelOuter, 11)
_SPIRV_OP(BuiltInTessLevelInner, 12)
_SPIRV_OP(BuiltInTessCoord, 13)
_SPIRV_OP(BuiltInPatchVertices, 14)
_SPIRV_OP(BuiltInFragCoord, 15)
_SPIRV_OP(BuiltInPointCoord, 16)
_SPIRV_OP(BuiltInFrontFacing, 17)
_SPIRV_OP(BuiltInSampleId, 18)
_SPIRV_OP(BuiltInSamplePosition, 19)
_SPIRV_OP(BuiltInSampleMask, 20)
_SPIRV_OP(BuiltInFragColor, 21)
_SPIRV_OP(BuiltInFragDepth, 22)
_SPIRV_OP(BuiltInHelperInvocation, 23)
_SPIRV_OP(BuiltInNumWorkgroups, 24)
_SPIRV_OP(BuiltInWorkgroupSize, 25)
_SPIRV_OP(BuiltInWorkgroupId, 26)
_SPIRV_OP(BuiltInLocalInvocationId, 27)
_SPIRV_OP(BuiltInGlobalInvocationId, 28)
_SPIRV_OP(BuiltInLocalInvocationIndex, 29)
_SPIRV_OP(BuiltInWorkDim, 30)
_SPIRV_OP(BuiltInGlobalSize, 31)
_SPIRV_OP(BuiltInEnqueuedWorkgroupSize, 32)
_SPIRV_OP(BuiltInGlobalOffset, 33)
_SPIRV_OP(BuiltInGlobalLinearId, 34)
_SPIRV_OP(BuiltInSubgroupSize, 36)
_SPIRV_OP(BuiltInSubgroupMaxSize, 37)
_SPIRV_OP(BuiltInNumSubgroups, 38)
_SPIRV_OP(BuiltInNumEnqueuedSubgroups, 39)
_SPIRV_OP(BuiltInSubgroupId, 40)
_SPIRV_OP(BuiltInSubgroupLocalInvocationId, 41)
_SPIRV_OP(BuiltInVertexIndex, 42)
_SPIRV_OP(BuiltInInstanceIndex, 43)
// Ballot extension
_SPIRV_OP(BuiltInSubgroupEqMaskKHR, 4416)
_SPIRV_OP(BuiltInSubgroupGeMaskKHR, 4417)
_SPIRV_OP(BuiltInSubgroupGtMaskKHR, 4418)
_SPIRV_OP(BuiltInSubgroupLeMaskKHR, 4419)
_SPIRV_OP(BuiltInSubgroupLtMaskKHR, 4420)
// SPV_INTEL_hw_thread_queries
_SPIRV_OP(BuiltInSubDeviceIDINTEL, 6135)
_SPIRV_OP(GlobalHWThreadIDINTEL, 6136)
