/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

/*
 * !!! DO NOT EDIT THIS FILE !!!
 *
 * This file was automagically crafted by GED's model parser.
 */

#include "xcoder/ged_internal_api.h"
#include "ged_decoding_tables.h"
#include "ged_mapping_tables.h"
#include "ged_encoding_masks_tables.h"
#include "ged_model_xe_hpc_a.h"

namespace GED_MODEL_NS_XE_HPC_A
{

/*!
 * Top level decoding table in the decoding chain for the format EU_INSTRUCTION_BASIC_ONE_SRC.
 * The indices for the table are the GED_INS_FIELD enumerator values.
 */
static ged_ins_decoding_table_t EU_INSTRUCTION_BASIC_ONE_SRCDecoding = DecodingTable643;

/*!
 * Top level encoding masks table in the encoding chain for the format EU_INSTRUCTION_BASIC_ONE_SRC.
 * The table is a list of encoding masks entries terminated by a NO_MASKS entry.
 */
static ged_instruction_masks_table_t EU_INSTRUCTION_BASIC_ONE_SRCEncodingMasks = EncodingMasksTable147;

/*!
 * Top level decoding table in the decoding chain for the format EU_INSTRUCTION_BASIC_THREE_SRC.
 * The indices for the table are the GED_INS_FIELD enumerator values.
 */
static ged_ins_decoding_table_t EU_INSTRUCTION_BASIC_THREE_SRCDecoding = DecodingTable644;

/*!
 * Top level encoding masks table in the encoding chain for the format EU_INSTRUCTION_BASIC_THREE_SRC.
 * The table is a list of encoding masks entries terminated by a NO_MASKS entry.
 */
static ged_instruction_masks_table_t EU_INSTRUCTION_BASIC_THREE_SRCEncodingMasks = EncodingMasksTable157;

/*!
 * Top level decoding table in the decoding chain for the format EU_INSTRUCTION_BASIC_TWO_SRC.
 * The indices for the table are the GED_INS_FIELD enumerator values.
 */
static ged_ins_decoding_table_t EU_INSTRUCTION_BASIC_TWO_SRCDecoding = DecodingTable645;

/*!
 * Top level encoding masks table in the encoding chain for the format EU_INSTRUCTION_BASIC_TWO_SRC.
 * The table is a list of encoding masks entries terminated by a NO_MASKS entry.
 */
static ged_instruction_masks_table_t EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks = EncodingMasksTable145;

/*!
 * Top level decoding table in the decoding chain for the format EU_INSTRUCTION_BFN.
 * The indices for the table are the GED_INS_FIELD enumerator values.
 */
static ged_ins_decoding_table_t EU_INSTRUCTION_BFNDecoding = DecodingTable646;

/*!
 * Top level encoding masks table in the encoding chain for the format EU_INSTRUCTION_BFN.
 * The table is a list of encoding masks entries terminated by a NO_MASKS entry.
 */
static ged_instruction_masks_table_t EU_INSTRUCTION_BFNEncodingMasks = EncodingMasksTable159;

/*!
 * Top level decoding table in the decoding chain for the format EU_INSTRUCTION_BRANCH_ONE_SRC.
 * The indices for the table are the GED_INS_FIELD enumerator values.
 */
static ged_ins_decoding_table_t EU_INSTRUCTION_BRANCH_ONE_SRCDecoding = DecodingTable647;

/*!
 * Top level encoding masks table in the encoding chain for the format EU_INSTRUCTION_BRANCH_ONE_SRC.
 * The table is a list of encoding masks entries terminated by a NO_MASKS entry.
 */
static ged_instruction_masks_table_t EU_INSTRUCTION_BRANCH_ONE_SRCEncodingMasks = EncodingMasksTable168;

/*!
 * Top level decoding table in the decoding chain for the format EU_INSTRUCTION_BRANCH_TWO_SRC.
 * The indices for the table are the GED_INS_FIELD enumerator values.
 */
static ged_ins_decoding_table_t EU_INSTRUCTION_BRANCH_TWO_SRCDecoding = DecodingTable648;

/*!
 * Top level encoding masks table in the encoding chain for the format EU_INSTRUCTION_BRANCH_TWO_SRC.
 * The table is a list of encoding masks entries terminated by a NO_MASKS entry.
 */
static ged_instruction_masks_table_t EU_INSTRUCTION_BRANCH_TWO_SRCEncodingMasks = EncodingMasksTable170;

/*!
 * Top level decoding table in the decoding chain for the format EU_INSTRUCTION_COMPACT_THREE_SRC.
 * The indices for the table are the GED_INS_FIELD enumerator values.
 */
static ged_ins_decoding_table_t EU_INSTRUCTION_COMPACT_THREE_SRCDecoding = DecodingTable649;

/*!
 * Top level encoding masks table in the encoding chain for the format EU_INSTRUCTION_COMPACT_THREE_SRC.
 * The table is a list of encoding masks entries terminated by a NO_MASKS entry.
 */
static ged_instruction_masks_table_t EU_INSTRUCTION_COMPACT_THREE_SRCEncodingMasks = EncodingMasksTable131;

/*!
 * Top level mapping table in the mapping chain for the compact instruction format EU_INSTRUCTION_COMPACT_THREE_SRC.
 * The indices for the table are the GED_INS_FIELD enumerator values.
 */
static ged_compact_mapping_table_t EU_INSTRUCTION_COMPACT_THREE_SRCMapping = MappingTable53;

/*!
 * Top level decoding table in the decoding chain for the format EU_INSTRUCTION_COMPACT_TWO_SRC.
 * The indices for the table are the GED_INS_FIELD enumerator values.
 */
static ged_ins_decoding_table_t EU_INSTRUCTION_COMPACT_TWO_SRCDecoding = DecodingTable650;

/*!
 * Top level encoding masks table in the encoding chain for the format EU_INSTRUCTION_COMPACT_TWO_SRC.
 * The table is a list of encoding masks entries terminated by a NO_MASKS entry.
 */
static ged_instruction_masks_table_t EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks = EncodingMasksTable131;

/*!
 * Top level mapping table in the mapping chain for the compact instruction format EU_INSTRUCTION_COMPACT_TWO_SRC.
 * The indices for the table are the GED_INS_FIELD enumerator values.
 */
static ged_compact_mapping_table_t EU_INSTRUCTION_COMPACT_TWO_SRCMapping = MappingTable54;

/*!
 * Top level decoding table in the decoding chain for the format EU_INSTRUCTION_DPAS_THREE_SRC.
 * The indices for the table are the GED_INS_FIELD enumerator values.
 */
static ged_ins_decoding_table_t EU_INSTRUCTION_DPAS_THREE_SRCDecoding = DecodingTable651;

/*!
 * Top level encoding masks table in the encoding chain for the format EU_INSTRUCTION_DPAS_THREE_SRC.
 * The table is a list of encoding masks entries terminated by a NO_MASKS entry.
 */
static ged_instruction_masks_table_t EU_INSTRUCTION_DPAS_THREE_SRCEncodingMasks = EncodingMasksTable158;

/*!
 * Top level decoding table in the decoding chain for the format EU_INSTRUCTION_ILLEGAL.
 * The indices for the table are the GED_INS_FIELD enumerator values.
 */
static ged_ins_decoding_table_t EU_INSTRUCTION_ILLEGALDecoding = DecodingTable622;

/*!
 * Top level encoding masks table in the encoding chain for the format EU_INSTRUCTION_ILLEGAL.
 * The table is a list of encoding masks entries terminated by a NO_MASKS entry.
 */
static ged_instruction_masks_table_t EU_INSTRUCTION_ILLEGALEncodingMasks = EncodingMasksTable11;

/*!
 * Top level decoding table in the decoding chain for the format EU_INSTRUCTION_MATH.
 * The indices for the table are the GED_INS_FIELD enumerator values.
 */
static ged_ins_decoding_table_t EU_INSTRUCTION_MATHDecoding = DecodingTable652;

/*!
 * Top level encoding masks table in the encoding chain for the format EU_INSTRUCTION_MATH.
 * The table is a list of encoding masks entries terminated by a NO_MASKS entry.
 */
static ged_instruction_masks_table_t EU_INSTRUCTION_MATHEncodingMasks = EncodingMasksTable166;

/*!
 * Top level decoding table in the decoding chain for the format EU_INSTRUCTION_NOP.
 * The indices for the table are the GED_INS_FIELD enumerator values.
 */
static ged_ins_decoding_table_t EU_INSTRUCTION_NOPDecoding = DecodingTable653;

/*!
 * Top level encoding masks table in the encoding chain for the format EU_INSTRUCTION_NOP.
 * The table is a list of encoding masks entries terminated by a NO_MASKS entry.
 */
static ged_instruction_masks_table_t EU_INSTRUCTION_NOPEncodingMasks = EncodingMasksTable151;

/*!
 * Top level decoding table in the decoding chain for the format EU_INSTRUCTION_SEND.
 * The indices for the table are the GED_INS_FIELD enumerator values.
 */
static ged_ins_decoding_table_t EU_INSTRUCTION_SENDDecoding = DecodingTable654;

/*!
 * Top level encoding masks table in the encoding chain for the format EU_INSTRUCTION_SEND.
 * The table is a list of encoding masks entries terminated by a NO_MASKS entry.
 */
static ged_instruction_masks_table_t EU_INSTRUCTION_SENDEncodingMasks = EncodingMasksTable160;

/*!
 * Top level decoding table in the decoding chain for the format EU_INSTRUCTION_SYNC.
 * The indices for the table are the GED_INS_FIELD enumerator values.
 */
static ged_ins_decoding_table_t EU_INSTRUCTION_SYNCDecoding = DecodingTable655;

/*!
 * Top level encoding masks table in the encoding chain for the format EU_INSTRUCTION_SYNC.
 * The table is a list of encoding masks entries terminated by a NO_MASKS entry.
 */
static ged_instruction_masks_table_t EU_INSTRUCTION_SYNCEncodingMasks = EncodingMasksTable150;

/*!
 * Top level decoding table in the decoding chain for the format oneSourceCompact.
 * The indices for the table are the GED_INS_FIELD enumerator values.
 */
static ged_ins_decoding_table_t oneSourceCompactDecoding = DecodingTable656;

/*!
 * Top level encoding masks table in the encoding chain for the format oneSourceCompact.
 * The table is a list of encoding masks entries terminated by a NO_MASKS entry.
 */
static ged_instruction_masks_table_t oneSourceCompactEncodingMasks = EncodingMasksTable131;

/*!
 * Top level mapping table in the mapping chain for the compact instruction format oneSourceCompact.
 * The indices for the table are the GED_INS_FIELD enumerator values.
 */
static ged_compact_mapping_table_t oneSourceCompactMapping = MappingTable55;
OpcodeTables Opcodes[128] =
{
    EU_INSTRUCTION_ILLEGALDecoding, EU_INSTRUCTION_ILLEGALEncodingMasks, NULL, NULL, NULL, // 0
    EU_INSTRUCTION_SYNCDecoding, EU_INSTRUCTION_SYNCEncodingMasks, oneSourceCompactDecoding, oneSourceCompactEncodingMasks, oneSourceCompactMapping, // 1
    NULL, NULL, NULL, NULL, NULL, // 2
    NULL, NULL, NULL, NULL, NULL, // 3
    NULL, NULL, NULL, NULL, NULL, // 4
    NULL, NULL, NULL, NULL, NULL, // 5
    NULL, NULL, NULL, NULL, NULL, // 6
    NULL, NULL, NULL, NULL, NULL, // 7
    NULL, NULL, NULL, NULL, NULL, // 8
    NULL, NULL, NULL, NULL, NULL, // 9
    NULL, NULL, NULL, NULL, NULL, // 10
    NULL, NULL, NULL, NULL, NULL, // 11
    NULL, NULL, NULL, NULL, NULL, // 12
    NULL, NULL, NULL, NULL, NULL, // 13
    NULL, NULL, NULL, NULL, NULL, // 14
    NULL, NULL, NULL, NULL, NULL, // 15
    NULL, NULL, NULL, NULL, NULL, // 16
    NULL, NULL, NULL, NULL, NULL, // 17
    NULL, NULL, NULL, NULL, NULL, // 18
    NULL, NULL, NULL, NULL, NULL, // 19
    NULL, NULL, NULL, NULL, NULL, // 20
    NULL, NULL, NULL, NULL, NULL, // 21
    NULL, NULL, NULL, NULL, NULL, // 22
    NULL, NULL, NULL, NULL, NULL, // 23
    NULL, NULL, NULL, NULL, NULL, // 24
    NULL, NULL, NULL, NULL, NULL, // 25
    NULL, NULL, NULL, NULL, NULL, // 26
    NULL, NULL, NULL, NULL, NULL, // 27
    NULL, NULL, NULL, NULL, NULL, // 28
    NULL, NULL, NULL, NULL, NULL, // 29
    NULL, NULL, NULL, NULL, NULL, // 30
    NULL, NULL, NULL, NULL, NULL, // 31
    EU_INSTRUCTION_BRANCH_ONE_SRCDecoding, EU_INSTRUCTION_BRANCH_ONE_SRCEncodingMasks, NULL, NULL, NULL, // 32
    EU_INSTRUCTION_BRANCH_ONE_SRCDecoding, EU_INSTRUCTION_BRANCH_ONE_SRCEncodingMasks, NULL, NULL, NULL, // 33
    EU_INSTRUCTION_BRANCH_TWO_SRCDecoding, EU_INSTRUCTION_BRANCH_TWO_SRCEncodingMasks, NULL, NULL, NULL, // 34
    EU_INSTRUCTION_BRANCH_TWO_SRCDecoding, EU_INSTRUCTION_BRANCH_TWO_SRCEncodingMasks, NULL, NULL, NULL, // 35
    EU_INSTRUCTION_BRANCH_TWO_SRCDecoding, EU_INSTRUCTION_BRANCH_TWO_SRCEncodingMasks, NULL, NULL, NULL, // 36
    EU_INSTRUCTION_BRANCH_ONE_SRCDecoding, EU_INSTRUCTION_BRANCH_ONE_SRCEncodingMasks, NULL, NULL, NULL, // 37
    NULL, NULL, NULL, NULL, NULL, // 38
    EU_INSTRUCTION_BRANCH_ONE_SRCDecoding, EU_INSTRUCTION_BRANCH_ONE_SRCEncodingMasks, NULL, NULL, NULL, // 39
    EU_INSTRUCTION_BRANCH_TWO_SRCDecoding, EU_INSTRUCTION_BRANCH_TWO_SRCEncodingMasks, NULL, NULL, NULL, // 40
    EU_INSTRUCTION_BRANCH_TWO_SRCDecoding, EU_INSTRUCTION_BRANCH_TWO_SRCEncodingMasks, NULL, NULL, NULL, // 41
    EU_INSTRUCTION_BRANCH_TWO_SRCDecoding, EU_INSTRUCTION_BRANCH_TWO_SRCEncodingMasks, NULL, NULL, NULL, // 42
    EU_INSTRUCTION_BRANCH_ONE_SRCDecoding, EU_INSTRUCTION_BRANCH_ONE_SRCEncodingMasks, NULL, NULL, NULL, // 43
    EU_INSTRUCTION_BRANCH_ONE_SRCDecoding, EU_INSTRUCTION_BRANCH_ONE_SRCEncodingMasks, NULL, NULL, NULL, // 44
    EU_INSTRUCTION_BRANCH_ONE_SRCDecoding, EU_INSTRUCTION_BRANCH_ONE_SRCEncodingMasks, NULL, NULL, NULL, // 45
    EU_INSTRUCTION_BRANCH_TWO_SRCDecoding, EU_INSTRUCTION_BRANCH_TWO_SRCEncodingMasks, NULL, NULL, NULL, // 46
    EU_INSTRUCTION_BRANCH_ONE_SRCDecoding, EU_INSTRUCTION_BRANCH_ONE_SRCEncodingMasks, NULL, NULL, NULL, // 47
    NULL, NULL, NULL, NULL, NULL, // 48
    EU_INSTRUCTION_SENDDecoding, EU_INSTRUCTION_SENDEncodingMasks, NULL, NULL, NULL, // 49
    EU_INSTRUCTION_SENDDecoding, EU_INSTRUCTION_SENDEncodingMasks, NULL, NULL, NULL, // 50
    NULL, NULL, NULL, NULL, NULL, // 51
    NULL, NULL, NULL, NULL, NULL, // 52
    NULL, NULL, NULL, NULL, NULL, // 53
    NULL, NULL, NULL, NULL, NULL, // 54
    NULL, NULL, NULL, NULL, NULL, // 55
    EU_INSTRUCTION_MATHDecoding, EU_INSTRUCTION_MATHEncodingMasks, NULL, NULL, NULL, // 56
    NULL, NULL, NULL, NULL, NULL, // 57
    NULL, NULL, NULL, NULL, NULL, // 58
    NULL, NULL, NULL, NULL, NULL, // 59
    NULL, NULL, NULL, NULL, NULL, // 60
    NULL, NULL, NULL, NULL, NULL, // 61
    NULL, NULL, NULL, NULL, NULL, // 62
    NULL, NULL, NULL, NULL, NULL, // 63
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 64
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 65
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 66
    EU_INSTRUCTION_BASIC_ONE_SRCDecoding, EU_INSTRUCTION_BASIC_ONE_SRCEncodingMasks, oneSourceCompactDecoding, oneSourceCompactEncodingMasks, oneSourceCompactMapping, // 67
    EU_INSTRUCTION_BASIC_ONE_SRCDecoding, EU_INSTRUCTION_BASIC_ONE_SRCEncodingMasks, oneSourceCompactDecoding, oneSourceCompactEncodingMasks, oneSourceCompactMapping, // 68
    EU_INSTRUCTION_BASIC_ONE_SRCDecoding, EU_INSTRUCTION_BASIC_ONE_SRCEncodingMasks, oneSourceCompactDecoding, oneSourceCompactEncodingMasks, oneSourceCompactMapping, // 69
    EU_INSTRUCTION_BASIC_ONE_SRCDecoding, EU_INSTRUCTION_BASIC_ONE_SRCEncodingMasks, oneSourceCompactDecoding, oneSourceCompactEncodingMasks, oneSourceCompactMapping, // 70
    EU_INSTRUCTION_BASIC_ONE_SRCDecoding, EU_INSTRUCTION_BASIC_ONE_SRCEncodingMasks, oneSourceCompactDecoding, oneSourceCompactEncodingMasks, oneSourceCompactMapping, // 71
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 72
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 73
    EU_INSTRUCTION_BASIC_ONE_SRCDecoding, EU_INSTRUCTION_BASIC_ONE_SRCEncodingMasks, oneSourceCompactDecoding, oneSourceCompactEncodingMasks, oneSourceCompactMapping, // 74
    EU_INSTRUCTION_BASIC_ONE_SRCDecoding, EU_INSTRUCTION_BASIC_ONE_SRCEncodingMasks, oneSourceCompactDecoding, oneSourceCompactEncodingMasks, oneSourceCompactMapping, // 75
    EU_INSTRUCTION_BASIC_ONE_SRCDecoding, EU_INSTRUCTION_BASIC_ONE_SRCEncodingMasks, oneSourceCompactDecoding, oneSourceCompactEncodingMasks, oneSourceCompactMapping, // 76
    EU_INSTRUCTION_BASIC_ONE_SRCDecoding, EU_INSTRUCTION_BASIC_ONE_SRCEncodingMasks, oneSourceCompactDecoding, oneSourceCompactEncodingMasks, oneSourceCompactMapping, // 77
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 78
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 79
    NULL, NULL, NULL, NULL, NULL, // 80
    NULL, NULL, NULL, NULL, NULL, // 81
    EU_INSTRUCTION_BASIC_THREE_SRCDecoding, EU_INSTRUCTION_BASIC_THREE_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_THREE_SRCDecoding, EU_INSTRUCTION_COMPACT_THREE_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_THREE_SRCMapping, // 82
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 83
    NULL, NULL, NULL, NULL, NULL, // 84
    NULL, NULL, NULL, NULL, NULL, // 85
    NULL, NULL, NULL, NULL, NULL, // 86
    NULL, NULL, NULL, NULL, NULL, // 87
    EU_INSTRUCTION_BASIC_THREE_SRCDecoding, EU_INSTRUCTION_BASIC_THREE_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_THREE_SRCDecoding, EU_INSTRUCTION_COMPACT_THREE_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_THREE_SRCMapping, // 88
    EU_INSTRUCTION_DPAS_THREE_SRCDecoding, EU_INSTRUCTION_DPAS_THREE_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_THREE_SRCDecoding, EU_INSTRUCTION_COMPACT_THREE_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_THREE_SRCMapping, // 89
    NULL, NULL, NULL, NULL, NULL, // 90
    EU_INSTRUCTION_BASIC_THREE_SRCDecoding, EU_INSTRUCTION_BASIC_THREE_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_THREE_SRCDecoding, EU_INSTRUCTION_COMPACT_THREE_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_THREE_SRCMapping, // 91
    NULL, NULL, NULL, NULL, NULL, // 92
    EU_INSTRUCTION_BASIC_THREE_SRCDecoding, EU_INSTRUCTION_BASIC_THREE_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_THREE_SRCDecoding, EU_INSTRUCTION_COMPACT_THREE_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_THREE_SRCMapping, // 93
    NULL, NULL, NULL, NULL, NULL, // 94
    NULL, NULL, NULL, NULL, NULL, // 95
    EU_INSTRUCTION_NOPDecoding, EU_INSTRUCTION_NOPEncodingMasks, NULL, NULL, NULL, // 96
    EU_INSTRUCTION_BASIC_ONE_SRCDecoding, EU_INSTRUCTION_BASIC_ONE_SRCEncodingMasks, oneSourceCompactDecoding, oneSourceCompactEncodingMasks, oneSourceCompactMapping, // 97
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 98
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 99
    EU_INSTRUCTION_BASIC_ONE_SRCDecoding, EU_INSTRUCTION_BASIC_ONE_SRCEncodingMasks, oneSourceCompactDecoding, oneSourceCompactEncodingMasks, oneSourceCompactMapping, // 100
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 101
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 102
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 103
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 104
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 105
    NULL, NULL, NULL, NULL, NULL, // 106
    EU_INSTRUCTION_BFNDecoding, EU_INSTRUCTION_BFNEncodingMasks, EU_INSTRUCTION_COMPACT_THREE_SRCDecoding, EU_INSTRUCTION_COMPACT_THREE_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_THREE_SRCMapping, // 107
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 108
    NULL, NULL, NULL, NULL, NULL, // 109
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 110
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 111
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 112
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 113
    EU_INSTRUCTION_BASIC_THREE_SRCDecoding, EU_INSTRUCTION_BASIC_THREE_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_THREE_SRCDecoding, EU_INSTRUCTION_COMPACT_THREE_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_THREE_SRCMapping, // 114
    NULL, NULL, NULL, NULL, NULL, // 115
    NULL, NULL, NULL, NULL, NULL, // 116
    NULL, NULL, NULL, NULL, NULL, // 117
    NULL, NULL, NULL, NULL, NULL, // 118
    EU_INSTRUCTION_BASIC_ONE_SRCDecoding, EU_INSTRUCTION_BASIC_ONE_SRCEncodingMasks, oneSourceCompactDecoding, oneSourceCompactEncodingMasks, oneSourceCompactMapping, // 119
    EU_INSTRUCTION_BASIC_THREE_SRCDecoding, EU_INSTRUCTION_BASIC_THREE_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_THREE_SRCDecoding, EU_INSTRUCTION_COMPACT_THREE_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_THREE_SRCMapping, // 120
    EU_INSTRUCTION_BASIC_TWO_SRCDecoding, EU_INSTRUCTION_BASIC_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCDecoding, EU_INSTRUCTION_COMPACT_TWO_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_TWO_SRCMapping, // 121
    EU_INSTRUCTION_BASIC_THREE_SRCDecoding, EU_INSTRUCTION_BASIC_THREE_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_THREE_SRCDecoding, EU_INSTRUCTION_COMPACT_THREE_SRCEncodingMasks, EU_INSTRUCTION_COMPACT_THREE_SRCMapping, // 122
    NULL, NULL, NULL, NULL, NULL, // 123
    NULL, NULL, NULL, NULL, NULL, // 124
    NULL, NULL, NULL, NULL, NULL, // 125
    NULL, NULL, NULL, NULL, NULL, // 126
    NULL, NULL, NULL, NULL, NULL // 127
}; // Opcodes[]
} // namespace GED_MODEL_NS_XE_HPC_A
