/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

/*
 * !!! DO NOT EDIT THIS FILE !!!
 *
 * This file was automagically crafted by GED's model parser.
 */


#ifndef GED_INS_FIELD_INTERNAL_H
#define GED_INS_FIELD_INTERNAL_H

#include "common/ged_types_internal.h"

/*!
 * Table mapping a GED_INS_FIELD to its GED_FIELD_TYPE.
 */
extern GED_FIELD_TYPE fieldTypesByField[129];

/*!
 * Table mapping GED_INS_FIELD to its name (string representation).
 */
extern const char* fieldNameByField[129];

/*!
 * Table mapping a GED_PSEUDO_FIELD to its GED_FIELD_TYPE.
 */
extern GED_FIELD_TYPE pseudoFieldTypesByField[45];

/*!
 * Table mapping GED_PSEUDO_FIELD to its name (string representation).
 */
extern const char* fieldNameByPseudoField[45];
#endif // GED_INS_FIELD_INTERNAL_H
