/*========================== begin_copyright_notice ============================

Copyright (C) 2023 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

//===----------------------------------------------------------------------===//
/// Listed below are STD libraries and compiler header(s) with
/// largest number of inclusions being done when building IGC
//===----------------------------------------------------------------------===//

// std
#include <algorithm>
#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cwchar>
#include <exception>
#include <fstream>
#include <functional>
#include <immintrin.h>
#include <intrin.h>
#include <ios>
#include <iosfwd>
#include <iostream>
#include <istream>
#include <limits>
#include <list>
#include <memory>
#include <mutex>
#include <new>
#include <optional>
#include <ostream>
#include <set>
#include <sstream>
#include <stdexcept>
#include <streambuf>
#include <string>
#include <system_error>
#include <type_traits>
#include <unordered_map>
#include <utility>
#include <vector>
#include <xhash>
#include <xiosbase>
#include <xlocale>
#include <xlocnum>
#include <xmemory>
#include <xstring>
#include <xutility>
#include <zmmintrin.h>

// misc
#include "Compiler/CodeGenPublic.h"
