/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

/*
 * !!! DO NOT EDIT THIS FILE !!!
 *
 * This file was automagically crafted by GED's model parser.
 */

#include "ged_enum_tables.h"
#include "ged_enum_interpreters.h"
const ged_unsigned_table_t EnumInterpretersTable0[3] =
{
    NULL, // 0
    NULL, // 1
    NULL // 2
}; // EnumInterpretersTable0[]
const ged_unsigned_table_t EnumInterpretersTable1[3] =
{
    unsignedTable23, // 0
    unsignedTable29, // 1
    unsignedTable33 // 2
}; // EnumInterpretersTable1[]
const ged_unsigned_table_t EnumInterpretersTable2[3] =
{
    unsignedTable24, // 0
    unsignedTable30, // 1
    unsignedTable33 // 2
}; // EnumInterpretersTable2[]
const ged_unsigned_table_t EnumInterpretersTable3[3] =
{
    unsignedTable25, // 0
    unsignedTable30, // 1
    unsignedTable33 // 2
}; // EnumInterpretersTable3[]
const ged_unsigned_table_t EnumInterpretersTable4[3] =
{
    unsignedTable25, // 0
    unsignedTable30, // 1
    NULL // 2
}; // EnumInterpretersTable4[]
const ged_unsigned_table_t EnumInterpretersTable5[3] =
{
    unsignedTable26, // 0
    unsignedTable31, // 1
    NULL // 2
}; // EnumInterpretersTable5[]
const ged_unsigned_table_t EnumInterpretersTable6[3] =
{
    unsignedTable27, // 0
    unsignedTable31, // 1
    NULL // 2
}; // EnumInterpretersTable6[]
const ged_unsigned_table_t EnumInterpretersTable7[3] =
{
    unsignedTable28, // 0
    unsignedTable32, // 1
    NULL // 2
}; // EnumInterpretersTable7[]
