/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

// Debug info ops
_OCL_EXT_OP(DebugInfoNone, 0)
_OCL_EXT_OP(CompileUnit, 1)
_OCL_EXT_OP(TypeBasic, 2)
_OCL_EXT_OP(TypePtr, 3)
_OCL_EXT_OP(TypeQualifier, 4)
_OCL_EXT_OP(TypeArray, 5)
_OCL_EXT_OP(TypeVector, 6)
_OCL_EXT_OP(TypeDef, 7)
_OCL_EXT_OP(TypeFunction, 8)
_OCL_EXT_OP(TypeEnum, 9)
_OCL_EXT_OP(TypeComposite, 10)
_OCL_EXT_OP(TypeMember, 11)
_OCL_EXT_OP(TypeInheritance, 12)
_OCL_EXT_OP(TypePtrToMember, 13)
_OCL_EXT_OP(TypeTemplate, 14)
_OCL_EXT_OP(TypeTemplateParameter, 15)
_OCL_EXT_OP(TypeTemplateTemplateParameter, 16)
_OCL_EXT_OP(TypeTemplateParameterPack, 17)
_OCL_EXT_OP(GlobalVariable, 18)
_OCL_EXT_OP(FuncDecl, 19)
_OCL_EXT_OP(Function, 20)
_OCL_EXT_OP(LexicalBlock, 21)
_OCL_EXT_OP(LexicalBlockDiscriminator, 22)
_OCL_EXT_OP(Scope, 23)
_OCL_EXT_OP(NoScope, 24)
_OCL_EXT_OP(InlinedAt, 25)
_OCL_EXT_OP(LocalVariable, 26)
_OCL_EXT_OP(InlinedVariable, 27)
_OCL_EXT_OP(DbgDcl, 28)
_OCL_EXT_OP(DbgVal, 29)
_OCL_EXT_OP(DbgOperation, 30)
_OCL_EXT_OP(DbgExpr, 31)
_OCL_EXT_OP(MacroDef, 32)
_OCL_EXT_OP(MacroUndef, 33)
_OCL_EXT_OP(ImportedEntity, 34)
_OCL_EXT_OP(Source, 35)
_OCL_EXT_OP(ModuleINTEL, 36)
