/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

#ifndef GED_ENUM_TYPES_INTERNAL_H
#define GED_ENUM_TYPES_INTERNAL_H

#include "common/ged_enum_types.h"


// String representation of the GED_RETURN_VALUE enum and the padding to be added in order for them to be aligned column-wise.
extern const char* gedReturnValueStrings[GED_RETURN_VALUE_SIZE];
extern const char* gedReturnValuePadding[GED_RETURN_VALUE_SIZE];

#endif // GED_ENUM_TYPES_INTERNAL_H
