/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

/*
 * !!! DO NOT EDIT THIS FILE !!!
 *
 * This file was automagically crafted by GED's model parser.
 */


#ifndef GED_RESTRICTIONS_H
#define GED_RESTRICTIONS_H

#include "common/ged_ins_restrictions.h"
extern const ged_field_restriction_t* RestrictionTable0[2];
extern const ged_field_restriction_t* RestrictionTable1[2];
extern const ged_field_restriction_t* RestrictionTable2[2];
extern const ged_field_restriction_t* RestrictionTable3[2];
extern const ged_field_restriction_t* RestrictionTable4[2];
extern const ged_field_restriction_t* RestrictionTable5[2];
extern const ged_field_restriction_t* RestrictionTable6[2];
extern const ged_field_restriction_t* RestrictionTable7[2];
extern const ged_field_restriction_t* RestrictionTable8[2];
extern const ged_field_restriction_t* RestrictionTable9[2];
extern const ged_field_restriction_t* RestrictionTable10[2];
extern const ged_field_restriction_t* RestrictionTable11[2];
extern const ged_field_restriction_t* RestrictionTable12[2];
extern const ged_field_restriction_t* RestrictionTable13[2];
extern const ged_field_restriction_t* RestrictionTable14[2];
extern const ged_field_restriction_t* RestrictionTable15[2];
extern const ged_field_restriction_t* RestrictionTable16[2];
extern const ged_field_restriction_t* RestrictionTable17[2];
extern const ged_field_restriction_t* RestrictionTable18[2];
extern const ged_field_restriction_t* RestrictionTable19[2];
extern const ged_field_restriction_t* RestrictionTable20[2];
extern const ged_field_restriction_t* RestrictionTable21[2];
extern const ged_field_restriction_t* RestrictionTable22[2];
extern const ged_field_restriction_t* RestrictionTable23[2];
extern const ged_field_restriction_t* RestrictionTable24[2];
extern const ged_field_restriction_t* RestrictionTable25[2];
extern const ged_field_restriction_t* RestrictionTable26[2];
extern const ged_field_restriction_t* RestrictionTable27[2];
extern const ged_field_restriction_t* RestrictionTable28[2];
extern const ged_field_restriction_t* RestrictionTable29[2];
extern const ged_field_restriction_t* RestrictionTable30[2];
extern const ged_field_restriction_t* RestrictionTable31[2];
extern const ged_field_restriction_t* RestrictionTable32[2];
extern const ged_field_restriction_t* RestrictionTable33[2];
extern const ged_field_restriction_t* RestrictionTable34[2];
extern const ged_field_restriction_t* RestrictionTable35[2];
extern const ged_field_restriction_t* RestrictionTable36[2];
extern const ged_field_restriction_t* RestrictionTable37[2];
extern const ged_field_restriction_t* RestrictionTable38[2];
extern const ged_field_restriction_t* RestrictionTable39[2];
extern const ged_field_restriction_t* RestrictionTable40[2];
extern const ged_field_restriction_t* RestrictionTable41[2];
extern const ged_field_restriction_t* RestrictionTable42[2];
extern const ged_field_restriction_t* RestrictionTable43[2];
extern const ged_field_restriction_t* RestrictionTable44[2];
extern const ged_field_restriction_t* RestrictionTable45[2];
extern const ged_field_restriction_t* RestrictionTable46[2];
extern const ged_field_restriction_t* RestrictionTable47[2];
extern const ged_field_restriction_t* RestrictionTable48[2];
extern const ged_field_restriction_t* RestrictionTable49[2];
extern const ged_field_restriction_t* RestrictionTable50[2];
extern const ged_field_restriction_t* RestrictionTable51[2];
extern const ged_field_restriction_t* RestrictionTable52[2];
extern const ged_field_restriction_t* RestrictionTable53[2];
extern const ged_field_restriction_t* RestrictionTable54[2];
extern const ged_field_restriction_t* RestrictionTable55[2];
extern const ged_field_restriction_t* RestrictionTable56[2];
extern const ged_field_restriction_t* RestrictionTable57[2];
extern const ged_field_restriction_t* RestrictionTable58[2];
extern const ged_field_restriction_t* RestrictionTable59[2];
extern const ged_field_restriction_t* RestrictionTable60[2];
extern const ged_field_restriction_t* RestrictionTable61[2];
extern const ged_field_restriction_t* RestrictionTable62[2];
extern const ged_field_restriction_t* RestrictionTable63[2];
extern const ged_field_restriction_t* RestrictionTable64[2];
extern const ged_field_restriction_t* RestrictionTable65[2];
extern const ged_field_restriction_t* RestrictionTable66[2];
extern const ged_field_restriction_t* RestrictionTable67[2];
extern const ged_field_restriction_t* RestrictionTable68[2];
extern const ged_field_restriction_t* RestrictionTable69[2];
extern const ged_field_restriction_t* RestrictionTable70[2];
extern const ged_field_restriction_t* RestrictionTable71[2];
extern const ged_field_restriction_t* RestrictionTable72[2];
extern const ged_field_restriction_t* RestrictionTable73[2];
extern const ged_field_restriction_t* RestrictionTable74[2];
extern const ged_field_restriction_t* RestrictionTable75[2];
extern const ged_field_restriction_t* RestrictionTable76[2];
extern const ged_field_restriction_t* RestrictionTable77[2];
extern const ged_field_restriction_t* RestrictionTable78[2];
extern const ged_field_restriction_t* RestrictionTable79[2];
extern const ged_field_restriction_t* RestrictionTable80[2];
extern const ged_field_restriction_t* RestrictionTable81[2];
extern const ged_field_restriction_t* RestrictionTable82[2];
extern const ged_field_restriction_t* RestrictionTable83[2];
extern const ged_field_restriction_t* RestrictionTable84[2];
extern const ged_field_restriction_t* RestrictionTable85[2];
extern const ged_field_restriction_t* RestrictionTable86[2];
extern const ged_field_restriction_t* RestrictionTable87[2];
extern const ged_field_restriction_t* RestrictionTable88[2];
extern const ged_field_restriction_t* RestrictionTable89[2];
extern const ged_field_restriction_t* RestrictionTable90[2];
extern const ged_field_restriction_t* RestrictionTable91[2];
extern const ged_field_restriction_t* RestrictionTable92[2];
extern const ged_field_restriction_t* RestrictionTable93[2];
extern const ged_field_restriction_t* RestrictionTable94[2];
extern const ged_field_restriction_t* RestrictionTable95[2];
extern const ged_field_restriction_t* RestrictionTable96[2];
extern const ged_field_restriction_t* RestrictionTable97[2];
extern const ged_field_restriction_t* RestrictionTable98[2];
extern const ged_field_restriction_t* RestrictionTable99[2];
extern const ged_field_restriction_t* RestrictionTable100[2];
extern const ged_field_restriction_t* RestrictionTable101[2];
extern const ged_field_restriction_t* RestrictionTable102[2];
extern const ged_field_restriction_t* RestrictionTable103[2];
extern const ged_field_restriction_t* RestrictionTable104[2];
extern const ged_field_restriction_t* RestrictionTable105[2];
extern const ged_field_restriction_t* RestrictionTable106[2];
extern const ged_field_restriction_t* RestrictionTable107[2];
extern const ged_field_restriction_t* RestrictionTable108[2];
extern const ged_field_restriction_t* RestrictionTable109[2];
extern const ged_field_restriction_t* RestrictionTable110[2];
extern const ged_field_restriction_t* RestrictionTable111[2];
extern const ged_field_restriction_t* RestrictionTable112[2];
extern const ged_field_restriction_t* RestrictionTable113[2];
extern const ged_field_restriction_t* RestrictionTable114[2];
extern const ged_field_restriction_t* RestrictionTable115[2];
extern const ged_field_restriction_t* RestrictionTable116[2];
extern const ged_field_restriction_t* RestrictionTable117[2];
extern const ged_field_restriction_t* RestrictionTable118[2];
extern const ged_field_restriction_t* RestrictionTable119[2];
extern const ged_field_restriction_t* RestrictionTable120[2];
extern const ged_field_restriction_t* RestrictionTable121[2];
extern const ged_field_restriction_t* RestrictionTable122[2];
extern const ged_field_restriction_t* RestrictionTable123[2];
extern const ged_field_restriction_t* RestrictionTable124[2];
extern const ged_field_restriction_t* RestrictionTable125[2];
extern const ged_field_restriction_t* RestrictionTable126[2];
extern const ged_field_restriction_t* RestrictionTable127[2];
extern const ged_field_restriction_t* RestrictionTable128[2];
extern const ged_field_restriction_t* RestrictionTable129[2];
extern const ged_field_restriction_t* RestrictionTable130[2];
extern const ged_field_restriction_t* RestrictionTable131[2];
extern const ged_field_restriction_t* RestrictionTable132[2];
extern const ged_field_restriction_t* RestrictionTable133[2];
extern const ged_field_restriction_t* RestrictionTable134[2];
extern const ged_field_restriction_t* RestrictionTable135[2];
extern const ged_field_restriction_t* RestrictionTable136[2];
extern const ged_field_restriction_t* RestrictionTable137[2];
extern const ged_field_restriction_t* RestrictionTable138[2];
extern const ged_field_restriction_t* RestrictionTable139[2];
extern const ged_field_restriction_t* RestrictionTable140[2];
extern const ged_field_restriction_t* RestrictionTable141[2];
extern const ged_field_restriction_t* RestrictionTable142[2];
extern const ged_field_restriction_t* RestrictionTable143[2];
extern const ged_field_restriction_t* RestrictionTable144[2];
extern const ged_field_restriction_t* RestrictionTable145[2];
extern const ged_field_restriction_t* RestrictionTable146[2];
extern const ged_field_restriction_t* RestrictionTable147[2];
extern const ged_field_restriction_t* RestrictionTable148[2];
extern const ged_field_restriction_t* RestrictionTable149[2];
extern const ged_field_restriction_t* RestrictionTable150[2];
extern const ged_field_restriction_t* RestrictionTable151[2];
extern const ged_field_restriction_t* RestrictionTable152[2];
extern const ged_field_restriction_t* RestrictionTable153[2];
extern const ged_field_restriction_t* RestrictionTable154[2];
extern const ged_field_restriction_t* RestrictionTable155[2];
extern const ged_field_restriction_t* RestrictionTable156[2];
extern const ged_field_restriction_t* RestrictionTable157[2];
extern const ged_field_restriction_t* RestrictionTable158[2];
extern const ged_field_restriction_t* RestrictionTable159[2];
extern const ged_field_restriction_t* RestrictionTable160[2];
extern const ged_field_restriction_t* RestrictionTable161[2];
extern const ged_field_restriction_t* RestrictionTable162[2];
extern const ged_field_restriction_t* RestrictionTable163[2];
extern const ged_field_restriction_t* RestrictionTable164[2];
extern const ged_field_restriction_t* RestrictionTable165[2];
extern const ged_field_restriction_t* RestrictionTable166[2];
extern const ged_field_restriction_t* RestrictionTable167[2];
extern const ged_field_restriction_t* RestrictionTable168[2];
extern const ged_field_restriction_t* RestrictionTable169[2];
extern const ged_field_restriction_t* RestrictionTable170[2];
extern const ged_field_restriction_t* RestrictionTable171[2];
extern const ged_field_restriction_t* RestrictionTable172[2];
extern const ged_field_restriction_t* RestrictionTable173[2];
extern const ged_field_restriction_t* RestrictionTable174[2];
extern const ged_field_restriction_t* RestrictionTable175[2];
extern const ged_field_restriction_t* RestrictionTable176[2];
extern const ged_field_restriction_t* RestrictionTable177[2];
extern const ged_field_restriction_t* RestrictionTable178[2];
extern const ged_field_restriction_t* RestrictionTable179[2];
extern const ged_field_restriction_t* RestrictionTable180[2];
extern const ged_field_restriction_t* RestrictionTable181[2];
extern const ged_field_restriction_t* RestrictionTable182[2];
extern const ged_field_restriction_t* RestrictionTable183[2];
extern const ged_field_restriction_t* RestrictionTable184[2];
extern const ged_field_restriction_t* RestrictionTable185[2];
extern const ged_field_restriction_t* RestrictionTable186[2];
extern const ged_field_restriction_t* RestrictionTable187[2];
extern const ged_field_restriction_t* RestrictionTable188[2];
extern const ged_field_restriction_t* RestrictionTable189[2];
extern const ged_field_restriction_t* RestrictionTable190[2];
extern const ged_field_restriction_t* RestrictionTable191[2];
extern const ged_field_restriction_t* RestrictionTable192[2];
extern const ged_field_restriction_t* RestrictionTable193[2];
extern const ged_field_restriction_t* RestrictionTable194[2];
extern const ged_field_restriction_t* RestrictionTable195[2];
extern const ged_field_restriction_t* RestrictionTable196[2];
extern const ged_field_restriction_t* RestrictionTable197[2];
extern const ged_field_restriction_t* RestrictionTable198[2];
extern const ged_field_restriction_t* RestrictionTable199[2];
extern const ged_field_restriction_t* RestrictionTable200[2];
extern const ged_field_restriction_t* RestrictionTable201[2];
extern const ged_field_restriction_t* RestrictionTable202[2];
extern const ged_field_restriction_t* RestrictionTable203[2];
extern const ged_field_restriction_t* RestrictionTable204[2];
extern const ged_field_restriction_t* RestrictionTable205[2];
extern const ged_field_restriction_t* RestrictionTable206[2];
extern const ged_field_restriction_t* RestrictionTable207[2];
extern const ged_field_restriction_t* RestrictionTable208[2];
extern const ged_field_restriction_t* RestrictionTable209[2];
extern const ged_field_restriction_t* RestrictionTable210[2];
extern const ged_field_restriction_t* RestrictionTable211[2];
extern const ged_field_restriction_t* RestrictionTable212[2];
#endif // GED_RESTRICTIONS_H
