/*========================== begin_copyright_notice ============================

Copyright (C) 2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

syntax = "proto3";

package IGC_METRICS;

message CodeRef {
  string pathToFile = 1;
  int32 line = 2;
}