/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

#undef CIF_FORWARD_DECLARE
#undef CIF_FORWARD_DECLARE_S
#undef CIF_FORWARD_DECLARE_C

#undef CIF_ULT_EXPOSE
#undef CIF_DECLARE_INTERFACE
#undef CIF_DEFINE_INTERFACE_VER
#undef CIF_DEFINE_INTERFACE_VER_WITH_COMPATIBILITY
#undef CIF_INHERIT_CONSTRUCTOR
#undef CIF_DECLARE_INTERFACE_DEPENDENCIES
#undef CIF_GENERATE_VERSIONS_LIST
#undef CIF_GENERATE_VERSIONS_LIST_AND_DECLARE_INTERFACE_DEPENDENCIES
#undef CIF_MARK_LATEST_VERSION
#undef CIF_DECLARE_INTERFACE_PIMPL
#undef CIF_PIMPL_DECLARE_CONSTRUCTOR
#undef CIF_GET_INTERFACE_CLASS
#undef CIF_GET_PIMPL
#undef CIF_EXPORT_ENTRY_POINTS_STATIC
#undef CIF_DEFINE_INTERFACE_TO_PIMPL_FORWARDING_CTOR_DTOR
#undef CIF_DEFINE_INTERFACE_VIEW_TO_PIMPL_CTOR_DTOR
#undef CIF_PIMPL_T
#undef CREATE_CIF_PIMPL
#undef CIF_CALLING_CONV
