/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

#include "../headers/resource.h"

/////////////////////////////////////////////////////////////////////////////
//
// DLL versioning section
//
#ifdef WINNT
#include <windows.h>
#endif

#include "../../../../../install/intcver.h"

#define VER_FILETYPE                 VFT_DLL
#define VER_FILESUBTYPE              VFT2_DRV_INSTALLABLE
#ifdef WINNT
#define VER_FILEOS                   VOS_NT_WINDOWS32
#else
#define VER_FILEOS                   VOS_DOS_WINDOWS32
#endif
#define VER_FILEDESCRIPTION         "OpenCL User Mode Driver for Intel(R) Graphics Technology\0"

#ifdef _AMD64_
#define VER_INTERNALNAME            "igdfcl64.dll"
#define VER_ORIGINALNAME            "igdfcl64.dll"
#if _DEBUG
#define VER_DUMP                    "../../../../../dump64/igc/bif/Debug/opencl_cth.h"
#elif _INTERNAL
#define VER_DUMP                    "../../../../../dump64/igc/bif/ReleaseInternal/opencl_cth.h"
#else
#define VER_DUMP                    "../../../../../dump64/igc/bif/Release/opencl_cth.h"
#endif
#else
#define VER_INTERNALNAME            "igdfcl32.dll"
#define VER_ORIGINALNAME            "igdfcl32.dll"
#if _DEBUG
#define VER_DUMP                    "../../../../../dump32/igc/bif/Debug/opencl_cth.h"
#elif _INTERNAL
#define VER_DUMP                    "../../../../../dump32/igc/bif/ReleaseInternal/opencl_cth.h"
#else
#define VER_DUMP                    "../../../../../dump32/igc/bif/Release/opencl_cth.h"
#endif
#endif


IDR_CTH_H               H            VER_DUMP


#include <verinfo.ver>

/////////////////////////////////////////////////////////////////////////////
