/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

/*
 * !!! DO NOT EDIT THIS FILE !!!
 *
 * This file was automagically crafted by GED's model parser.
 */


#ifndef GED_MAPPING_TABLES_H
#define GED_MAPPING_TABLES_H

#include "common/ged_compact_mapping_table.h"
#include "ged_ins_field.h"
extern const ged_compact_mapping_entry_t MappingTable0[4];
extern const ged_compact_mapping_entry_t MappingTable1[4];
extern const ged_compact_mapping_entry_t MappingTable2[4];
extern const ged_compact_mapping_entry_t MappingTable3[4];
extern const ged_compact_mapping_entry_t MappingTable4[4];
extern const ged_compact_mapping_entry_t MappingTable5[4];
extern const ged_compact_mapping_entry_t MappingTable6[4];
extern const ged_compact_mapping_entry_t MappingTable7[4];
extern const ged_compact_mapping_entry_t MappingTable8[16];
extern const ged_compact_mapping_entry_t MappingTable9[16];
extern const ged_compact_mapping_entry_t MappingTable10[16];
extern const ged_compact_mapping_entry_t MappingTable11[16];
extern const ged_compact_mapping_entry_t MappingTable12[16];
extern const ged_compact_mapping_entry_t MappingTable13[16];
extern const ged_compact_mapping_entry_t MappingTable14[2];
extern const ged_compact_mapping_entry_t MappingTable15[2];
extern const ged_compact_mapping_entry_t MappingTable16[2];
extern const ged_compact_mapping_entry_t MappingTable17[2];
extern const ged_compact_mapping_entry_t MappingTable18[16];
extern const ged_compact_mapping_entry_t MappingTable19[2];
extern const ged_compact_mapping_entry_t MappingTable20[2];
extern const ged_compact_mapping_entry_t MappingTable21[2];
extern const ged_compact_mapping_entry_t MappingTable22[2];
extern const ged_compact_mapping_entry_t MappingTable23[2];
extern const ged_compact_mapping_entry_t MappingTable24[2];
extern const ged_compact_mapping_entry_t MappingTable25[128];
extern const ged_compact_mapping_entry_t MappingTable26[128];
extern const ged_compact_mapping_entry_t MappingTable27[2];
extern const ged_compact_mapping_entry_t MappingTable28[2];
extern const ged_compact_mapping_entry_t MappingTable29[128];
extern const ged_compact_mapping_entry_t MappingTable30[128];
extern const ged_compact_mapping_entry_t MappingTable31[129];
extern const ged_compact_mapping_entry_t MappingTable32[129];
extern const ged_compact_mapping_entry_t MappingTable33[129];
extern const ged_compact_mapping_entry_t MappingTable34[129];
extern const ged_compact_mapping_entry_t MappingTable35[129];
extern const ged_compact_mapping_entry_t MappingTable36[129];
extern const ged_compact_mapping_entry_t MappingTable37[129];
extern const ged_compact_mapping_entry_t MappingTable38[129];
extern const ged_compact_mapping_entry_t MappingTable39[129];
extern const ged_compact_mapping_entry_t MappingTable40[129];
extern const ged_compact_mapping_entry_t MappingTable41[129];
extern const ged_compact_mapping_entry_t MappingTable42[129];
extern const ged_compact_mapping_entry_t MappingTable43[129];
extern const ged_compact_mapping_entry_t MappingTable44[129];
extern const ged_compact_mapping_entry_t MappingTable45[129];
extern const ged_compact_mapping_entry_t MappingTable46[129];
extern const ged_compact_mapping_entry_t MappingTable47[129];
extern const ged_compact_mapping_entry_t MappingTable48[129];
extern const ged_compact_mapping_entry_t MappingTable49[129];
extern const ged_compact_mapping_entry_t MappingTable50[129];
extern const ged_compact_mapping_entry_t MappingTable51[129];
extern const ged_compact_mapping_entry_t MappingTable52[129];
extern const ged_compact_mapping_entry_t MappingTable53[129];
extern const ged_compact_mapping_entry_t MappingTable54[129];
extern const ged_compact_mapping_entry_t MappingTable55[129];
extern const ged_compact_mapping_entry_t MappingTable56[129];
extern const ged_compact_mapping_entry_t MappingTable57[129];
extern const ged_compact_mapping_entry_t MappingTable58[129];
#endif // GED_MAPPING_TABLES_H
