/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

/*
 * !!! DO NOT EDIT THIS FILE !!!
 *
 * This file was automagically crafted by GED's model parser.
 */


#ifndef GED_ENUM_TABLES_H
#define GED_ENUM_TABLES_H

#include "common/ged_types_internal.h"
#include "ged_enumerations.h"
extern const ged_unsigned_enum_value_t* unsignedTable0[8];
extern const ged_unsigned_enum_value_t* unsignedTable1[8];
extern const ged_unsigned_enum_value_t* unsignedTable2[4];
extern const ged_unsigned_enum_value_t* unsignedTable3[4];
extern const ged_unsigned_enum_value_t* unsignedTable4[2];
extern const ged_unsigned_enum_value_t* unsignedTable5[8];
extern const ged_unsigned_enum_value_t* unsignedTable6[8];
extern const ged_unsigned_enum_value_t* unsignedTable7[16];
extern const ged_unsigned_enum_value_t* unsignedTable8[4];
extern const ged_unsigned_enum_value_t* unsignedTable9[8];
extern const ged_unsigned_enum_value_t* unsignedTable10[4];
extern const ged_unsigned_enum_value_t* unsignedTable11[4];
extern const ged_unsigned_enum_value_t* unsignedTable12[16];
extern const ged_unsigned_enum_value_t* unsignedTable13[4];
extern const ged_unsigned_enum_value_t* unsignedTable14[4];
extern const ged_unsigned_enum_value_t* unsignedTable15[8];
extern const ged_unsigned_enum_value_t* unsignedTable16[8];
extern const ged_unsigned_enum_value_t* unsignedTable17[4];
extern const ged_unsigned_enum_value_t* unsignedTable18[16];
extern const ged_unsigned_enum_value_t* unsignedTable19[8];
extern const ged_unsigned_enum_value_t* unsignedTable20[16];
extern const ged_unsigned_enum_value_t* unsignedTable21[16];
extern const ged_unsigned_enum_value_t* unsignedTable22[8];
extern const ged_unsigned_enum_value_t* unsignedTable23[19];
extern const ged_unsigned_enum_value_t* unsignedTable24[19];
extern const ged_unsigned_enum_value_t* unsignedTable25[19];
extern const ged_unsigned_enum_value_t* unsignedTable26[19];
extern const ged_unsigned_enum_value_t* unsignedTable27[19];
extern const ged_unsigned_enum_value_t* unsignedTable28[19];
extern const ged_unsigned_enum_value_t* unsignedTable29[19];
extern const ged_unsigned_enum_value_t* unsignedTable30[19];
extern const ged_unsigned_enum_value_t* unsignedTable31[19];
extern const ged_unsigned_enum_value_t* unsignedTable32[19];
extern const ged_unsigned_enum_value_t* unsignedTable33[14];
extern const GED_ACCESS_MODE* AccessModeTable0[2];
extern const GED_ACCESS_MODE* AccessModeTable1[2];
extern const GED_ACC_WR_CTRL* AccWrCtrlTable0[2];
extern const GED_ADDR_MODE* AddrModeTable0[2];
extern const GED_ADDR_MODE* AddrModeTable1[2];
extern const GED_ARCH_REG* ArchRegTable0[16];
extern const GED_ARCH_REG* ArchRegTable1[16];
extern const GED_ARCH_REG* ArchRegTable2[16];
extern const GED_ARCH_REG* ArchRegTable3[16];
extern const GED_ATOMIC_OPERATION_TYPE* AtomicOperationTypeTable0[16];
extern const GED_ATOMIC_OPERATION_TYPE* AtomicOperationTypeTable1[16];
extern const GED_BLOCK_SIZE* BlockSizeTable0[8];
extern const GED_BRANCH_CTRL* BranchCtrlTable0[2];
extern const GED_CHANNEL_MASK* ChannelMaskTable0[2];
extern const GED_CHANNEL_MASK* ChannelMaskTable1[2];
extern const GED_CHANNEL_MASK* ChannelMaskTable2[2];
extern const GED_CHANNEL_MASK* ChannelMaskTable3[2];
extern const GED_CHANNEL_MODE* ChannelModeTable0[2];
extern const GED_CHANNEL_OFFSET* ChannelOffsetTable0[8];
extern const GED_CHANNEL_OFFSET* ChannelOffsetTable1[4];
extern const GED_COND_MODIFIER* CondModifierTable0[16];
extern const GED_COND_MODIFIER* CondModifierTable1[4];
extern const GED_DATA_TYPE* DataTypeTable0[16];
extern const GED_DATA_TYPE* DataTypeTable1[16];
extern const GED_DATA_TYPE* DataTypeTable2[16];
extern const GED_DATA_TYPE* DataTypeTable3[16];
extern const GED_DATA_TYPE* DataTypeTable4[16];
extern const GED_DATA_TYPE* DataTypeTable5[8];
extern const GED_DATA_TYPE* DataTypeTable6[2];
extern const GED_DATA_TYPE* DataTypeTable7[8];
extern const GED_DATA_TYPE* DataTypeTable8[8];
extern const GED_DATA_TYPE* DataTypeTable9[16];
extern const GED_DATA_TYPE* DataTypeTable10[16];
extern const GED_DATA_TYPE* DataTypeTable11[16];
extern const GED_DATA_TYPE* DataTypeTable12[8];
extern const GED_DATA_TYPE* DataTypeTable13[16];
extern const GED_DATA_TYPE* DataTypeTable14[16];
extern const GED_DATA_TYPE* DataTypeTable15[16];
extern const GED_DATA_TYPE* DataTypeTable16[16];
extern const GED_DATA_TYPE* DataTypeTable17[8];
extern const GED_DATA_TYPE* DataTypeTable18[8];
extern const GED_DATA_TYPE* DataTypeTable19[16];
extern const GED_DATA_TYPE* DataTypeTable20[8];
extern const GED_DATA_TYPE* DataTypeTable21[8];
extern const GED_DATA_TYPE* DataTypeTable22[8];
extern const GED_DATA_TYPE* DataTypeTable23[8];
extern const GED_DATA_TYPE* DataTypeTable24[16];
extern const GED_DATA_TYPE* DataTypeTable25[16];
extern const GED_DATA_TYPE* DataTypeTable26[16];
extern const GED_DATA_TYPE* DataTypeTable27[16];
extern const GED_DATA_TYPE* DataTypeTable28[8];
extern const GED_DATA_TYPE* DataTypeTable29[8];
extern const GED_DATA_TYPE* DataTypeTable30[16];
extern const GED_DATA_TYPE* DataTypeTable31[16];
extern const GED_DATA_TYPE* DataTypeTable32[16];
extern const GED_DATA_TYPE* DataTypeTable33[8];
extern const GED_DATA_TYPE* DataTypeTable34[8];
extern const GED_DATA_TYPE* DataTypeTable35[8];
extern const GED_DATA_TYPE* DataTypeTable36[8];
extern const GED_DATA_TYPE* DataTypeTable37[8];
extern const GED_DATA_TYPE* DataTypeTable38[4];
extern const GED_DEBUG_CTRL* DebugCtrlTable0[2];
extern const GED_DEP_CTRL* DepCtrlTable0[4];
extern const GED_DP_ADDR_SIZE* DpAddrSizeTable0[4];
extern const GED_DP_ADDR_SURFACE_TYPE* DpAddrSurfaceTypeTable0[4];
extern const GED_DP_DATA_SIZE* DpDataSizeTable0[8];
extern const GED_DP_FENCE_SCOPE* DpFenceScopeTable0[8];
extern const GED_DP_FENCE_SCOPE* DpFenceScopeTable1[8];
extern const GED_DP_FLUSH_TYPE* DpFlushTypeTable0[8];
extern const GED_DP_FLUSH_TYPE* DpFlushTypeTable1[8];
extern const GED_DP_OPCODE* DpOpcodeTable0[64];
extern const GED_DP_OPCODE* DpOpcodeTable1[64];
extern const GED_DP_TRANSPOSE* DpTransposeTable0[2];
extern const GED_DP_VECT_SIZE* DpVectSizeTable0[8];
extern const GED_DST_CHAN_EN* DstChanEnTable0[16];
extern const GED_EOT* EOTTable0[2];
extern const GED_EXEC_MASK_OFFSET_CTRL* ExecMaskOffsetCtrlTable0[8];
extern const GED_EXEC_MASK_OFFSET_CTRL* ExecMaskOffsetCtrlTable1[8];
extern const GED_EXEC_MASK_OFFSET_CTRL* ExecMaskOffsetCtrlTable2[8];
extern const GED_EXECUTION_DATA_TYPE* ExecutionDataTypeTable0[2];
extern const GED_FUSION_CTRL* FusionCtrlTable0[2];
extern const GED_HEADER_PRESENT* HeaderPresentTable0[2];
extern const GED_MASK_CTRL* MaskCtrlTable0[2];
extern const GED_MASK_CTRL* MaskCtrlTable1[2];
extern const GED_MATH_FC* MathFCTable0[16];
extern const GED_MATH_FC* MathFCTable1[16];
extern const GED_MATH_FC* MathFCTable2[16];
extern const GED_MATH_FC* MathFCTable3[16];
extern const GED_MATH_MACRO_EXT* MathMacroExtTable0[16];
extern const GED_MESSAGE_TYPE* MessageTypeTable0[2];
extern const GED_MESSAGE_TYPE* MessageTypeTable1[16];
extern const GED_MESSAGE_TYPE* MessageTypeTable2[32];
extern const GED_MESSAGE_TYPE* MessageTypeTable3[32];
extern const GED_MESSAGE_TYPE* MessageTypeTable4[32];
extern const GED_MESSAGE_TYPE* MessageTypeTable5[32];
extern const GED_MESSAGE_TYPE* MessageTypeTable6[16];
extern const GED_MESSAGE_TYPE* MessageTypeTable7[32];
extern const GED_MESSAGE_TYPE* MessageTypeTable8[32];
extern const GED_MESSAGE_TYPE* MessageTypeTable9[32];
extern const GED_MESSAGE_TYPE* MessageTypeTable10[32];
extern const GED_MESSAGE_TYPE* MessageTypeTable11[16];
extern const GED_MESSAGE_TYPE* MessageTypeTable12[32];
extern const GED_MESSAGE_TYPE* MessageTypeTable13[16];
extern const GED_MESSAGE_TYPE* MessageTypeTable14[16];
extern const GED_MESSAGE_TYPE* MessageTypeTable15[16];
extern const GED_MESSAGE_TYPE* MessageTypeTable16[16];
extern const GED_MESSAGE_TYPE* MessageTypeTable17[16];
extern const GED_MESSAGE_TYPE* MessageTypeTable18[16];
extern const GED_MESSAGE_TYPE* MessageTypeTable19[16];
extern const GED_MESSAGE_TYPE* MessageTypeTable20[16];
extern const GED_MESSAGE_TYPE* MessageTypeTable21[32];
extern const GED_MESSAGE_TYPE* MessageTypeTable22[32];
extern const GED_NO_SRC_DEP_SET* NoSrcDepSetTable0[2];
extern const GED_OPCODE* OpcodeTable0[128];
extern const GED_OPCODE* OpcodeTable1[128];
extern const GED_OPCODE* OpcodeTable2[128];
extern const GED_OPCODE* OpcodeTable3[128];
extern const GED_OPCODE* OpcodeTable4[128];
extern const GED_OPCODE* OpcodeTable5[128];
extern const GED_OPCODE* OpcodeTable6[128];
extern const GED_OPCODE* OpcodeTable7[128];
extern const GED_OPCODE* OpcodeTable8[128];
extern const GED_OPCODE* OpcodeTable9[128];
extern const GED_OPCODE* OpcodeTable10[128];
extern const GED_PRECISION* PrecisionTable0[32];
extern const GED_PRECISION* PrecisionTable1[32];
extern const GED_PRECISION* PrecisionTable2[32];
extern const GED_PRED_CTRL* PredCtrlTable0[16];
extern const GED_PRED_CTRL* PredCtrlTable1[16];
extern const GED_PRED_CTRL* PredCtrlTable2[4];
extern const GED_PRED_INV* PredInvTable0[2];
extern const GED_REG_FILE* RegFileTable0[4];
extern const GED_REG_FILE* RegFileTable1[4];
extern const GED_REG_FILE* RegFileTable2[4];
extern const GED_REG_FILE* RegFileTable3[4];
extern const GED_REG_FILE* RegFileTable4[4];
extern const GED_REG_FILE* RegFileTable5[4];
extern const GED_REG_FILE* RegFileTable6[2];
extern const GED_REG_FILE* RegFileTable7[2];
extern const GED_REG_FILE* RegFileTable8[2];
extern const GED_REG_FILE* RegFileTable9[2];
extern const GED_REG_FILE* RegFileTable10[2];
extern const GED_REG_FILE* RegFileTable11[2];
extern const GED_REG_FILE* RegFileTable12[2];
extern const GED_REG_FILE* RegFileTable13[4];
extern const GED_REG_FILE* RegFileTable14[4];
extern const GED_REP_CTRL* RepCtrlTable0[2];
extern const GED_RETURN_DATA_CONTROL* ReturnDataControlTable0[2];
extern const GED_SATURATE* SaturateTable0[2];
extern const GED_SFID* SFIDTable0[16];
extern const GED_SFID* SFIDTable1[16];
extern const GED_SFID* SFIDTable2[16];
extern const GED_SFID* SFIDTable3[16];
extern const GED_SFID* SFIDTable4[16];
extern const GED_SIMDMODE* SIMDModeTable0[4];
extern const GED_SIMDMODE* SIMDModeTable1[2];
extern const GED_SLOT_GROUP* SlotGroupTable0[4];
extern const GED_SLOT_GROUP* SlotGroupTable1[2];
extern const GED_SRC_MOD* SrcModTable0[4];
extern const GED_SRC_MOD* SrcModTable1[4];
extern const GED_SUB_BYTE_PRECISION* SubBytePrecisionTable0[4];
extern const GED_SUB_BYTE_PRECISION* SubBytePrecisionTable1[4];
extern const GED_SUB_BYTE_PRECISION* SubBytePrecisionTable2[4];
extern const GED_SUB_BYTE_PRECISION* SubBytePrecisionTable3[4];
extern const GED_SUB_BYTE_PRECISION* SubBytePrecisionTable4[4];
extern const GED_SUB_FUNC_ID* SubFuncIDTable0[8];
extern const GED_SUB_FUNC_ID* SubFuncIDTable1[8];
extern const GED_SWIZZLE* SwizzleTable0[4];
extern const GED_SYNC_FC* SyncFCTable0[16];
extern const GED_SYNC_FC* SyncFCTable1[16];
extern const GED_THREAD_CTRL* ThreadCtrlTable0[4];
extern const GED_THREAD_CTRL* ThreadCtrlTable1[4];
extern const GED_THREAD_CTRL* ThreadCtrlTable2[2];
extern const GED_THREAD_CTRL* ThreadCtrlTable3[4];
#endif // GED_ENUM_TABLES_H
