/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

/*
 * !!! DO NOT EDIT THIS FILE !!!
 *
 * This file was automagically crafted by GED's model parser.
 */


#ifndef GED_ENUMERATIONS_INTERNAL_H
#define GED_ENUMERATIONS_INTERNAL_H

extern const char* AccessModeEnumeration[3];
extern const char* AccWrCtrlEnumeration[3];
extern const char* AddrModeEnumeration[3];
extern const char* ArchRegEnumeration[16];
extern const char* AtomicOperationTypeEnumeration[16];
extern const char* BlockSizeEnumeration[6];
extern const char* BranchCtrlEnumeration[3];
extern const char* ChannelMaskEnumeration[6];
extern const char* ChannelModeEnumeration[3];
extern const char* ChannelOffsetEnumeration[9];
extern const char* CondModifierEnumeration[10];
extern const char* DataTypeEnumeration[20];
extern const char* DebugCtrlEnumeration[3];
extern const char* DepCtrlEnumeration[5];
extern const char* DpAddrSizeEnumeration[4];
extern const char* DpAddrSurfaceTypeEnumeration[5];
extern const char* DpDataSizeEnumeration[8];
extern const char* DpFenceScopeEnumeration[8];
extern const char* DpFlushTypeEnumeration[7];
extern const char* DpOpcodeEnumeration[28];
extern const char* DpTransposeEnumeration[3];
extern const char* DpVectSizeEnumeration[9];
extern const char* DstChanEnEnumeration[17];
extern const char* EOTEnumeration[3];
extern const char* ExecMaskOffsetCtrlEnumeration[15];
extern const char* ExecutionDataTypeEnumeration[3];
extern const char* FusionCtrlEnumeration[3];
extern const char* HeaderPresentEnumeration[3];
extern const char* MaskCtrlEnumeration[3];
extern const char* MathFCEnumeration[15];
extern const char* MathMacroExtEnumeration[10];
extern const char* MessageTypeEnumeration[78];
extern const char* NoSrcDepSetEnumeration[3];
extern const char* OpcodeEnumeration[83];
extern const char* PrecisionEnumeration[14];
extern const char* PredCtrlEnumeration[21];
extern const char* PredInvEnumeration[3];
extern const char* RegFileEnumeration[4];
extern const char* RepCtrlEnumeration[3];
extern const char* ReturnDataControlEnumeration[3];
extern const char* SaturateEnumeration[3];
extern const char* SFIDEnumeration[22];
extern const char* SIMDModeEnumeration[4];
extern const char* SlotGroupEnumeration[4];
extern const char* SrcModEnumeration[5];
extern const char* SubBytePrecisionEnumeration[8];
extern const char* SubFuncIDEnumeration[8];
extern const char* SwizzleEnumeration[5];
extern const char* SyncFCEnumeration[8];
extern const char* ThreadCtrlEnumeration[5];
#endif // GED_ENUMERATIONS_INTERNAL_H
