/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

/*
 * !!! DO NOT EDIT THIS FILE !!!
 *
 * This file was automagically crafted by GED's model parser.
 */

#include "ged_fragments.h"
#include "ged_restrictions.h"
#include "ged_decoding_tables.h"
const ged_ins_field_entry_t DecodingTable0[8] =
{
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable19 }, // 0
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable19 }, // 1
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable18 }, // 2
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable19 }, // 3
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable20 }, // 4
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable19 }, // 5
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable19 }, // 6
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable19 } // 7
}; // DecodingTable0[]
const ged_ins_field_entry_t DecodingTable1[2] =
{
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  16,  19,   0,  16,    0xf0000 } */ { 0x10001310, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable21 }, // 0
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  16,  19,   0,  16,    0xf0000 } */ { 0x10001310, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable22 } // 1
}; // DecodingTable1[]
const ged_ins_field_entry_t DecodingTable2[4] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  44,  46,   1,  12,     0x7000 } */ {  0xc012e2c, (void*)(g_uintptr_t)    0x7000 }, RestrictionTable13 }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  44,  46,   1,  12,     0x7000 } */ {  0xc012e2c, (void*)(g_uintptr_t)    0x7000 }, RestrictionTable13 }, // 1
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  44,  46,   1,  12,     0x7000 } */ {  0xc012e2c, (void*)(g_uintptr_t)    0x7000 }, RestrictionTable13 }, // 2
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  44,  46,   1,  12,     0x7000 } */ {  0xc012e2c, (void*)(g_uintptr_t)    0x7000 }, RestrictionTable23 } // 3
}; // DecodingTable2[]
const ged_ins_field_entry_t DecodingTable3[2] =
{
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  48,  51,   1,  16,    0xf0000 } */ { 0x10013330, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable24 } // 1
}; // DecodingTable3[]
const ged_ins_field_entry_t DecodingTable4[2] =
{
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  48,  52,   1,  16,   0x1f0000 } */ { 0x10013430, (void*)(g_uintptr_t)  0x1f0000 }, NULL }, // 0
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  52,  52,   1,  16,   0x100000 } */ { 0x10013434, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable25 } // 1
}; // DecodingTable4[]
const ged_ins_field_entry_t DecodingTable5[2] =
{
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_AccessMode,                        DecodingTable4 }, NULL }, // 0
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable5[]
const ged_ins_field_entry_t DecodingTable6[2] =
{
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {  48,  57,   1,  16,  0x3ff0000 } */ { 0x10013930, (void*)(g_uintptr_t) 0x3ff0000 }, NULL }, // 0
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {  52,  57,   1,  16,  0x3f00000 } */ { 0x10013934, (void*)(g_uintptr_t) 0x3f00000 }, RestrictionTable25 } // 1
}; // DecodingTable6[]
const ged_ins_field_entry_t DecodingTable7[2] =
{
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_AccessMode,                        DecodingTable6 }, NULL } // 1
}; // DecodingTable7[]
const ged_ins_field_entry_t DecodingTable8[4] =
{
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  53,  60,   1,  21, 0x1fe00000 } */ { 0x15013c35, (void*)(g_uintptr_t)0x1fe00000 }, NULL }, // 0
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  53,  60,   1,  21, 0x1fe00000 } */ { 0x15013c35, (void*)(g_uintptr_t)0x1fe00000 }, RestrictionTable26 }, // 1
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable8[]
const ged_ins_field_entry_t DecodingTable9[2] =
{
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstRegFile,                        DecodingTable8 }, NULL }, // 0
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable9[]
const ged_ins_field_entry_t DecodingTable10[2] =
{
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  58,  60,   1,  26, 0x1c000000 } */ { 0x1a013c3a, (void*)(g_uintptr_t)0x1c000000 }, NULL } // 1
}; // DecodingTable10[]
const ged_ins_field_entry_t DecodingTable11[2] =
{
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  61,  62,   1,  29, 0x60000000 } */ { 0x1d013e3d, (void*)(g_uintptr_t)0x60000000 }, RestrictionTable27 }, // 0
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable11[]
const ged_ins_field_entry_t DecodingTable12[2] =
{
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      8, { 2, FragmentsTable2 }, NULL } // 1
}; // DecodingTable12[]
const ged_ins_field_entry_t DecodingTable13[2] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  64,  68,   2,   0,       0x1f } */ {    0x24440, (void*)(g_uintptr_t)      0x1f }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  68,  68,   2,   0,       0x10 } */ {    0x24444, (void*)(g_uintptr_t)      0x10 }, RestrictionTable25 } // 1
}; // DecodingTable13[]
const ged_ins_field_entry_t DecodingTable14[2] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_AccessMode,                        DecodingTable13 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable14[]
const ged_ins_field_entry_t DecodingTable15[2] =
{
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {  64,  73,   2,   0,      0x3ff } */ {    0x24940, (void*)(g_uintptr_t)     0x3ff }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {  68,  73,   2,   0,      0x3f0 } */ {    0x24944, (void*)(g_uintptr_t)     0x3f0 }, RestrictionTable25 } // 1
}; // DecodingTable15[]
const ged_ins_field_entry_t DecodingTable16[2] =
{
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_AccessMode,                        DecodingTable15 }, NULL } // 1
}; // DecodingTable16[]
const ged_ins_field_entry_t DecodingTable17[4] =
{
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  69,  76,   2,   5,     0x1fe0 } */ {  0x5024c45, (void*)(g_uintptr_t)    0x1fe0 }, NULL }, // 0
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  69,  76,   2,   5,     0x1fe0 } */ {  0x5024c45, (void*)(g_uintptr_t)    0x1fe0 }, RestrictionTable26 }, // 1
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable17[]
const ged_ins_field_entry_t DecodingTable18[2] =
{
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable17 }, NULL }, // 0
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable18[]
const ged_ins_field_entry_t DecodingTable19[2] =
{
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  74,  76,   2,  10,     0x1c00 } */ {  0xa024c4a, (void*)(g_uintptr_t)    0x1c00 }, NULL } // 1
}; // DecodingTable19[]
const ged_ins_field_entry_t DecodingTable20[2] =
{
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  80,  81,   2,  16,    0x30000 } */ { 0x10025150, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 0
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable20[]
const ged_ins_field_entry_t DecodingTable21[2] =
{
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  82,  84,   2,  18,   0x1c0000 } */ { 0x12025452, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable28 }, // 0
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable21[]
const ged_ins_field_entry_t DecodingTable22[2] =
{
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      8, { 2, FragmentsTable3 }, NULL } // 1
}; // DecodingTable22[]
const ged_ins_field_entry_t DecodingTable23[4] =
{
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable22 }, NULL }, // 1
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable23[]
const ged_ins_field_entry_t DecodingTable24[4] =
{
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 109, 110,   3,  13,     0x6000 } */ {  0xd036e6d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 1
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable24[]
const ged_ins_field_entry_t DecodingTable25[4] =
{
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 111, 111,   3,  15,     0x8000 } */ {  0xf036f6f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 1
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable25[]
const ged_ins_field_entry_t DecodingTable26[2] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  96, 100,   3,   0,       0x1f } */ {    0x36460, (void*)(g_uintptr_t)      0x1f }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* { 100, 100,   3,   0,       0x10 } */ {    0x36464, (void*)(g_uintptr_t)      0x10 }, RestrictionTable25 } // 1
}; // DecodingTable26[]
const ged_ins_field_entry_t DecodingTable27[2] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_AccessMode,                        DecodingTable26 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable27[]
const ged_ins_field_entry_t DecodingTable28[4] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable27 }, NULL }, // 1
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable28[]
const ged_ins_field_entry_t DecodingTable29[2] =
{
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {  96, 105,   3,   0,      0x3ff } */ {    0x36960, (void*)(g_uintptr_t)     0x3ff }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* { 100, 105,   3,   0,      0x3f0 } */ {    0x36964, (void*)(g_uintptr_t)     0x3f0 }, RestrictionTable25 } // 1
}; // DecodingTable29[]
const ged_ins_field_entry_t DecodingTable30[2] =
{
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_AccessMode,                        DecodingTable29 }, NULL } // 1
}; // DecodingTable30[]
const ged_ins_field_entry_t DecodingTable31[4] =
{
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable30 }, NULL }, // 1
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable31[]
const ged_ins_field_entry_t DecodingTable32[2] =
{
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 101, 108,   3,   5,     0x1fe0 } */ {  0x5036c65, (void*)(g_uintptr_t)    0x1fe0 }, RestrictionTable26 }, // 0
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable32[]
const ged_ins_field_entry_t DecodingTable33[4] =
{
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable32 }, NULL }, // 1
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable33[]
const ged_ins_field_entry_t DecodingTable34[2] =
{
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* { 106, 108,   3,  10,     0x1c00 } */ {  0xa036c6a, (void*)(g_uintptr_t)    0x1c00 }, NULL } // 1
}; // DecodingTable34[]
const ged_ins_field_entry_t DecodingTable35[4] =
{
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable34 }, NULL }, // 1
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable35[]
const ged_ins_field_entry_t DecodingTable36[2] =
{
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 0
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable36[]
const ged_ins_field_entry_t DecodingTable37[4] =
{
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable36 }, NULL }, // 1
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable37[]
const ged_ins_field_entry_t DecodingTable38[2] =
{
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* { 114, 116,   3,  18,   0x1c0000 } */ { 0x12037472, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable28 }, // 0
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable38[]
const ged_ins_field_entry_t DecodingTable39[4] =
{
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable38 }, NULL }, // 1
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable39[]
const ged_ins_field_entry_t DecodingTable40[4] =
{
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* { 117, 120,   3,  21,  0x1e00000 } */ { 0x15037875, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 1
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable40[]
const ged_ins_field_entry_t DecodingTable41[8] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable30 }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable31 }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable32 }, // 3
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 4
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 5
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 6
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 } // 7
}; // DecodingTable41[]
const ged_ins_field_entry_t DecodingTable42[4] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1DataType,                      DecodingTable41 }, NULL } // 3
}; // DecodingTable42[]
const ged_ins_field_entry_t DecodingTable43[4] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  39,  41,   1,   7,      0x380 } */ {  0x7012927, (void*)(g_uintptr_t)     0x380 }, RestrictionTable13 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  39,  41,   1,   7,      0x380 } */ {  0x7012927, (void*)(g_uintptr_t)     0x380 }, RestrictionTable13 }, // 1
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  39,  41,   1,   7,      0x380 } */ {  0x7012927, (void*)(g_uintptr_t)     0x380 }, RestrictionTable13 }, // 2
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  39,  41,   1,   7,      0x380 } */ {  0x7012927, (void*)(g_uintptr_t)     0x380 }, RestrictionTable23 } // 3
}; // DecodingTable43[]
const ged_ins_field_entry_t DecodingTable44[4] =
{
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 0
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 1
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 2
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable44[]
const ged_ins_field_entry_t DecodingTable45[4] =
{
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 0
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 1
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 2
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable45[]
const ged_ins_field_entry_t DecodingTable46[4] =
{
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 0
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 1
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 2
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable46[]
const ged_ins_field_entry_t DecodingTable47[4] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 1
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 2
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable47[]
const ged_ins_field_entry_t DecodingTable48[4] =
{
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable16 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable16 }, NULL }, // 1
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable16 }, NULL }, // 2
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable48[]
const ged_ins_field_entry_t DecodingTable49[4] =
{
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 0
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 1
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 2
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable49[]
const ged_ins_field_entry_t DecodingTable50[4] =
{
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable19 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable19 }, NULL }, // 1
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable19 }, NULL }, // 2
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable50[]
const ged_ins_field_entry_t DecodingTable51[4] =
{
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 0
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 1
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 2
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable51[]
const ged_ins_field_entry_t DecodingTable52[4] =
{
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 0
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 1
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 2
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable52[]
const ged_ins_field_entry_t DecodingTable53[4] =
{
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 0
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 1
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 2
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable53[]
const ged_ins_field_entry_t DecodingTable54[4] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0DataType,                      DecodingTable41 }, NULL } // 3
}; // DecodingTable54[]
const ged_ins_field_entry_t DecodingTable55[2] =
{
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  49,  52,   1,  17,   0x1e0000 } */ { 0x11013431, (void*)(g_uintptr_t)  0x1e0000 }, RestrictionTable24 } // 1
}; // DecodingTable55[]
const ged_ins_field_entry_t DecodingTable56[2] =
{
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  65,  72,   2,   1,      0x1fe } */ {  0x1024841, (void*)(g_uintptr_t)     0x1fe }, NULL }, // 0
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable56[]
const ged_ins_field_entry_t DecodingTable57[2] =
{
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RepCtrl,                       DecodingTable56 }, NULL } // 1
}; // DecodingTable57[]
const ged_ins_field_entry_t DecodingTable58[2] =
{
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  86,  93,   2,  22, 0x3fc00000 } */ { 0x16025d56, (void*)(g_uintptr_t)0x3fc00000 }, NULL }, // 0
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable58[]
const ged_ins_field_entry_t DecodingTable59[2] =
{
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RepCtrl,                       DecodingTable58 }, NULL } // 1
}; // DecodingTable59[]
const ged_ins_field_entry_t DecodingTable60[2] =
{
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 107, 114,   3,  11,    0x7f800 } */ {  0xb03726b, (void*)(g_uintptr_t)   0x7f800 }, NULL }, // 0
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable60[]
const ged_ins_field_entry_t DecodingTable61[2] =
{
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src2RepCtrl,                       DecodingTable60 }, NULL } // 1
}; // DecodingTable61[]
const ged_ins_field_entry_t DecodingTable62[4] =
{
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  44,  46,   1,  12,     0x7000 } */ {  0xc012e2c, (void*)(g_uintptr_t)    0x7000 }, RestrictionTable45 }, // 0
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  44,  46,   1,  12,     0x7000 } */ {  0xc012e2c, (void*)(g_uintptr_t)    0x7000 }, RestrictionTable45 }, // 1
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  44,  46,   1,  12,     0x7000 } */ {  0xc012e2c, (void*)(g_uintptr_t)    0x7000 }, RestrictionTable45 }, // 2
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  44,  46,   1,  12,     0x7000 } */ {  0xc012e2c, (void*)(g_uintptr_t)    0x7000 }, RestrictionTable23 } // 3
}; // DecodingTable62[]
const ged_ins_field_entry_t DecodingTable63[2] =
{
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  96, 100,   3,   0,       0x1f } */ {    0x36460, (void*)(g_uintptr_t)      0x1f }, RestrictionTable0 }, // 0
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 100, 100,   3,   4,       0x10 } */ {  0x4036464, (void*)(g_uintptr_t)      0x10 }, RestrictionTable0 } // 1
}; // DecodingTable63[]
const ged_ins_field_entry_t DecodingTable64[4] =
{
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_AccessMode,                        DecodingTable63 }, NULL }, // 0
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable64[]
const ged_ins_field_entry_t DecodingTable65[4] =
{
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 101, 108,   3,   5,     0x1fe0 } */ {  0x5036c65, (void*)(g_uintptr_t)    0x1fe0 }, RestrictionTable46 }, // 0
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable65[]
const ged_ins_field_entry_t DecodingTable66[2] =
{
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 0
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable66[]
const ged_ins_field_entry_t DecodingTable67[4] =
{
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable66 }, NULL }, // 0
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable67[]
const ged_ins_field_entry_t DecodingTable68[2] =
{
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* { 114, 116,   3,  18,   0x1c0000 } */ { 0x12037472, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable28 }, // 0
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable68[]
const ged_ins_field_entry_t DecodingTable69[4] =
{
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable68 }, NULL }, // 0
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable69[]
const ged_ins_field_entry_t DecodingTable70[4] =
{
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* { 117, 120,   3,  21,  0x1e00000 } */ { 0x15037875, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 0
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable70[]
const ged_ins_field_entry_t DecodingTable71[4] =
{
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    29, /* {  96, 124,   3,   0, 0x1fffffff } */ {    0x37c60, (void*)(g_uintptr_t)0x1fffffff }, NULL } // 3
}; // DecodingTable71[]
const ged_ins_field_entry_t DecodingTable72[16] =
{
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 0
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 1
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 2
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 3
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 7
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 8
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 9
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 10
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 11
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 12
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 13
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 14
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL } // 15
}; // DecodingTable72[]
const ged_ins_field_entry_t DecodingTable73[16] =
{
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable48 }, // 1
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable48 }, // 2
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable48 }, // 3
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable48 }, // 4
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable48 }, // 5
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable48 }, // 6
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable48 }, // 7
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable38 }, // 9
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable38 }, // 10
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable38 }, // 11
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable38 }, // 12
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable38 }, // 13
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 15
}; // DecodingTable73[]
const ged_ins_field_entry_t DecodingTable74[4] =
{
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable22 }, NULL }, // 0
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable22 }, NULL }, // 1
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable74[]
const ged_ins_field_entry_t DecodingTable75[4] =
{
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 109, 110,   3,  13,     0x6000 } */ {  0xd036e6d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 0
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 109, 110,   3,  13,     0x6000 } */ {  0xd036e6d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 1
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable75[]
const ged_ins_field_entry_t DecodingTable76[4] =
{
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 111, 111,   3,  15,     0x8000 } */ {  0xf036f6f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable39 }, // 0
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 111, 111,   3,  15,     0x8000 } */ {  0xf036f6f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable39 }, // 1
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable76[]
const ged_ins_field_entry_t DecodingTable77[4] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable27 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable27 }, NULL }, // 1
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable77[]
const ged_ins_field_entry_t DecodingTable78[4] =
{
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable30 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable30 }, NULL }, // 1
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable78[]
const ged_ins_field_entry_t DecodingTable79[4] =
{
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable32 }, NULL }, // 0
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable32 }, NULL }, // 1
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable79[]
const ged_ins_field_entry_t DecodingTable80[4] =
{
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable34 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable34 }, NULL }, // 1
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable80[]
const ged_ins_field_entry_t DecodingTable81[4] =
{
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable36 }, NULL }, // 0
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable36 }, NULL }, // 1
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable81[]
const ged_ins_field_entry_t DecodingTable82[4] =
{
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable38 }, NULL }, // 0
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable38 }, NULL }, // 1
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable82[]
const ged_ins_field_entry_t DecodingTable83[4] =
{
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* { 117, 120,   3,  21,  0x1e00000 } */ { 0x15037875, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 0
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* { 117, 120,   3,  21,  0x1e00000 } */ { 0x15037875, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 1
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable83[]
const ged_ins_field_entry_t DecodingTable84[128] =
{
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 0
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 1
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 2
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 3
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 4
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 7
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 8
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 9
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 10
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 11
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 12
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 13
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 14
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 15
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 16
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 17
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 18
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 19
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 20
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 21
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 22
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 23
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 24
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 25
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 26
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 27
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 28
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 29
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 30
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 31
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 32
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 33
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 34
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 35
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 36
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 37
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 38
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 39
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 40
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 41
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 42
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 43
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 44
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 45
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 46
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 47
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 48
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 49
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 50
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 51
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 52
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 53
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 54
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 55
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 56
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 57
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 58
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 59
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 60
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 61
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 62
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 63
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 64
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 65
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 66
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 67
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 68
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 69
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 70
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 71
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 72
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 73
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 74
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 75
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 76
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 77
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 78
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 79
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 80
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 81
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 82
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 83
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 84
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 85
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 86
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 87
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 88
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 89
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 90
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 91
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 92
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 93
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 94
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 95
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 96
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 97
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 98
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 99
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 100
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 101
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 102
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 103
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 104
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 105
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 106
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 107
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 108
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 109
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 110
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 111
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 112
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 113
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 114
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 115
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 116
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 117
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 118
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 119
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 120
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 121
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 122
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 123
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 124
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 125
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 126
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL } // 127
}; // DecodingTable84[]
const ged_ins_field_entry_t DecodingTable85[4] =
{
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 0
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 1
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 2
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL } // 3
}; // DecodingTable85[]
const ged_ins_field_entry_t DecodingTable86[4] =
{
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  96, 111,   3,   0,     0xffff } */ {    0x36f60, (void*)(g_uintptr_t)    0xffff }, NULL } // 3
}; // DecodingTable86[]
const ged_ins_field_entry_t DecodingTable87[4] =
{
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 112, 127,   3,  16, 0xffff0000 } */ { 0x10037f70, (void*)(g_uintptr_t)0xffff0000 }, NULL } // 3
}; // DecodingTable87[]
const ged_ins_field_entry_t DecodingTable88[4] =
{
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 0
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 1
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 2
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, NULL } // 3
}; // DecodingTable88[]
const ged_ins_field_entry_t DecodingTable89[4] =
{
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, NULL } // 3
}; // DecodingTable89[]
const ged_ins_field_entry_t DecodingTable90[4] =
{
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  96, 100,   3,   0,       0x1f } */ {    0x36460, (void*)(g_uintptr_t)      0x1f }, NULL }, // 0
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable90[]
const ged_ins_field_entry_t DecodingTable91[4] =
{
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  64,  95,   2,   0, 0xffffffff } */ {    0x25f40, (void*)(g_uintptr_t)0xffffffff }, NULL } // 3
}; // DecodingTable91[]
const ged_ins_field_entry_t DecodingTable92[2] =
{
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      8, { 2, FragmentsTable2 }, RestrictionTable0 } // 1
}; // DecodingTable92[]
const ged_ins_field_entry_t DecodingTable93[2] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  64,  68,   2,   0,       0x1f } */ {    0x24440, (void*)(g_uintptr_t)      0x1f }, RestrictionTable0 }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  68,  68,   2,   4,       0x10 } */ {  0x4024444, (void*)(g_uintptr_t)      0x10 }, RestrictionTable0 } // 1
}; // DecodingTable93[]
const ged_ins_field_entry_t DecodingTable94[2] =
{
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  80,  81,   2,  16,    0x30000 } */ { 0x10025150, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable77 }, // 0
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable94[]
const ged_ins_field_entry_t DecodingTable95[2] =
{
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  82,  84,   2,  18,   0x1c0000 } */ { 0x12025452, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable78 }, // 0
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable95[]
const ged_ins_field_entry_t DecodingTable96[8] =
{
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable19 }, // 0
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable19 }, // 1
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable18 }, // 2
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable19 }, // 3
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable20 }, // 4
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable19 }, // 5
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable19 }, // 6
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable19 } // 7
}; // DecodingTable96[]
const ged_ins_field_entry_t DecodingTable97[4] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  91,  94,   2,  27, 0x78000000 } */ { 0x1b025e5b, (void*)(g_uintptr_t)0x78000000 }, RestrictionTable84 }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  91,  94,   2,  27, 0x78000000 } */ { 0x1b025e5b, (void*)(g_uintptr_t)0x78000000 }, RestrictionTable84 }, // 1
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  91,  94,   2,  27, 0x78000000 } */ { 0x1b025e5b, (void*)(g_uintptr_t)0x78000000 }, RestrictionTable84 }, // 2
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  91,  94,   2,  27, 0x78000000 } */ { 0x1b025e5b, (void*)(g_uintptr_t)0x78000000 }, RestrictionTable85 } // 3
}; // DecodingTable97[]
const ged_ins_field_entry_t DecodingTable98[2] =
{
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_FRAGMENTED,     10, { 2, FragmentsTable11 }, NULL }, // 0
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_FRAGMENTED,     10, { 2, FragmentsTable10 }, RestrictionTable25 } // 1
}; // DecodingTable98[]
const ged_ins_field_entry_t DecodingTable99[2] =
{
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_AccessMode,                        DecodingTable98 }, NULL } // 1
}; // DecodingTable99[]
const ged_ins_field_entry_t DecodingTable100[2] =
{
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  57,  60,   1,  25, 0x1e000000 } */ { 0x19013c39, (void*)(g_uintptr_t)0x1e000000 }, NULL } // 1
}; // DecodingTable100[]
const ged_ins_field_entry_t DecodingTable101[2] =
{
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,     10, { 2, FragmentsTable13 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,     10, { 2, FragmentsTable12 }, RestrictionTable25 } // 1
}; // DecodingTable101[]
const ged_ins_field_entry_t DecodingTable102[2] =
{
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_AccessMode,                        DecodingTable101 }, NULL } // 1
}; // DecodingTable102[]
const ged_ins_field_entry_t DecodingTable103[2] =
{
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  73,  76,   2,   9,     0x1e00 } */ {  0x9024c49, (void*)(g_uintptr_t)    0x1e00 }, NULL } // 1
}; // DecodingTable103[]
const ged_ins_field_entry_t DecodingTable104[2] =
{
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,     10, { 2, FragmentsTable15 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,     10, { 2, FragmentsTable14 }, RestrictionTable25 } // 1
}; // DecodingTable104[]
const ged_ins_field_entry_t DecodingTable105[2] =
{
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_AccessMode,                        DecodingTable104 }, NULL } // 1
}; // DecodingTable105[]
const ged_ins_field_entry_t DecodingTable106[4] =
{
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable105 }, NULL }, // 1
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable106[]
const ged_ins_field_entry_t DecodingTable107[2] =
{
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* { 105, 108,   3,   9,     0x1e00 } */ {  0x9036c69, (void*)(g_uintptr_t)    0x1e00 }, NULL } // 1
}; // DecodingTable107[]
const ged_ins_field_entry_t DecodingTable108[4] =
{
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable107 }, NULL }, // 1
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable108[]
const ged_ins_field_entry_t DecodingTable109[16] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable30 }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable31 }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable32 }, // 3
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 4
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 5
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 6
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 7
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable31 }, // 11
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 15
}; // DecodingTable109[]
const ged_ins_field_entry_t DecodingTable110[4] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1DataType,                      DecodingTable109 }, NULL } // 3
}; // DecodingTable110[]
const ged_ins_field_entry_t DecodingTable111[4] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  43,  46,   1,  11,     0x7800 } */ {  0xb012e2b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable84 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  43,  46,   1,  11,     0x7800 } */ {  0xb012e2b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable84 }, // 1
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  43,  46,   1,  11,     0x7800 } */ {  0xb012e2b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable84 }, // 2
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  43,  46,   1,  11,     0x7800 } */ {  0xb012e2b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable86 } // 3
}; // DecodingTable111[]
const ged_ins_field_entry_t DecodingTable112[4] =
{
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable102 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable102 }, NULL }, // 1
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable102 }, NULL }, // 2
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable112[]
const ged_ins_field_entry_t DecodingTable113[4] =
{
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 1
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 2
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable113[]
const ged_ins_field_entry_t DecodingTable114[16] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable30 }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable31 }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable32 }, // 3
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 4
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 5
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 6
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 7
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_FRAGMENTED,     64, { 2, FragmentsTable9 }, RestrictionTable87 }, // 8
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_FRAGMENTED,     64, { 2, FragmentsTable9 }, RestrictionTable73 }, // 9
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_FRAGMENTED,     64, { 2, FragmentsTable9 }, RestrictionTable73 }, // 10
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable31 }, // 11
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 15
}; // DecodingTable114[]
const ged_ins_field_entry_t DecodingTable115[4] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0DataType,                      DecodingTable114 }, NULL } // 3
}; // DecodingTable115[]
const ged_ins_field_entry_t DecodingTable116[4] =
{
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  91,  94,   2,  27, 0x78000000 } */ { 0x1b025e5b, (void*)(g_uintptr_t)0x78000000 }, RestrictionTable91 }, // 0
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  91,  94,   2,  27, 0x78000000 } */ { 0x1b025e5b, (void*)(g_uintptr_t)0x78000000 }, RestrictionTable91 }, // 1
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  91,  94,   2,  27, 0x78000000 } */ { 0x1b025e5b, (void*)(g_uintptr_t)0x78000000 }, RestrictionTable91 }, // 2
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  91,  94,   2,  27, 0x78000000 } */ { 0x1b025e5b, (void*)(g_uintptr_t)0x78000000 }, RestrictionTable85 } // 3
}; // DecodingTable116[]
const ged_ins_field_entry_t DecodingTable117[16] =
{
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 0
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 1
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 2
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 3
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 7
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 8
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 9
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 10
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 11
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 12
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 13
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 14
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL } // 15
}; // DecodingTable117[]
const ged_ins_field_entry_t DecodingTable118[16] =
{
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable14 }, // 1
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable14 }, // 2
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable14 }, // 3
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable14 }, // 4
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable14 }, // 5
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable14 }, // 6
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable14 }, // 7
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable38 }, // 9
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable38 }, // 10
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable38 }, // 11
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable38 }, // 12
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable38 }, // 13
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable38 }, // 14
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable14 } // 15
}; // DecodingTable118[]
const ged_ins_field_entry_t DecodingTable119[16] =
{
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 9
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 10
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 11
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 12
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 13
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 14
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 15
}; // DecodingTable119[]
const ged_ins_field_entry_t DecodingTable120[4] =
{
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable39 }, // 0
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable39 }, // 1
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable39 }, // 2
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable120[]
const ged_ins_field_entry_t DecodingTable121[16] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable115 }, NULL }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable115 }, NULL }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable115 }, NULL }, // 3
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable115 }, NULL }, // 4
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable115 }, NULL }, // 5
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable115 }, NULL }, // 6
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable115 }, NULL }, // 7
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable110 }, NULL }, // 9
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable110 }, NULL }, // 10
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable110 }, NULL }, // 11
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable110 }, NULL }, // 12
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable110 }, NULL }, // 13
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable110 }, NULL }, // 14
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable115 }, NULL } // 15
}; // DecodingTable121[]
const ged_ins_field_entry_t DecodingTable122[16] =
{
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  48,  51,   1,  16,    0xf0000 } */ { 0x10013330, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable90 }, // 14
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  48,  51,   1,  16,    0xf0000 } */ { 0x10013330, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable90 } // 15
}; // DecodingTable122[]
const ged_ins_field_entry_t DecodingTable123[2] =
{
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_MathFC,                            DecodingTable122 }, NULL } // 1
}; // DecodingTable123[]
const ged_ins_field_entry_t DecodingTable124[16] =
{
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  64,  67,   2,   0,        0xf } */ {    0x24340, (void*)(g_uintptr_t)       0xf }, RestrictionTable90 }, // 14
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  64,  67,   2,   0,        0xf } */ {    0x24340, (void*)(g_uintptr_t)       0xf }, RestrictionTable90 } // 15
}; // DecodingTable124[]
const ged_ins_field_entry_t DecodingTable125[2] =
{
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_MathFC,                            DecodingTable124 }, NULL } // 1
}; // DecodingTable125[]
const ged_ins_field_entry_t DecodingTable126[16] =
{
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  96,  99,   3,   0,        0xf } */ {    0x36360, (void*)(g_uintptr_t)       0xf }, RestrictionTable90 }, // 14
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 15
}; // DecodingTable126[]
const ged_ins_field_entry_t DecodingTable127[2] =
{
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_MathFC,                            DecodingTable126 }, NULL } // 1
}; // DecodingTable127[]
const ged_ins_field_entry_t DecodingTable128[2] =
{
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  80,  81,   2,  16,    0x30000 } */ { 0x10025150, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable94 }, // 0
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable128[]
const ged_ins_field_entry_t DecodingTable129[2] =
{
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  82,  84,   2,  18,   0x1c0000 } */ { 0x12025452, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable95 }, // 0
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable129[]
const ged_ins_field_entry_t DecodingTable130[8] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  36,  36,   1,   4,       0x10 } */ {  0x4012424, (void*)(g_uintptr_t)      0x10 }, RestrictionTable104 }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  36,  36,   1,   4,       0x10 } */ {  0x4012424, (void*)(g_uintptr_t)      0x10 }, RestrictionTable104 }, // 4
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 7
}; // DecodingTable130[]
const ged_ins_field_entry_t DecodingTable131[8] =
{
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable104 }, // 0
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable104 }, // 4
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 7
}; // DecodingTable131[]
const ged_ins_field_entry_t DecodingTable132[4] =
{
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    31, /* {  96, 126,   3,   0, 0x7fffffff } */ {    0x37e60, (void*)(g_uintptr_t)0x7fffffff }, NULL } // 3
}; // DecodingTable132[]
const ged_ins_field_entry_t DecodingTable133[2] =
{
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,     10, { 2, FragmentsTable12 }, RestrictionTable25 } // 1
}; // DecodingTable133[]
const ged_ins_field_entry_t DecodingTable134[2] =
{
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_FRAGMENTED,     10, { 2, FragmentsTable21 }, RestrictionTable25 } // 1
}; // DecodingTable134[]
const ged_ins_field_entry_t DecodingTable135[2] =
{
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,     10, { 2, FragmentsTable22 }, RestrictionTable25 } // 1
}; // DecodingTable135[]
const ged_ins_field_entry_t DecodingTable136[2] =
{
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  44,  51,   1,  12,    0xff000 } */ {  0xc01332c, (void*)(g_uintptr_t)   0xff000 }, RestrictionTable0 }, // 0
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  44,  51,   1,  12,    0xff000 } */ {  0xc01332c, (void*)(g_uintptr_t)   0xff000 }, RestrictionTable26 } // 1
}; // DecodingTable136[]
const ged_ins_field_entry_t DecodingTable137[2] =
{
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  96, 100,   3,   0,       0x1f } */ {    0x36460, (void*)(g_uintptr_t)      0x1f }, NULL } // 1
}; // DecodingTable137[]
const ged_ins_field_entry_t DecodingTable138[2] =
{
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 16 }, NULL } // 1
}; // DecodingTable138[]
const ged_ins_field_entry_t DecodingTable139[2] =
{
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    31, /* {  96, 126,   3,   0, 0x7fffffff } */ {    0x37e60, (void*)(g_uintptr_t)0x7fffffff }, NULL }, // 0
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable139[]
const ged_ins_field_entry_t DecodingTable140[2] =
{
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,     32, { 4, FragmentsTable23 }, RestrictionTable110 }, // 0
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable140[]
const ged_ins_field_entry_t DecodingTable141[2] =
{
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  80,  95,   2,  16, 0xffff0000 } */ { 0x10025f50, (void*)(g_uintptr_t)0xffff0000 }, NULL }, // 0
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable141[]
const ged_ins_field_entry_t DecodingTable142[2] =
{
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  80,  82,   2,  15,    0x70000 } */ {  0xf025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable103 } // 1
}; // DecodingTable142[]
const ged_ins_field_entry_t DecodingTable143[2] =
{
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 16 }, NULL } // 1
}; // DecodingTable143[]
const ged_ins_field_entry_t DecodingTable144[2] =
{
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    12, /* {  96, 107,   3,   0,      0xfff } */ {    0x36b60, (void*)(g_uintptr_t)     0xfff }, NULL } // 1
}; // DecodingTable144[]
const ged_ins_field_entry_t DecodingTable145[2] =
{
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 108, 109,   3,  12,     0x3000 } */ {  0xc036d6c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable111 } // 1
}; // DecodingTable145[]
const ged_ins_field_entry_t DecodingTable146[2] =
{
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 111, 111,   3,  15,     0x8000 } */ {  0xf036f6f, (void*)(g_uintptr_t)    0x8000 }, NULL } // 1
}; // DecodingTable146[]
const ged_ins_field_entry_t DecodingTable147[2] =
{
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 112, 112,   3,  16,    0x10000 } */ { 0x10037070, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable112 } // 1
}; // DecodingTable147[]
const ged_ins_field_entry_t DecodingTable148[2] =
{
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 113, 113,   3,  17,    0x20000 } */ { 0x11037171, (void*)(g_uintptr_t)   0x20000 }, RestrictionTable101 } // 1
}; // DecodingTable148[]
const ged_ins_field_entry_t DecodingTable149[4] =
{
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable22 }, NULL }, // 0
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable22 }, NULL }, // 1
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable22 }, NULL }, // 2
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable149[]
const ged_ins_field_entry_t DecodingTable150[4] =
{
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 109, 110,   3,  13,     0x6000 } */ {  0xd036e6d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 0
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 109, 110,   3,  13,     0x6000 } */ {  0xd036e6d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 1
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 109, 110,   3,  13,     0x6000 } */ {  0xd036e6d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 2
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable150[]
const ged_ins_field_entry_t DecodingTable151[4] =
{
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 111, 111,   3,  15,     0x8000 } */ {  0xf036f6f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 0
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 111, 111,   3,  15,     0x8000 } */ {  0xf036f6f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 1
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 111, 111,   3,  15,     0x8000 } */ {  0xf036f6f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 2
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable151[]
const ged_ins_field_entry_t DecodingTable152[4] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable27 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable27 }, NULL }, // 1
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable27 }, NULL }, // 2
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable152[]
const ged_ins_field_entry_t DecodingTable153[4] =
{
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable105 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable105 }, NULL }, // 1
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable105 }, NULL }, // 2
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable153[]
const ged_ins_field_entry_t DecodingTable154[4] =
{
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable32 }, NULL }, // 0
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable32 }, NULL }, // 1
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable32 }, NULL }, // 2
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable154[]
const ged_ins_field_entry_t DecodingTable155[4] =
{
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable107 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable107 }, NULL }, // 1
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable107 }, NULL }, // 2
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable155[]
const ged_ins_field_entry_t DecodingTable156[4] =
{
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable36 }, NULL }, // 0
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable36 }, NULL }, // 1
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable36 }, NULL }, // 2
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable156[]
const ged_ins_field_entry_t DecodingTable157[4] =
{
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable38 }, NULL }, // 0
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable38 }, NULL }, // 1
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable38 }, NULL }, // 2
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable157[]
const ged_ins_field_entry_t DecodingTable158[4] =
{
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* { 117, 120,   3,  21,  0x1e00000 } */ { 0x15037875, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 0
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* { 117, 120,   3,  21,  0x1e00000 } */ { 0x15037875, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 1
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* { 117, 120,   3,  21,  0x1e00000 } */ { 0x15037875, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 2
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable158[]
const ged_ins_field_entry_t DecodingTable159[4] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_AccessMode,                        DecodingTable26 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_AccessMode,                        DecodingTable26 }, NULL }, // 1
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_AccessMode,                        DecodingTable26 }, NULL }, // 2
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable159[]
const ged_ins_field_entry_t DecodingTable160[4] =
{
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 101, 108,   3,   5,     0x1fe0 } */ {  0x5036c65, (void*)(g_uintptr_t)    0x1fe0 }, RestrictionTable26 }, // 0
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 101, 108,   3,   5,     0x1fe0 } */ {  0x5036c65, (void*)(g_uintptr_t)    0x1fe0 }, RestrictionTable26 }, // 1
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 101, 108,   3,   5,     0x1fe0 } */ {  0x5036c65, (void*)(g_uintptr_t)    0x1fe0 }, RestrictionTable26 }, // 2
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable160[]
const ged_ins_field_entry_t DecodingTable161[4] =
{
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 111, 111,   3,  15,     0x8000 } */ {  0xf036f6f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable39 }, // 0
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 111, 111,   3,  15,     0x8000 } */ {  0xf036f6f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable39 }, // 1
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 111, 111,   3,  15,     0x8000 } */ {  0xf036f6f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable39 }, // 2
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable161[]
const ged_ins_field_entry_t DecodingTable162[2] =
{
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  36,  36,   1,   4,       0x10 } */ {  0x4012424, (void*)(g_uintptr_t)      0x10 }, RestrictionTable119 }, // 0
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 } // 1
}; // DecodingTable162[]
const ged_ins_field_entry_t DecodingTable163[2] =
{
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  43,  43,   1,  11,      0x800 } */ {  0xb012b2b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable120 }, // 0
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable119 } // 1
}; // DecodingTable163[]
const ged_ins_field_entry_t DecodingTable164[2] =
{
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  44,  44,   1,  12,     0x1000 } */ {  0xc012c2c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable119 }, // 0
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable119 } // 1
}; // DecodingTable164[]
const ged_ins_field_entry_t DecodingTable165[2] =
{
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  54,  55,   1,  19,   0xc00000 } */ { 0x13013736, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable121 }, // 0
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  53,  55,   1,  19,   0xe00000 } */ { 0x13013735, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable36 } // 1
}; // DecodingTable165[]
const ged_ins_field_entry_t DecodingTable166[4] =
{
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, RestrictionTable26 }, // 0
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, RestrictionTable122 }, // 1
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, RestrictionTable122 }, // 2
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, RestrictionTable122 } // 3
}; // DecodingTable166[]
const ged_ins_field_entry_t DecodingTable167[2] =
{
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  49,  49,   1,  17,    0x20000 } */ { 0x11013131, (void*)(g_uintptr_t)   0x20000 }, RestrictionTable123 }, // 0
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable167[]
const ged_ins_field_entry_t DecodingTable168[2] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  71,  75,   2,   7,      0xf80 } */ {  0x7024b47, (void*)(g_uintptr_t)     0xf80 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable168[]
const ged_ins_field_entry_t DecodingTable169[2] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable168 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable16 }, RestrictionTable103 } // 1
}; // DecodingTable169[]
const ged_ins_field_entry_t DecodingTable170[2] =
{
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  76,  83,   2,  12,    0xff000 } */ {  0xc02534c, (void*)(g_uintptr_t)   0xff000 }, RestrictionTable26 }, // 0
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable170[]
const ged_ins_field_entry_t DecodingTable171[2] =
{
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable170 }, NULL }, // 0
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  76,  83,   2,  12,    0xff000 } */ {  0xc02534c, (void*)(g_uintptr_t)   0xff000 }, RestrictionTable26 } // 1
}; // DecodingTable171[]
const ged_ins_field_entry_t DecodingTable172[2] =
{
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  69,  70,   2,   5,       0x60 } */ {  0x5024645, (void*)(g_uintptr_t)      0x60 }, RestrictionTable27 }, // 0
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable172[]
const ged_ins_field_entry_t DecodingTable173[2] =
{
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable172 }, NULL }, // 0
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable173[]
const ged_ins_field_entry_t DecodingTable174[2] =
{
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  67,  68,   2,   3,       0x18 } */ {  0x3024443, (void*)(g_uintptr_t)      0x18 }, RestrictionTable124 }, // 0
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable174[]
const ged_ins_field_entry_t DecodingTable175[2] =
{
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable174 }, NULL }, // 0
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 } // 1
}; // DecodingTable175[]
const ged_ins_field_entry_t DecodingTable176[2] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable24 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 3, FragmentsTable17 }, RestrictionTable103 } // 1
}; // DecodingTable176[]
const ged_ins_field_entry_t DecodingTable177[2] =
{
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  97, 104,   3,   1,      0x1fe } */ {  0x1036861, (void*)(g_uintptr_t)     0x1fe }, RestrictionTable26 }, // 0
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  97, 104,   3,   1,      0x1fe } */ {  0x1036861, (void*)(g_uintptr_t)     0x1fe }, RestrictionTable122 } // 1
}; // DecodingTable177[]
const ged_ins_field_entry_t DecodingTable178[2] =
{
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  90,  91,   2,  26,  0xc000000 } */ { 0x1a025b5a, (void*)(g_uintptr_t) 0xc000000 }, RestrictionTable27 }, // 0
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable178[]
const ged_ins_field_entry_t DecodingTable179[2] =
{
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  88,  89,   2,  24,  0x3000000 } */ { 0x18025958, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable124 }, // 0
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 } // 1
}; // DecodingTable179[]
const ged_ins_field_entry_t DecodingTable180[2] =
{
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  43,  45,   1,  11,     0x3800 } */ {  0xb012d2b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable89 } // 1
}; // DecodingTable180[]
const ged_ins_field_entry_t DecodingTable181[2] =
{
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  64,  64,   2,   0,        0x1 } */ {    0x24040, (void*)(g_uintptr_t)       0x1 }, RestrictionTable37 } // 1
}; // DecodingTable181[]
const ged_ins_field_entry_t DecodingTable182[2] =
{
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  85,  85,   2,  21,   0x200000 } */ { 0x15025555, (void*)(g_uintptr_t)  0x200000 }, RestrictionTable37 } // 1
}; // DecodingTable182[]
const ged_ins_field_entry_t DecodingTable183[2] =
{
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 106, 106,   3,  10,      0x400 } */ {  0xa036a6a, (void*)(g_uintptr_t)     0x400 }, RestrictionTable37 } // 1
}; // DecodingTable183[]
const ged_ins_field_entry_t DecodingTable184[2] =
{
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  45,  45,   1,  13,     0x2000 } */ {  0xd012d2d, (void*)(g_uintptr_t)    0x2000 }, RestrictionTable120 }, // 0
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable119 } // 1
}; // DecodingTable184[]
const ged_ins_field_entry_t DecodingTable185[2] =
{
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* { 113, 117,   3,  17,   0x3e0000 } */ { 0x11037571, (void*)(g_uintptr_t)  0x3e0000 }, NULL }, // 0
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable185[]
const ged_ins_field_entry_t DecodingTable186[2] =
{
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src2RegFile,                       DecodingTable185 }, NULL }, // 0
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable18 }, RestrictionTable103 } // 1
}; // DecodingTable186[]
const ged_ins_field_entry_t DecodingTable187[2] =
{
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 118, 125,   3,  22, 0x3fc00000 } */ { 0x16037d76, (void*)(g_uintptr_t)0x3fc00000 }, RestrictionTable26 }, // 0
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable187[]
const ged_ins_field_entry_t DecodingTable188[2] =
{
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src2RegFile,                       DecodingTable187 }, NULL }, // 0
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 118, 125,   3,  22, 0x3fc00000 } */ { 0x16037d76, (void*)(g_uintptr_t)0x3fc00000 }, RestrictionTable26 } // 1
}; // DecodingTable188[]
const ged_ins_field_entry_t DecodingTable189[2] =
{
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable125 }, // 0
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable189[]
const ged_ins_field_entry_t DecodingTable190[2] =
{
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  46,  48,   1,  14,    0x1c000 } */ {  0xe01302e, (void*)(g_uintptr_t)   0x1c000 }, RestrictionTable126 }, // 0
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  46,  48,   1,  14,    0x1c000 } */ {  0xe01302e, (void*)(g_uintptr_t)   0x1c000 }, RestrictionTable127 } // 1
}; // DecodingTable190[]
const ged_ins_field_entry_t DecodingTable191[2] =
{
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable190 }, NULL }, // 0
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  46,  48,   1,  14,    0x1c000 } */ {  0xe01302e, (void*)(g_uintptr_t)   0x1c000 }, RestrictionTable89 } // 1
}; // DecodingTable191[]
const ged_ins_field_entry_t DecodingTable192[2] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  64,  66,   2,   0,        0x7 } */ {    0x24240, (void*)(g_uintptr_t)       0x7 }, RestrictionTable126 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  64,  66,   2,   0,        0x7 } */ {    0x24240, (void*)(g_uintptr_t)       0x7 }, RestrictionTable127 } // 1
}; // DecodingTable192[]
const ged_ins_field_entry_t DecodingTable193[2] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable192 }, NULL }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable193[]
const ged_ins_field_entry_t DecodingTable194[2] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  85,  87,   2,  21,   0xe00000 } */ { 0x15025755, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable126 }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  85,  87,   2,  21,   0xe00000 } */ { 0x15025755, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable127 } // 1
}; // DecodingTable194[]
const ged_ins_field_entry_t DecodingTable195[2] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable194 }, NULL }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_SrcDataType,                       DecodingTable130 }, NULL } // 1
}; // DecodingTable195[]
const ged_ins_field_entry_t DecodingTable196[2] =
{
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* { 106, 108,   3,  10,     0x1c00 } */ {  0xa036c6a, (void*)(g_uintptr_t)    0x1c00 }, RestrictionTable126 }, // 0
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* { 106, 108,   3,  10,     0x1c00 } */ {  0xa036c6a, (void*)(g_uintptr_t)    0x1c00 }, RestrictionTable127 } // 1
}; // DecodingTable196[]
const ged_ins_field_entry_t DecodingTable197[2] =
{
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable196 }, NULL }, // 0
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_SrcDataType,                       DecodingTable131 }, NULL } // 1
}; // DecodingTable197[]
const ged_ins_field_entry_t DecodingTable198[8] =
{
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  67,  82,   2,   3,    0x7fff8 } */ {  0x3025243, (void*)(g_uintptr_t)   0x7fff8 }, RestrictionTable128 }, // 0
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  67,  82,   2,   3,    0x7fff8 } */ {  0x3025243, (void*)(g_uintptr_t)   0x7fff8 }, RestrictionTable129 }, // 1
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  67,  82,   2,   3,    0x7fff8 } */ {  0x3025243, (void*)(g_uintptr_t)   0x7fff8 }, RestrictionTable128 }, // 2
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  67,  82,   2,   3,    0x7fff8 } */ {  0x3025243, (void*)(g_uintptr_t)   0x7fff8 }, RestrictionTable129 }, // 3
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  67,  82,   2,   3,    0x7fff8 } */ {  0x3025243, (void*)(g_uintptr_t)   0x7fff8 }, RestrictionTable128 }, // 4
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  67,  82,   2,   3,    0x7fff8 } */ {  0x3025243, (void*)(g_uintptr_t)   0x7fff8 }, RestrictionTable129 }, // 5
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  67,  82,   2,   3,    0x7fff8 } */ {  0x3025243, (void*)(g_uintptr_t)   0x7fff8 }, RestrictionTable129 }, // 6
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  67,  82,   2,   3,    0x7fff8 } */ {  0x3025243, (void*)(g_uintptr_t)   0x7fff8 }, RestrictionTable129 } // 7
}; // DecodingTable198[]
const ged_ins_field_entry_t DecodingTable199[2] =
{
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src0DataType,                      DecodingTable198 }, NULL }, // 0
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  67,  82,   2,   3,    0x7fff8 } */ {  0x3025243, (void*)(g_uintptr_t)   0x7fff8 }, RestrictionTable128 } // 1
}; // DecodingTable199[]
const ged_ins_field_entry_t DecodingTable200[2] =
{
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable199 }, NULL } // 1
}; // DecodingTable200[]
const ged_ins_field_entry_t DecodingTable201[2] =
{
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src0RegFile,                       DecodingTable200 }, NULL }, // 0
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable201[]
const ged_ins_field_entry_t DecodingTable202[8] =
{
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 109, 124,   3,  13, 0x1fffe000 } */ {  0xd037c6d, (void*)(g_uintptr_t)0x1fffe000 }, RestrictionTable128 }, // 0
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 109, 124,   3,  13, 0x1fffe000 } */ {  0xd037c6d, (void*)(g_uintptr_t)0x1fffe000 }, RestrictionTable129 }, // 1
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 109, 124,   3,  13, 0x1fffe000 } */ {  0xd037c6d, (void*)(g_uintptr_t)0x1fffe000 }, RestrictionTable128 }, // 2
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 109, 124,   3,  13, 0x1fffe000 } */ {  0xd037c6d, (void*)(g_uintptr_t)0x1fffe000 }, RestrictionTable129 }, // 3
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 109, 124,   3,  13, 0x1fffe000 } */ {  0xd037c6d, (void*)(g_uintptr_t)0x1fffe000 }, RestrictionTable128 }, // 4
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 109, 124,   3,  13, 0x1fffe000 } */ {  0xd037c6d, (void*)(g_uintptr_t)0x1fffe000 }, RestrictionTable129 }, // 5
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 109, 124,   3,  13, 0x1fffe000 } */ {  0xd037c6d, (void*)(g_uintptr_t)0x1fffe000 }, RestrictionTable129 }, // 6
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 109, 124,   3,  13, 0x1fffe000 } */ {  0xd037c6d, (void*)(g_uintptr_t)0x1fffe000 }, RestrictionTable129 } // 7
}; // DecodingTable202[]
const ged_ins_field_entry_t DecodingTable203[2] =
{
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src2DataType,                      DecodingTable202 }, NULL }, // 0
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 109, 124,   3,  13, 0x1fffe000 } */ {  0xd037c6d, (void*)(g_uintptr_t)0x1fffe000 }, RestrictionTable128 } // 1
}; // DecodingTable203[]
const ged_ins_field_entry_t DecodingTable204[2] =
{
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable203 }, NULL } // 1
}; // DecodingTable204[]
const ged_ins_field_entry_t DecodingTable205[2] =
{
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src2RegFile,                       DecodingTable204 }, NULL }, // 0
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable205[]
const ged_ins_field_entry_t DecodingTable206[2] =
{
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 111, 112,   3,  15,    0x18000 } */ {  0xf03706f, (void*)(g_uintptr_t)   0x18000 }, RestrictionTable27 }, // 0
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable206[]
const ged_ins_field_entry_t DecodingTable207[2] =
{
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src2RegFile,                       DecodingTable206 }, NULL }, // 0
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable207[]
const ged_ins_field_entry_t DecodingTable208[2] =
{
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  36,  36,   1,   4,       0x10 } */ {  0x4012424, (void*)(g_uintptr_t)      0x10 }, RestrictionTable119 }, // 0
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable119 } // 1
}; // DecodingTable208[]
const ged_ins_field_entry_t DecodingTable209[2] =
{
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, RestrictionTable26 }, // 0
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, RestrictionTable122 } // 1
}; // DecodingTable209[]
const ged_ins_field_entry_t DecodingTable210[2] =
{
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  52,  52,   1,  16,   0x100000 } */ { 0x10013434, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable25 }, // 0
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable210[]
const ged_ins_field_entry_t DecodingTable211[2] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  68,  68,   2,   0,       0x10 } */ {    0x24444, (void*)(g_uintptr_t)      0x10 }, RestrictionTable25 }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable211[]
const ged_ins_field_entry_t DecodingTable212[4] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  91,  94,   2,  27, 0x78000000 } */ { 0x1b025e5b, (void*)(g_uintptr_t)0x78000000 }, RestrictionTable132 }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  91,  94,   2,  27, 0x78000000 } */ { 0x1b025e5b, (void*)(g_uintptr_t)0x78000000 }, RestrictionTable132 }, // 1
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  91,  94,   2,  27, 0x78000000 } */ { 0x1b025e5b, (void*)(g_uintptr_t)0x78000000 }, RestrictionTable132 }, // 2
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  91,  94,   2,  27, 0x78000000 } */ { 0x1b025e5b, (void*)(g_uintptr_t)0x78000000 }, RestrictionTable133 } // 3
}; // DecodingTable212[]
const ged_ins_field_entry_t DecodingTable213[2] =
{
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  48,  52,   1,  16,   0x1f0000 } */ { 0x10013430, (void*)(g_uintptr_t)  0x1f0000 }, NULL }, // 0
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable213[]
const ged_ins_field_entry_t DecodingTable214[2] =
{
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_FRAGMENTED,     10, { 2, FragmentsTable11 }, NULL } // 1
}; // DecodingTable214[]
const ged_ins_field_entry_t DecodingTable215[2] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  64,  68,   2,   0,       0x1f } */ {    0x24440, (void*)(g_uintptr_t)      0x1f }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable215[]
const ged_ins_field_entry_t DecodingTable216[2] =
{
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,     10, { 2, FragmentsTable13 }, NULL } // 1
}; // DecodingTable216[]
const ged_ins_field_entry_t DecodingTable217[2] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  96, 100,   3,   0,       0x1f } */ {    0x36460, (void*)(g_uintptr_t)      0x1f }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable217[]
const ged_ins_field_entry_t DecodingTable218[4] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable217 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable217 }, NULL }, // 1
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable217 }, NULL }, // 2
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable218[]
const ged_ins_field_entry_t DecodingTable219[2] =
{
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,     10, { 2, FragmentsTable15 }, NULL } // 1
}; // DecodingTable219[]
const ged_ins_field_entry_t DecodingTable220[4] =
{
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable219 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable219 }, NULL }, // 1
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable219 }, NULL }, // 2
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable220[]
const ged_ins_field_entry_t DecodingTable221[4] =
{
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 0
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 1
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 2
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable221[]
const ged_ins_field_entry_t DecodingTable222[4] =
{
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* { 114, 116,   3,  18,   0x1c0000 } */ { 0x12037472, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable28 }, // 0
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* { 114, 116,   3,  18,   0x1c0000 } */ { 0x12037472, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable28 }, // 1
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* { 114, 116,   3,  18,   0x1c0000 } */ { 0x12037472, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable28 }, // 2
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable222[]
const ged_ins_field_entry_t DecodingTable223[16] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable30 }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable31 }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable32 }, // 3
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 4
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 5
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable31 }, // 8
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 9
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 11
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 15
}; // DecodingTable223[]
const ged_ins_field_entry_t DecodingTable224[4] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1DataType,                      DecodingTable223 }, NULL } // 3
}; // DecodingTable224[]
const ged_ins_field_entry_t DecodingTable225[4] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  43,  46,   1,  11,     0x7800 } */ {  0xb012e2b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable132 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  43,  46,   1,  11,     0x7800 } */ {  0xb012e2b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable132 }, // 1
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  43,  46,   1,  11,     0x7800 } */ {  0xb012e2b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable132 }, // 2
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  43,  46,   1,  11,     0x7800 } */ {  0xb012e2b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable134 } // 3
}; // DecodingTable225[]
const ged_ins_field_entry_t DecodingTable226[4] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable215 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable215 }, NULL }, // 1
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable215 }, NULL }, // 2
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable226[]
const ged_ins_field_entry_t DecodingTable227[4] =
{
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable216 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable216 }, NULL }, // 1
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable216 }, NULL }, // 2
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable227[]
const ged_ins_field_entry_t DecodingTable228[4] =
{
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  80,  81,   2,  16,    0x30000 } */ { 0x10025150, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 0
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  80,  81,   2,  16,    0x30000 } */ { 0x10025150, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 1
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  80,  81,   2,  16,    0x30000 } */ { 0x10025150, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 2
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable228[]
const ged_ins_field_entry_t DecodingTable229[4] =
{
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  82,  84,   2,  18,   0x1c0000 } */ { 0x12025452, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable28 }, // 0
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  82,  84,   2,  18,   0x1c0000 } */ { 0x12025452, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable28 }, // 1
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  82,  84,   2,  18,   0x1c0000 } */ { 0x12025452, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable28 }, // 2
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable229[]
const ged_ins_field_entry_t DecodingTable230[16] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable30 }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable31 }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable32 }, // 3
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 4
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 5
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_FRAGMENTED,     64, { 2, FragmentsTable9 }, RestrictionTable87 }, // 6
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_FRAGMENTED,     64, { 2, FragmentsTable9 }, RestrictionTable73 }, // 7
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable31 }, // 8
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 9
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_FRAGMENTED,     64, { 2, FragmentsTable9 }, RestrictionTable73 }, // 10
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 11
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 15
}; // DecodingTable230[]
const ged_ins_field_entry_t DecodingTable231[4] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0DataType,                      DecodingTable230 }, NULL } // 3
}; // DecodingTable231[]
const ged_ins_field_entry_t DecodingTable232[4] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  96, 100,   3,   0,       0x1f } */ {    0x36460, (void*)(g_uintptr_t)      0x1f }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  96, 100,   3,   0,       0x1f } */ {    0x36460, (void*)(g_uintptr_t)      0x1f }, NULL }, // 1
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  96, 100,   3,   0,       0x1f } */ {    0x36460, (void*)(g_uintptr_t)      0x1f }, NULL }, // 2
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable232[]
const ged_ins_field_entry_t DecodingTable233[2] =
{
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  46,  48,   1,  14,    0x1c000 } */ {  0xe01302e, (void*)(g_uintptr_t)   0x1c000 }, RestrictionTable126 }, // 0
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  46,  48,   1,  14,    0x1c000 } */ {  0xe01302e, (void*)(g_uintptr_t)   0x1c000 }, RestrictionTable135 } // 1
}; // DecodingTable233[]
const ged_ins_field_entry_t DecodingTable234[2] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  64,  66,   2,   0,        0x7 } */ {    0x24240, (void*)(g_uintptr_t)       0x7 }, RestrictionTable126 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  64,  66,   2,   0,        0x7 } */ {    0x24240, (void*)(g_uintptr_t)       0x7 }, RestrictionTable135 } // 1
}; // DecodingTable234[]
const ged_ins_field_entry_t DecodingTable235[8] =
{
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  43,  43,   1,  11,      0x800 } */ {  0xb012b2b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable120 }, // 0
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  43,  43,   1,  11,      0x800 } */ {  0xb012b2b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable120 }, // 1
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  43,  43,   1,  11,      0x800 } */ {  0xb012b2b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable120 }, // 2
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  43,  43,   1,  11,      0x800 } */ {  0xb012b2b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable119 }, // 3
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  43,  43,   1,  11,      0x800 } */ {  0xb012b2b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable120 }, // 4
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  43,  43,   1,  11,      0x800 } */ {  0xb012b2b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable120 }, // 5
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  43,  43,   1,  11,      0x800 } */ {  0xb012b2b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable120 }, // 6
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  43,  43,   1,  11,      0x800 } */ {  0xb012b2b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable120 } // 7
}; // DecodingTable235[]
const ged_ins_field_entry_t DecodingTable236[2] =
{
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  43,  43,   1,  11,      0x800 } */ {  0xb012b2b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable120 }, // 0
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0DataType,                      DecodingTable235 }, NULL } // 1
}; // DecodingTable236[]
const ged_ins_field_entry_t DecodingTable237[2] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  85,  87,   2,  21,   0xe00000 } */ { 0x15025755, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable126 }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  85,  87,   2,  21,   0xe00000 } */ { 0x15025755, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable135 } // 1
}; // DecodingTable237[]
const ged_ins_field_entry_t DecodingTable238[8] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  71,  75,   2,   7,      0xf80 } */ {  0x7024b47, (void*)(g_uintptr_t)     0xf80 }, NULL }, // 3
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 7
}; // DecodingTable238[]
const ged_ins_field_entry_t DecodingTable239[2] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0DataType,                      DecodingTable238 }, NULL } // 1
}; // DecodingTable239[]
const ged_ins_field_entry_t DecodingTable240[2] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  71,  75,   2,   7,      0xf80 } */ {  0x7024b47, (void*)(g_uintptr_t)     0xf80 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable239 }, NULL } // 1
}; // DecodingTable240[]
const ged_ins_field_entry_t DecodingTable241[8] =
{
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  76,  83,   2,  12,    0xff000 } */ {  0xc02534c, (void*)(g_uintptr_t)   0xff000 }, RestrictionTable122 }, // 3
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 7
}; // DecodingTable241[]
const ged_ins_field_entry_t DecodingTable242[2] =
{
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0DataType,                      DecodingTable241 }, NULL } // 1
}; // DecodingTable242[]
const ged_ins_field_entry_t DecodingTable243[2] =
{
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  76,  83,   2,  12,    0xff000 } */ {  0xc02534c, (void*)(g_uintptr_t)   0xff000 }, RestrictionTable26 }, // 0
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable242 }, NULL } // 1
}; // DecodingTable243[]
const ged_ins_field_entry_t DecodingTable244[8] =
{
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  69,  70,   2,   5,       0x60 } */ {  0x5024645, (void*)(g_uintptr_t)      0x60 }, RestrictionTable27 }, // 3
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 7
}; // DecodingTable244[]
const ged_ins_field_entry_t DecodingTable245[2] =
{
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0DataType,                      DecodingTable244 }, NULL } // 1
}; // DecodingTable245[]
const ged_ins_field_entry_t DecodingTable246[2] =
{
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  69,  70,   2,   5,       0x60 } */ {  0x5024645, (void*)(g_uintptr_t)      0x60 }, RestrictionTable27 }, // 0
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable245 }, NULL } // 1
}; // DecodingTable246[]
const ged_ins_field_entry_t DecodingTable247[8] =
{
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  67,  68,   2,   3,       0x18 } */ {  0x3024443, (void*)(g_uintptr_t)      0x18 }, RestrictionTable124 }, // 3
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 7
}; // DecodingTable247[]
const ged_ins_field_entry_t DecodingTable248[2] =
{
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0DataType,                      DecodingTable247 }, NULL } // 1
}; // DecodingTable248[]
const ged_ins_field_entry_t DecodingTable249[2] =
{
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  67,  68,   2,   3,       0x18 } */ {  0x3024443, (void*)(g_uintptr_t)      0x18 }, RestrictionTable124 }, // 0
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable248 }, NULL } // 1
}; // DecodingTable249[]
const ged_ins_field_entry_t DecodingTable250[2] =
{
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* { 106, 108,   3,  10,     0x1c00 } */ {  0xa036c6a, (void*)(g_uintptr_t)    0x1c00 }, RestrictionTable126 }, // 0
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* { 106, 108,   3,  10,     0x1c00 } */ {  0xa036c6a, (void*)(g_uintptr_t)    0x1c00 }, RestrictionTable135 } // 1
}; // DecodingTable250[]
const ged_ins_field_entry_t DecodingTable251[8] =
{
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  67,  82,   2,   3,    0x7fff8 } */ {  0x3025243, (void*)(g_uintptr_t)   0x7fff8 }, RestrictionTable128 }, // 0
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  67,  82,   2,   3,    0x7fff8 } */ {  0x3025243, (void*)(g_uintptr_t)   0x7fff8 }, RestrictionTable128 }, // 1
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  67,  82,   2,   3,    0x7fff8 } */ {  0x3025243, (void*)(g_uintptr_t)   0x7fff8 }, RestrictionTable128 }, // 2
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  67,  82,   2,   3,    0x7fff8 } */ {  0x3025243, (void*)(g_uintptr_t)   0x7fff8 }, RestrictionTable128 }, // 4
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  67,  82,   2,   3,    0x7fff8 } */ {  0x3025243, (void*)(g_uintptr_t)   0x7fff8 }, RestrictionTable128 }, // 5
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  67,  82,   2,   3,    0x7fff8 } */ {  0x3025243, (void*)(g_uintptr_t)   0x7fff8 }, RestrictionTable128 }, // 6
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  67,  82,   2,   3,    0x7fff8 } */ {  0x3025243, (void*)(g_uintptr_t)   0x7fff8 }, RestrictionTable128 } // 7
}; // DecodingTable251[]
const ged_ins_field_entry_t DecodingTable252[2] =
{
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src0DataType,                      DecodingTable198 }, NULL }, // 0
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src0DataType,                      DecodingTable251 }, NULL } // 1
}; // DecodingTable252[]
const ged_ins_field_entry_t DecodingTable253[2] =
{
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable252 }, NULL } // 1
}; // DecodingTable253[]
const ged_ins_field_entry_t DecodingTable254[8] =
{
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  72,  75,   2,   8,      0xf00 } */ {  0x8024b48, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable90 }, // 3
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 7
}; // DecodingTable254[]
const ged_ins_field_entry_t DecodingTable255[2] =
{
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0DataType,                      DecodingTable254 }, NULL } // 1
}; // DecodingTable255[]
const ged_ins_field_entry_t DecodingTable256[2] =
{
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  72,  75,   2,   8,      0xf00 } */ {  0x8024b48, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable90 }, // 0
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable255 }, NULL } // 1
}; // DecodingTable256[]
const ged_ins_field_entry_t DecodingTable257[2] =
{
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* { 114, 117,   3,  18,   0x3c0000 } */ { 0x12037572, (void*)(g_uintptr_t)  0x3c0000 }, RestrictionTable90 }, // 0
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable257[]
const ged_ins_field_entry_t DecodingTable258[2] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  68,  68,   2,   0,       0x10 } */ {    0x24444, (void*)(g_uintptr_t)      0x10 }, RestrictionTable25 }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL } // 1
}; // DecodingTable258[]
const ged_ins_field_entry_t DecodingTable259[16] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable231 }, NULL }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable231 }, NULL }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable231 }, NULL }, // 3
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable231 }, NULL }, // 4
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable231 }, NULL }, // 5
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable231 }, NULL }, // 6
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable231 }, NULL }, // 7
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable224 }, NULL }, // 9
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable224 }, NULL }, // 10
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable224 }, NULL }, // 11
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable224 }, NULL }, // 12
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable224 }, NULL }, // 13
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable224 }, NULL }, // 14
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable231 }, NULL } // 15
}; // DecodingTable259[]
const ged_ins_field_entry_t DecodingTable260[2] =
{
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  64,  67,   2,   0,        0xf } */ {    0x24340, (void*)(g_uintptr_t)       0xf }, RestrictionTable90 }, // 0
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable260[]
const ged_ins_field_entry_t DecodingTable261[16] =
{
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable260 }, NULL }, // 14
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable260 }, NULL } // 15
}; // DecodingTable261[]
const ged_ins_field_entry_t DecodingTable262[2] =
{
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  96,  99,   3,   0,        0xf } */ {    0x36360, (void*)(g_uintptr_t)       0xf }, RestrictionTable90 }, // 0
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable262[]
const ged_ins_field_entry_t DecodingTable263[4] =
{
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable262 }, NULL }, // 1
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable263[]
const ged_ins_field_entry_t DecodingTable264[16] =
{
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable263 }, NULL }, // 14
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable263 }, NULL } // 15
}; // DecodingTable264[]
const ged_ins_field_entry_t DecodingTable265[2] =
{
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 0
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable265[]
const ged_ins_field_entry_t DecodingTable266[2] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable140 } // 1
}; // DecodingTable266[]
const ged_ins_field_entry_t DecodingTable267[2] =
{
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 0
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable267[]
const ged_ins_field_entry_t DecodingTable268[2] =
{
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {  50,  59,   1,  18,  0xffc0000 } */ { 0x12013b32, (void*)(g_uintptr_t) 0xffc0000 }, NULL } // 1
}; // DecodingTable268[]
const ged_ins_field_entry_t DecodingTable269[2] =
{
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 0
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL } // 1
}; // DecodingTable269[]
const ged_ins_field_entry_t DecodingTable270[2] =
{
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstRegFile,                        DecodingTable269 }, NULL }, // 0
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable270[]
const ged_ins_field_entry_t DecodingTable271[2] =
{
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  60,  63,   1,  28, 0xf0000000 } */ { 0x1c013f3c, (void*)(g_uintptr_t)0xf0000000 }, NULL } // 1
}; // DecodingTable271[]
const ged_ins_field_entry_t DecodingTable272[4] =
{
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  80,  80,   2,  16,    0x10000 } */ { 0x10025050, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable15 }, // 0
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  80,  80,   2,  16,    0x10000 } */ { 0x10025050, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable15 }, // 1
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable272[]
const ged_ins_field_entry_t DecodingTable273[2] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable273[]
const ged_ins_field_entry_t DecodingTable274[2] =
{
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {  66,  75,   2,   2,      0xffc } */ {  0x2024b42, (void*)(g_uintptr_t)     0xffc }, NULL } // 1
}; // DecodingTable274[]
const ged_ins_field_entry_t DecodingTable275[4] =
{
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 0
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 1
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable275[]
const ged_ins_field_entry_t DecodingTable276[2] =
{
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable275 }, NULL }, // 0
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable276[]
const ged_ins_field_entry_t DecodingTable277[2] =
{
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  76,  79,   2,  12,     0xf000 } */ {  0xc024f4c, (void*)(g_uintptr_t)    0xf000 }, NULL } // 1
}; // DecodingTable277[]
const ged_ins_field_entry_t DecodingTable278[4] =
{
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 120, 121,   3,  24,  0x3000000 } */ { 0x18037978, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable16 }, // 0
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 120, 121,   3,  24,  0x3000000 } */ { 0x18037978, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable16 }, // 1
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable278[]
const ged_ins_field_entry_t DecodingTable279[4] =
{
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 112, 112,   3,  16,    0x10000 } */ { 0x10037070, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable15 }, // 0
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 112, 112,   3,  16,    0x10000 } */ { 0x10037070, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable15 }, // 1
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable279[]
const ged_ins_field_entry_t DecodingTable280[2] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable280[]
const ged_ins_field_entry_t DecodingTable281[4] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable280 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable280 }, NULL }, // 1
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable281[]
const ged_ins_field_entry_t DecodingTable282[2] =
{
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {  98, 107,   3,   2,      0xffc } */ {  0x2036b62, (void*)(g_uintptr_t)     0xffc }, NULL } // 1
}; // DecodingTable282[]
const ged_ins_field_entry_t DecodingTable283[2] =
{
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 104, 111,   3,   8,     0xff00 } */ {  0x8036f68, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 0
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable283[]
const ged_ins_field_entry_t DecodingTable284[2] =
{
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* { 108, 111,   3,  12,     0xf000 } */ {  0xc036f6c, (void*)(g_uintptr_t)    0xf000 }, NULL } // 1
}; // DecodingTable284[]
const ged_ins_field_entry_t DecodingTable285[4] =
{
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 0
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 1
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable285[]
const ged_ins_field_entry_t DecodingTable286[4] =
{
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* { 113, 115,   3,  17,    0xe0000 } */ { 0x11037371, (void*)(g_uintptr_t)   0xe0000 }, RestrictionTable28 }, // 0
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* { 113, 115,   3,  17,    0xe0000 } */ { 0x11037371, (void*)(g_uintptr_t)   0xe0000 }, RestrictionTable28 }, // 1
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable286[]
const ged_ins_field_entry_t DecodingTable287[4] =
{
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* { 116, 119,   3,  20,   0xf00000 } */ { 0x14037774, (void*)(g_uintptr_t)  0xf00000 }, RestrictionTable17 }, // 0
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* { 116, 119,   3,  20,   0xf00000 } */ { 0x14037774, (void*)(g_uintptr_t)  0xf00000 }, RestrictionTable17 }, // 1
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable287[]
const ged_ins_field_entry_t DecodingTable288[16] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable31 }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 4
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable32 }, // 5
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable30 }, // 6
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 8
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable31 }, // 9
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 10
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 15
}; // DecodingTable288[]
const ged_ins_field_entry_t DecodingTable289[4] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1DataType,                      DecodingTable288 }, NULL }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1DataType,                      DecodingTable288 }, NULL } // 3
}; // DecodingTable289[]
const ged_ins_field_entry_t DecodingTable290[2] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable141 } // 1
}; // DecodingTable290[]
const ged_ins_field_entry_t DecodingTable291[16] =
{
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 0
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 1
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 2
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 4
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 5
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 6
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 8
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 9
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 10
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 15
}; // DecodingTable291[]
const ged_ins_field_entry_t DecodingTable292[4] =
{
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 0
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 1
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0DataType,                      DecodingTable291 }, NULL }, // 2
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0DataType,                      DecodingTable291 }, NULL } // 3
}; // DecodingTable292[]
const ged_ins_field_entry_t DecodingTable293[4] =
{
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 0
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 1
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable293[]
const ged_ins_field_entry_t DecodingTable294[4] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable273 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable273 }, NULL }, // 1
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable294[]
const ged_ins_field_entry_t DecodingTable295[4] =
{
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable274 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable274 }, NULL }, // 1
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable295[]
const ged_ins_field_entry_t DecodingTable296[4] =
{
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable277 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable277 }, NULL }, // 1
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable296[]
const ged_ins_field_entry_t DecodingTable297[4] =
{
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 0
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 1
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable297[]
const ged_ins_field_entry_t DecodingTable298[4] =
{
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  81,  83,   2,  17,    0xe0000 } */ { 0x11025351, (void*)(g_uintptr_t)   0xe0000 }, RestrictionTable28 }, // 0
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  81,  83,   2,  17,    0xe0000 } */ { 0x11025351, (void*)(g_uintptr_t)   0xe0000 }, RestrictionTable28 }, // 1
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable298[]
const ged_ins_field_entry_t DecodingTable299[4] =
{
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  84,  87,   2,  20,   0xf00000 } */ { 0x14025754, (void*)(g_uintptr_t)  0xf00000 }, RestrictionTable17 }, // 0
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  84,  87,   2,  20,   0xf00000 } */ { 0x14025754, (void*)(g_uintptr_t)  0xf00000 }, RestrictionTable17 }, // 1
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable299[]
const ged_ins_field_entry_t DecodingTable300[16] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable31 }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_FRAGMENTED,     64, { 2, FragmentsTable28 }, RestrictionTable87 }, // 3
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 4
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable32 }, // 5
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable30 }, // 6
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_FRAGMENTED,     64, { 2, FragmentsTable28 }, RestrictionTable73 }, // 7
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 8
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable31 }, // 9
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 10
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_FRAGMENTED,     64, { 2, FragmentsTable28 }, RestrictionTable73 }, // 11
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 15
}; // DecodingTable300[]
const ged_ins_field_entry_t DecodingTable301[4] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0DataType,                      DecodingTable300 }, NULL }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0DataType,                      DecodingTable300 }, NULL } // 3
}; // DecodingTable301[]
const ged_ins_field_entry_t DecodingTable302[16] =
{
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 0
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 1
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable146 }, // 2
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable146 }, // 3
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 4
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 5
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 6
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 7
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 8
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 9
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 10
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 11
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 12
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 13
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 14
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 } // 15
}; // DecodingTable302[]
const ged_ins_field_entry_t DecodingTable303[4] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 1
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable140 }, // 2
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable140 } // 3
}; // DecodingTable303[]
const ged_ins_field_entry_t DecodingTable304[4] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable0 }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 1
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable304[]
const ged_ins_field_entry_t DecodingTable305[4] =
{
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, RestrictionTable0 }, // 0
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 1
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable305[]
const ged_ins_field_entry_t DecodingTable306[4] =
{
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 0
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 1
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable306[]
const ged_ins_field_entry_t DecodingTable307[4] =
{
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable144 }, // 0
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable144 }, // 1
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable307[]
const ged_ins_field_entry_t DecodingTable308[4] =
{
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable50 }, // 0
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable50 }, // 1
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable308[]
const ged_ins_field_entry_t DecodingTable309[4] =
{
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 0 }, RestrictionTable145 }, // 0
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 0 }, RestrictionTable145 }, // 1
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable309[]
const ged_ins_field_entry_t DecodingTable310[4] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 } // 3
}; // DecodingTable310[]
const ged_ins_field_entry_t DecodingTable311[4] =
{
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  87,  87,   2,  23,   0x800000 } */ { 0x17025757, (void*)(g_uintptr_t)  0x800000 }, RestrictionTable0 }, // 0
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  87,  87,   2,  23,   0x800000 } */ { 0x17025757, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 1
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable311[]
const ged_ins_field_entry_t DecodingTable312[2] =
{
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 } // 1
}; // DecodingTable312[]
const ged_ins_field_entry_t DecodingTable313[2] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 } // 1
}; // DecodingTable313[]
const ged_ins_field_entry_t DecodingTable314[2] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 } // 1
}; // DecodingTable314[]
const ged_ins_field_entry_t DecodingTable315[4] =
{
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 0
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 1
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable315[]
const ged_ins_field_entry_t DecodingTable316[2] =
{
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 104, 111,   3,   8,     0xff00 } */ {  0x8036f68, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 0
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 104, 111,   3,   8,     0xff00 } */ {  0x8036f68, (void*)(g_uintptr_t)    0xff00 }, NULL } // 1
}; // DecodingTable316[]
const ged_ins_field_entry_t DecodingTable317[4] =
{
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* { 115, 119,   3,  19,   0xf80000 } */ { 0x13037773, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 0
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* { 115, 119,   3,  19,   0xf80000 } */ { 0x13037773, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 1
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable317[]
const ged_ins_field_entry_t DecodingTable318[4] =
{
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 120, 127,   3,  24, 0xff000000 } */ { 0x18037f78, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 0
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 120, 127,   3,  24, 0xff000000 } */ { 0x18037f78, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 1
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable318[]
const ged_ins_field_entry_t DecodingTable319[2] =
{
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 } // 1
}; // DecodingTable319[]
const ged_ins_field_entry_t DecodingTable320[8] =
{
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  64,  79,   2,   0,     0xffff } */ {    0x24f40, (void*)(g_uintptr_t)    0xffff }, RestrictionTable128 }, // 0
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  64,  79,   2,   0,     0xffff } */ {    0x24f40, (void*)(g_uintptr_t)    0xffff }, RestrictionTable128 }, // 1
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  64,  79,   2,   0,     0xffff } */ {    0x24f40, (void*)(g_uintptr_t)    0xffff }, RestrictionTable128 }, // 2
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  64,  79,   2,   0,     0xffff } */ {    0x24f40, (void*)(g_uintptr_t)    0xffff }, RestrictionTable128 }, // 3
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  64,  79,   2,   0,     0xffff } */ {    0x24f40, (void*)(g_uintptr_t)    0xffff }, RestrictionTable129 }, // 4
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  64,  79,   2,   0,     0xffff } */ {    0x24f40, (void*)(g_uintptr_t)    0xffff }, RestrictionTable129 }, // 5
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  64,  79,   2,   0,     0xffff } */ {    0x24f40, (void*)(g_uintptr_t)    0xffff }, RestrictionTable129 }, // 6
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  64,  79,   2,   0,     0xffff } */ {    0x24f40, (void*)(g_uintptr_t)    0xffff }, RestrictionTable129 } // 7
}; // DecodingTable320[]
const ged_ins_field_entry_t DecodingTable321[2] =
{
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src0DataType,                      DecodingTable320 }, NULL }, // 0
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  64,  79,   2,   0,     0xffff } */ {    0x24f40, (void*)(g_uintptr_t)    0xffff }, RestrictionTable128 } // 1
}; // DecodingTable321[]
const ged_ins_field_entry_t DecodingTable322[4] =
{
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable321 }, NULL }, // 2
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable321 }, NULL } // 3
}; // DecodingTable322[]
const ged_ins_field_entry_t DecodingTable323[8] =
{
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 112, 127,   3,  16, 0xffff0000 } */ { 0x10037f70, (void*)(g_uintptr_t)0xffff0000 }, RestrictionTable128 }, // 0
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 112, 127,   3,  16, 0xffff0000 } */ { 0x10037f70, (void*)(g_uintptr_t)0xffff0000 }, RestrictionTable128 }, // 1
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 112, 127,   3,  16, 0xffff0000 } */ { 0x10037f70, (void*)(g_uintptr_t)0xffff0000 }, RestrictionTable128 }, // 2
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 112, 127,   3,  16, 0xffff0000 } */ { 0x10037f70, (void*)(g_uintptr_t)0xffff0000 }, RestrictionTable128 }, // 3
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 112, 127,   3,  16, 0xffff0000 } */ { 0x10037f70, (void*)(g_uintptr_t)0xffff0000 }, RestrictionTable129 }, // 4
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 112, 127,   3,  16, 0xffff0000 } */ { 0x10037f70, (void*)(g_uintptr_t)0xffff0000 }, RestrictionTable129 }, // 5
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 112, 127,   3,  16, 0xffff0000 } */ { 0x10037f70, (void*)(g_uintptr_t)0xffff0000 }, RestrictionTable129 }, // 6
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 112, 127,   3,  16, 0xffff0000 } */ { 0x10037f70, (void*)(g_uintptr_t)0xffff0000 }, RestrictionTable129 } // 7
}; // DecodingTable323[]
const ged_ins_field_entry_t DecodingTable324[2] =
{
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src2DataType,                      DecodingTable323 }, NULL }, // 0
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 112, 127,   3,  16, 0xffff0000 } */ { 0x10037f70, (void*)(g_uintptr_t)0xffff0000 }, RestrictionTable128 } // 1
}; // DecodingTable324[]
const ged_ins_field_entry_t DecodingTable325[4] =
{
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable324 }, NULL }, // 2
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable324 }, NULL } // 3
}; // DecodingTable325[]
const ged_ins_field_entry_t DecodingTable326[4] =
{
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 0
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 1
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable326[]
const ged_ins_field_entry_t DecodingTable327[2] =
{
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_FRAGMENTED,     32, { 5, FragmentsTable34 }, NULL }, // 0
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable327[]
const ged_ins_field_entry_t DecodingTable328[2] =
{
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,     32, { 5, FragmentsTable35 }, RestrictionTable152 }, // 0
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable328[]
const ged_ins_field_entry_t DecodingTable329[2] =
{
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  42,   1,   7,      0x700 } */ {  0x7012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable103 } // 1
}; // DecodingTable329[]
const ged_ins_field_entry_t DecodingTable330[16] =
{
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 0
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 1
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 2
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 3
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 7
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 8
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 9
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 10
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 11
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 12
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 13
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 14
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL } // 15
}; // DecodingTable330[]
const ged_ins_field_entry_t DecodingTable331[16] =
{
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 1
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 2
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 3
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 4
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 5
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 6
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 7
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable12 }, // 11
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable12 }, // 12
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable12 }, // 13
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable12 }, // 14
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 } // 15
}; // DecodingTable331[]
const ged_ins_field_entry_t DecodingTable332[16] =
{
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable27 }, RestrictionTable139 }, // 11
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable27 }, RestrictionTable139 }, // 12
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable27 }, RestrictionTable139 }, // 13
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable27 }, RestrictionTable139 }, // 14
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 15
}; // DecodingTable332[]
const ged_ins_field_entry_t DecodingTable333[16] =
{
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 0
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 1
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 2
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 3
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 4
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 5
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 6
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 7
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 8
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 9
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 10
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 11
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 12
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 13
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 14
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL } // 15
}; // DecodingTable333[]
const ged_ins_field_entry_t DecodingTable334[16] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 1
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 2
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 3
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 4
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 5
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 6
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 7
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 8
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 9
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 10
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 11
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 12
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 13
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL } // 15
}; // DecodingTable334[]
const ged_ins_field_entry_t DecodingTable335[4] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_MathFC,                            DecodingTable334 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_MathFC,                            DecodingTable334 }, NULL }, // 1
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable335[]
const ged_ins_field_entry_t DecodingTable336[4] =
{
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  80,  80,   2,  16,    0x10000 } */ { 0x10025050, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable39 }, // 0
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  80,  80,   2,  16,    0x10000 } */ { 0x10025050, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable39 }, // 1
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable336[]
const ged_ins_field_entry_t DecodingTable337[16] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 1
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 2
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 3
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 4
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 5
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 6
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 7
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 8
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 9
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 10
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 11
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 12
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 13
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL } // 15
}; // DecodingTable337[]
const ged_ins_field_entry_t DecodingTable338[4] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_MathFC,                            DecodingTable337 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_MathFC,                            DecodingTable337 }, NULL }, // 1
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable338[]
const ged_ins_field_entry_t DecodingTable339[4] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable338 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable338 }, NULL }, // 1
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable339[]
const ged_ins_field_entry_t DecodingTable340[4] =
{
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 0
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 1
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable340[]
const ged_ins_field_entry_t DecodingTable341[4] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0DataType,                      DecodingTable288 }, NULL }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0DataType,                      DecodingTable288 }, NULL } // 3
}; // DecodingTable341[]
const ged_ins_field_entry_t DecodingTable342[16] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable341 }, NULL }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable341 }, NULL }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable341 }, NULL }, // 3
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable341 }, NULL }, // 4
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable341 }, NULL }, // 5
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable341 }, NULL }, // 6
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable341 }, NULL }, // 7
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable289 }, NULL }, // 11
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable289 }, NULL }, // 12
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable289 }, NULL }, // 13
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable289 }, NULL }, // 14
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable341 }, NULL } // 15
}; // DecodingTable342[]
const ged_ins_field_entry_t DecodingTable343[16] =
{
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  51,  54,   1,  19,   0x780000 } */ { 0x13013633, (void*)(g_uintptr_t)  0x780000 }, RestrictionTable90 }, // 14
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  51,  54,   1,  19,   0x780000 } */ { 0x13013633, (void*)(g_uintptr_t)  0x780000 }, RestrictionTable90 } // 15
}; // DecodingTable343[]
const ged_ins_field_entry_t DecodingTable344[4] =
{
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  67,  70,   2,   3,       0x78 } */ {  0x3024643, (void*)(g_uintptr_t)      0x78 }, RestrictionTable90 }, // 0
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  67,  70,   2,   3,       0x78 } */ {  0x3024643, (void*)(g_uintptr_t)      0x78 }, RestrictionTable90 }, // 1
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable344[]
const ged_ins_field_entry_t DecodingTable345[16] =
{
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable344 }, NULL }, // 14
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable344 }, NULL } // 15
}; // DecodingTable345[]
const ged_ins_field_entry_t DecodingTable346[4] =
{
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  99, 102,   3,   3,       0x78 } */ {  0x3036663, (void*)(g_uintptr_t)      0x78 }, RestrictionTable90 }, // 0
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  99, 102,   3,   3,       0x78 } */ {  0x3036663, (void*)(g_uintptr_t)      0x78 }, RestrictionTable90 }, // 1
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable346[]
const ged_ins_field_entry_t DecodingTable347[16] =
{
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable346 }, NULL }, // 14
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable346 }, NULL } // 15
}; // DecodingTable347[]
const ged_ins_field_entry_t DecodingTable348[128] =
{
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 0
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 1
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 2
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 3
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 4
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 5
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 6
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 7
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 8
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 9
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 10
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 11
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 12
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 13
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 16
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 17
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 18
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 19
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 20
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 21
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 22
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 23
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 24
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 25
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 26
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 27
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 28
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 29
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 30
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 31
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable50 }, // 32
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 33
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 34
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 35
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 36
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 37
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 38
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 39
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 40
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 41
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 42
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 43
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 44
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 45
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 46
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 47
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 48
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 49
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 50
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 51
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 52
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 53
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 54
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 55
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 56
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 57
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 58
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 59
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 60
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 61
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 62
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 63
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 64
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 65
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 66
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 67
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 68
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 69
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 70
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 71
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 72
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 73
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 74
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 75
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 76
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 77
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 78
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 79
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 80
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 81
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 82
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 83
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 84
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 85
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 86
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 87
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 88
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 89
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 90
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 91
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 92
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 93
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 94
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 95
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 96
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 97
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 98
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 99
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 100
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 101
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 102
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 103
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 104
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 105
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 106
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 107
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 108
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 109
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 110
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 111
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 112
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 113
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 114
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 115
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 116
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 117
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 118
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 119
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 120
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 121
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 122
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 123
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 124
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 125
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 126
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 } // 127
}; // DecodingTable348[]
const ged_ins_field_entry_t DecodingTable349[4] =
{
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable143 }, // 0
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable143 }, // 1
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable154 }, // 2
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable154 } // 3
}; // DecodingTable349[]
const ged_ins_field_entry_t DecodingTable350[4] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 1
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable350[]
const ged_ins_field_entry_t DecodingTable351[2] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable351[]
const ged_ins_field_entry_t DecodingTable352[2] =
{
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     8, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable352[]
const ged_ins_field_entry_t DecodingTable353[2] =
{
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable144 }, // 0
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable353[]
const ged_ins_field_entry_t DecodingTable354[2] =
{
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable50 }, // 0
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable354[]
const ged_ins_field_entry_t DecodingTable355[2] =
{
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 0 }, RestrictionTable145 }, // 0
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable355[]
const ged_ins_field_entry_t DecodingTable356[4] =
{
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, NULL }, // 2
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, NULL } // 3
}; // DecodingTable356[]
const ged_ins_field_entry_t DecodingTable357[2] =
{
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  64,  95,   2,   0, 0xffffffff } */ {    0x25f40, (void*)(g_uintptr_t)0xffffffff }, NULL } // 1
}; // DecodingTable357[]
const ged_ins_field_entry_t DecodingTable358[128] =
{
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 0
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 1
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 2
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 3
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 4
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 5
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 6
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 7
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 8
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 9
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 10
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 11
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 12
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 13
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 14
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 15
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 16
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 17
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 18
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 19
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 22
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 23
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 24
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 25
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 26
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 27
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 28
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 29
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 30
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 31
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 32
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 33
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 34
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 35
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 36
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 37
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 38
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 39
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 40
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 41
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 42
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 43
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 44
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 45
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 46
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 47
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 48
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 49
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 50
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 51
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 52
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 53
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 54
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 55
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 56
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 57
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 58
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 59
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 60
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 61
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 62
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 63
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable159 }, // 64
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable159 }, // 65
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 66
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 67
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 68
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 69
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 70
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 71
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable159 }, // 72
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 73
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 74
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 75
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 76
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 77
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 78
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 79
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 80
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 81
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 82
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 83
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 84
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 85
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 86
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 87
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 88
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 89
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 90
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 91
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 92
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 93
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 94
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 95
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 96
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 97
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable159 }, // 98
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 99
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 100
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 101
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 102
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 103
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 104
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 105
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 106
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 107
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 108
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 109
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 110
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 111
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable159 }, // 112
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 113
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 114
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 115
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 116
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 117
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 118
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 119
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 120
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 121
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 122
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 123
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 124
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 125
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 126
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 } // 127
}; // DecodingTable358[]
const ged_ins_field_entry_t DecodingTable359[128] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 1
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 2
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 3
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 4
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 5
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 6
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 7
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 8
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 9
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 10
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 11
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 12
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 13
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 14
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 15
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 16
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 17
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 18
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 19
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 20
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 23
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 24
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 25
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 26
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 27
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 28
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 29
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 30
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 31
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 32
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 33
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 34
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 35
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 36
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 37
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 38
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 39
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 40
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 41
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 42
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 43
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 44
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 45
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 46
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 47
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 48
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 49
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 50
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 51
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 52
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 53
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 54
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 55
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 56
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 57
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 58
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 59
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 60
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 61
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 62
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 63
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable159 }, // 64
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable159 }, // 65
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 66
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 67
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 68
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 69
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 70
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 71
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable159 }, // 72
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 73
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 74
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 75
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 76
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 77
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 78
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 79
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 80
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 81
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 82
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 83
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 84
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 85
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 86
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 87
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 88
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 89
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 90
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 91
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 92
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 93
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 94
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 95
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 96
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 97
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable159 }, // 98
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 99
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 100
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 101
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 102
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 103
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 104
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 105
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 106
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 107
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 108
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 109
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 110
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 111
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable159 }, // 112
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 113
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 114
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 115
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 116
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 117
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 118
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 119
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 120
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 121
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 122
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 123
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 124
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 125
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 126
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 } // 127
}; // DecodingTable359[]
const ged_ins_field_entry_t DecodingTable360[128] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 1
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 2
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 3
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 4
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 5
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 6
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 7
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 8
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 9
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 10
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 11
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 12
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 13
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 14
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 15
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 16
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 17
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 18
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 19
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 20
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 22
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 23
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 25
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 26
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 27
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 28
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 29
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 30
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 31
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 32
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 33
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 34
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 35
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 36
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 37
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 38
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 39
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 40
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 41
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 42
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 43
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 44
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 45
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 46
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 47
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 48
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 49
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 50
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 51
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 52
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 53
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 54
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 55
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 56
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 57
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 58
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 59
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 60
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 61
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 62
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 63
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable159 }, // 64
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable159 }, // 65
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 66
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 67
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 68
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 69
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 70
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 71
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable159 }, // 72
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 73
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 74
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 75
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 76
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 77
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 78
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 79
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 80
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 81
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 82
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 83
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 84
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 85
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 86
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 87
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 88
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 89
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 90
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 91
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 92
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 93
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 94
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 95
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 96
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 97
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable159 }, // 98
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 99
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 100
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 101
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 102
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 103
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 104
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 105
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 106
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 107
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 108
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 109
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 110
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 111
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable159 }, // 112
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 113
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 114
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 115
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 116
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 117
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 118
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 119
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 120
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 121
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 122
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 123
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 124
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 125
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 126
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 } // 127
}; // DecodingTable360[]
const ged_ins_field_entry_t DecodingTable361[2] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable360 }, NULL }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable140 } // 1
}; // DecodingTable361[]
const ged_ins_field_entry_t DecodingTable362[128] =
{
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 0
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 1
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 2
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 3
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 4
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 5
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 6
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 7
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 8
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 9
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 10
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 11
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 12
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 13
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 14
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 15
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 16
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 17
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 18
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 19
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 22
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 23
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 24
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 25
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 26
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 27
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 28
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 29
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 30
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 31
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 32
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 33
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 34
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 35
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 36
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 37
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 38
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 39
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 40
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 41
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 42
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 43
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 44
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 45
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 46
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 47
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 48
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 49
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 50
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 51
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 52
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 53
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 54
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 55
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 56
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 57
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 58
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 59
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 60
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 61
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 62
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 63
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 64
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 65
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 66
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 67
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 68
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 69
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 70
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 71
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 72
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 73
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 74
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 75
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 76
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 77
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 78
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 79
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 80
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 81
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 82
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 83
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 84
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 85
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 86
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 87
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 88
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 89
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 90
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 91
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 92
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 93
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 94
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 95
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 96
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable159 }, // 97
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 98
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 99
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 100
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 101
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 102
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 103
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 104
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 105
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 106
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 107
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 108
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 109
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 110
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 111
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 112
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 113
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 114
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 115
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 116
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 117
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 118
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 119
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 120
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 121
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 122
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 123
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 124
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 125
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 126
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 } // 127
}; // DecodingTable362[]
const ged_ins_field_entry_t DecodingTable363[128] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 1
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 2
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 3
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 4
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 5
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 6
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 7
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 8
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 9
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 10
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 11
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 12
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 13
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 14
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 15
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 16
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 17
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 18
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 19
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 20
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 23
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 24
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 25
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 26
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 27
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 28
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 29
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 30
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 31
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 32
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 33
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 34
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 35
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 36
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 37
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 38
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 39
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 40
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 41
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 42
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 43
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 44
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 45
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 46
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 47
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 48
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 49
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 50
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 51
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 52
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 53
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 54
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 55
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 56
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 57
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 58
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 59
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 60
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 61
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 62
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 63
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 64
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 65
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 66
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 67
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 68
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 69
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 70
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 71
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 72
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 73
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 74
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 75
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 76
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 77
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 78
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 79
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 80
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 81
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 82
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 83
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 84
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 85
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 86
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 87
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 88
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 89
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 90
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 91
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 92
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 93
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 94
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 95
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 96
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable159 }, // 97
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 98
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 99
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 100
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 101
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 102
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 103
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 104
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 105
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 106
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 107
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 108
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 109
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 110
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 111
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 112
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 113
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 114
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 115
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 116
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 117
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 118
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 119
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 120
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 121
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 122
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 123
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 124
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 125
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 126
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 } // 127
}; // DecodingTable363[]
const ged_ins_field_entry_t DecodingTable364[2] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable363 }, NULL }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable141 } // 1
}; // DecodingTable364[]
const ged_ins_field_entry_t DecodingTable365[128] =
{
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 0
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 1
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 2
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 3
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 4
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 5
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 6
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 7
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 8
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 9
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 10
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 11
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 12
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 13
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 14
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 17
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 18
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 19
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 20
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 21
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 22
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 23
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 24
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 25
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 26
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 27
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 28
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 29
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 30
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 31
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 32
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 33
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 34
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 35
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 36
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 37
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 38
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 39
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 40
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 41
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 42
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 43
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 44
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 45
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 46
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 47
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 48
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 49
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 50
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 51
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 52
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 53
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 54
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 55
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 56
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 57
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 58
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 59
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 60
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 61
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 62
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 63
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 64
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 65
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 66
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 67
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 68
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 69
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 70
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 71
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 72
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 73
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 74
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 75
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 76
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 77
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 78
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 79
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 80
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 81
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 82
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 83
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 84
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 85
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 86
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 87
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 88
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 89
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 90
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 91
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 92
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 93
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 94
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 95
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 96
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 97
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 98
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 99
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 100
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 101
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 102
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 103
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 104
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 105
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 106
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable160 }, // 107
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 108
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 109
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 110
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 111
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 112
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 113
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 114
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 115
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 116
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 117
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 118
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 119
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 120
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 121
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 122
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 123
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 124
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 125
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 126
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 } // 127
}; // DecodingTable365[]
const ged_ins_field_entry_t DecodingTable366[128] =
{
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 0
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 1
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 2
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 3
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 4
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 5
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 6
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 7
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 8
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 9
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 10
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 11
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 12
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 13
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 14
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 15
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 16
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 17
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 18
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 19
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 21
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 22
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 23
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 24
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 25
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 26
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 27
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 28
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 29
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 30
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 31
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 32
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 33
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 34
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 35
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 36
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 37
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 38
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 39
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 40
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 41
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 42
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 43
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 44
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 45
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 46
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 47
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 48
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 49
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 50
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 51
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 52
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 53
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 54
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 55
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 56
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 57
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 58
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 59
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 60
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 61
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 62
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 63
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 64
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 65
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 66
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 67
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 68
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 69
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 70
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 71
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 72
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 73
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 74
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 75
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 76
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 77
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 78
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 79
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 80
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 81
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 82
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 83
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 84
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 85
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 86
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 87
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 88
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable162 }, // 89
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable162 }, // 90
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable161 }, // 91
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 92
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 93
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 94
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 95
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 96
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 97
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 98
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 99
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 100
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 101
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 102
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 103
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 104
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 105
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 106
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 107
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 108
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 109
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 110
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 111
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 112
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 113
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 114
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 115
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 116
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 117
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 118
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 119
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 120
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 121
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 122
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 123
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 124
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 125
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 126
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 } // 127
}; // DecodingTable366[]
const ged_ins_field_entry_t DecodingTable367[2] =
{
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable366 }, NULL } // 1
}; // DecodingTable367[]
const ged_ins_field_entry_t DecodingTable368[128] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 1
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 2
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 3
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 4
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 5
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 6
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 7
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 8
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 9
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 10
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 11
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 12
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 13
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 14
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 15
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 16
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 17
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 18
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 19
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 20
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 21
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 23
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 24
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 25
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 26
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 27
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 28
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 29
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 30
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 31
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 32
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 33
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 34
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 35
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 36
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 37
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 38
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 39
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 40
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 41
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 42
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 43
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 44
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 45
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 46
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 47
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 48
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 49
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 50
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 51
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 52
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 53
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 54
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 55
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 56
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 57
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 58
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 59
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 60
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 61
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 62
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 63
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 64
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 65
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 66
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 67
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 68
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 69
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 70
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 71
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 72
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 73
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 74
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 75
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 76
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 77
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 78
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 79
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 80
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 81
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 82
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 83
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 84
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 85
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 86
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 87
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 88
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable162 }, // 89
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable162 }, // 90
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable161 }, // 91
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 92
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 93
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 94
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 95
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 96
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 97
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 98
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 99
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 100
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 101
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 102
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 103
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 104
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 105
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 106
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 107
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 108
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 109
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 110
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 111
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 112
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 113
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 114
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 115
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 116
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 117
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 118
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 119
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 120
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 121
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 122
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 123
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 124
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 125
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 126
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 } // 127
}; // DecodingTable368[]
const ged_ins_field_entry_t DecodingTable369[2] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable368 }, NULL } // 1
}; // DecodingTable369[]
const ged_ins_field_entry_t DecodingTable370[2] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable163 }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable164 } // 1
}; // DecodingTable370[]
const ged_ins_field_entry_t DecodingTable371[128] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 1
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 2
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 3
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 4
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 5
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 6
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 7
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 8
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 9
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 10
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 11
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 12
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 13
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 14
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 15
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 16
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 17
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 18
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 19
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 20
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 21
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 22
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 23
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 25
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 26
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 27
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 28
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 29
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 30
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 31
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 32
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 33
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 34
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 35
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 36
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 37
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 38
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 39
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 40
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 41
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 42
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 43
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 44
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 45
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 46
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 47
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 48
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 49
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 50
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 51
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 52
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 53
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 54
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 55
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 56
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 57
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 58
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 59
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 60
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 61
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 62
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 63
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 64
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 65
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 66
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 67
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 68
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 69
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 70
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 71
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 72
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 73
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 74
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 75
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 76
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 77
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 78
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 79
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 80
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 81
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 82
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 83
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 84
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 85
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 86
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 87
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 88
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 89
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 90
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable161 }, // 91
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 92
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 93
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 94
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 95
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 96
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 97
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 98
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 99
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 100
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 101
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 102
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 103
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 104
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 105
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 106
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 107
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 108
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 109
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 110
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 111
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 112
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 113
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 114
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 115
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 116
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 117
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 118
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 119
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 120
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 121
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 122
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 123
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 124
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 125
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 }, // 126
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable149 } // 127
}; // DecodingTable371[]
const ged_ins_field_entry_t DecodingTable372[2] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable371 }, NULL } // 1
}; // DecodingTable372[]
const ged_ins_field_entry_t DecodingTable373[128] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 1
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 2
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 3
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 4
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 5
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 6
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 7
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 8
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 9
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 10
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 11
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 12
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 13
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 14
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 15
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 16
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 17
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 18
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 19
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 20
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 21
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 22
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 23
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 25
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 26
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 27
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 28
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 29
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 30
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 31
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 32
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 33
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 34
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 35
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 36
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 37
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 38
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 39
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 40
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 41
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 42
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 43
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 44
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 45
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 46
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 47
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 48
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 49
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 50
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 51
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 52
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 53
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 54
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 55
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 56
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 57
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 58
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 59
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 60
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 61
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 62
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 63
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 64
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 65
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 66
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 67
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 68
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 69
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 70
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 71
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 72
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 73
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 74
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 75
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 76
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 77
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 78
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 79
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 80
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 81
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 82
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 83
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 84
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 85
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 86
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 87
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 88
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable370 }, NULL }, // 89
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable370 }, NULL }, // 90
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 91
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 92
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 93
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 94
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 95
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 96
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 97
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 98
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 99
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 100
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 101
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 102
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 103
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 104
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 105
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 106
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 107
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 108
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 109
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 110
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 111
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 112
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 113
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 114
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 115
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 116
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 117
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 118
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 119
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 120
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 121
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 122
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 123
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 124
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 125
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 126
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL } // 127
}; // DecodingTable373[]
const ged_ins_field_entry_t DecodingTable374[128] =
{
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 0
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 1
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 2
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 3
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 4
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 5
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 6
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 7
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 8
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 9
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 10
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 11
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 12
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 13
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 14
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 15
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 16
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 17
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 18
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 19
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 20
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 21
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 22
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 23
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 24
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 25
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 26
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 27
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 28
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 29
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 31
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 32
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 33
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 34
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 35
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 36
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 37
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 38
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 39
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 40
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 41
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 42
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 43
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 44
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 45
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 46
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 47
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 48
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 49
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 50
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 51
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 52
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 53
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 54
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 55
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 56
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 57
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 58
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 59
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 60
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 61
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 62
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 63
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 64
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 65
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 66
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 67
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 68
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 69
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 70
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 71
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 72
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 73
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 74
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 75
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 76
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 77
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 78
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 79
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 80
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 81
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 82
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 83
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 84
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 85
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 86
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 87
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 88
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable123 }, // 89
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable123 }, // 90
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 91
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 92
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 93
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 94
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 95
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 96
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 97
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 98
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 99
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 100
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 101
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 102
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 103
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 104
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 105
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 106
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 107
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 108
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 109
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 110
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 111
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 112
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 113
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 114
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 115
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 116
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 117
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 118
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 119
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 120
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 121
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 122
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 123
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 124
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 125
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 126
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 } // 127
}; // DecodingTable374[]
const ged_ins_field_entry_t DecodingTable375[128] =
{
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 0
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 1
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 2
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 3
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 4
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 5
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 6
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 7
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 8
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 9
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 10
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 11
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 12
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 13
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 14
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 15
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 16
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 17
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 18
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 19
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 20
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 21
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 22
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 23
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 24
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 25
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 26
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 27
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 28
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 29
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 30
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 31
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 32
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 33
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 34
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 35
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 36
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 39
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 40
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 41
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 42
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 43
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 44
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 45
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 46
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 47
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 48
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 49
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 50
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 51
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 52
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 53
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 54
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 55
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 56
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 57
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 58
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 59
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 60
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 61
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 62
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 63
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 64
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 65
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 66
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 67
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 68
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 69
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 70
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 71
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 72
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 73
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 74
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 75
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 76
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 77
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 78
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 79
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 80
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 81
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 82
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 83
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 84
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 85
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 86
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 87
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 88
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 91
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 92
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 93
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 94
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 95
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 96
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 97
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 98
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 99
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 100
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 101
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 102
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 103
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 104
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 105
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 106
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 108
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 109
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 110
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 111
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 112
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 113
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 114
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 115
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 116
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 117
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 118
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 119
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 120
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 121
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 122
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 123
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 124
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 125
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 126
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 } // 127
}; // DecodingTable375[]
const ged_ins_field_entry_t DecodingTable376[128] =
{
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 0
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 1
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 2
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 3
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 4
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 5
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 6
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 7
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 8
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 9
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 10
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 11
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 12
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 13
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 14
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 15
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 16
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 17
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 18
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 19
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 20
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 21
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 22
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 23
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 24
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 25
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 26
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 27
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 28
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 29
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 30
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 32
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 33
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 34
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 35
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 36
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 37
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 38
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 40
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 41
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 42
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 43
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 44
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 45
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 46
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 47
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 48
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 49
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 50
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 51
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 52
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 53
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 54
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 55
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 56
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 57
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 58
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 59
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 60
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 61
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 62
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 63
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 64
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 65
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 66
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 67
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 68
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 69
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 70
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 71
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 72
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 73
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 74
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 75
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 76
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 77
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 78
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 79
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 80
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 81
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 82
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 83
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 84
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 85
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 86
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 87
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 88
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable27 }, // 89
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable27 }, // 90
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 91
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 92
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 93
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 94
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 95
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 96
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 97
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 98
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 99
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 100
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 101
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 102
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 103
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 104
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 105
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 106
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 107
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 108
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 109
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 110
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 111
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 112
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 113
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 114
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 115
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 116
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 117
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 118
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 119
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 120
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 121
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 122
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 123
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 124
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 125
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 126
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 } // 127
}; // DecodingTable376[]
const ged_ins_field_entry_t DecodingTable377[4] =
{
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable376 }, NULL }, // 0
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable376 }, NULL }, // 1
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable377[]
const ged_ins_field_entry_t DecodingTable378[128] =
{
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 0
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 1
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 2
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 3
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 4
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 5
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 6
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 7
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 8
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 9
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 10
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 11
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 12
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 13
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 14
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 15
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 16
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 17
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 18
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 19
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 20
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 21
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 22
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 23
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 24
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 25
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 26
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 27
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 28
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 29
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 30
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 31
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 32
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 33
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 34
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 35
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 36
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 37
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 38
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 39
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 40
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 42
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 43
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 44
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 45
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 46
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 47
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 48
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 49
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 50
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 51
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 52
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 53
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 54
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 55
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 56
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 57
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 58
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 59
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 60
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 61
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 62
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 63
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 64
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 65
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 66
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 67
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 68
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 69
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 70
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 71
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 72
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 73
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 74
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 75
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 76
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 77
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 78
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 79
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 80
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 81
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 82
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 83
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 84
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 85
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 86
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 87
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 88
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable147 }, // 89
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable147 }, // 90
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 91
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 92
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 93
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 94
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 95
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 96
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 97
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 98
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 99
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 100
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 101
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 102
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 103
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 104
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 105
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 106
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 107
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 108
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 109
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 110
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 111
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 112
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 113
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 114
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 115
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 116
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 117
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 118
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 119
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 120
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 121
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 122
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 123
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 124
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 125
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 126
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 } // 127
}; // DecodingTable378[]
const ged_ins_field_entry_t DecodingTable379[4] =
{
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable378 }, NULL }, // 0
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable378 }, NULL }, // 1
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable379[]
const ged_ins_field_entry_t DecodingTable380[128] =
{
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 0
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 1
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 2
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 3
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 4
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 5
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 6
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 7
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 8
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 9
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 10
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 11
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 12
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 13
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 14
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 15
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 16
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 17
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 18
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 19
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 20
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 21
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 22
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 23
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 24
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 25
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 26
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 27
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 28
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 29
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 30
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 31
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 32
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 33
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 34
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 35
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 36
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 37
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 39
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 40
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 41
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 42
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 43
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 44
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 45
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 46
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 47
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 48
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 50
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 51
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 52
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 53
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 54
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 55
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 56
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 57
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 58
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 59
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 60
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 61
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 62
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 63
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 64
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 65
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 66
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 67
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 68
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 69
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 70
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 71
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 72
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 73
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 74
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 75
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 76
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 77
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 78
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 79
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 80
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 81
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 82
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 83
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 84
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 85
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 86
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 87
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 88
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 91
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 92
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 93
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 94
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 95
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 96
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 97
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 98
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 99
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 100
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 101
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 102
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 103
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 104
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 105
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 106
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 108
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 109
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 110
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 111
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 112
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 113
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 114
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 115
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 116
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 117
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 118
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 119
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 120
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 121
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 122
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 123
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 124
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 125
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 126
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 } // 127
}; // DecodingTable380[]
const ged_ins_field_entry_t DecodingTable381[128] =
{
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 0
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 1
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 2
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 3
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 4
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 5
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 6
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 7
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 8
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 9
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 10
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 11
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 12
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 13
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 14
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 15
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 16
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 17
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 18
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 19
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 20
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 21
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 22
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 23
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 24
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 25
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 26
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 27
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 28
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 29
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 30
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 32
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 33
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 34
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 35
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 36
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 37
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 38
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 39
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 40
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 41
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 42
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 43
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 44
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 45
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 46
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 47
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 48
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 49
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 50
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 52
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 53
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 54
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 55
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 56
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 57
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 58
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 59
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 60
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 61
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 62
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 63
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 64
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 65
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 66
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 67
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 68
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 69
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 70
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 71
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 72
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 73
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 74
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 75
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 76
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 77
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 78
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 79
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 80
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 81
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 82
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 83
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 84
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 85
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 86
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 87
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 88
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable27 }, // 89
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable27 }, // 90
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 91
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 92
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 93
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 94
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 95
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 96
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 97
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 98
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 99
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 100
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 101
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 102
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 103
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 104
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 105
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 106
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 107
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 108
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 109
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 110
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 111
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 112
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 113
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 114
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 115
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 116
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 117
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 118
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 119
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 120
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 121
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 122
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 123
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 124
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 125
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 126
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 } // 127
}; // DecodingTable381[]
const ged_ins_field_entry_t DecodingTable382[128] =
{
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 0
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 1
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 2
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 3
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 4
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 5
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 6
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 7
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 8
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 9
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 10
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 11
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 12
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 13
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 14
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 15
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 16
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 17
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 18
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 19
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 20
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 21
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 22
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 23
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 24
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 25
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 26
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 27
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 28
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 29
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 30
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 31
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 32
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 33
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 34
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 35
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 36
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 37
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 38
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 39
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 40
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 41
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 42
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 43
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 44
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 45
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 46
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 47
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 48
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 49
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 50
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 51
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 52
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 54
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 55
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 56
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 57
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 58
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 59
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 60
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 61
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 62
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 63
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 64
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 65
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 66
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 67
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 68
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 69
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 70
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 71
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 72
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 73
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 74
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 75
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 76
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 77
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 78
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 79
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 80
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 81
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 82
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 83
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 84
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 85
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 86
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 87
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 88
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable147 }, // 89
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable147 }, // 90
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 91
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 92
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 93
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 94
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 95
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 96
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 97
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 98
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 99
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 100
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 101
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 102
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 103
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 104
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 105
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 106
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 107
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 108
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 109
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 110
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 111
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 112
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 113
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 114
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 115
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 116
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 117
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 118
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 119
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 120
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 121
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 122
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 123
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 124
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 125
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 126
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 } // 127
}; // DecodingTable382[]
const ged_ins_field_entry_t DecodingTable383[128] =
{
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 0
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 1
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 2
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 3
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 4
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 5
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 6
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 7
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 8
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 9
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 10
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 11
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 12
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 13
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 14
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 15
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 16
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 17
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 18
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 19
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 20
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 21
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 22
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 23
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 24
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 25
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 26
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 27
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 28
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 29
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 30
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 31
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 32
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 33
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 34
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 35
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 36
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 37
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 39
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 40
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 41
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 42
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 43
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 44
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 45
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 46
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 47
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 48
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 49
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 50
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 51
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 52
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 53
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 54
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 56
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 57
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 58
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 59
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 60
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 61
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 62
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 63
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 64
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 65
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 66
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 67
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 68
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 69
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 70
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 71
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 72
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 73
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 74
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 75
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 76
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 77
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 78
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 79
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 80
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 81
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 82
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 83
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 84
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 85
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 86
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 87
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 88
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 91
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 92
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 93
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 94
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 95
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 96
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 97
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 98
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 99
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 100
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 101
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 102
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 103
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 104
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 105
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 106
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 108
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 109
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 110
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 111
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 112
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 113
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 114
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 115
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 116
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 117
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 118
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 119
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 120
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 121
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 122
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 123
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 124
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 125
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 126
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 } // 127
}; // DecodingTable383[]
const ged_ins_field_entry_t DecodingTable384[2] =
{
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable163 }, // 0
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable164 } // 1
}; // DecodingTable384[]
const ged_ins_field_entry_t DecodingTable385[128] =
{
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 0
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 1
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 2
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 3
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 4
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 5
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 6
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 7
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 8
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 9
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 10
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 11
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 12
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 13
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 14
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 15
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 16
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 17
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 18
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 19
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 20
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 21
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 22
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 23
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 24
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 25
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 26
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 27
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 28
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 29
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 30
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 31
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 32
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 33
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 34
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 35
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 36
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 37
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 38
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 39
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 40
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 41
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 42
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 43
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 44
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 45
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 46
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 47
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 48
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 49
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 50
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 51
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 52
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 53
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 54
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 55
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 56
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 57
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 58
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 59
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 60
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 61
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 62
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 63
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 64
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 65
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 66
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 67
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 68
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 69
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 70
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 71
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 72
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 73
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 74
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 75
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 76
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 77
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 78
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 79
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 80
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 81
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 82
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 83
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 84
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 85
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 86
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 87
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 88
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 89
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 90
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable161 }, // 91
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 92
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 94
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 95
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 96
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 97
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 98
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 99
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 100
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 101
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 102
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 103
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 104
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 105
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 106
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 107
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 108
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 109
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 110
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 111
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 112
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 113
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 114
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 115
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 116
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 117
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 118
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 119
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 120
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 121
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 122
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 123
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 124
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 125
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 }, // 126
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable149 } // 127
}; // DecodingTable385[]
const ged_ins_field_entry_t DecodingTable386[2] =
{
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable385 }, NULL } // 1
}; // DecodingTable386[]
const ged_ins_field_entry_t DecodingTable387[128] =
{
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 0
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 1
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 2
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 3
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 4
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 5
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 6
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 7
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 8
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 9
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 10
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 11
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 12
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 13
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 14
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 15
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 16
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 17
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 18
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 19
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 20
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 21
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 22
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 23
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 24
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 25
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 26
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 27
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 28
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 29
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 30
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 31
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 32
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 33
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 34
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 35
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 36
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 37
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 38
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 39
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 40
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 41
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 42
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 43
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 44
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 45
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 46
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 47
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 48
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 49
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 50
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 51
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 52
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 53
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 54
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 55
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 56
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 57
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 58
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 59
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 60
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 61
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 62
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 63
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 64
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 65
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 66
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 67
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 68
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 69
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 70
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 71
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 72
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 73
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 74
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 75
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 76
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 77
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 78
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 79
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 80
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 81
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 82
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 83
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 84
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 85
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 86
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 87
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 88
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable384 }, NULL }, // 89
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable384 }, NULL }, // 90
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 91
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 92
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 94
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 95
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 96
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 97
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 98
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 99
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 100
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 101
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 102
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 103
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 104
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 105
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 106
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 107
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 108
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 109
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 110
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 111
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 112
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 113
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 114
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 115
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 116
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 117
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 118
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 119
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 120
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 121
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 122
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 123
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 124
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 125
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 126
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL } // 127
}; // DecodingTable387[]
const ged_ins_field_entry_t DecodingTable388[128] =
{
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 0
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 1
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 2
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 3
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 4
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 5
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 6
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 7
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 8
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 9
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 10
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 11
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 12
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 13
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 14
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 15
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 16
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 17
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 18
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 19
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 20
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 21
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 22
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 23
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 24
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 25
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 26
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 27
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 28
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 29
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 30
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 32
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 33
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 34
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 35
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 36
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 37
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 38
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 39
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 40
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 41
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 42
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 43
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 44
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 45
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 46
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 47
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 48
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 49
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 50
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 51
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 52
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 53
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 54
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 55
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 56
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 57
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 58
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 59
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 60
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 61
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 62
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 63
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 64
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 65
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 66
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 67
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 68
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 69
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 70
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 71
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 72
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 73
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 74
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 75
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 76
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 77
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 78
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 79
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 80
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 81
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 82
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 83
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 84
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 85
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 86
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 87
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 88
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable27 }, // 89
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable27 }, // 90
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 91
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 92
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 93
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 94
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 95
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 96
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 97
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 98
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 99
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 100
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 101
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 102
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 103
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 104
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 105
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 106
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 107
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 109
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 110
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 111
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 112
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 113
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 114
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 115
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 116
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 117
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 118
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 119
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 120
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 121
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 122
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 123
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 124
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 125
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 126
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 } // 127
}; // DecodingTable388[]
const ged_ins_field_entry_t DecodingTable389[4] =
{
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable388 }, NULL }, // 0
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable388 }, NULL }, // 1
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable389[]
const ged_ins_field_entry_t DecodingTable390[128] =
{
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 18
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 29
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  43,  45,   1,  11,     0x3800 } */ {  0xb012d2b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable165 }, // 89
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  43,  45,   1,  11,     0x3800 } */ {  0xb012d2b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable165 }, // 90
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 127
}; // DecodingTable390[]
const ged_ins_field_entry_t DecodingTable391[2] =
{
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable39 }, RestrictionTable166 }, // 0
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable39 }, RestrictionTable167 } // 1
}; // DecodingTable391[]
const ged_ins_field_entry_t DecodingTable392[128] =
{
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 18
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 29
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable391 }, NULL }, // 89
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable391 }, NULL }, // 90
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 127
}; // DecodingTable392[]
const ged_ins_field_entry_t DecodingTable393[8] =
{
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable168 }, // 0
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable169 }, // 4
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 7
}; // DecodingTable393[]
const ged_ins_field_entry_t DecodingTable394[2] =
{
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src2DataType,                      DecodingTable393 }, NULL }, // 0
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable170 } // 1
}; // DecodingTable394[]
const ged_ins_field_entry_t DecodingTable395[128] =
{
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 18
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 29
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable394 }, NULL }, // 89
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable394 }, NULL }, // 90
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 127
}; // DecodingTable395[]
const ged_ins_field_entry_t DecodingTable396[2] =
{
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  86,  90,   2,  22,  0x7c00000 } */ { 0x16025a56, (void*)(g_uintptr_t) 0x7c00000 }, RestrictionTable166 }, // 0
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  86,  90,   2,  22,  0x7c00000 } */ { 0x16025a56, (void*)(g_uintptr_t) 0x7c00000 }, RestrictionTable167 } // 1
}; // DecodingTable396[]
const ged_ins_field_entry_t DecodingTable397[128] =
{
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 18
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 29
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable396 }, NULL }, // 89
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable396 }, NULL }, // 90
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 127
}; // DecodingTable397[]
const ged_ins_field_entry_t DecodingTable398[8] =
{
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable168 }, // 0
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable169 }, // 4
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 7
}; // DecodingTable398[]
const ged_ins_field_entry_t DecodingTable399[2] =
{
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1DataType,                      DecodingTable398 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable170 } // 1
}; // DecodingTable399[]
const ged_ins_field_entry_t DecodingTable400[128] =
{
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 18
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 29
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable399 }, NULL }, // 89
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable399 }, NULL }, // 90
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 127
}; // DecodingTable400[]
const ged_ins_field_entry_t DecodingTable401[128] =
{
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 18
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 29
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_FRAGMENTED,      8, { 2, FragmentsTable40 }, NULL }, // 107
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 127
}; // DecodingTable401[]
const ged_ins_field_entry_t DecodingTable402[2] =
{
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  39,  39,   1,   7,       0x80 } */ {  0x7012727, (void*)(g_uintptr_t)      0x80 }, NULL } // 1
}; // DecodingTable402[]
const ged_ins_field_entry_t DecodingTable403[2] =
{
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, NULL } // 1
}; // DecodingTable403[]
const ged_ins_field_entry_t DecodingTable404[2] =
{
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_ExBSO,                             DecodingTable403 }, NULL } // 1
}; // DecodingTable404[]
const ged_ins_field_entry_t DecodingTable405[2] =
{
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL } // 1
}; // DecodingTable405[]
const ged_ins_field_entry_t DecodingTable406[2] =
{
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 0
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_ExBSO,                             DecodingTable405 }, NULL } // 1
}; // DecodingTable406[]
const ged_ins_field_entry_t DecodingTable407[4] =
{
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable346 }, NULL }, // 0
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable346 }, NULL }, // 1
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable407[]
const ged_ins_field_entry_t DecodingTable408[16] =
{
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable407 }, NULL }, // 14
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable407 }, NULL } // 15
}; // DecodingTable408[]
const ged_ins_field_entry_t DecodingTable409[2] =
{
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,     32, { 4, FragmentsTable44 }, RestrictionTable174 }, // 0
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable409[]
const ged_ins_field_entry_t DecodingTable410[2] =
{
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, NULL }, // 0
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_ExBSO,                             DecodingTable403 }, NULL } // 1
}; // DecodingTable410[]
const ged_ins_field_entry_t DecodingTable411[2] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_MathFC,                            DecodingTable334 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable411[]
const ged_ins_field_entry_t DecodingTable412[2] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_MathFC,                            DecodingTable337 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable412[]
const ged_ins_field_entry_t DecodingTable413[2] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1IsImm,                         DecodingTable412 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable413[]
const ged_ins_field_entry_t DecodingTable414[2] =
{
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  67,  70,   2,   3,       0x78 } */ {  0x3024643, (void*)(g_uintptr_t)      0x78 }, RestrictionTable90 }, // 0
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable414[]
const ged_ins_field_entry_t DecodingTable415[16] =
{
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0IsImm,                         DecodingTable414 }, NULL }, // 14
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0IsImm,                         DecodingTable414 }, NULL } // 15
}; // DecodingTable415[]
const ged_ins_field_entry_t DecodingTable416[16] =
{
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  99, 102,   3,   3,       0x78 } */ {  0x3036663, (void*)(g_uintptr_t)      0x78 }, RestrictionTable90 }, // 14
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  99, 102,   3,   3,       0x78 } */ {  0x3036663, (void*)(g_uintptr_t)      0x78 }, RestrictionTable90 } // 15
}; // DecodingTable416[]
const ged_ins_field_entry_t DecodingTable417[2] =
{
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_MathFC,                            DecodingTable416 }, NULL }, // 0
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable417[]
const ged_ins_field_entry_t DecodingTable418[2] =
{
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable417 }, NULL }, // 0
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable418[]
const ged_ins_field_entry_t DecodingTable419[128] =
{
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 0
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 1
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 2
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 3
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 4
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 5
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 6
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 7
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 8
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 9
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 10
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 11
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 12
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 13
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 14
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 15
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 16
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 17
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 18
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 19
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 22
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 23
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 24
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 25
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 26
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 27
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 28
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 29
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 30
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 31
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 32
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 33
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 34
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 35
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 36
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 37
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 38
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 39
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 40
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 41
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 42
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 43
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 44
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 45
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 46
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 47
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 48
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 49
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 50
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 51
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 52
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 53
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 54
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 55
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 56
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 57
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 58
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 59
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 60
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 61
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 62
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 63
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable159 }, // 64
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable159 }, // 65
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 66
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 67
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 68
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 69
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 70
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 71
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable159 }, // 72
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 73
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 74
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 75
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 76
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 77
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 78
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 79
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 80
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 81
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 82
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable189 }, // 83
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 84
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 85
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 86
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 87
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 88
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 89
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 90
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 91
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 92
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 93
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 94
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 95
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 96
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 97
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable159 }, // 98
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 99
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 100
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 101
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 102
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 103
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 104
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 105
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 106
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 107
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 108
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 109
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 110
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 111
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable159 }, // 112
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 113
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 114
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 115
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 116
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 117
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 118
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 119
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 120
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 121
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 122
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 123
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 124
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 125
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 126
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 } // 127
}; // DecodingTable419[]
const ged_ins_field_entry_t DecodingTable420[128] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 1
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 2
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 3
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 4
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 5
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 6
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 7
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 8
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 9
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 10
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 11
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 12
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 13
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 14
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 15
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 16
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 17
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 18
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 19
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 20
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 23
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 24
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 25
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 26
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 27
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 28
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 29
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 30
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 31
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 32
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 33
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 34
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 35
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 36
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 37
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 38
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 39
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 40
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 41
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 42
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 43
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 44
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 45
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 46
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 47
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 48
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 49
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 50
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 51
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 52
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 53
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 54
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 55
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 56
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 57
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 58
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 59
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 60
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 61
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 62
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 63
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable159 }, // 64
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable159 }, // 65
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 66
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 67
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 68
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 69
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 70
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 71
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable159 }, // 72
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 73
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 74
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 75
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 76
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 77
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 78
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 79
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 80
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 81
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 82
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable189 }, // 83
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 84
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 85
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 86
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 87
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 88
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 89
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 90
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 91
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 92
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 93
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 94
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 95
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 96
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 97
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable159 }, // 98
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 99
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 100
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 101
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 102
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 103
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 104
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 105
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 106
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 107
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 108
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 109
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 110
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 111
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable159 }, // 112
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 113
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 114
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 115
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 116
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 117
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 118
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 119
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 120
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 121
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 122
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 123
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 124
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 125
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 126
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 } // 127
}; // DecodingTable420[]
const ged_ins_field_entry_t DecodingTable421[2] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable360 }, NULL }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable190 } // 1
}; // DecodingTable421[]
const ged_ins_field_entry_t DecodingTable422[128] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 1
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 2
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 3
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 4
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 5
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 6
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 7
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 8
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 9
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 10
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 11
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 12
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 13
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 14
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 15
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 16
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 17
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 18
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 19
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 20
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 21
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 22
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 23
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 25
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 26
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 27
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 28
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 29
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 30
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 31
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 32
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 33
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 34
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 35
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 36
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 37
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 38
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 39
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 40
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 41
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 42
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 43
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 44
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 45
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 46
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 47
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 48
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 49
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 50
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 51
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 52
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 53
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 54
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 55
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 56
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 57
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 58
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 59
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 60
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 61
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 62
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 63
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 64
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 65
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 66
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 67
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 68
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 69
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 70
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 71
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 72
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 73
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 74
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 75
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 76
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 77
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 78
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 79
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 80
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 81
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 82
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable189 }, // 83
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 84
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 85
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 86
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 87
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 88
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 89
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 90
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 91
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 92
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 93
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 94
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 95
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 96
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 97
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 98
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 99
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 100
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 101
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 102
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 103
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 104
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 105
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 106
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 107
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 108
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 109
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 110
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 111
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 112
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 113
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 114
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 115
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 116
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 117
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 118
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 119
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 120
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 121
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 122
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 123
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 124
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 125
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL }, // 126
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable421 }, NULL } // 127
}; // DecodingTable422[]
const ged_ins_field_entry_t DecodingTable423[2] =
{
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable45 }, NULL }, // 0
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable423[]
const ged_ins_field_entry_t DecodingTable424[2] =
{
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_FRAGMENTED,     11, { 2, FragmentsTable46 }, NULL } // 1
}; // DecodingTable424[]
const ged_ins_field_entry_t DecodingTable425[2] =
{
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 0
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable425[]
const ged_ins_field_entry_t DecodingTable426[2] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable426[]
const ged_ins_field_entry_t DecodingTable427[2] =
{
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,     11, { 2, FragmentsTable48 }, NULL } // 1
}; // DecodingTable427[]
const ged_ins_field_entry_t DecodingTable428[2] =
{
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 0
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable428[]
const ged_ins_field_entry_t DecodingTable429[2] =
{
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 0
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable429[]
const ged_ins_field_entry_t DecodingTable430[2] =
{
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 120, 121,   3,  24,  0x3000000 } */ { 0x18037978, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable16 }, // 0
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable430[]
const ged_ins_field_entry_t DecodingTable431[2] =
{
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 112, 112,   3,  16,    0x10000 } */ { 0x10037070, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable15 }, // 0
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable431[]
const ged_ins_field_entry_t DecodingTable432[2] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable432[]
const ged_ins_field_entry_t DecodingTable433[2] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable432 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable433[]
const ged_ins_field_entry_t DecodingTable434[2] =
{
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    11, /* {  98, 107,   3,   1,      0xffc } */ {  0x1036b62, (void*)(g_uintptr_t)     0xffc }, RestrictionTable103 } // 1
}; // DecodingTable434[]
const ged_ins_field_entry_t DecodingTable435[2] =
{
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     11, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable434 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable435[]
const ged_ins_field_entry_t DecodingTable436[2] =
{
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable283 }, NULL }, // 0
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable436[]
const ged_ins_field_entry_t DecodingTable437[2] =
{
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable284 }, NULL }, // 0
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable437[]
const ged_ins_field_entry_t DecodingTable438[2] =
{
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 0
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable438[]
const ged_ins_field_entry_t DecodingTable439[2] =
{
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* { 113, 115,   3,  17,    0xe0000 } */ { 0x11037371, (void*)(g_uintptr_t)   0xe0000 }, RestrictionTable28 }, // 0
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable439[]
const ged_ins_field_entry_t DecodingTable440[2] =
{
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* { 116, 118,   3,  20,   0x700000 } */ { 0x14037674, (void*)(g_uintptr_t)  0x700000 }, RestrictionTable188 }, // 0
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable440[]
const ged_ins_field_entry_t DecodingTable441[16] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable31 }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 4
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable32 }, // 5
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable30 }, // 6
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 8
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable31 }, // 9
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 10
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable31 }, // 13
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 15
}; // DecodingTable441[]
const ged_ins_field_entry_t DecodingTable442[2] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1DataType,                      DecodingTable441 }, NULL } // 1
}; // DecodingTable442[]
const ged_ins_field_entry_t DecodingTable443[128] =
{
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 0
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 1
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 2
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 3
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 4
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 5
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 6
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 7
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 8
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 9
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 10
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 11
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 12
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 13
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 14
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 15
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 16
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 17
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 18
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 19
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 22
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 23
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 24
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 25
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 26
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 27
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 28
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 29
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 30
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 31
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 32
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 33
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 34
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 35
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 36
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 37
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 38
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 39
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 40
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 41
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 42
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 43
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 44
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 45
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 46
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 47
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 48
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 49
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 50
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 51
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 52
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 53
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 54
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 55
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 56
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 57
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 58
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 59
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 60
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 61
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 62
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 63
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 64
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 65
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 66
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 67
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 68
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 69
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 70
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 71
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 72
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 73
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 74
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 75
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 76
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 77
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 78
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 79
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 80
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 81
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 82
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 83
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 84
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 85
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 86
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 87
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 88
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 89
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 90
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 91
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 92
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 93
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 94
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 95
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 96
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable191 }, // 97
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 98
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 99
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 100
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 101
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 102
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 103
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 104
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 105
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 106
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 107
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 108
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 109
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 110
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 111
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 112
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 113
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 114
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 115
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 116
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 117
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 118
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 119
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 120
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 121
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 122
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 123
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 124
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 125
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 126
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 } // 127
}; // DecodingTable443[]
const ged_ins_field_entry_t DecodingTable444[128] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 1
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 2
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 3
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 4
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 5
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 6
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 7
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 8
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 9
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 10
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 11
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 12
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 13
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 14
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 15
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 16
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 17
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 18
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 19
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 20
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 23
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 24
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 25
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 26
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 27
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 28
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 29
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 30
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 31
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 32
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 33
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 34
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 35
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 36
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 37
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 38
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 39
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 40
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 41
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 42
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 43
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 44
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 45
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 46
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 47
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 48
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 49
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 50
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 51
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 52
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 53
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 54
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 55
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 56
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 57
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 58
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 59
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 60
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 61
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 62
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 63
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 64
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 65
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 66
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 67
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 68
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 69
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 70
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 71
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 72
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 73
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 74
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 75
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 76
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 77
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 78
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 79
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 80
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 81
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 82
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 83
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 84
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 85
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 86
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 87
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 88
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 89
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 90
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 91
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 92
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 93
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 94
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 95
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 96
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable191 }, // 97
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 98
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 99
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 100
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 101
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 102
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 103
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 104
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 105
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 106
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 107
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 108
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 109
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 110
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 111
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 112
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 113
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 114
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 115
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 116
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 117
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 118
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 119
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 120
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 121
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 122
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 123
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 124
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 125
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 126
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 } // 127
}; // DecodingTable444[]
const ged_ins_field_entry_t DecodingTable445[2] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable444 }, NULL }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable192 } // 1
}; // DecodingTable445[]
const ged_ins_field_entry_t DecodingTable446[16] =
{
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 0
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 1
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 2
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 4
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 5
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 6
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 8
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 9
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 10
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 12
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 13
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 14
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 } // 15
}; // DecodingTable446[]
const ged_ins_field_entry_t DecodingTable447[2] =
{
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 0
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0DataType,                      DecodingTable446 }, NULL } // 1
}; // DecodingTable447[]
const ged_ins_field_entry_t DecodingTable448[2] =
{
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  80,  80,   2,  16,    0x10000 } */ { 0x10025050, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable15 }, // 0
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable448[]
const ged_ins_field_entry_t DecodingTable449[2] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable426 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable449[]
const ged_ins_field_entry_t DecodingTable450[2] =
{
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     11, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable427 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable450[]
const ged_ins_field_entry_t DecodingTable451[2] =
{
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0IsImm,                         DecodingTable428 }, NULL }, // 0
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable451[]
const ged_ins_field_entry_t DecodingTable452[2] =
{
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable277 }, NULL }, // 0
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable452[]
const ged_ins_field_entry_t DecodingTable453[2] =
{
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 0
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable453[]
const ged_ins_field_entry_t DecodingTable454[2] =
{
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  81,  83,   2,  17,    0xe0000 } */ { 0x11025351, (void*)(g_uintptr_t)   0xe0000 }, RestrictionTable28 }, // 0
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable454[]
const ged_ins_field_entry_t DecodingTable455[2] =
{
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  84,  86,   2,  20,   0x700000 } */ { 0x14025654, (void*)(g_uintptr_t)  0x700000 }, RestrictionTable188 }, // 0
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable455[]
const ged_ins_field_entry_t DecodingTable456[16] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable31 }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_FRAGMENTED,     64, { 2, FragmentsTable28 }, RestrictionTable87 }, // 3
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 4
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable32 }, // 5
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable30 }, // 6
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_FRAGMENTED,     64, { 2, FragmentsTable28 }, RestrictionTable73 }, // 7
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 8
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable31 }, // 9
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 }, // 10
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_FRAGMENTED,     64, { 2, FragmentsTable28 }, RestrictionTable73 }, // 11
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable31 }, // 13
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 15
}; // DecodingTable456[]
const ged_ins_field_entry_t DecodingTable457[2] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0DataType,                      DecodingTable456 }, NULL } // 1
}; // DecodingTable457[]
const ged_ins_field_entry_t DecodingTable458[16] =
{
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 0
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 1
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable146 }, // 2
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable146 }, // 3
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 4
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 5
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 6
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 7
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 8
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 9
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 10
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 11
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 12
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 }, // 13
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable146 }, // 14
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable48 } // 15
}; // DecodingTable458[]
const ged_ins_field_entry_t DecodingTable459[2] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable190 } // 1
}; // DecodingTable459[]
const ged_ins_field_entry_t DecodingTable460[16] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, RestrictionTable196 }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, RestrictionTable196 }, // 1
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, RestrictionTable196 }, // 2
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, RestrictionTable196 }, // 3
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, RestrictionTable196 }, // 4
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, RestrictionTable196 }, // 5
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, RestrictionTable196 }, // 6
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, RestrictionTable196 }, // 7
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, RestrictionTable196 }, // 8
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, RestrictionTable196 }, // 9
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, RestrictionTable196 }, // 10
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, RestrictionTable196 }, // 11
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, RestrictionTable196 }, // 12
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, RestrictionTable196 }, // 13
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, RestrictionTable195 }, // 14
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, RestrictionTable196 } // 15
}; // DecodingTable460[]
const ged_ins_field_entry_t DecodingTable461[4] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_SyncFC,                            DecodingTable460 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable461[]
const ged_ins_field_entry_t DecodingTable462[16] =
{
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, RestrictionTable0 }, // 0
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, RestrictionTable0 }, // 1
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, RestrictionTable0 }, // 2
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, RestrictionTable0 }, // 3
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, RestrictionTable0 }, // 4
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, RestrictionTable0 }, // 5
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, RestrictionTable0 }, // 6
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, RestrictionTable0 }, // 7
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, RestrictionTable0 }, // 8
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, RestrictionTable0 }, // 9
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, RestrictionTable0 }, // 10
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, RestrictionTable0 }, // 11
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, RestrictionTable0 }, // 12
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, RestrictionTable0 }, // 13
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, RestrictionTable197 }, // 14
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, RestrictionTable0 } // 15
}; // DecodingTable462[]
const ged_ins_field_entry_t DecodingTable463[4] =
{
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_SyncFC,                            DecodingTable462 }, NULL }, // 0
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 3
}; // DecodingTable463[]
const ged_ins_field_entry_t DecodingTable464[2] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, RestrictionTable29 } // 1
}; // DecodingTable464[]
const ged_ins_field_entry_t DecodingTable465[128] =
{
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 0
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 1
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 2
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 3
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 4
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 5
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 6
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 7
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 8
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 9
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 10
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 11
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 12
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 13
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 14
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 15
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 16
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 17
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 18
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 19
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 21
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 22
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 23
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 24
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 25
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 26
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 27
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 28
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 29
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 30
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 31
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 32
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 33
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 34
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 35
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 36
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 37
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 38
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 39
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 40
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 41
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 42
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 43
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 44
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 45
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 46
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 47
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 48
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 49
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 50
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 51
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 52
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 53
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 54
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 55
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 56
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 57
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 58
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 59
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 60
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 61
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 62
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 63
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 64
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 65
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 66
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 67
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 68
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 69
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 70
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 71
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 72
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 73
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 74
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 75
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 76
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 77
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 78
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 79
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 80
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 81
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 82
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 83
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 84
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 85
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 86
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 87
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 88
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 89
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 90
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable198 }, // 91
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 92
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 93
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 94
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 95
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 96
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 97
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 98
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 99
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 100
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 101
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 102
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 103
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 104
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 105
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 106
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 107
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 108
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 109
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 110
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 111
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 112
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 113
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 114
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 115
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 116
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 117
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 118
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 119
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 120
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 121
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 122
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 123
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 124
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 125
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 }, // 126
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 } // 127
}; // DecodingTable465[]
const ged_ins_field_entry_t DecodingTable466[2] =
{
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable465 }, NULL } // 1
}; // DecodingTable466[]
const ged_ins_field_entry_t DecodingTable467[128] =
{
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 0
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 1
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 2
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 3
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 4
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 5
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 6
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 7
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 8
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 9
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 10
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 11
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 12
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 13
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 14
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 15
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 16
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 17
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 18
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 19
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 20
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 22
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 23
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 24
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 25
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 26
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 27
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 28
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 29
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 30
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 31
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 32
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 33
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 34
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 35
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 36
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 37
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 38
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 39
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 40
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 41
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 42
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 43
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 44
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 45
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 46
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 47
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 48
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 49
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 50
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 51
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 52
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 53
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 54
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 55
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 56
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 57
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 58
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 59
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 60
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 61
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 62
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 63
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 64
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 65
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 66
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 67
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 68
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 69
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 70
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 71
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 72
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 73
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 74
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 75
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 76
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 77
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 78
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 79
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 80
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 81
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 82
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 83
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 84
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 85
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 86
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 87
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 88
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 89
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 90
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 91
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 92
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable12 }, // 93
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 94
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 95
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 96
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 97
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 98
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 99
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 100
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 101
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 102
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 103
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 104
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 105
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 106
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 107
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 108
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 109
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 110
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 111
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 112
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 113
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 114
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 115
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 116
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 117
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 118
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 119
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 120
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 121
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 122
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 123
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 124
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 125
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 126
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 } // 127
}; // DecodingTable467[]
const ged_ins_field_entry_t DecodingTable468[128] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 1
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 2
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 3
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 4
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 5
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 6
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 7
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 8
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 9
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 10
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 11
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 12
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 13
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 14
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 15
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 16
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 17
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 18
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 19
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 20
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 21
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 23
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 24
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 25
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 26
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 27
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 28
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 29
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 30
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 31
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 32
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 33
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 34
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 35
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 36
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 37
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 38
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 39
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 40
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 41
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 42
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 43
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 44
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 45
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 46
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 47
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 48
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 49
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 50
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 51
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 52
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 53
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 54
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 55
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 56
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 57
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 58
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 59
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 60
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 61
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 62
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 63
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 64
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 65
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 66
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 67
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 68
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 69
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 70
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 71
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 72
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 73
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 74
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 75
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 76
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 77
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 78
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 79
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 80
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 81
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 82
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 83
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 84
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 85
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 86
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 87
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 88
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 89
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 90
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable198 }, // 91
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 92
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 93
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 94
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 95
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 96
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 97
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 98
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 99
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 100
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 101
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 102
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 103
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 104
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 105
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 106
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 107
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 108
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 109
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 110
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 111
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 112
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 113
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 114
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 115
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 116
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 117
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 118
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 119
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 120
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 121
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 122
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 123
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 124
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 125
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 }, // 126
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 } // 127
}; // DecodingTable468[]
const ged_ins_field_entry_t DecodingTable469[2] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable468 }, NULL } // 1
}; // DecodingTable469[]
const ged_ins_field_entry_t DecodingTable470[128] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 1
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 2
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 3
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 4
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 5
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 6
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 7
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 8
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 9
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 10
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 11
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 12
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 13
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 14
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 15
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 16
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 17
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 18
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 19
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 20
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 21
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 22
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 23
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 25
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 26
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 27
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 28
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 29
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 30
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 31
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 32
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 33
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 34
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 35
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 36
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 37
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 38
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 39
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 40
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 41
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 42
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 43
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 44
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 45
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 46
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 47
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 48
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 49
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 50
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 51
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 52
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 53
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 54
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 55
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 56
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 57
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 58
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 59
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 60
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 61
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 62
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 63
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 64
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 65
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 66
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 67
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 68
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 69
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 70
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 71
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 72
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 73
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 74
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 75
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 76
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 77
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 78
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 79
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 80
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 81
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 82
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 83
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 84
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 85
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 86
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 87
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 88
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 89
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 90
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable198 }, // 91
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 92
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 93
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 94
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 95
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 96
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 97
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 98
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 99
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 100
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 101
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 102
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 103
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 104
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 105
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 106
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 107
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 108
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 109
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 110
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 111
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 112
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 113
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 114
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 115
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 116
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 117
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 118
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 119
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 120
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 121
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 122
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 123
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 124
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 125
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 }, // 126
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 } // 127
}; // DecodingTable470[]
const ged_ins_field_entry_t DecodingTable471[2] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable470 }, NULL } // 1
}; // DecodingTable471[]
const ged_ins_field_entry_t DecodingTable472[128] =
{
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 0
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 1
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 2
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 3
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 4
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 5
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 6
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 7
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 8
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 9
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 10
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 11
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 12
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 13
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 14
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 15
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 16
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 17
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 18
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 19
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 20
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 21
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 22
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 23
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 24
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 25
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 27
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 28
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 29
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 30
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 31
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 32
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 33
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 34
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 35
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 36
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 37
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 38
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 39
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 40
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 41
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 42
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 43
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 44
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 45
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 46
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 47
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 48
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 49
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 50
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 51
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 52
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 53
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 54
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 55
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 56
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 57
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 58
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 59
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 60
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 61
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 62
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 63
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 64
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 65
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 66
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 67
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 68
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 69
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 70
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 71
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 72
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 73
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 74
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 75
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 76
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 77
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 78
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 79
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 80
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 81
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 82
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 83
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 84
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 85
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 86
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 87
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 88
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 89
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 90
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 91
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 92
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 94
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 95
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 96
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 97
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 98
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 99
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 100
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 101
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 102
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 103
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 104
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 105
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 106
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 107
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 108
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 109
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 110
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 111
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 112
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 113
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 114
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 115
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 116
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 117
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 118
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 119
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 120
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 121
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 122
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 123
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 124
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 125
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 126
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 } // 127
}; // DecodingTable472[]
const ged_ins_field_entry_t DecodingTable473[128] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 1
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 2
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 3
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 4
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 5
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 6
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 7
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 8
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 9
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 10
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 11
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 12
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 13
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 14
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 15
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 16
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 17
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 18
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 19
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 20
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 21
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 22
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 23
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 24
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 25
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 26
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 27
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 28
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 29
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 30
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 31
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 32
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 34
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 35
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 36
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 37
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 38
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 39
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 40
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 41
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 42
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 43
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 44
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 45
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 46
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 47
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 48
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 49
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 50
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 51
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 52
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 53
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 54
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 55
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 56
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 57
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 58
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 59
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 60
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 61
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 62
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 63
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 64
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 65
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 66
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 67
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 68
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 69
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 70
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 71
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 72
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 73
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 74
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 75
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 76
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 77
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 78
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 79
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 80
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 81
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 82
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 83
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 84
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 85
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 86
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 87
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 88
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 89
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 90
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 91
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 92
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 94
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 95
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 96
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 97
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 98
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 99
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 100
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 101
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 102
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 103
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 104
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 105
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 106
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 107
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 108
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 109
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 110
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 111
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 112
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 113
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 114
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 115
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 116
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 117
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 118
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 119
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 120
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 121
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 122
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 123
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 124
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 125
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 126
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 } // 127
}; // DecodingTable473[]
const ged_ins_field_entry_t DecodingTable474[2] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Opcode,                            DecodingTable473 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable474[]
const ged_ins_field_entry_t DecodingTable475[2] =
{
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable33 }, RestrictionTable147 }, // 0
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable475[]
const ged_ins_field_entry_t DecodingTable476[128] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 1
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 2
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 3
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 4
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 5
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 6
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 7
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 8
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 9
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 10
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 11
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 12
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 13
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 14
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 15
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 16
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 17
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 18
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 19
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 20
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 21
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 22
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 23
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 24
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 25
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 26
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 27
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 28
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 29
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 30
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 31
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 32
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 33
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 34
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 35
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 36
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 37
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 38
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 39
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 40
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 41
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 42
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 43
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 44
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 46
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 47
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 48
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 49
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 50
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 51
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 52
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 53
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 54
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 55
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 56
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 57
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 58
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 59
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 60
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 61
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 62
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 63
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 64
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 65
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 66
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 67
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 68
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 69
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 70
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 71
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 72
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 73
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 74
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 75
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 76
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 77
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 78
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 79
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 80
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 81
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 82
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 83
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 84
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 85
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 86
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 87
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 88
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 89
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 90
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 91
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 92
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 94
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 95
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 96
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 97
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 98
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 99
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 100
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 101
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 102
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 103
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 104
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 105
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 106
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 107
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 108
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 109
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 110
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 111
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 112
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 113
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 114
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 115
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 116
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 117
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 118
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 119
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 120
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 121
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 122
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 123
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 124
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 125
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 126
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 } // 127
}; // DecodingTable476[]
const ged_ins_field_entry_t DecodingTable477[2] =
{
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 0
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable477[]
const ged_ins_field_entry_t DecodingTable478[2] =
{
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* { 115, 119,   3,  18,   0xf80000 } */ { 0x12037773, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 0
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable478[]
const ged_ins_field_entry_t DecodingTable479[128] =
{
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 0
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 1
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 2
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 3
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 4
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 5
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 6
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 7
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 8
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 9
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 10
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 11
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 12
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 13
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 14
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 15
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 16
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 17
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 18
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 19
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 20
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 21
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 22
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 23
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 24
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 25
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 26
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 27
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 28
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 29
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 30
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 31
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 32
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 33
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 34
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 35
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 36
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 37
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 38
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 39
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 40
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 41
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 42
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 43
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 44
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 45
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 46
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 47
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 48
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 49
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 50
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 51
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 52
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 53
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 54
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 55
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 56
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 57
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 58
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 59
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 60
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 62
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 63
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 64
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 65
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 66
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 67
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 68
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 69
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 70
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 71
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 72
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 73
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 74
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 75
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 76
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 77
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 78
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 79
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 80
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 81
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 82
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 83
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 84
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 85
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 86
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 87
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 88
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 89
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 90
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 91
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 92
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 94
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 95
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 96
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 97
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 98
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 99
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 100
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 101
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 102
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 103
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 104
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 105
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 106
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 107
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 108
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 109
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 110
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 111
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 112
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 113
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 114
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 115
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 116
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 117
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 118
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 119
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 121
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 122
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 123
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 124
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 125
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 126
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL } // 127
}; // DecodingTable479[]
const ged_ins_field_entry_t DecodingTable480[2] =
{
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 120, 127,   3,  24, 0xff000000 } */ { 0x18037f78, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 0
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable480[]
const ged_ins_field_entry_t DecodingTable481[128] =
{
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 0
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 1
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 2
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 3
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 4
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 5
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 6
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 7
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 8
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 9
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 10
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 11
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 12
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 13
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 14
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 15
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 16
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 17
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 18
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 19
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 20
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 21
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 22
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 23
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 24
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 25
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 26
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 27
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 28
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 29
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 30
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 31
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 32
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 33
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 34
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 35
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 36
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 37
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 38
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 39
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 40
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 41
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 42
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 43
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 44
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 45
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 46
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 47
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 48
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 49
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 50
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 51
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 52
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 53
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 54
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 55
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 56
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 57
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 58
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 59
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 60
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 61
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 62
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 64
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 65
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 66
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 67
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 68
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 69
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 70
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 71
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 72
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 73
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 74
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 75
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 76
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 77
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 78
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 79
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 80
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 81
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 82
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 83
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 84
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 85
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 86
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 87
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 88
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 89
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 90
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 91
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 92
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable12 }, // 93
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 94
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 95
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 96
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 97
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 98
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 99
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 100
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 101
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 102
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 103
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 104
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 105
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 106
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 107
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 108
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 109
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 110
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 111
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 112
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 113
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 114
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 115
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 116
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 117
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 118
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 119
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 120
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 121
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 122
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 123
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 124
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 125
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 126
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 } // 127
}; // DecodingTable481[]
const ged_ins_field_entry_t DecodingTable482[128] =
{
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 18
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 29
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  51,  54,   1,  19,   0x780000 } */ { 0x13013633, (void*)(g_uintptr_t)  0x780000 }, RestrictionTable90 }, // 93
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 127
}; // DecodingTable482[]
const ged_ins_field_entry_t DecodingTable483[128] =
{
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 18
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 29
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  67,  70,   2,   3,       0x78 } */ {  0x3024643, (void*)(g_uintptr_t)      0x78 }, RestrictionTable90 }, // 93
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 127
}; // DecodingTable483[]
const ged_ins_field_entry_t DecodingTable484[2] =
{
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable483 }, NULL }, // 0
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable484[]
const ged_ins_field_entry_t DecodingTable485[128] =
{
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 18
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 29
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  99, 102,   3,   3,       0x78 } */ {  0x3036663, (void*)(g_uintptr_t)      0x78 }, RestrictionTable90 }, // 93
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 127
}; // DecodingTable485[]
const ged_ins_field_entry_t DecodingTable486[128] =
{
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 18
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 29
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* { 115, 118,   3,  19,   0x780000 } */ { 0x13037673, (void*)(g_uintptr_t)  0x780000 }, RestrictionTable90 }, // 93
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 127
}; // DecodingTable486[]
const ged_ins_field_entry_t DecodingTable487[2] =
{
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable486 }, NULL }, // 0
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable487[]
const ged_ins_field_entry_t DecodingTable488[128] =
{
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 0
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 1
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 2
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 3
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 4
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 5
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 6
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 7
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 8
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 9
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 10
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 11
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 12
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 13
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 14
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 15
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 16
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 17
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 18
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 19
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 20
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 21
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 22
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 23
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 24
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 25
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 26
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 27
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 28
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 29
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 30
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 31
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 32
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 33
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 34
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 35
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 36
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 37
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 38
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 39
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 40
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 41
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 42
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 43
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 44
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 45
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 46
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 47
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 48
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 49
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 50
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 51
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 52
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 53
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 54
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 55
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 56
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 57
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 58
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 59
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 60
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 61
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 62
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 63
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 64
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 65
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 66
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 67
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 68
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 69
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 70
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 71
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 72
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 73
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 74
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 75
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 76
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 77
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 78
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 79
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 80
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 81
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 82
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 83
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 84
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 85
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 86
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 87
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 88
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 89
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 90
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable198 }, // 91
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 92
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 94
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 95
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 96
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 97
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 98
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 99
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 100
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 101
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 102
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 103
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 104
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 105
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 106
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 107
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 108
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 109
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 110
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 111
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 112
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 113
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 114
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 115
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 116
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 117
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 118
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 119
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 120
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 121
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 122
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 123
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 124
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 125
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 }, // 126
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 } // 127
}; // DecodingTable488[]
const ged_ins_field_entry_t DecodingTable489[2] =
{
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable488 }, NULL } // 1
}; // DecodingTable489[]
const ged_ins_field_entry_t DecodingTable490[2] =
{
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable321 }, NULL } // 1
}; // DecodingTable490[]
const ged_ins_field_entry_t DecodingTable491[2] =
{
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable324 }, NULL } // 1
}; // DecodingTable491[]
const ged_ins_field_entry_t DecodingTable492[2] =
{
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* { 112, 113,   3,  16,    0x30000 } */ { 0x10037170, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 0
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable492[]
const ged_ins_field_entry_t DecodingTable493[2] =
{
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable162 } // 1
}; // DecodingTable493[]
const ged_ins_field_entry_t DecodingTable494[2] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable162 } // 1
}; // DecodingTable494[]
const ged_ins_field_entry_t DecodingTable495[8] =
{
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable200 }, // 0
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable201 }, // 4
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 7
}; // DecodingTable495[]
const ged_ins_field_entry_t DecodingTable496[2] =
{
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src2DataType,                      DecodingTable495 }, NULL }, // 0
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable170 } // 1
}; // DecodingTable496[]
const ged_ins_field_entry_t DecodingTable497[8] =
{
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable200 }, // 0
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable201 }, // 4
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 7
}; // DecodingTable497[]
const ged_ins_field_entry_t DecodingTable498[2] =
{
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1DataType,                      DecodingTable497 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable170 } // 1
}; // DecodingTable498[]
const ged_ins_field_entry_t DecodingTable499[2] =
{
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable199 } // 1
}; // DecodingTable499[]
const ged_ins_field_entry_t DecodingTable500[2] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable199 } // 1
}; // DecodingTable500[]
const ged_ins_field_entry_t DecodingTable501[2] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable199 } // 1
}; // DecodingTable501[]
const ged_ins_field_entry_t DecodingTable502[2] =
{
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable199 } // 1
}; // DecodingTable502[]
const ged_ins_field_entry_t DecodingTable503[16] =
{
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable45 }, NULL }, // 0
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable45 }, NULL }, // 1
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable45 }, NULL }, // 2
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable45 }, NULL }, // 3
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable45 }, NULL }, // 4
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable45 }, NULL }, // 5
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable45 }, NULL }, // 6
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable45 }, NULL }, // 7
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable45 }, NULL }, // 8
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable45 }, NULL }, // 9
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable45 }, NULL }, // 10
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable45 }, NULL }, // 11
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable45 }, NULL }, // 12
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable45 }, NULL }, // 13
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     6, { 0 }, NULL }, // 14
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     6, { 0 }, NULL } // 15
}; // DecodingTable503[]
const ged_ins_field_entry_t DecodingTable504[16] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, NULL }, // 1
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, NULL }, // 2
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, NULL }, // 3
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, NULL }, // 4
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, NULL }, // 5
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, NULL }, // 6
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, NULL }, // 7
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, NULL }, // 8
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, NULL }, // 9
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, NULL }, // 10
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, NULL }, // 11
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, NULL }, // 12
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable47 }, NULL }, // 13
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     6, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     6, { 0 }, NULL } // 15
}; // DecodingTable504[]
const ged_ins_field_entry_t DecodingTable505[2] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_MathFC,                            DecodingTable504 }, NULL }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable505[]
const ged_ins_field_entry_t DecodingTable506[16] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 1
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 2
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 3
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 4
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 5
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 6
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 7
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 8
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 9
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 10
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 11
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 12
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 13
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     6, { 0 }, NULL }, // 14
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     6, { 0 }, NULL } // 15
}; // DecodingTable506[]
const ged_ins_field_entry_t DecodingTable507[2] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_MathFC,                            DecodingTable506 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable507[]
const ged_ins_field_entry_t DecodingTable508[2] =
{
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src1IsImm,                         DecodingTable507 }, NULL }, // 0
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable508[]
const ged_ins_field_entry_t DecodingTable509[2] =
{
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1IsImm,                         DecodingTable283 }, NULL }, // 0
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable509[]
const ged_ins_field_entry_t DecodingTable510[2] =
{
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1IsImm,                         DecodingTable430 }, NULL }, // 0
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable510[]
const ged_ins_field_entry_t DecodingTable511[2] =
{
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1IsImm,                         DecodingTable438 }, NULL }, // 0
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable511[]
const ged_ins_field_entry_t DecodingTable512[2] =
{
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1IsImm,                         DecodingTable439 }, NULL }, // 0
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable512[]
const ged_ins_field_entry_t DecodingTable513[2] =
{
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1IsImm,                         DecodingTable440 }, NULL }, // 0
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable513[]
const ged_ins_field_entry_t DecodingTable514[2] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0DataType,                      DecodingTable288 }, NULL } // 1
}; // DecodingTable514[]
const ged_ins_field_entry_t DecodingTable515[2] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1DataType,                      DecodingTable288 }, NULL } // 1
}; // DecodingTable515[]
const ged_ins_field_entry_t DecodingTable516[16] =
{
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0IsImm,                         DecodingTable514 }, NULL }, // 1
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0IsImm,                         DecodingTable514 }, NULL }, // 2
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0IsImm,                         DecodingTable514 }, NULL }, // 3
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0IsImm,                         DecodingTable514 }, NULL }, // 4
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0IsImm,                         DecodingTable514 }, NULL }, // 5
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0IsImm,                         DecodingTable514 }, NULL }, // 6
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0IsImm,                         DecodingTable514 }, NULL }, // 7
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1IsImm,                         DecodingTable515 }, NULL }, // 11
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1IsImm,                         DecodingTable515 }, NULL }, // 12
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1IsImm,                         DecodingTable515 }, NULL }, // 13
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1IsImm,                         DecodingTable515 }, NULL }, // 14
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0IsImm,                         DecodingTable514 }, NULL } // 15
}; // DecodingTable516[]
const ged_ins_field_entry_t DecodingTable517[128] =
{
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 0
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 1
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 2
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 3
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 4
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 5
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 6
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 7
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 8
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 9
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 10
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 11
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 12
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 13
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 16
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 17
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 18
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 19
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 20
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 21
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 22
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 23
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 24
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 25
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 26
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 27
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 28
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 29
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 30
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 31
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable50 }, // 32
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 33
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 34
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 35
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 36
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 37
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 38
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 39
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 40
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 41
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 42
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 43
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 44
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 45
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 46
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 47
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 48
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 49
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 50
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 51
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 52
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 53
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 54
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 55
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 56
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 57
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 58
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 59
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 60
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 61
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 62
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 63
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 64
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 65
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 66
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 67
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 68
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 69
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 70
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 71
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 72
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 73
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 74
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 75
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 76
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 77
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 78
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 79
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 80
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 81
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 82
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 83
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 84
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 85
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 86
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 87
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 88
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 89
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 90
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 91
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 92
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 93
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 94
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 95
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 96
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 97
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 98
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 99
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 100
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 101
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 102
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 103
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 104
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 105
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 106
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 107
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 108
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 109
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 110
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 111
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 112
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 113
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 114
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 115
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 116
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 117
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 118
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 119
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 120
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 121
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 122
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 123
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 124
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 125
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 126
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 } // 127
}; // DecodingTable517[]
const ged_ins_field_entry_t DecodingTable518[2] =
{
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 0
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 1
}; // DecodingTable518[]
const ged_ins_field_entry_t DecodingTable519[2] =
{
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 0
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, NULL } // 1
}; // DecodingTable519[]
const ged_ins_field_entry_t DecodingTable520[2] =
{
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable143 }, // 0
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable154 } // 1
}; // DecodingTable520[]
const ged_ins_field_entry_t DecodingTable521[128] =
{
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 0
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 1
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 2
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 3
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 4
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 5
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 6
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 7
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 8
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 9
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 10
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 11
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 12
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 13
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 14
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 15
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 16
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 17
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 18
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 19
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 22
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 23
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 24
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 25
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 26
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 27
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 28
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 29
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 30
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 31
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 32
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 33
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 34
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 35
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 36
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 37
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 38
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 39
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 40
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 41
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 42
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 43
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 44
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 45
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 46
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 47
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 48
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 49
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 50
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 51
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 52
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 53
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 54
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 55
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 56
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 57
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 58
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 59
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 60
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 61
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 62
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 63
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable159 }, // 64
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable159 }, // 65
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 66
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 67
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 68
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 69
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 70
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 71
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable159 }, // 72
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 73
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 74
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 75
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 76
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 77
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 78
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 79
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 80
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 81
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 82
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable189 }, // 83
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable205 }, // 84
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 85
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 86
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 87
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 88
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 89
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 90
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 91
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 92
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 93
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 94
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 95
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 96
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 97
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable159 }, // 98
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 99
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 100
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 101
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 102
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 103
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 104
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 105
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 106
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 107
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 108
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 109
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 110
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 111
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable159 }, // 112
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 113
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 114
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 115
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 116
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 117
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 118
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 119
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 120
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 121
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 122
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 123
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 124
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 125
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 126
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 } // 127
}; // DecodingTable521[]
const ged_ins_field_entry_t DecodingTable522[128] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 1
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 2
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 3
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 4
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 5
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 6
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 7
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 8
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 9
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 10
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 11
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 12
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 13
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 14
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 15
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 16
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 17
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 18
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 19
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 20
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 23
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 24
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 25
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 26
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 27
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 28
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 29
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 30
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 31
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 32
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 33
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 34
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 35
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 36
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 37
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 38
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 39
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 40
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 41
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 42
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 43
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 44
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 45
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 46
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 47
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 48
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 49
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 50
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 51
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 52
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 53
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 54
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 55
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 56
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 57
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 58
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 59
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 60
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 61
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 62
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 63
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable159 }, // 64
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable159 }, // 65
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 66
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 67
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 68
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 69
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 70
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 71
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable159 }, // 72
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 73
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 74
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 75
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 76
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 77
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 78
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 79
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 80
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 81
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 82
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable189 }, // 83
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable206 }, // 84
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 85
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 86
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 87
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 88
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 89
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 90
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 91
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 92
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 93
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 94
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 95
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 96
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 97
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable159 }, // 98
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 99
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 100
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 101
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 102
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 103
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 104
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 105
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 106
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 107
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 108
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 109
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 110
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 111
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable159 }, // 112
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 113
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 114
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 115
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 116
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 117
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 118
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 119
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 120
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 121
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 122
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 123
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 124
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 125
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 126
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 } // 127
}; // DecodingTable522[]
const ged_ins_field_entry_t DecodingTable523[128] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 1
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 2
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 3
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 4
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 5
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 6
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 7
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 8
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 9
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 10
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 11
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 12
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 13
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 14
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 15
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 16
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 17
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 18
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 19
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 20
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 22
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 23
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 25
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 26
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 27
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 28
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 29
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 30
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 31
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 32
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 33
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 34
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 35
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 36
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 37
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 38
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 39
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 40
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 41
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 42
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 43
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 44
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 45
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 46
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 47
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 48
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 49
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 50
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 51
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 52
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 53
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 54
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 55
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 56
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 57
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 58
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 59
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 60
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 61
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 62
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 63
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable159 }, // 64
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable159 }, // 65
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 66
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 67
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 68
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 69
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 70
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 71
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable159 }, // 72
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 73
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 74
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 75
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 76
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 77
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 78
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 79
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 80
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 81
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 82
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 83
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable206 }, // 84
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 85
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 86
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 87
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 88
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 89
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 90
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 91
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 92
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 93
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 94
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 95
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 96
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 97
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable159 }, // 98
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 99
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 100
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 101
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 102
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 103
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 104
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 105
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 106
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 107
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 108
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 109
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 110
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 111
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable159 }, // 112
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 113
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 114
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 115
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 116
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 117
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 118
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 119
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 120
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 121
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 122
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 123
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 124
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 125
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 }, // 126
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable138 } // 127
}; // DecodingTable523[]
const ged_ins_field_entry_t DecodingTable524[2] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable523 }, NULL }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable190 } // 1
}; // DecodingTable524[]
const ged_ins_field_entry_t DecodingTable525[128] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 1
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 2
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 3
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 4
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 5
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 6
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 7
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 8
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 9
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 10
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 11
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 12
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 13
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 14
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 15
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 16
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 17
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 18
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 19
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 20
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 21
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 22
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 23
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 25
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 26
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 27
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 28
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 29
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 30
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 31
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 32
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 33
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 34
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 35
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 36
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 37
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 38
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 39
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 40
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 41
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 42
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 43
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 44
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 45
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 46
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 47
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 48
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 49
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 50
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 51
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 52
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 53
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 54
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 55
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 56
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 57
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 58
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 59
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 60
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 61
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 62
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 63
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 64
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 65
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 66
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 67
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 68
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 69
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 70
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 71
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 72
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 73
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 74
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 75
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 76
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 77
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 78
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 79
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 80
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 81
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 82
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  88,  91,   2,  24,  0xf000000 } */ { 0x18025b58, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable189 }, // 83
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 84
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 85
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 86
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 87
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 88
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 89
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 90
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 91
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 92
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 93
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 94
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 95
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 96
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 97
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 98
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 99
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 100
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 101
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 102
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 103
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 104
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 105
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 106
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 107
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 108
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 109
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 110
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 111
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 112
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 113
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 114
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 115
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 116
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 117
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 118
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 119
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 120
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 121
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 122
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 123
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 124
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 125
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL }, // 126
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable524 }, NULL } // 127
}; // DecodingTable525[]
const ged_ins_field_entry_t DecodingTable526[128] =
{
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 0
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 1
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 2
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 3
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 4
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 5
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 6
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 7
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 8
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 9
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 10
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 11
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 12
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 13
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 14
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 15
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 16
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 17
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 18
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 19
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 22
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 23
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 24
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 25
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 26
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 27
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 28
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 29
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 30
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 31
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 32
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 33
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 34
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 35
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 36
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 37
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 38
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 39
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 40
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 41
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 42
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 43
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 44
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 45
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 46
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 47
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 48
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 49
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 50
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 51
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 52
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 53
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 54
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 55
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 56
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 57
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 58
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 59
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 60
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 61
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 62
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 63
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 64
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 65
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 66
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 67
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 68
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 69
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 70
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 71
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 72
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 73
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 74
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 75
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 76
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 77
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 78
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 79
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 80
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 81
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 82
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 83
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 84
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 85
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 86
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 87
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 88
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 89
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 90
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 91
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 92
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 93
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 94
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 95
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 96
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable207 }, // 97
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 98
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 99
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 100
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 101
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 102
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 103
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 104
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 105
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 106
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 107
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 108
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 109
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 110
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 111
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 112
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 113
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 114
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 115
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 116
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 117
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 118
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 119
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 120
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 121
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 122
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 123
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 124
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 125
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 126
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 } // 127
}; // DecodingTable526[]
const ged_ins_field_entry_t DecodingTable527[128] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 1
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 2
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 3
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 4
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 5
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 6
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 7
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 8
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 9
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 10
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 11
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 12
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 13
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 14
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 15
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 16
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 17
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 18
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 19
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 20
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 23
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 24
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 25
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 26
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 27
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 28
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 29
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 30
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 31
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 32
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 33
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 34
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 35
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 36
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 37
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 38
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 39
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 40
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 41
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 42
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 43
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 44
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 45
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 46
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 47
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 48
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 49
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 50
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 51
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 52
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 53
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 54
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 55
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 56
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 57
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 58
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 59
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 60
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 61
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 62
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 63
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 64
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 65
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 66
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 67
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 68
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 69
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 70
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 71
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 72
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 73
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 74
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 75
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 76
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 77
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 78
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 79
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 80
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 81
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 82
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 83
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 84
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 85
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 86
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 87
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 88
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 89
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 90
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 91
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 92
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 93
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 94
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 95
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 96
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable207 }, // 97
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 98
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 99
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 100
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 101
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 102
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 103
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 104
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 105
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 106
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 107
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 108
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 109
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 110
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 111
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 112
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 113
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 114
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 115
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 116
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 117
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 118
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 119
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 120
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 121
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 122
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 123
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 124
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 125
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 126
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 } // 127
}; // DecodingTable527[]
const ged_ins_field_entry_t DecodingTable528[2] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable527 }, NULL }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable192 } // 1
}; // DecodingTable528[]
const ged_ins_field_entry_t DecodingTable529[2] =
{
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable198 } // 1
}; // DecodingTable529[]
const ged_ins_field_entry_t DecodingTable530[2] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable198 } // 1
}; // DecodingTable530[]
const ged_ins_field_entry_t DecodingTable531[2] =
{
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable163 }, // 0
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  88,  90,   2,  24,  0x7000000 } */ { 0x18025a58, (void*)(g_uintptr_t) 0x7000000 }, RestrictionTable210 } // 1
}; // DecodingTable531[]
const ged_ins_field_entry_t DecodingTable532[2] =
{
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable163 }, // 0
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  80,  82,   2,  16,    0x70000 } */ { 0x10025250, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable210 } // 1
}; // DecodingTable532[]
const ged_ins_field_entry_t DecodingTable533[2] =
{
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable39 }, RestrictionTable166 }, // 0
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable39 }, RestrictionTable211 } // 1
}; // DecodingTable533[]
const ged_ins_field_entry_t DecodingTable534[2] =
{
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  86,  90,   2,  22,  0x7c00000 } */ { 0x16025a56, (void*)(g_uintptr_t) 0x7c00000 }, RestrictionTable166 }, // 0
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  86,  90,   2,  22,  0x7c00000 } */ { 0x16025a56, (void*)(g_uintptr_t) 0x7c00000 }, RestrictionTable211 } // 1
}; // DecodingTable534[]
const ged_ins_field_entry_t DecodingTable535[128] =
{
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 0
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 1
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 2
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 3
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 4
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 5
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 6
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 7
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 8
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 9
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 10
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 11
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 12
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 13
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 14
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 15
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 16
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 17
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 18
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 19
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 21
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 22
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 23
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 24
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 25
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 26
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 27
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 28
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 29
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 30
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 31
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 32
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 33
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 34
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 35
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 36
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 37
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 38
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 39
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 40
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 41
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 42
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 43
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 44
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 45
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 46
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 47
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 48
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 49
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 50
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 51
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 52
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 53
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 54
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 55
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 56
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 57
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 58
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 59
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 60
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 61
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 62
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 63
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 64
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 65
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 66
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 67
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 68
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 69
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 70
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 71
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 72
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 73
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 74
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 75
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 76
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 77
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 78
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 79
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 80
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 81
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 82
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 83
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 84
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 85
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 86
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 87
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 88
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 89
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 90
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable161 }, // 91
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 92
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 93
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 94
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 95
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 96
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 97
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 98
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 99
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 100
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 101
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 102
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 103
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 104
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 105
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 106
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 107
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 108
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 109
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 110
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 111
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 112
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 113
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 114
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 115
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 116
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 117
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 118
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 119
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 120
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 121
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 122
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 123
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 124
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 125
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 }, // 126
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable149 } // 127
}; // DecodingTable535[]
const ged_ins_field_entry_t DecodingTable536[2] =
{
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  36,  38,   1,   4,       0x70 } */ {  0x4012624, (void*)(g_uintptr_t)      0x70 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable535 }, NULL } // 1
}; // DecodingTable536[]
const ged_ins_field_entry_t DecodingTable537[128] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 1
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 2
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 3
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 4
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 5
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 6
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 7
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 8
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 9
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 10
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 11
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 12
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 13
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 14
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 15
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 16
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 17
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 18
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 19
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 20
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 21
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 23
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 24
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 25
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 26
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 27
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 28
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 29
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 30
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 31
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 32
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 33
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 34
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 35
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 36
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 37
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 38
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 39
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 40
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 41
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 42
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 43
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 44
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 45
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 46
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 47
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 48
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 49
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 50
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 51
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 52
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 53
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 54
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 55
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 56
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 57
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 58
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 59
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 60
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 61
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 62
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 63
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 64
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 65
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 66
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 67
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 68
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 69
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 70
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 71
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 72
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 73
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 74
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 75
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 76
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 77
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 78
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 79
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 80
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 81
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 82
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 83
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 84
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 85
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 86
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 87
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 88
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 89
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 90
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable161 }, // 91
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 92
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 93
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 94
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 95
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 96
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 97
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 98
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 99
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 100
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 101
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 102
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 103
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 104
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 105
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 106
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 107
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 108
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 109
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 110
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 111
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 112
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 113
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 114
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 115
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 116
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 117
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 118
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 119
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 120
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 121
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 122
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 123
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 124
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 125
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 }, // 126
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable149 } // 127
}; // DecodingTable537[]
const ged_ins_field_entry_t DecodingTable538[2] =
{
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, RestrictionTable148 }, // 0
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable537 }, NULL } // 1
}; // DecodingTable538[]
const ged_ins_field_entry_t DecodingTable539[128] =
{
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 0
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 1
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 2
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 3
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 4
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 5
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 6
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 7
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 8
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 9
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 10
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 11
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 12
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 13
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 14
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 15
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 16
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 17
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 18
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 19
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 20
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 21
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 22
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 23
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 24
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 25
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 26
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 27
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 28
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 29
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 30
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 31
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 32
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 33
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 34
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 35
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 36
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 39
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 40
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 41
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 42
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 43
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 44
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 45
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 46
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 47
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 48
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 49
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 50
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 51
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 52
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 53
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 54
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 55
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 56
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 57
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 58
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 59
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 60
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 61
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 62
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 63
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 64
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 65
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 66
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 67
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 68
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 69
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 70
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 71
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 72
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 73
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 74
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 75
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 76
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 77
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 78
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 79
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 80
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 81
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 82
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 83
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 84
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 85
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 86
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 87
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 88
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 89
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 90
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 91
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 92
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 93
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 94
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 95
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 96
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 97
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 98
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 99
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 100
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 101
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 102
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 103
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 104
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 105
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 106
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 108
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 109
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 110
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 111
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 112
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 113
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 114
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 115
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 116
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 117
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 118
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 119
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 120
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 121
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 122
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 123
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 124
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 125
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 126
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 } // 127
}; // DecodingTable539[]
const ged_ins_field_entry_t DecodingTable540[128] =
{
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 0
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 1
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 2
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 3
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 4
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 5
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 6
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 7
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 8
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 9
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 10
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 11
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 12
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 13
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 14
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 15
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 16
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 17
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 18
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 19
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 20
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 21
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 22
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 23
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 24
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 25
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 26
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 27
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 28
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 29
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 30
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 31
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 32
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 33
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 34
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 35
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 36
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 37
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 39
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 40
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 41
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 42
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 43
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 44
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 45
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 46
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 47
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 48
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 50
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 51
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 52
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 53
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 54
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 55
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 56
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 57
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 58
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 59
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 60
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 61
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 62
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 63
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 64
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 65
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 66
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 67
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 68
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 69
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 70
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 71
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 72
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 73
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 74
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 75
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 76
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 77
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 78
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 79
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 80
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 81
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 82
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 83
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 84
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 85
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 86
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 87
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 88
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 89
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 90
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 91
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 92
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 93
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 94
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 95
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 96
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 97
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 98
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 99
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 100
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 101
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 102
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 103
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 104
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 105
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 106
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 108
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 109
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 110
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 111
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 112
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 113
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 114
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 115
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 116
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 117
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 118
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 119
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 120
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 121
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 122
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 123
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 124
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 125
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 126
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 } // 127
}; // DecodingTable540[]
const ged_ins_field_entry_t DecodingTable541[128] =
{
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 0
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 1
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 2
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 3
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 4
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 5
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 6
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 7
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 8
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 9
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 10
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 11
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 12
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 13
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 14
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 15
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 16
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 17
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 18
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 19
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 20
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 21
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 22
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 23
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 24
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 25
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 26
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 27
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 28
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 29
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 30
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 31
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 32
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 33
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 34
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 35
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 36
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 37
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 39
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 40
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 41
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 42
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 43
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 44
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 45
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 46
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 47
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 48
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 49
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 50
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 51
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 52
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 53
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 54
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 56
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 57
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 58
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 59
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 60
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 61
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 62
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 63
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 64
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 65
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 66
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 67
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 68
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 69
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 70
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 71
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 72
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 73
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 74
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 75
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 76
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 77
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 78
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 79
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 80
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 81
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 82
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 83
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 84
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 85
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 86
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 87
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 88
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 89
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 90
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 91
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 92
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 93
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 94
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 95
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 96
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 97
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 98
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 99
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 100
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 101
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 102
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 103
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 104
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 105
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 106
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 108
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 109
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 110
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 111
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 112
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 113
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 114
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 115
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 116
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 117
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 118
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 119
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 120
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 121
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 122
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 123
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 124
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 125
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 126
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 } // 127
}; // DecodingTable541[]
const ged_ins_field_entry_t DecodingTable542[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 127, 127,   3,  31, 0x80000000 } */ { 0x1f037f7f, (void*)(g_uintptr_t)0x80000000 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable542[]
const ged_ins_field_entry_t DecodingTable543[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable85 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  10,  11,   0,  10,      0xc00 } */ {  0xa000b0a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  32,  33,   1,   0,        0x3 } */ {    0x12120, (void*)(g_uintptr_t)       0x3 }, RestrictionTable48 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  34,  36,   1,   2,       0x1c } */ {  0x2012422, (void*)(g_uintptr_t)      0x1c }, RestrictionTable13 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable43 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable2 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  48,  52,   1,  16,   0x1f0000 } */ { 0x10013430, (void*)(g_uintptr_t)  0x1f0000 }, RestrictionTable0 }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable7 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  53,  60,   1,  21, 0x1fe00000 } */ { 0x15013c35, (void*)(g_uintptr_t)0x1fe00000 }, RestrictionTable0 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DstAddrMode,                       DecodingTable10 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable44 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable47 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable48 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable50 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable45 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable46 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable51 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable52 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  89,  89,   2,  25,  0x2000000 } */ { 0x19025959, (void*)(g_uintptr_t) 0x2000000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  90,  90,   2,  26,  0x4000000 } */ { 0x1a025a5a, (void*)(g_uintptr_t) 0x4000000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src0RegFile,                       DecodingTable86 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src0RegFile,                       DecodingTable87 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable543[]
const ged_ins_field_entry_t DecodingTable544[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable85 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  10,  11,   0,  10,      0xc00 } */ {  0xa000b0a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  32,  33,   1,   0,        0x3 } */ {    0x12120, (void*)(g_uintptr_t)       0x3 }, RestrictionTable48 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  34,  36,   1,   2,       0x1c } */ {  0x2012422, (void*)(g_uintptr_t)      0x1c }, RestrictionTable13 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable43 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable2 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  48,  52,   1,  16,   0x1f0000 } */ { 0x10013430, (void*)(g_uintptr_t)  0x1f0000 }, RestrictionTable0 }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable7 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  53,  60,   1,  21, 0x1fe00000 } */ { 0x15013c35, (void*)(g_uintptr_t)0x1fe00000 }, RestrictionTable0 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DstAddrMode,                       DecodingTable10 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable44 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable47 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable48 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable50 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable45 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable46 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable51 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable52 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  89,  89,   2,  25,  0x2000000 } */ { 0x19025959, (void*)(g_uintptr_t) 0x2000000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  90,  90,   2,  26,  0x4000000 } */ { 0x1a025a5a, (void*)(g_uintptr_t) 0x4000000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src0RegFile,                       DecodingTable86 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable544[]
const ged_ins_field_entry_t DecodingTable545[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable88 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable49 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  10,  11,   0,  10,      0xc00 } */ {  0xa000b0a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable50 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  32,  33,   1,   0,        0x3 } */ {    0x12120, (void*)(g_uintptr_t)       0x3 }, RestrictionTable48 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  34,  36,   1,   2,       0x1c } */ {  0x2012422, (void*)(g_uintptr_t)      0x1c }, RestrictionTable13 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable48 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  39,  41,   1,   7,      0x380 } */ {  0x7012927, (void*)(g_uintptr_t)     0x380 }, RestrictionTable13 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  44,  46,   1,  12,     0x7000 } */ {  0xc012e2c, (void*)(g_uintptr_t)    0x7000 }, RestrictionTable51 }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable7 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  53,  60,   1,  21, 0x1fe00000 } */ { 0x15013c35, (void*)(g_uintptr_t)0x1fe00000 }, RestrictionTable52 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DstAddrMode,                       DecodingTable10 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable16 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  69,  76,   2,   5,     0x1fe0 } */ {  0x5024c45, (void*)(g_uintptr_t)    0x1fe0 }, RestrictionTable52 }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable19 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  89,  89,   2,  25,  0x2000000 } */ { 0x19025959, (void*)(g_uintptr_t) 0x2000000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  90,  90,   2,  26,  0x4000000 } */ { 0x1a025a5a, (void*)(g_uintptr_t) 0x4000000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable23 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable28 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1RegFile,                       DecodingTable31 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable33 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable35 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable25 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable37 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable39 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable40 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable89 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable545[]
const ged_ins_field_entry_t DecodingTable546[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   7,   7,   0,   7,       0x80 } */ {  0x7000707, (void*)(g_uintptr_t)      0x80 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  40,  47,   1,   8,     0xff00 } */ {  0x8012f28, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  48,  55,   1,  16,   0xff0000 } */ { 0x10013730, (void*)(g_uintptr_t)  0xff0000 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {   8,  12,   0,   8,     0x1f00 } */ {  0x8000c08, (void*)(g_uintptr_t)    0x1f00 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  13,  17,   0,  13,    0x3e000 } */ {  0xd00110d, (void*)(g_uintptr_t)   0x3e000 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  18,  22,   0,  18,   0x7c0000 } */ { 0x12001612, (void*)(g_uintptr_t)  0x7c0000 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable8 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  35,  39,   1,   3,       0xf8 } */ {  0x3012723, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable546[]
const ged_ins_field_entry_t DecodingTable547[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Opcode,                            DecodingTable84 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  10,  11,   0,  10,      0xc00 } */ {  0xa000b0a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  32,  33,   1,   0,        0x3 } */ {    0x12120, (void*)(g_uintptr_t)       0x3 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  34,  36,   1,   2,       0x1c } */ {  0x2012422, (void*)(g_uintptr_t)      0x1c }, RestrictionTable13 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable43 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable2 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable7 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DstAddrMode,                       DecodingTable10 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable16 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable19 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  89,  89,   2,  25,  0x2000000 } */ { 0x19025959, (void*)(g_uintptr_t) 0x2000000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  90,  90,   2,  26,  0x4000000 } */ { 0x1a025a5a, (void*)(g_uintptr_t) 0x4000000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  96, 111,   3,   0,     0xffff } */ {    0x36f60, (void*)(g_uintptr_t)    0xffff }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable547[]
const ged_ins_field_entry_t DecodingTable548[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  10,  11,   0,  10,      0xc00 } */ {  0xa000b0a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  32,  33,   1,   0,        0x3 } */ {    0x12120, (void*)(g_uintptr_t)       0x3 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  34,  36,   1,   2,       0x1c } */ {  0x2012422, (void*)(g_uintptr_t)      0x1c }, RestrictionTable13 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable43 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable2 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable7 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DstAddrMode,                       DecodingTable10 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  89,  89,   2,  25,  0x2000000 } */ { 0x19025959, (void*)(g_uintptr_t) 0x2000000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  90,  90,   2,  26,  0x4000000 } */ { 0x1a025a5a, (void*)(g_uintptr_t) 0x4000000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  96, 111,   3,   0,     0xffff } */ {    0x36f60, (void*)(g_uintptr_t)    0xffff }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 112, 127,   3,  16, 0xffff0000 } */ { 0x10037f70, (void*)(g_uintptr_t)0xffff0000 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable548[]
const ged_ins_field_entry_t DecodingTable549[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable549[]
const ged_ins_field_entry_t DecodingTable550[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable72 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  10,  11,   0,  10,      0xc00 } */ {  0xa000b0a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  32,  33,   1,   0,        0x3 } */ {    0x12120, (void*)(g_uintptr_t)       0x3 }, RestrictionTable38 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  34,  36,   1,   2,       0x1c } */ {  0x2012422, (void*)(g_uintptr_t)      0x1c }, RestrictionTable13 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable38 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  39,  41,   1,   7,      0x380 } */ {  0x7012927, (void*)(g_uintptr_t)     0x380 }, RestrictionTable13 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable73 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable2 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable7 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DstAddrMode,                       DecodingTable10 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable16 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable19 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  89,  89,   2,  25,  0x2000000 } */ { 0x19025959, (void*)(g_uintptr_t) 0x2000000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  90,  90,   2,  26,  0x4000000 } */ { 0x1a025a5a, (void*)(g_uintptr_t) 0x4000000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable74 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable77 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1RegFile,                       DecodingTable78 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable79 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable80 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable75 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable76 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable81 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable82 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable83 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable47 }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable550[]
const ged_ins_field_entry_t DecodingTable551[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   7,   7,   0,   7,       0x80 } */ {  0x7000707, (void*)(g_uintptr_t)      0x80 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  40,  47,   1,   8,     0xff00 } */ {  0x8012f28, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  48,  55,   1,  16,   0xff0000 } */ { 0x10013730, (void*)(g_uintptr_t)  0xff0000 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable47 }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {   8,  12,   0,   8,     0x1f00 } */ {  0x8000c08, (void*)(g_uintptr_t)    0x1f00 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  13,  17,   0,  13,    0x3e000 } */ {  0xd00110d, (void*)(g_uintptr_t)   0x3e000 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  18,  22,   0,  18,   0x7c0000 } */ { 0x12001612, (void*)(g_uintptr_t)  0x7c0000 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable8 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  35,  39,   1,   3,       0xf8 } */ {  0x3012723, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable551[]
const ged_ins_field_entry_t DecodingTable552[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable552[]
const ged_ins_field_entry_t DecodingTable553[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  10,  11,   0,  10,      0xc00 } */ {  0xa000b0a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  32,  33,   1,   0,        0x3 } */ {    0x12120, (void*)(g_uintptr_t)       0x3 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  34,  36,   1,   2,       0x1c } */ {  0x2012422, (void*)(g_uintptr_t)      0x1c }, RestrictionTable13 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable43 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable2 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable7 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DstAddrMode,                       DecodingTable10 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable44 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable47 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable48 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable50 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable45 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable46 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable51 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable52 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  89,  89,   2,  25,  0x2000000 } */ { 0x19025959, (void*)(g_uintptr_t) 0x2000000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  90,  90,   2,  26,  0x4000000 } */ { 0x1a025a5a, (void*)(g_uintptr_t) 0x4000000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable54 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable553[]
const ged_ins_field_entry_t DecodingTable554[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   7,   7,   0,   7,       0x80 } */ {  0x7000707, (void*)(g_uintptr_t)      0x80 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  40,  47,   1,   8,     0xff00 } */ {  0x8012f28, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  48,  55,   1,  16,   0xff0000 } */ { 0x10013730, (void*)(g_uintptr_t)  0xff0000 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {   8,  12,   0,   8,     0x1f00 } */ {  0x8000c08, (void*)(g_uintptr_t)    0x1f00 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  13,  17,   0,  13,    0x3e000 } */ {  0xd00110d, (void*)(g_uintptr_t)   0x3e000 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  18,  22,   0,  18,   0x7c0000 } */ { 0x12001612, (void*)(g_uintptr_t)  0x7c0000 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable8 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  35,  39,   1,   3,       0xf8 } */ {  0x3012723, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable554[]
const ged_ins_field_entry_t DecodingTable555[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  10,  11,   0,  10,      0xc00 } */ {  0xa000b0a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  32,  33,   1,   0,        0x3 } */ {    0x12120, (void*)(g_uintptr_t)       0x3 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  34,  36,   1,   2,       0x1c } */ {  0x2012422, (void*)(g_uintptr_t)      0x1c }, RestrictionTable13 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable43 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable7 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DstAddrMode,                       DecodingTable10 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable16 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable19 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  89,  89,   2,  25,  0x2000000 } */ { 0x19025959, (void*)(g_uintptr_t) 0x2000000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  90,  90,   2,  26,  0x4000000 } */ { 0x1a025a5a, (void*)(g_uintptr_t) 0x4000000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable41 }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable42 }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DescRegFile,                       DecodingTable62 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DescRegFile,                       DecodingTable64 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DescRegFile,                       DecodingTable65 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_DescRegFile,                       DecodingTable67 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DescRegFile,                       DecodingTable69 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DescRegFile,                       DecodingTable70 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     29, { GED_INS_FIELD_DescRegFile,                       DecodingTable71 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable7 }, RestrictionTable43 }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 127, 127,   3,  31, 0x80000000 } */ { 0x1f037f7f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable44 }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable555[]
const ged_ins_field_entry_t DecodingTable556[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   7,   7,   0,   7,       0x80 } */ {  0x7000707, (void*)(g_uintptr_t)      0x80 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  40,  47,   1,   8,     0xff00 } */ {  0x8012f28, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  48,  55,   1,  16,   0xff0000 } */ { 0x10013730, (void*)(g_uintptr_t)  0xff0000 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable41 }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {   8,  12,   0,   8,     0x1f00 } */ {  0x8000c08, (void*)(g_uintptr_t)    0x1f00 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  13,  17,   0,  13,    0x3e000 } */ {  0xd00110d, (void*)(g_uintptr_t)   0x3e000 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  18,  22,   0,  18,   0x7c0000 } */ { 0x12001612, (void*)(g_uintptr_t)  0x7c0000 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable8 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  35,  39,   1,   3,       0xf8 } */ {  0x3012723, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable556[]
const ged_ins_field_entry_t DecodingTable557[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  10,  11,   0,  10,      0xc00 } */ {  0xa000b0a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable35 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable55 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  53,  55,   1,  19,   0xe00000 } */ { 0x13013735, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable36 }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, RestrictionTable26 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable57 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  73,  75,   2,   7,      0xe00 } */ {  0x7024b49, (void*)(g_uintptr_t)     0xe00 }, RestrictionTable36 }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  76,  83,   2,  12,    0xff000 } */ {  0xc02534c, (void*)(g_uintptr_t)   0xff000 }, RestrictionTable26 }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  36,  37,   1,   4,       0x30 } */ {  0x4012524, (void*)(g_uintptr_t)      0x30 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable59 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable6 }, RestrictionTable36 }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  97, 104,   3,   1,      0x1fe } */ {  0x1036861, (void*)(g_uintptr_t)     0x1fe }, RestrictionTable26 }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  38,  39,   1,   6,       0xc0 } */ {  0x6012726, (void*)(g_uintptr_t)      0xc0 }, RestrictionTable16 }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  40,  41,   1,   8,      0x300 } */ {  0x8012928, (void*)(g_uintptr_t)     0x300 }, RestrictionTable16 }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable35 }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  64,  64,   2,   0,        0x1 } */ {    0x24040, (void*)(g_uintptr_t)       0x1 }, RestrictionTable37 }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  85,  85,   2,  21,   0x200000 } */ { 0x15025555, (void*)(g_uintptr_t)  0x200000 }, RestrictionTable37 }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 106, 106,   3,  10,      0x400 } */ {  0xa036a6a, (void*)(g_uintptr_t)     0x400 }, RestrictionTable37 }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable61 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* { 115, 117,   3,  17,   0x380000 } */ { 0x11037573, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable36 }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 118, 125,   3,  22, 0x3fc00000 } */ { 0x16037d76, (void*)(g_uintptr_t)0x3fc00000 }, RestrictionTable26 }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable557[]
const ged_ins_field_entry_t DecodingTable558[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  10,  11,   0,  10,      0xc00 } */ {  0xa000b0a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  32,  33,   1,   0,        0x3 } */ {    0x12120, (void*)(g_uintptr_t)       0x3 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  34,  36,   1,   2,       0x1c } */ {  0x2012422, (void*)(g_uintptr_t)      0x1c }, RestrictionTable13 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable12 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  39,  41,   1,   7,      0x380 } */ {  0x7012927, (void*)(g_uintptr_t)     0x380 }, RestrictionTable13 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable2 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable7 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DstAddrMode,                       DecodingTable10 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable16 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable19 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  89,  89,   2,  25,  0x2000000 } */ { 0x19025959, (void*)(g_uintptr_t) 0x2000000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  90,  90,   2,  26,  0x4000000 } */ { 0x1a025a5a, (void*)(g_uintptr_t) 0x4000000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable23 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable28 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1RegFile,                       DecodingTable31 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable33 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable35 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable24 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable25 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable37 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable39 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable40 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable42 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable558[]
const ged_ins_field_entry_t DecodingTable559[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  10,  11,   0,  10,      0xc00 } */ {  0xa000b0a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  32,  33,   1,   0,        0x3 } */ {    0x12120, (void*)(g_uintptr_t)       0x3 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  34,  36,   1,   2,       0x1c } */ {  0x2012422, (void*)(g_uintptr_t)      0x1c }, RestrictionTable13 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable43 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable2 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable7 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DstAddrMode,                       DecodingTable10 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable16 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable19 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  89,  89,   2,  25,  0x2000000 } */ { 0x19025959, (void*)(g_uintptr_t) 0x2000000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  90,  90,   2,  26,  0x4000000 } */ { 0x1a025a5a, (void*)(g_uintptr_t) 0x4000000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable54 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable559[]
const ged_ins_field_entry_t DecodingTable560[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable85 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  10,  11,   0,  10,      0xc00 } */ {  0xa000b0a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  32,  33,   1,   0,        0x3 } */ {    0x12120, (void*)(g_uintptr_t)       0x3 }, RestrictionTable48 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  34,  36,   1,   2,       0x1c } */ {  0x2012422, (void*)(g_uintptr_t)      0x1c }, RestrictionTable13 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable43 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable2 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable7 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  53,  60,   1,  21, 0x1fe00000 } */ { 0x15013c35, (void*)(g_uintptr_t)0x1fe00000 }, RestrictionTable52 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DstAddrMode,                       DecodingTable10 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable44 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable47 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable48 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable50 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable46 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable51 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable52 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable89 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable91 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable560[]
const ged_ins_field_entry_t DecodingTable561[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable85 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  10,  11,   0,  10,      0xc00 } */ {  0xa000b0a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  32,  33,   1,   0,        0x3 } */ {    0x12120, (void*)(g_uintptr_t)       0x3 }, RestrictionTable48 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  34,  36,   1,   2,       0x1c } */ {  0x2012422, (void*)(g_uintptr_t)      0x1c }, RestrictionTable13 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable43 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable2 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable7 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  53,  60,   1,  21, 0x1fe00000 } */ { 0x15013c35, (void*)(g_uintptr_t)0x1fe00000 }, RestrictionTable52 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DstAddrMode,                       DecodingTable10 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable44 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable47 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable48 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable50 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable46 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable51 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable52 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  89,  89,   2,  25,  0x2000000 } */ { 0x19025959, (void*)(g_uintptr_t) 0x2000000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  90,  90,   2,  26,  0x4000000 } */ { 0x1a025a5a, (void*)(g_uintptr_t) 0x4000000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable89 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable561[]
const ged_ins_field_entry_t DecodingTable562[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable88 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  10,  11,   0,  10,      0xc00 } */ {  0xa000b0a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  32,  33,   1,   0,        0x3 } */ {    0x12120, (void*)(g_uintptr_t)       0x3 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  34,  36,   1,   2,       0x1c } */ {  0x2012422, (void*)(g_uintptr_t)      0x1c }, RestrictionTable13 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable48 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  39,  41,   1,   7,      0x380 } */ {  0x7012927, (void*)(g_uintptr_t)     0x380 }, RestrictionTable13 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable2 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable7 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DstAddrMode,                       DecodingTable10 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable92 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_AccessMode,                        DecodingTable93 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  69,  76,   2,   5,     0x1fe0 } */ {  0x5024c45, (void*)(g_uintptr_t)    0x1fe0 }, RestrictionTable0 }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable75 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable94 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable95 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable76 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  89,  89,   2,  25,  0x2000000 } */ { 0x19025959, (void*)(g_uintptr_t) 0x2000000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  90,  90,   2,  26,  0x4000000 } */ { 0x1a025a5a, (void*)(g_uintptr_t) 0x4000000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable23 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable28 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1RegFile,                       DecodingTable31 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable33 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable35 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable24 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable25 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable37 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable39 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable40 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable89 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable562[]
const ged_ins_field_entry_t DecodingTable563[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  10,  11,   0,  10,      0xc00 } */ {  0xa000b0a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  32,  33,   1,   0,        0x3 } */ {    0x12120, (void*)(g_uintptr_t)       0x3 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  34,  36,   1,   2,       0x1c } */ {  0x2012422, (void*)(g_uintptr_t)      0x1c }, RestrictionTable13 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable43 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable2 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable7 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DstAddrMode,                       DecodingTable10 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable16 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable19 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  89,  89,   2,  25,  0x2000000 } */ { 0x19025959, (void*)(g_uintptr_t) 0x2000000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  90,  90,   2,  26,  0x4000000 } */ { 0x1a025a5a, (void*)(g_uintptr_t) 0x4000000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  96, 111,   3,   0,     0xffff } */ {    0x36f60, (void*)(g_uintptr_t)    0xffff }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable563[]
const ged_ins_field_entry_t DecodingTable564[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  10,  11,   0,  10,      0xc00 } */ {  0xa000b0a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  32,  33,   1,   0,        0x3 } */ {    0x12120, (void*)(g_uintptr_t)       0x3 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  34,  36,   1,   2,       0x1c } */ {  0x2012422, (void*)(g_uintptr_t)      0x1c }, RestrictionTable13 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable43 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable34 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable2 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable7 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DstAddrMode,                       DecodingTable10 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  89,  89,   2,  25,  0x2000000 } */ { 0x19025959, (void*)(g_uintptr_t) 0x2000000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  90,  90,   2,  26,  0x4000000 } */ { 0x1a025a5a, (void*)(g_uintptr_t) 0x4000000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* {  96, 111,   3,   0,     0xffff } */ {    0x36f60, (void*)(g_uintptr_t)    0xffff }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    16, /* { 112, 127,   3,  16, 0xffff0000 } */ { 0x10037f70, (void*)(g_uintptr_t)0xffff0000 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable564[]
const ged_ins_field_entry_t DecodingTable565[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  10,  11,   0,  10,      0xc00 } */ {  0xa000b0a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  32,  33,   1,   0,        0x3 } */ {    0x12120, (void*)(g_uintptr_t)       0x3 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  34,  36,   1,   2,       0x1c } */ {  0x2012422, (void*)(g_uintptr_t)      0x1c }, RestrictionTable13 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable43 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable2 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable7 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DstAddrMode,                       DecodingTable10 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_FRAGMENTED,     64, { 2, FragmentsTable9 }, RestrictionTable73 }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable565[]
const ged_ins_field_entry_t DecodingTable566[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  10,  11,   0,  10,      0xc00 } */ {  0xa000b0a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      3, { 2, FragmentsTable1 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  32,  33,   1,   0,        0x3 } */ {    0x12120, (void*)(g_uintptr_t)       0x3 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  34,  36,   1,   2,       0x1c } */ {  0x2012422, (void*)(g_uintptr_t)      0x1c }, RestrictionTable13 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable43 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable7 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DstAddrMode,                       DecodingTable10 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable16 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable19 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  89,  89,   2,  25,  0x2000000 } */ { 0x19025959, (void*)(g_uintptr_t) 0x2000000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  90,  90,   2,  26,  0x4000000 } */ { 0x1a025a5a, (void*)(g_uintptr_t) 0x4000000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable74 }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  42,  43,   1,  10,      0xc00 } */ {  0xa012b2a, (void*)(g_uintptr_t)     0xc00 }, RestrictionTable42 }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_DescRegFile,                       DecodingTable62 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DescRegFile,                       DecodingTable90 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DescRegFile,                       DecodingTable65 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     29, { GED_INS_FIELD_DescRegFile,                       DecodingTable71 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable7 }, RestrictionTable43 }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 127, 127,   3,  31, 0x80000000 } */ { 0x1f037f7f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable44 }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, RestrictionTable42 }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable566[]
const ged_ins_field_entry_t DecodingTable567[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   7,   7,   0,   7,       0x80 } */ {  0x7000707, (void*)(g_uintptr_t)      0x80 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  40,  47,   1,   8,     0xff00 } */ {  0x8012f28, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  48,  55,   1,  16,   0xff0000 } */ { 0x10013730, (void*)(g_uintptr_t)  0xff0000 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable74 }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {   8,  12,   0,   8,     0x1f00 } */ {  0x8000c08, (void*)(g_uintptr_t)    0x1f00 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  13,  17,   0,  13,    0x3e000 } */ {  0xd00110d, (void*)(g_uintptr_t)   0x3e000 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  18,  22,   0,  18,   0x7c0000 } */ { 0x12001612, (void*)(g_uintptr_t)  0x7c0000 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable8 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  35,  39,   1,   3,       0xf8 } */ {  0x3012723, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable567[]
const ged_ins_field_entry_t DecodingTable568[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable85 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable48 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  53,  60,   1,  21, 0x1fe00000 } */ { 0x15013c35, (void*)(g_uintptr_t)0x1fe00000 }, RestrictionTable52 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable44 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable47 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable112 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable113 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable46 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable51 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable52 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable89 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable91 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable568[]
const ged_ins_field_entry_t DecodingTable569[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable85 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable48 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable97 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  53,  60,   1,  21, 0x1fe00000 } */ { 0x15013c35, (void*)(g_uintptr_t)0x1fe00000 }, RestrictionTable52 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable44 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable47 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable112 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable113 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable46 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable51 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable52 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable89 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable569[]
const ged_ins_field_entry_t DecodingTable570[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable88 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable48 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  43,  46,   1,  11,     0x7800 } */ {  0xb012e2b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable91 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable97 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable92 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_AccessMode,                        DecodingTable93 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  69,  76,   2,   5,     0x1fe0 } */ {  0x5024c45, (void*)(g_uintptr_t)    0x1fe0 }, RestrictionTable0 }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable75 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable128 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable129 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable93 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable23 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable28 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1RegFile,                       DecodingTable106 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable33 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable108 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable24 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable25 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable37 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable39 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable40 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable89 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable570[]
const ged_ins_field_entry_t DecodingTable571[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable88 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable49 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable50 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable48 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable48 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  43,  46,   1,  11,     0x7800 } */ {  0xb012e2b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable84 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  91,  94,   2,  27, 0x78000000 } */ { 0x1b025e5b, (void*)(g_uintptr_t)0x78000000 }, RestrictionTable96 }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  53,  60,   1,  21, 0x1fe00000 } */ { 0x15013c35, (void*)(g_uintptr_t)0x1fe00000 }, RestrictionTable52 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable102 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  69,  76,   2,   5,     0x1fe0 } */ {  0x5024c45, (void*)(g_uintptr_t)    0x1fe0 }, RestrictionTable52 }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable23 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable28 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1RegFile,                       DecodingTable106 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable33 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable108 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable25 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable37 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable39 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable40 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable89 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable571[]
const ged_ins_field_entry_t DecodingTable572[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable97 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable102 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable572[]
const ged_ins_field_entry_t DecodingTable573[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  64,  95,   2,   0, 0xffffffff } */ {    0x25f40, (void*)(g_uintptr_t)0xffffffff }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable97 }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable573[]
const ged_ins_field_entry_t DecodingTable574[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  64,  95,   2,   0, 0xffffffff } */ {    0x25f40, (void*)(g_uintptr_t)0xffffffff }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable574[]
const ged_ins_field_entry_t DecodingTable575[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable117 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable38 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable118 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable119 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable97 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable44 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable47 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable112 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable113 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable45 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable120 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable51 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable52 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable23 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable28 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1RegFile,                       DecodingTable106 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable33 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable108 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable24 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable25 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable37 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable39 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable40 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_MathFC,                            DecodingTable121 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable92 }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable123 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable125 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable127 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable575[]
const ged_ins_field_entry_t DecodingTable576[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   7,   7,   0,   7,       0x80 } */ {  0x7000707, (void*)(g_uintptr_t)      0x80 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  40,  47,   1,   8,     0xff00 } */ {  0x8012f28, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  48,  55,   1,  16,   0xff0000 } */ { 0x10013730, (void*)(g_uintptr_t)  0xff0000 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable92 }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {   8,  12,   0,   8,     0x1f00 } */ {  0x8000c08, (void*)(g_uintptr_t)    0x1f00 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  13,  17,   0,  13,    0x3e000 } */ {  0xd00110d, (void*)(g_uintptr_t)   0x3e000 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  18,  22,   0,  18,   0x7c0000 } */ { 0x12001612, (void*)(g_uintptr_t)  0x7c0000 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable8 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  35,  39,   1,   3,       0xf8 } */ {  0x3012723, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable576[]
const ged_ins_field_entry_t DecodingTable577[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable44 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable47 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable112 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable113 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable45 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable46 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable51 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable52 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable115 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable577[]
const ged_ins_field_entry_t DecodingTable578[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable12 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable102 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable74 }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable42 }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DescRegFile,                       DecodingTable116 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DescRegFile,                       DecodingTable90 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DescRegFile,                       DecodingTable65 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     29, { GED_INS_FIELD_DescRegFile,                       DecodingTable71 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable7 }, RestrictionTable43 }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 127, 127,   3,  31, 0x80000000 } */ { 0x1f037f7f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable44 }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, RestrictionTable42 }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable578[]
const ged_ins_field_entry_t DecodingTable579[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable88 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  46,  48,   1,  14,    0x1c000 } */ {  0xe01302e, (void*)(g_uintptr_t)   0x1c000 }, RestrictionTable89 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable55 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  53,  55,   1,  19,   0xe00000 } */ { 0x13013735, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable36 }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, RestrictionTable26 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable57 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  73,  75,   2,   7,      0xe00 } */ {  0x7024b49, (void*)(g_uintptr_t)     0xe00 }, RestrictionTable36 }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  76,  83,   2,  12,    0xff000 } */ {  0xc02534c, (void*)(g_uintptr_t)   0xff000 }, RestrictionTable26 }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable59 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable6 }, RestrictionTable36 }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  97, 104,   3,   1,      0x1fe } */ {  0x1036861, (void*)(g_uintptr_t)     0x1fe }, RestrictionTable26 }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  39,  40,   1,   7,      0x180 } */ {  0x7012827, (void*)(g_uintptr_t)     0x180 }, RestrictionTable16 }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable16 }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  43,  45,   1,  11,     0x3800 } */ {  0xb012d2b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable89 }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  64,  64,   2,   0,        0x1 } */ {    0x24040, (void*)(g_uintptr_t)       0x1 }, RestrictionTable37 }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  85,  85,   2,  21,   0x200000 } */ { 0x15025555, (void*)(g_uintptr_t)  0x200000 }, RestrictionTable37 }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 106, 106,   3,  10,      0x400 } */ {  0xa036a6a, (void*)(g_uintptr_t)     0x400 }, RestrictionTable37 }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable61 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* { 115, 117,   3,  17,   0x380000 } */ { 0x11037573, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable36 }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 118, 125,   3,  22, 0x3fc00000 } */ { 0x16037d76, (void*)(g_uintptr_t)0x3fc00000 }, RestrictionTable26 }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable579[]
const ged_ins_field_entry_t DecodingTable580[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  12,  18,   0,  12,    0x7f000 } */ {  0xc00120c, (void*)(g_uintptr_t)   0x7f000 }, RestrictionTable98 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  34,  36,   1,   2,       0x1c } */ {  0x2012422, (void*)(g_uintptr_t)      0x1c }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {  43,  49,   1,  11,    0x3f800 } */ {  0xb01312b, (void*)(g_uintptr_t)   0x3f800 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  37,  39,   1,   5,       0xe0 } */ {  0x5012725, (void*)(g_uintptr_t)      0xe0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {  50,  56,   1,  18,  0x1fc0000 } */ { 0x12013832, (void*)(g_uintptr_t) 0x1fc0000 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable37 }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable37 }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable37 }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  40,  42,   1,   8,      0x700 } */ {  0x8012a28, (void*)(g_uintptr_t)     0x700 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {  57,  63,   1,  25, 0xfe000000 } */ { 0x19013f39, (void*)(g_uintptr_t)0xfe000000 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   8,   9,   0,   8,      0x300 } */ {  0x8000908, (void*)(g_uintptr_t)     0x300 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  10,  11,   0,  10,      0xc00 } */ {  0xa000b0a, (void*)(g_uintptr_t)     0xc00 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable580[]
const ged_ins_field_entry_t DecodingTable581[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable88 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  46,  48,   1,  14,    0x1c000 } */ {  0xe01302e, (void*)(g_uintptr_t)   0x1c000 }, RestrictionTable89 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable55 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  53,  55,   1,  19,   0xe00000 } */ { 0x13013735, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable36 }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, RestrictionTable26 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable57 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  73,  75,   2,   7,      0xe00 } */ {  0x7024b49, (void*)(g_uintptr_t)     0xe00 }, RestrictionTable36 }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  76,  83,   2,  12,    0xff000 } */ {  0xc02534c, (void*)(g_uintptr_t)   0xff000 }, RestrictionTable26 }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable59 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable6 }, RestrictionTable36 }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  97, 104,   3,   1,      0x1fe } */ {  0x1036861, (void*)(g_uintptr_t)     0x1fe }, RestrictionTable26 }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  39,  40,   1,   7,      0x180 } */ {  0x7012827, (void*)(g_uintptr_t)     0x180 }, RestrictionTable16 }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable16 }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  43,  45,   1,  11,     0x3800 } */ {  0xb012d2b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable89 }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  64,  64,   2,   0,        0x1 } */ {    0x24040, (void*)(g_uintptr_t)       0x1 }, RestrictionTable37 }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  85,  85,   2,  21,   0x200000 } */ { 0x15025555, (void*)(g_uintptr_t)  0x200000 }, RestrictionTable37 }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 106, 106,   3,  10,      0x400 } */ {  0xa036a6a, (void*)(g_uintptr_t)     0x400 }, RestrictionTable37 }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable61 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* { 115, 117,   3,  17,   0x380000 } */ { 0x11037573, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable36 }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 118, 125,   3,  22, 0x3fc00000 } */ { 0x16037d76, (void*)(g_uintptr_t)0x3fc00000 }, RestrictionTable26 }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  49,  52,   1,  17,   0x1e0000 } */ { 0x11013431, (void*)(g_uintptr_t)  0x1e0000 }, RestrictionTable90 }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  65,  68,   2,   1,       0x1e } */ {  0x1024441, (void*)(g_uintptr_t)      0x1e }, RestrictionTable90 }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  86,  89,   2,  22,  0x3c00000 } */ { 0x16025956, (void*)(g_uintptr_t) 0x3c00000 }, RestrictionTable90 }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* { 107, 110,   3,  11,     0x7800 } */ {  0xb036e6b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable90 }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable581[]
const ged_ins_field_entry_t DecodingTable582[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable12 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  43,  46,   1,  11,     0x7800 } */ {  0xb012e2b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable84 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable97 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable102 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable23 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable28 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1RegFile,                       DecodingTable106 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable33 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable108 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable24 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable25 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable37 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable39 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable40 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable110 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable582[]
const ged_ins_field_entry_t DecodingTable583[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable44 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable47 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable112 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable113 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable45 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable46 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable51 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable52 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable115 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable583[]
const ged_ins_field_entry_t DecodingTable584[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable12 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable102 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable74 }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable42 }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DescRegFile,                       DecodingTable116 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DescRegFile,                       DecodingTable90 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DescRegFile,                       DecodingTable65 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     31, { GED_INS_FIELD_DescRegFile,                       DecodingTable132 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,      6, { 2, FragmentsTable7 }, RestrictionTable43 }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 127, 127,   3,  31, 0x80000000 } */ { 0x1f037f7f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable44 }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, RestrictionTable42 }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable584[]
const ged_ins_field_entry_t DecodingTable585[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable88 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  46,  48,   1,  14,    0x1c000 } */ {  0xe01302e, (void*)(g_uintptr_t)   0x1c000 }, RestrictionTable89 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_SrcDataType,                       DecodingTable130 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable55 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  53,  55,   1,  19,   0xe00000 } */ { 0x13013735, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable36 }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, RestrictionTable26 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable57 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable16 }, RestrictionTable103 }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  76,  83,   2,  12,    0xff000 } */ {  0xc02534c, (void*)(g_uintptr_t)   0xff000 }, RestrictionTable26 }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable59 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 3, FragmentsTable17 }, RestrictionTable103 }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  97, 104,   3,   1,      0x1fe } */ {  0x1036861, (void*)(g_uintptr_t)     0x1fe }, RestrictionTable26 }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  39,  40,   1,   7,      0x180 } */ {  0x7012827, (void*)(g_uintptr_t)     0x180 }, RestrictionTable16 }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable16 }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  43,  45,   1,  11,     0x3800 } */ {  0xb012d2b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable89 }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  64,  64,   2,   0,        0x1 } */ {    0x24040, (void*)(g_uintptr_t)       0x1 }, RestrictionTable37 }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  85,  85,   2,  21,   0x200000 } */ { 0x15025555, (void*)(g_uintptr_t)  0x200000 }, RestrictionTable37 }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 106, 106,   3,  10,      0x400 } */ {  0xa036a6a, (void*)(g_uintptr_t)     0x400 }, RestrictionTable37 }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable61 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable18 }, RestrictionTable103 }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 118, 125,   3,  22, 0x3fc00000 } */ { 0x16037d76, (void*)(g_uintptr_t)0x3fc00000 }, RestrictionTable26 }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_SrcDataType,                       DecodingTable131 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable585[]
const ged_ins_field_entry_t DecodingTable586[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable88 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  46,  48,   1,  14,    0x1c000 } */ {  0xe01302e, (void*)(g_uintptr_t)   0x1c000 }, RestrictionTable89 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_SrcDataType,                       DecodingTable130 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable55 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  53,  55,   1,  19,   0xe00000 } */ { 0x13013735, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable36 }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, RestrictionTable26 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable57 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable16 }, RestrictionTable103 }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  76,  83,   2,  12,    0xff000 } */ {  0xc02534c, (void*)(g_uintptr_t)   0xff000 }, RestrictionTable26 }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable59 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 3, FragmentsTable17 }, RestrictionTable103 }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  97, 104,   3,   1,      0x1fe } */ {  0x1036861, (void*)(g_uintptr_t)     0x1fe }, RestrictionTable26 }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  39,  40,   1,   7,      0x180 } */ {  0x7012827, (void*)(g_uintptr_t)     0x180 }, RestrictionTable16 }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable16 }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  43,  45,   1,  11,     0x3800 } */ {  0xb012d2b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable89 }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  64,  64,   2,   0,        0x1 } */ {    0x24040, (void*)(g_uintptr_t)       0x1 }, RestrictionTable37 }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  85,  85,   2,  21,   0x200000 } */ { 0x15025555, (void*)(g_uintptr_t)  0x200000 }, RestrictionTable37 }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 106, 106,   3,  10,      0x400 } */ {  0xa036a6a, (void*)(g_uintptr_t)     0x400 }, RestrictionTable37 }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable61 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable18 }, RestrictionTable103 }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 118, 125,   3,  22, 0x3fc00000 } */ { 0x16037d76, (void*)(g_uintptr_t)0x3fc00000 }, RestrictionTable26 }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  49,  52,   1,  17,   0x1e0000 } */ { 0x11013431, (void*)(g_uintptr_t)  0x1e0000 }, RestrictionTable90 }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  65,  68,   2,   1,       0x1e } */ {  0x1024441, (void*)(g_uintptr_t)      0x1e }, RestrictionTable90 }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  86,  89,   2,  22,  0x3c00000 } */ { 0x16025956, (void*)(g_uintptr_t) 0x3c00000 }, RestrictionTable90 }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* { 107, 110,   3,  11,     0x7800 } */ {  0xb036e6b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable90 }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_SrcDataType,                       DecodingTable131 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable586[]
const ged_ins_field_entry_t DecodingTable587[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable38 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable133 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable105 }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable42 }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DescRegFile,                       DecodingTable90 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DescRegFile,                       DecodingTable65 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     31, { GED_INS_FIELD_DescRegFile,                       DecodingTable132 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,     32, { 6, FragmentsTable20 }, RestrictionTable107 }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 127, 127,   3,  31, 0x80000000 } */ { 0x1f037f7f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable44 }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, RestrictionTable42 }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable106 }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_FRAGMENTED,     16, { 4, FragmentsTable19 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable587[]
const ged_ins_field_entry_t DecodingTable588[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable5 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 0 }, RestrictionTable91 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  36,  36,   1,   4,       0x10 } */ {  0x4012424, (void*)(g_uintptr_t)      0x10 }, RestrictionTable108 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 0 }, RestrictionTable91 }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable134 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable135 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable136 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable105 }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  77,  77,   2,  13,     0x2000 } */ {  0xd024d4d, (void*)(g_uintptr_t)    0x2000 }, RestrictionTable109 }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DescRegFile,                       DecodingTable137 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DescRegFile,                       DecodingTable138 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     31, { GED_INS_FIELD_DescRegFile,                       DecodingTable139 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable140 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 127, 127,   3,  31, 0x80000000 } */ { 0x1f037f7f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable44 }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  61,  61,   1,  29, 0x20000000 } */ { 0x1d013d3d, (void*)(g_uintptr_t)0x20000000 }, RestrictionTable109 }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable106 }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable141 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  64,  67,   2,   0,        0xf } */ {    0x24340, (void*)(g_uintptr_t)       0xf }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable142 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable143 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 114, 114,   3,  18,    0x40000 } */ { 0x12037272, (void*)(g_uintptr_t)   0x40000 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     12, { GED_INS_FIELD_MsgDescCategory,                   DecodingTable144 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MsgDescCategory,                   DecodingTable145 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_MsgDescCategory,                   DecodingTable146 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_MsgDescCategory,                   DecodingTable147 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_MsgDescCategory,                   DecodingTable148 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable588[]
const ged_ins_field_entry_t DecodingTable589[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable85 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable48 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  53,  60,   1,  21, 0x1fe00000 } */ { 0x15013c35, (void*)(g_uintptr_t)0x1fe00000 }, RestrictionTable52 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable44 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable47 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable112 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable113 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable46 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable51 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable52 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable89 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable91 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable589[]
const ged_ins_field_entry_t DecodingTable590[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable85 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable48 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable97 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  53,  60,   1,  21, 0x1fe00000 } */ { 0x15013c35, (void*)(g_uintptr_t)0x1fe00000 }, RestrictionTable52 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable44 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable47 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable112 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable113 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable46 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable51 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable52 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable89 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable590[]
const ged_ins_field_entry_t DecodingTable591[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable88 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable48 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  43,  46,   1,  11,     0x7800 } */ {  0xb012e2b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable91 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable97 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable92 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_AccessMode,                        DecodingTable93 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  69,  76,   2,   5,     0x1fe0 } */ {  0x5024c45, (void*)(g_uintptr_t)    0x1fe0 }, RestrictionTable0 }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable75 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable128 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable129 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable93 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable149 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable152 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1RegFile,                       DecodingTable153 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable154 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable155 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable150 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable151 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable156 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable157 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable158 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable89 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable591[]
const ged_ins_field_entry_t DecodingTable592[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable88 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable49 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable50 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable48 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable48 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  43,  46,   1,  11,     0x7800 } */ {  0xb012e2b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable84 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  91,  94,   2,  27, 0x78000000 } */ { 0x1b025e5b, (void*)(g_uintptr_t)0x78000000 }, RestrictionTable96 }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  53,  60,   1,  21, 0x1fe00000 } */ { 0x15013c35, (void*)(g_uintptr_t)0x1fe00000 }, RestrictionTable52 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable102 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  69,  76,   2,   5,     0x1fe0 } */ {  0x5024c45, (void*)(g_uintptr_t)    0x1fe0 }, RestrictionTable52 }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable149 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable152 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1RegFile,                       DecodingTable153 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable154 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable155 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable151 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable156 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable157 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable158 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable89 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable592[]
const ged_ins_field_entry_t DecodingTable593[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable97 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable102 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable593[]
const ged_ins_field_entry_t DecodingTable594[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  64,  95,   2,   0, 0xffffffff } */ {    0x25f40, (void*)(g_uintptr_t)0xffffffff }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable97 }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable594[]
const ged_ins_field_entry_t DecodingTable595[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  64,  95,   2,   0, 0xffffffff } */ {    0x25f40, (void*)(g_uintptr_t)0xffffffff }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable595[]
const ged_ins_field_entry_t DecodingTable596[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable117 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable38 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable118 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable119 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable97 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable44 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable47 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable112 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable113 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable45 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable120 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable51 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable52 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable149 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable152 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1RegFile,                       DecodingTable153 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable154 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable155 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable150 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable151 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable156 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable157 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable158 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_MathFC,                            DecodingTable121 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable92 }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable123 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable125 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable127 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable596[]
const ged_ins_field_entry_t DecodingTable597[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable12 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  43,  46,   1,  11,     0x7800 } */ {  0xb012e2b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable84 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable42 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable97 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable102 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable149 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable159 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable160 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable150 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable161 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable156 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable157 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable158 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable110 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable597[]
const ged_ins_field_entry_t DecodingTable598[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable44 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable47 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable112 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable113 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable45 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable46 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable51 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable52 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable115 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable598[]
const ged_ins_field_entry_t DecodingTable599[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable130 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable38 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable133 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable105 }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable42 }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DescRegFile,                       DecodingTable90 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DescRegFile,                       DecodingTable65 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     31, { GED_INS_FIELD_DescRegFile,                       DecodingTable132 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,     32, { 6, FragmentsTable20 }, RestrictionTable107 }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 127, 127,   3,  31, 0x80000000 } */ { 0x1f037f7f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable44 }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, RestrictionTable42 }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable106 }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_FRAGMENTED,     16, { 4, FragmentsTable19 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable599[]
const ged_ins_field_entry_t DecodingTable600[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable130 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 0 }, RestrictionTable91 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  36,  36,   1,   4,       0x10 } */ {  0x4012424, (void*)(g_uintptr_t)      0x10 }, RestrictionTable108 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 0 }, RestrictionTable91 }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable210 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable134 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable211 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable135 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable136 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable105 }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  77,  77,   2,  13,     0x2000 } */ {  0xd024d4d, (void*)(g_uintptr_t)    0x2000 }, RestrictionTable109 }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DescRegFile,                       DecodingTable137 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DescRegFile,                       DecodingTable138 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    31, /* {  96, 126,   3,   0, 0x7fffffff } */ {    0x37e60, (void*)(g_uintptr_t)0x7fffffff }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable140 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 127, 127,   3,  31, 0x80000000 } */ { 0x1f037f7f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable44 }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  61,  61,   1,  29, 0x20000000 } */ { 0x1d013d3d, (void*)(g_uintptr_t)0x20000000 }, RestrictionTable109 }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable106 }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable141 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  64,  67,   2,   0,        0xf } */ {    0x24340, (void*)(g_uintptr_t)       0xf }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable142 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable143 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable600[]
const ged_ins_field_entry_t DecodingTable601[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable162 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable191 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_AccessMode,                        DecodingTable163 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable193 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_AccessMode,                        DecodingTable164 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable195 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable55 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_AccessMode,                        DecodingTable165 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstRegFile,                        DecodingTable166 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_AccessMode,                        DecodingTable167 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable57 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_AccessMode,                        DecodingTable169 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable171 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable173 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable175 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable59 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_AccessMode,                        DecodingTable176 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable177 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  39,  40,   1,   7,      0x180 } */ {  0x7012827, (void*)(g_uintptr_t)     0x180 }, RestrictionTable16 }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable178 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable179 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable16 }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable180 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_AccessMode,                        DecodingTable181 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_AccessMode,                        DecodingTable182 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_AccessMode,                        DecodingTable183 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable61 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_AccessMode,                        DecodingTable186 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable188 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_AccessMode,                        DecodingTable184 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable197 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_AccessMode,                        DecodingTable189 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_AccessMode,                        DecodingTable201 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_AccessMode,                        DecodingTable205 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable207 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable601[]
const ged_ins_field_entry_t DecodingTable602[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable88 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_AccessMode,                        DecodingTable208 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable191 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_AccessMode,                        DecodingTable163 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable193 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_AccessMode,                        DecodingTable164 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable195 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable55 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_AccessMode,                        DecodingTable165 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstRegFile,                        DecodingTable209 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_AccessMode,                        DecodingTable167 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable57 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_AccessMode,                        DecodingTable169 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable171 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable173 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable175 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable59 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_AccessMode,                        DecodingTable176 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable177 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  39,  40,   1,   7,      0x180 } */ {  0x7012827, (void*)(g_uintptr_t)     0x180 }, RestrictionTable16 }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable178 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable179 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable16 }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable180 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_AccessMode,                        DecodingTable181 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_AccessMode,                        DecodingTable182 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_AccessMode,                        DecodingTable183 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable61 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_AccessMode,                        DecodingTable186 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable188 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_AccessMode,                        DecodingTable184 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  49,  52,   1,  17,   0x1e0000 } */ { 0x11013431, (void*)(g_uintptr_t)  0x1e0000 }, RestrictionTable90 }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  65,  68,   2,   1,       0x1e } */ {  0x1024441, (void*)(g_uintptr_t)      0x1e }, RestrictionTable90 }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  86,  89,   2,  22,  0x3c00000 } */ { 0x16025956, (void*)(g_uintptr_t) 0x3c00000 }, RestrictionTable90 }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* { 107, 110,   3,  11,     0x7800 } */ {  0xb036e6b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable90 }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable197 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_AccessMode,                        DecodingTable189 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_AccessMode,                        DecodingTable201 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_AccessMode,                        DecodingTable205 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable207 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable602[]
const ged_ins_field_entry_t DecodingTable603[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable12 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  43,  46,   1,  11,     0x7800 } */ {  0xb012e2b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable84 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable97 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_AccessMode,                        DecodingTable12 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable14 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable102 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable20 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_AccessMode,                        DecodingTable21 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable149 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable152 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1RegFile,                       DecodingTable153 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable154 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable155 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable150 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable151 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable156 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable157 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable158 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable110 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable603[]
const ged_ins_field_entry_t DecodingTable604[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable1 }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable1 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable84 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable111 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_AccessMode,                        DecodingTable3 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable5 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable99 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_AccessMode,                        DecodingTable11 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable44 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable47 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable112 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable113 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable45 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable46 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable51 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable52 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable115 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable604[]
const ged_ins_field_entry_t DecodingTable605[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  16,  19,   0,  16,    0xf0000 } */ { 0x10001310, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable48 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable132 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable225 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable213 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable214 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  53,  60,   1,  21, 0x1fe00000 } */ { 0x15013c35, (void*)(g_uintptr_t)0x1fe00000 }, RestrictionTable52 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  61,  62,   1,  29, 0x60000000 } */ { 0x1d013e3d, (void*)(g_uintptr_t)0x60000000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable226 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable227 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable113 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable46 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable228 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable229 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable89 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable91 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable605[]
const ged_ins_field_entry_t DecodingTable606[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  16,  19,   0,  16,    0xf0000 } */ { 0x10001310, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable48 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable132 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable225 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable212 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable213 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable214 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  53,  60,   1,  21, 0x1fe00000 } */ { 0x15013c35, (void*)(g_uintptr_t)0x1fe00000 }, RestrictionTable52 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  61,  62,   1,  29, 0x60000000 } */ { 0x1d013e3d, (void*)(g_uintptr_t)0x60000000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable226 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable227 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable113 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable46 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable228 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable229 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable89 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable606[]
const ged_ins_field_entry_t DecodingTable607[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  16,  19,   0,  16,    0xf0000 } */ { 0x10001310, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable132 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable48 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  43,  46,   1,  11,     0x7800 } */ {  0xb012e2b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable91 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable212 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable213 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable214 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  61,  62,   1,  29, 0x60000000 } */ { 0x1d013e3d, (void*)(g_uintptr_t)0x60000000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  64,  68,   2,   0,       0x1f } */ {    0x24440, (void*)(g_uintptr_t)      0x1f }, RestrictionTable0 }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  69,  76,   2,   5,     0x1fe0 } */ {  0x5024c45, (void*)(g_uintptr_t)    0x1fe0 }, RestrictionTable0 }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable75 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  80,  81,   2,  16,    0x30000 } */ { 0x10025150, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable94 }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  82,  84,   2,  18,   0x1c0000 } */ { 0x12025452, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable136 }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable93 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable218 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1RegFile,                       DecodingTable220 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable154 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable155 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable150 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable151 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable221 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable222 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable158 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable89 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable607[]
const ged_ins_field_entry_t DecodingTable608[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable49 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  16,  19,   0,  16,    0xf0000 } */ { 0x10001310, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable50 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable48 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable132 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable48 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  43,  46,   1,  11,     0x7800 } */ {  0xb012e2b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable132 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  91,  94,   2,  27, 0x78000000 } */ { 0x1b025e5b, (void*)(g_uintptr_t)0x78000000 }, RestrictionTable96 }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable213 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable214 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  53,  60,   1,  21, 0x1fe00000 } */ { 0x15013c35, (void*)(g_uintptr_t)0x1fe00000 }, RestrictionTable52 }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  61,  62,   1,  29, 0x60000000 } */ { 0x1d013e3d, (void*)(g_uintptr_t)0x60000000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable215 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable216 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  69,  76,   2,   5,     0x1fe0 } */ {  0x5024c45, (void*)(g_uintptr_t)    0x1fe0 }, RestrictionTable52 }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  80,  81,   2,  16,    0x30000 } */ { 0x10025150, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  82,  84,   2,  18,   0x1c0000 } */ { 0x12025452, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable28 }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable218 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1RegFile,                       DecodingTable220 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable154 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable155 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable151 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable221 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable222 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable158 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable89 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable608[]
const ged_ins_field_entry_t DecodingTable609[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  16,  19,   0,  16,    0xf0000 } */ { 0x10001310, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable132 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable225 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable212 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable213 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable214 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  61,  62,   1,  29, 0x60000000 } */ { 0x1d013e3d, (void*)(g_uintptr_t)0x60000000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable215 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable216 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  80,  81,   2,  16,    0x30000 } */ { 0x10025150, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  82,  84,   2,  18,   0x1c0000 } */ { 0x12025452, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable28 }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable609[]
const ged_ins_field_entry_t DecodingTable610[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  16,  19,   0,  16,    0xf0000 } */ { 0x10001310, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable132 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable225 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable213 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable214 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  61,  62,   1,  29, 0x60000000 } */ { 0x1d013e3d, (void*)(g_uintptr_t)0x60000000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  64,  95,   2,   0, 0xffffffff } */ {    0x25f40, (void*)(g_uintptr_t)0xffffffff }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable97 }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable610[]
const ged_ins_field_entry_t DecodingTable611[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  16,  19,   0,  16,    0xf0000 } */ { 0x10001310, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable132 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable225 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable213 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable214 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  61,  62,   1,  29, 0x60000000 } */ { 0x1d013e3d, (void*)(g_uintptr_t)0x60000000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  96, 127,   3,   0, 0xffffffff } */ {    0x37f60, (void*)(g_uintptr_t)0xffffffff }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    32, /* {  64,  95,   2,   0, 0xffffffff } */ {    0x25f40, (void*)(g_uintptr_t)0xffffffff }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable611[]
const ged_ins_field_entry_t DecodingTable612[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable117 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  16,  19,   0,  16,    0xf0000 } */ { 0x10001310, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable38 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable132 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable118 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable225 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable119 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable212 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable213 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable214 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  61,  62,   1,  29, 0x60000000 } */ { 0x1d013e3d, (void*)(g_uintptr_t)0x60000000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable226 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable227 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable113 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable45 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable120 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable228 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable229 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable218 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1RegFile,                       DecodingTable220 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable154 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable155 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable150 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable151 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable221 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable222 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable158 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_MathFC,                            DecodingTable259 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable92 }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_MathFC,                            DecodingTable122 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_MathFC,                            DecodingTable261 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_MathFC,                            DecodingTable264 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable612[]
const ged_ins_field_entry_t DecodingTable613[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  16,  19,   0,  16,    0xf0000 } */ { 0x10001310, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable132 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable12 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  43,  46,   1,  11,     0x7800 } */ {  0xb012e2b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable132 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable42 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable212 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable213 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable214 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  61,  62,   1,  29, 0x60000000 } */ { 0x1d013e3d, (void*)(g_uintptr_t)0x60000000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable215 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable216 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  80,  81,   2,  16,    0x30000 } */ { 0x10025150, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  82,  84,   2,  18,   0x1c0000 } */ { 0x12025452, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable28 }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable232 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable160 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable150 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable161 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable221 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable222 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable158 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable224 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable613[]
const ged_ins_field_entry_t DecodingTable614[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  16,  19,   0,  16,    0xf0000 } */ { 0x10001310, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable132 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable225 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable213 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable214 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  61,  62,   1,  29, 0x60000000 } */ { 0x1d013e3d, (void*)(g_uintptr_t)0x60000000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable226 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable227 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable113 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable45 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable46 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable228 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable229 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable231 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable614[]
const ged_ins_field_entry_t DecodingTable615[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable130 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  16,  19,   0,  16,    0xf0000 } */ { 0x10001310, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable132 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable38 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable225 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable213 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable214 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  61,  62,   1,  29, 0x60000000 } */ { 0x1d013e3d, (void*)(g_uintptr_t)0x60000000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable215 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable133 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable105 }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable42 }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DescRegFile,                       DecodingTable90 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DescRegFile,                       DecodingTable65 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     31, { GED_INS_FIELD_DescRegFile,                       DecodingTable132 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_FRAGMENTED,     32, { 6, FragmentsTable20 }, RestrictionTable107 }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 127, 127,   3,  31, 0x80000000 } */ { 0x1f037f7f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable44 }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, RestrictionTable42 }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable106 }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_FRAGMENTED,     16, { 4, FragmentsTable19 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable615[]
const ged_ins_field_entry_t DecodingTable616[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable130 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  16,  19,   0,  16,    0xf0000 } */ { 0x10001310, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable132 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable38 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 0 }, RestrictionTable91 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  36,  36,   1,   4,       0x10 } */ {  0x4012424, (void*)(g_uintptr_t)      0x10 }, RestrictionTable108 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 0 }, RestrictionTable91 }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable210 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable134 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable258 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable135 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable136 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable105 }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  77,  77,   2,  13,     0x2000 } */ {  0xd024d4d, (void*)(g_uintptr_t)    0x2000 }, RestrictionTable109 }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DescRegFile,                       DecodingTable137 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DescRegFile,                       DecodingTable138 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    31, /* {  96, 126,   3,   0, 0x7fffffff } */ {    0x37e60, (void*)(g_uintptr_t)0x7fffffff }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable140 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* { 127, 127,   3,  31, 0x80000000 } */ { 0x1f037f7f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable44 }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  61,  61,   1,  29, 0x20000000 } */ { 0x1d013d3d, (void*)(g_uintptr_t)0x20000000 }, RestrictionTable109 }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable106 }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable141 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  64,  67,   2,   0,        0xf } */ {    0x24340, (void*)(g_uintptr_t)       0xf }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable142 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable143 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable616[]
const ged_ins_field_entry_t DecodingTable617[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  16,  19,   0,  16,    0xf0000 } */ { 0x10001310, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  36,  36,   1,   4,       0x10 } */ {  0x4012424, (void*)(g_uintptr_t)      0x10 }, RestrictionTable119 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable233 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable236 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable234 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  44,  44,   1,  12,     0x1000 } */ {  0xc012c2c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable119 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable237 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  54,  55,   1,  19,   0xc00000 } */ { 0x13013736, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable121 }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstRegFile,                        DecodingTable209 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  49,  49,   1,  17,    0x20000 } */ { 0x11013131, (void*)(g_uintptr_t)   0x20000 }, RestrictionTable123 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable240 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable243 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable246 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable249 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable24 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable177 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  39,  40,   1,   7,      0x180 } */ {  0x7012827, (void*)(g_uintptr_t)     0x180 }, RestrictionTable16 }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  90,  91,   2,  26,  0xc000000 } */ { 0x1a025b5a, (void*)(g_uintptr_t) 0xc000000 }, RestrictionTable27 }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  88,  89,   2,  24,  0x3000000 } */ { 0x18025958, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable124 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable16 }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src2RegFile,                       DecodingTable185 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src2RegFile,                       DecodingTable187 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  45,  45,   1,  13,     0x2000 } */ {  0xd012d2d, (void*)(g_uintptr_t)    0x2000 }, RestrictionTable120 }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable250 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable125 }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src0RegFile,                       DecodingTable253 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src2RegFile,                       DecodingTable204 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src2RegFile,                       DecodingTable206 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable617[]
const ged_ins_field_entry_t DecodingTable618[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  16,  19,   0,  16,    0xf0000 } */ { 0x10001310, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  36,  36,   1,   4,       0x10 } */ {  0x4012424, (void*)(g_uintptr_t)      0x10 }, RestrictionTable119 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable233 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable236 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable234 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  44,  44,   1,  12,     0x1000 } */ {  0xc012c2c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable119 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable237 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  54,  55,   1,  19,   0xc00000 } */ { 0x13013736, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable121 }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstRegFile,                        DecodingTable209 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  49,  49,   1,  17,    0x20000 } */ { 0x11013131, (void*)(g_uintptr_t)   0x20000 }, RestrictionTable123 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable240 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable243 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  37,  38,   1,   5,       0x60 } */ {  0x5012625, (void*)(g_uintptr_t)      0x60 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable246 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable249 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable24 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable177 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  39,  40,   1,   7,      0x180 } */ {  0x7012827, (void*)(g_uintptr_t)     0x180 }, RestrictionTable16 }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  90,  91,   2,  26,  0xc000000 } */ { 0x1a025b5a, (void*)(g_uintptr_t) 0xc000000 }, RestrictionTable27 }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  88,  89,   2,  24,  0x3000000 } */ { 0x18025958, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable124 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable16 }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src2RegFile,                       DecodingTable185 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src2RegFile,                       DecodingTable187 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  45,  45,   1,  13,     0x2000 } */ {  0xd012d2d, (void*)(g_uintptr_t)    0x2000 }, RestrictionTable120 }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 3 }, RestrictionTable17 }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  52,  55,   1,  20,   0xf00000 } */ { 0x14013734, (void*)(g_uintptr_t)  0xf00000 }, RestrictionTable90 }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable256 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_FRAGMENTED,      4, { 2, FragmentsTable25 }, RestrictionTable90 }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src2RegFile,                       DecodingTable257 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable250 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable125 }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src0RegFile,                       DecodingTable253 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src2RegFile,                       DecodingTable204 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src2RegFile,                       DecodingTable206 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable618[]
const ged_ins_field_entry_t DecodingTable619[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  16,  19,   0,  16,    0xf0000 } */ { 0x10001310, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable132 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable12 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  43,  46,   1,  11,     0x7800 } */ {  0xb012e2b, (void*)(g_uintptr_t)    0x7800 }, RestrictionTable132 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  89,  90,   2,  25,  0x6000000 } */ { 0x19025a59, (void*)(g_uintptr_t) 0x6000000 }, RestrictionTable14 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable212 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable213 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable214 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  61,  62,   1,  29, 0x60000000 } */ { 0x1d013e3d, (void*)(g_uintptr_t)0x60000000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable215 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable216 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable18 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable103 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  77,  78,   2,  13,     0x6000 } */ {  0xd024e4d, (void*)(g_uintptr_t)    0x6000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  79,  79,   2,  15,     0x8000 } */ {  0xf024f4f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  80,  81,   2,  16,    0x30000 } */ { 0x10025150, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  82,  84,   2,  18,   0x1c0000 } */ { 0x12025452, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable28 }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  85,  88,   2,  21,  0x1e00000 } */ { 0x15025855, (void*)(g_uintptr_t) 0x1e00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable218 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1RegFile,                       DecodingTable220 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable154 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable155 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable150 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable151 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable221 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable222 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable158 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable224 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable619[]
const ged_ins_field_entry_t DecodingTable620[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   9,  10,   0,   9,      0x600 } */ {  0x9000a09, (void*)(g_uintptr_t)     0x600 }, RestrictionTable3 }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecSize,                          DecodingTable96 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  11,  13,   0,  11,     0x3800 } */ {  0xb000d0b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  14,  15,   0,  14,     0xc000 } */ {  0xe000f0e, (void*)(g_uintptr_t)    0xc000 }, RestrictionTable118 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  16,  19,   0,  16,    0xf0000 } */ { 0x10001310, (void*)(g_uintptr_t)   0xf0000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  20,  20,   0,  20,   0x100000 } */ { 0x14001414, (void*)(g_uintptr_t)  0x100000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  21,  23,   0,  21,   0xe00000 } */ { 0x15001715, (void*)(g_uintptr_t)  0xe00000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  35,  36,   1,   3,       0x18 } */ {  0x3012423, (void*)(g_uintptr_t)      0x18 }, RestrictionTable12 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  37,  40,   1,   5,      0x1e0 } */ {  0x5012825, (void*)(g_uintptr_t)     0x1e0 }, RestrictionTable132 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  41,  42,   1,   9,      0x600 } */ {  0x9012a29, (void*)(g_uintptr_t)     0x600 }, RestrictionTable34 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable225 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable213 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable214 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable9 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable100 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  61,  62,   1,  29, 0x60000000 } */ { 0x1d013e3d, (void*)(g_uintptr_t)0x60000000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  63,  63,   1,  31, 0x80000000 } */ { 0x1f013f3f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable226 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable227 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable49 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable113 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable45 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable46 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable228 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable229 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable53 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable231 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable620[]
const ged_ins_field_entry_t DecodingTable621[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable348 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable349 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstRegFile,                        DecodingTable269 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable350 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable275 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable307 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable308 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable309 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable351 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable352 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable353 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable354 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable355 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable356 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable357 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable97 }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable621[]
const ged_ins_field_entry_t DecodingTable622[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable622[]
const ged_ins_field_entry_t DecodingTable623[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable330 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_DstAddrMode,                       DecodingTable265 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable331 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable303 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable332 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable266 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_MathFC,                            DecodingTable333 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable270 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  48,  49,   1,  16,    0x30000 } */ { 0x10013130, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable335 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable275 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable293 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable336 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable297 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable298 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable299 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable339 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable283 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable278 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable340 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable285 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable286 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable287 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_MathFC,                            DecodingTable342 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable153 }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_MathFC,                            DecodingTable343 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_MathFC,                            DecodingTable345 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_MathFC,                            DecodingTable347 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable623[]
const ged_ins_field_entry_t DecodingTable624[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable624[]
const ged_ins_field_entry_t DecodingTable625[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable292 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_DstAddrMode,                       DecodingTable265 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0IsImm,                         DecodingTable290 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable267 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable268 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable270 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable271 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  48,  49,   1,  16,    0x30000 } */ { 0x10013130, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable294 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable295 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable276 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable296 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable293 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable272 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable297 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable298 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable299 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable301 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  46,  46,   1,  14,     0x4000 } */ {  0xe012e2e, (void*)(g_uintptr_t)    0x4000 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable625[]
const ged_ins_field_entry_t DecodingTable626[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   7,   7,   0,   7,       0x80 } */ {  0x7000707, (void*)(g_uintptr_t)      0x80 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  16,  23,   0,  16,   0xff0000 } */ { 0x10001710, (void*)(g_uintptr_t)  0xff0000 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  40,  47,   1,   8,     0xff00 } */ {  0x8012f28, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  24,  28,   0,  24, 0x1f000000 } */ { 0x18001c18, (void*)(g_uintptr_t)0x1f000000 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable8 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  35,  39,   1,   3,       0xf8 } */ {  0x3012723, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  48,  51,   1,  16,    0xf0000 } */ { 0x10013330, (void*)(g_uintptr_t)   0xf0000 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  52,  55,   1,  20,   0xf00000 } */ { 0x14013734, (void*)(g_uintptr_t)  0xf00000 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable8 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    12, /* {  52,  63,   1,  20, 0xfff00000 } */ { 0x14013f34, (void*)(g_uintptr_t)0xfff00000 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable626[]
const ged_ins_field_entry_t DecodingTable627[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 2 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  66,  66,   2,   2,        0x4 } */ {  0x2024242, (void*)(g_uintptr_t)       0x4 }, RestrictionTable108 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 2 }, RestrictionTable151 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  98,  98,   3,   2,        0x4 } */ {  0x2036262, (void*)(g_uintptr_t)       0x4 }, RestrictionTable108 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 2 }, RestrictionTable151 }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstRegFile,                        DecodingTable269 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 104, 111,   3,   8,     0xff00 } */ {  0x8036f68, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable105 }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable109 }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_DescRegFile,                       DecodingTable327 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable328 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable44 }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  49,  49,   1,  17,    0x20000 } */ { 0x11013131, (void*)(g_uintptr_t)   0x20000 }, RestrictionTable109 }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable329 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable150 }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable627[]
const ged_ins_field_entry_t DecodingTable628[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable143 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_SyncFC,                            DecodingTable302 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable303 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable143 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     8, { 0 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  48,  49,   1,  16,    0x30000 } */ { 0x10013130, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable77 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable304 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable305 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable75 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable306 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable307 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable308 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable309 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     8, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable75 }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable144 }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable50 }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 0 }, RestrictionTable145 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable310 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  46,  46,   1,  14,     0x4000 } */ {  0xe012e2e, (void*)(g_uintptr_t)    0x4000 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable311 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable142 }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable628[]
const ged_ins_field_entry_t DecodingTable629[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable312 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable313 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  98,  98,   3,   2,        0x4 } */ {  0x2036262, (void*)(g_uintptr_t)       0x4 }, RestrictionTable108 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable314 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  54,  55,   1,  19,   0xc00000 } */ { 0x13013736, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable121 }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstRegFile,                        DecodingTable269 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable275 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable297 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable315 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable316 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  84,  85,   2,  20,   0x300000 } */ { 0x14025554, (void*)(g_uintptr_t)  0x300000 }, RestrictionTable16 }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src2RegFile,                       DecodingTable317 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src2RegFile,                       DecodingTable318 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable319 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  39,  39,   1,   7,       0x80 } */ {  0x7012727, (void*)(g_uintptr_t)      0x80 }, RestrictionTable125 }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src0RegFile,                       DecodingTable322 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src2RegFile,                       DecodingTable325 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src2RegFile,                       DecodingTable326 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable629[]
const ged_ins_field_entry_t DecodingTable630[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   7,   7,   0,   7,       0x80 } */ {  0x7000707, (void*)(g_uintptr_t)      0x80 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  16,  23,   0,  16,   0xff0000 } */ { 0x10001710, (void*)(g_uintptr_t)  0xff0000 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  40,  47,   1,   8,     0xff00 } */ {  0x8012f28, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  48,  55,   1,  16,   0xff0000 } */ { 0x10013730, (void*)(g_uintptr_t)  0xff0000 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  24,  28,   0,  24, 0x1f000000 } */ { 0x18001c18, (void*)(g_uintptr_t)0x1f000000 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  35,  39,   1,   3,       0xf8 } */ {  0x3012723, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable8 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable630[]
const ged_ins_field_entry_t DecodingTable631[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_DstAddrMode,                       DecodingTable265 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable12 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  40,  43,   1,   8,      0xf00 } */ {  0x8012b28, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable138 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable27 }, RestrictionTable139 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable266 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable267 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable268 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable270 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable271 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  48,  49,   1,  16,    0x30000 } */ { 0x10013130, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable273 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable274 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable276 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable277 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable272 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  81,  83,   2,  17,    0xe0000 } */ { 0x11025351, (void*)(g_uintptr_t)   0xe0000 }, RestrictionTable28 }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  84,  87,   2,  20,   0xf00000 } */ { 0x14025754, (void*)(g_uintptr_t)  0xf00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable281 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable282 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable283 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable284 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable278 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable279 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable285 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable286 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable287 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable289 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable631[]
const ged_ins_field_entry_t DecodingTable632[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   7,   7,   0,   7,       0x80 } */ {  0x7000707, (void*)(g_uintptr_t)      0x80 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  16,  23,   0,  16,   0xff0000 } */ { 0x10001710, (void*)(g_uintptr_t)  0xff0000 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  40,  47,   1,   8,     0xff00 } */ {  0x8012f28, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  24,  28,   0,  24, 0x1f000000 } */ { 0x18001c18, (void*)(g_uintptr_t)0x1f000000 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable8 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  35,  39,   1,   3,       0xf8 } */ {  0x3012723, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  48,  51,   1,  16,    0xf0000 } */ { 0x10013330, (void*)(g_uintptr_t)   0xf0000 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  52,  55,   1,  20,   0xf00000 } */ { 0x14013734, (void*)(g_uintptr_t)  0xf00000 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable8 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    12, /* {  52,  63,   1,  20, 0xfff00000 } */ { 0x14013f34, (void*)(g_uintptr_t)0xfff00000 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable632[]
const ged_ins_field_entry_t DecodingTable633[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable348 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable358 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable359 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable349 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable350 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable275 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable307 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable308 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable309 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable351 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable352 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable353 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable354 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable355 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable356 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable357 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable97 }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable633[]
const ged_ins_field_entry_t DecodingTable634[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable330 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_DstAddrMode,                       DecodingTable265 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable358 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable331 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable303 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable332 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable361 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_MathFC,                            DecodingTable333 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable270 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  48,  49,   1,  16,    0x30000 } */ { 0x10013130, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable335 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable275 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable293 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable336 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable297 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable298 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable299 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable339 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable283 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable278 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable340 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable285 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable286 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable287 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_MathFC,                            DecodingTable342 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable153 }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_MathFC,                            DecodingTable343 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_MathFC,                            DecodingTable345 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_MathFC,                            DecodingTable408 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable634[]
const ged_ins_field_entry_t DecodingTable635[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable292 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_DstAddrMode,                       DecodingTable265 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable362 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0IsImm,                         DecodingTable364 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable267 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable268 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable270 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable271 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  48,  49,   1,  16,    0x30000 } */ { 0x10013130, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable294 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0RegFile,                       DecodingTable295 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable276 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable296 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable293 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable272 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable297 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable298 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable299 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0RegFile,                       DecodingTable301 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  46,  46,   1,  14,     0x4000 } */ {  0xe012e2e, (void*)(g_uintptr_t)    0x4000 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable635[]
const ged_ins_field_entry_t DecodingTable636[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 2 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  66,  66,   2,   2,        0x4 } */ {  0x2024242, (void*)(g_uintptr_t)       0x4 }, RestrictionTable108 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 2 }, RestrictionTable151 }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  98,  98,   3,   2,        0x4 } */ {  0x2036262, (void*)(g_uintptr_t)       0x4 }, RestrictionTable108 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     4, { 2 }, RestrictionTable151 }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 104, 111,   3,   8,     0xff00 } */ {  0x8036f68, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable105 }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable109 }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_DescRegFile,                       DecodingTable327 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable328 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable44 }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  49,  49,   1,  17,    0x20000 } */ { 0x11013131, (void*)(g_uintptr_t)   0x20000 }, RestrictionTable109 }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable329 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable150 }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable402 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable404 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable406 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable636[]
const ged_ins_field_entry_t DecodingTable637[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable365 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable367 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable369 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  98,  98,   3,   2,        0x4 } */ {  0x2036262, (void*)(g_uintptr_t)       0x4 }, RestrictionTable108 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable373 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstRegFile,                        DecodingTable269 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Opcode,                            DecodingTable374 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable275 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable375 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable377 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable379 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable316 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable380 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable381 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable382 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable383 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src2RegFile,                       DecodingTable317 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src2RegFile,                       DecodingTable318 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable387 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  39,  39,   1,   7,       0x80 } */ {  0x7012727, (void*)(g_uintptr_t)      0x80 }, RestrictionTable125 }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src0RegFile,                       DecodingTable322 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src2RegFile,                       DecodingTable325 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src2RegFile,                       DecodingTable389 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable390 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  48,  49,   1,  16,    0x30000 } */ { 0x10013130, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable111 }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Opcode,                            DecodingTable392 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable395 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Opcode,                            DecodingTable397 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable400 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Opcode,                            DecodingTable401 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable637[]
const ged_ins_field_entry_t DecodingTable638[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable367 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable12 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable369 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  98,  98,   3,   2,        0x4 } */ {  0x2036262, (void*)(g_uintptr_t)       0x4 }, RestrictionTable108 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable373 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstRegFile,                        DecodingTable269 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Opcode,                            DecodingTable374 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable275 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable375 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable377 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable379 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable316 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable380 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable381 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable382 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable383 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src2RegFile,                       DecodingTable318 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable12 }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  51,  54,   1,  19,   0x780000 } */ { 0x13013633, (void*)(g_uintptr_t)  0x780000 }, RestrictionTable90 }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  67,  70,   2,   3,       0x78 } */ {  0x3024643, (void*)(g_uintptr_t)      0x78 }, RestrictionTable90 }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  99, 102,   3,   3,       0x78 } */ {  0x3036663, (void*)(g_uintptr_t)      0x78 }, RestrictionTable90 }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* { 115, 118,   3,  19,   0x780000 } */ { 0x13037673, (void*)(g_uintptr_t)  0x780000 }, RestrictionTable90 }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable387 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  39,  39,   1,   7,       0x80 } */ {  0x7012727, (void*)(g_uintptr_t)      0x80 }, RestrictionTable125 }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src2RegFile,                       DecodingTable326 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable638[]
const ged_ins_field_entry_t DecodingTable639[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_DstAddrMode,                       DecodingTable265 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable358 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable12 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable359 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable27 }, RestrictionTable139 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable361 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_DstAddrMode,                       DecodingTable267 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable268 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable270 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable271 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  48,  49,   1,  16,    0x30000 } */ { 0x10013130, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable273 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable274 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable276 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable277 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable16 }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable272 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  81,  83,   2,  17,    0xe0000 } */ { 0x11025351, (void*)(g_uintptr_t)   0xe0000 }, RestrictionTable28 }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  84,  87,   2,  20,   0xf00000 } */ { 0x14025754, (void*)(g_uintptr_t)  0xf00000 }, RestrictionTable17 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable281 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable282 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable283 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable284 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable278 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable279 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable285 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable286 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable287 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable289 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable639[]
const ged_ins_field_entry_t DecodingTable640[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable348 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable358 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable359 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable349 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstRegFile,                        DecodingTable269 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0RegFile,                       DecodingTable350 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable275 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable307 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable308 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable309 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src1RegFile,                       DecodingTable351 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable352 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable353 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable354 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable355 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0RegFile,                       DecodingTable356 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable357 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable97 }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable640[]
const ged_ins_field_entry_t DecodingTable641[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable330 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_DstAddrMode,                       DecodingTable265 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable358 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable331 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable303 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable332 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable361 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_MathFC,                            DecodingTable333 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_DstAddrMode,                       DecodingTable268 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable270 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable271 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  48,  49,   1,  16,    0x30000 } */ { 0x10013130, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0IsImm,                         DecodingTable411 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable274 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable276 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable277 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable293 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0RegFile,                       DecodingTable272 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable297 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0RegFile,                       DecodingTable298 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0RegFile,                       DecodingTable299 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src0IsImm,                         DecodingTable413 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     10, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable282 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable283 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1AddrMode,                      DecodingTable284 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable278 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1RegFile,                       DecodingTable279 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1RegFile,                       DecodingTable285 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1RegFile,                       DecodingTable286 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1RegFile,                       DecodingTable287 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_MathFC,                            DecodingTable342 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable175 }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_MathFC,                            DecodingTable343 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_MathFC,                            DecodingTable415 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0IsImm,                         DecodingTable418 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  46,  46,   1,  14,     0x4000 } */ {  0xe012e2e, (void*)(g_uintptr_t)    0x4000 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable641[]
const ged_ins_field_entry_t DecodingTable642[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  66,  66,   2,   2,        0x4 } */ {  0x2024242, (void*)(g_uintptr_t)       0x4 }, RestrictionTable108 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  98,  98,   3,   2,        0x4 } */ {  0x2036262, (void*)(g_uintptr_t)       0x4 }, RestrictionTable108 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstRegFile,                        DecodingTable269 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 104, 111,   3,   8,     0xff00 } */ {  0x8036f68, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable173 }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable109 }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_DescRegFile,                       DecodingTable327 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable409 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable44 }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  49,  49,   1,  17,    0x20000 } */ { 0x11013131, (void*)(g_uintptr_t)   0x20000 }, RestrictionTable109 }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable329 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable150 }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable402 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable410 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable406 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable642[]
const ged_ins_field_entry_t DecodingTable643[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  24,  25,   0,  24,  0x3000000 } */ { 0x18001918, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable186 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  26,  27,   0,  26,  0xc000000 } */ { 0x1a001b1a, (void*)(g_uintptr_t) 0xc000000 }, RestrictionTable187 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0IsImm,                         DecodingTable447 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_DstAddrMode,                       DecodingTable265 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable443 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0IsImm,                         DecodingTable445 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_DstAddrMode,                       DecodingTable423 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     11, { GED_INS_FIELD_DstAddrMode,                       DecodingTable424 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable425 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable271 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  48,  49,   1,  16,    0x30000 } */ { 0x10013130, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src0IsImm,                         DecodingTable449 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     11, { GED_INS_FIELD_Src0IsImm,                         DecodingTable450 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable451 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0IsImm,                         DecodingTable452 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0IsImm,                         DecodingTable429 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0IsImm,                         DecodingTable448 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0IsImm,                         DecodingTable453 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0IsImm,                         DecodingTable454 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0IsImm,                         DecodingTable455 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  21,  21,   0,  21,   0x200000 } */ { 0x15001515, (void*)(g_uintptr_t)  0x200000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  22,  23,   0,  22,   0xc00000 } */ { 0x16001716, (void*)(g_uintptr_t)  0xc00000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0IsImm,                         DecodingTable457 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  46,  46,   1,  14,     0x4000 } */ {  0xe012e2e, (void*)(g_uintptr_t)    0x4000 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable643[]
const ged_ins_field_entry_t DecodingTable644[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  24,  25,   0,  24,  0x3000000 } */ { 0x18001918, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable186 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  26,  27,   0,  26,  0xc000000 } */ { 0x1a001b1a, (void*)(g_uintptr_t) 0xc000000 }, RestrictionTable187 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable466 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable467 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable469 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  98,  98,   3,   2,        0x4 } */ {  0x2036262, (void*)(g_uintptr_t)       0x4 }, RestrictionTable108 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable471 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Opcode,                            DecodingTable472 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src0IsImm,                         DecodingTable474 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0IsImm,                         DecodingTable428 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0IsImm,                         DecodingTable429 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0IsImm,                         DecodingTable453 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0IsImm,                         DecodingTable475 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  21,  21,   0,  21,   0x200000 } */ { 0x15001515, (void*)(g_uintptr_t)  0x200000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  22,  23,   0,  22,   0xc00000 } */ { 0x16001716, (void*)(g_uintptr_t)  0xc00000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Opcode,                            DecodingTable476 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable316 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  86,  87,   2,  22,   0xc00000 } */ { 0x16025756, (void*)(g_uintptr_t)  0xc00000 }, RestrictionTable16 }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src2IsImm,                         DecodingTable477 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Opcode,                            DecodingTable479 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src2IsImm,                         DecodingTable480 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable481 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable482 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0IsImm,                         DecodingTable484 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable485 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src2IsImm,                         DecodingTable487 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable489 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  39,  39,   1,   7,       0x80 } */ {  0x7012727, (void*)(g_uintptr_t)      0x80 }, RestrictionTable125 }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src0IsImm,                         DecodingTable490 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src2IsImm,                         DecodingTable491 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src2IsImm,                         DecodingTable492 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  46,  46,   1,  14,     0x4000 } */ {  0xe012e2e, (void*)(g_uintptr_t)    0x4000 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, NULL } // 128
}; // DecodingTable644[]
const ged_ins_field_entry_t DecodingTable645[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  24,  25,   0,  24,  0x3000000 } */ { 0x18001918, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable186 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  26,  27,   0,  26,  0xc000000 } */ { 0x1a001b1a, (void*)(g_uintptr_t) 0xc000000 }, RestrictionTable187 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_DstAddrMode,                       DecodingTable265 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable419 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable12 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable420 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable27 }, RestrictionTable139 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable422 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_DstAddrMode,                       DecodingTable423 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     11, { GED_INS_FIELD_DstAddrMode,                       DecodingTable424 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable425 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable271 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  48,  49,   1,  16,    0x30000 } */ { 0x10013130, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable426 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     11, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable427 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable428 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable277 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0IsImm,                         DecodingTable429 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  80,  80,   2,  16,    0x10000 } */ { 0x10025050, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  81,  83,   2,  17,    0xe0000 } */ { 0x11025351, (void*)(g_uintptr_t)   0xe0000 }, RestrictionTable28 }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  84,  86,   2,  20,   0x700000 } */ { 0x14025654, (void*)(g_uintptr_t)  0x700000 }, RestrictionTable188 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  21,  21,   0,  21,   0x200000 } */ { 0x15001515, (void*)(g_uintptr_t)  0x200000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  22,  23,   0,  22,   0xc00000 } */ { 0x16001716, (void*)(g_uintptr_t)  0xc00000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src1IsImm,                         DecodingTable433 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     11, { GED_INS_FIELD_Src1IsImm,                         DecodingTable435 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1IsImm,                         DecodingTable436 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable437 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1IsImm,                         DecodingTable430 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1IsImm,                         DecodingTable431 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1IsImm,                         DecodingTable438 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1IsImm,                         DecodingTable439 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1IsImm,                         DecodingTable440 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1IsImm,                         DecodingTable442 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  46,  46,   1,  14,     0x4000 } */ {  0xe012e2e, (void*)(g_uintptr_t)    0x4000 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable645[]
const ged_ins_field_entry_t DecodingTable646[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  24,  25,   0,  24,  0x3000000 } */ { 0x18001918, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable186 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  26,  27,   0,  26,  0xc000000 } */ { 0x1a001b1a, (void*)(g_uintptr_t) 0xc000000 }, RestrictionTable187 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  44,  45,   1,  12,     0x3000 } */ {  0xc012d2c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable160 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable499 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable500 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  98,  98,   3,   2,        0x4 } */ {  0x2036262, (void*)(g_uintptr_t)       0x4 }, RestrictionTable108 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable501 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0IsImm,                         DecodingTable428 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0IsImm,                         DecodingTable453 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0IsImm,                         DecodingTable475 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  21,  21,   0,  21,   0x200000 } */ { 0x15001515, (void*)(g_uintptr_t)  0x200000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  22,  23,   0,  22,   0xc00000 } */ { 0x16001716, (void*)(g_uintptr_t)  0xc00000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable316 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src2IsImm,                         DecodingTable480 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable502 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  39,  39,   1,   7,       0x80 } */ {  0x7012727, (void*)(g_uintptr_t)      0x80 }, RestrictionTable125 }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src0IsImm,                         DecodingTable490 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src2IsImm,                         DecodingTable491 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src2IsImm,                         DecodingTable492 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  46,  46,   1,  14,     0x4000 } */ {  0xe012e2e, (void*)(g_uintptr_t)    0x4000 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_FRAGMENTED,      8, { 2, FragmentsTable40 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, NULL } // 128
}; // DecodingTable646[]
const ged_ins_field_entry_t DecodingTable647[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  24,  25,   0,  24,  0x3000000 } */ { 0x18001918, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable186 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  26,  27,   0,  26,  0xc000000 } */ { 0x1a001b1a, (void*)(g_uintptr_t) 0xc000000 }, RestrictionTable187 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable517 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable419 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable420 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable108 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src0IsImm,                         DecodingTable518 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0IsImm,                         DecodingTable428 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable27 }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable28 }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable188 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  21,  21,   0,  21,   0x200000 } */ { 0x15001515, (void*)(g_uintptr_t)  0x200000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  22,  23,   0,  22,   0xc00000 } */ { 0x16001716, (void*)(g_uintptr_t)  0xc00000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     6, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     8, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable27 }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable28 }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable188 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0IsImm,                         DecodingTable519 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable97 }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  46,  46,   1,  14,     0x4000 } */ {  0xe012e2e, (void*)(g_uintptr_t)    0x4000 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable647[]
const ged_ins_field_entry_t DecodingTable648[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  24,  25,   0,  24,  0x3000000 } */ { 0x18001918, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable186 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  26,  27,   0,  26,  0xc000000 } */ { 0x1a001b1a, (void*)(g_uintptr_t) 0xc000000 }, RestrictionTable187 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable419 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable420 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0IsImm,                         DecodingTable520 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src0IsImm,                         DecodingTable518 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0IsImm,                         DecodingTable428 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable144 }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable50 }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable194 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  21,  21,   0,  21,   0x200000 } */ { 0x15001515, (void*)(g_uintptr_t)  0x200000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  22,  23,   0,  22,   0xc00000 } */ { 0x16001716, (void*)(g_uintptr_t)  0xc00000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     6, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     8, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable144 }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable50 }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable194 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0IsImm,                         DecodingTable519 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable357 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable97 }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  46,  46,   1,  14,     0x4000 } */ {  0xe012e2e, (void*)(g_uintptr_t)    0x4000 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable648[]
const ged_ins_field_entry_t DecodingTable649[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   7,   7,   0,   7,       0x80 } */ {  0x7000707, (void*)(g_uintptr_t)      0x80 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  32,  39,   1,   0,       0xff } */ {    0x12720, (void*)(g_uintptr_t)      0xff }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  40,  47,   1,   8,     0xff00 } */ {  0x8012f28, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  48,  55,   1,  16,   0xff0000 } */ { 0x10013730, (void*)(g_uintptr_t)  0xff0000 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  18,  21,   0,  18,   0x3c0000 } */ { 0x12001512, (void*)(g_uintptr_t)  0x3c0000 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable51 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  22,  25,   0,  22,  0x3c00000 } */ { 0x16001916, (void*)(g_uintptr_t) 0x3c00000 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable649[]
const ged_ins_field_entry_t DecodingTable650[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   7,   7,   0,   7,       0x80 } */ {  0x7000707, (void*)(g_uintptr_t)      0x80 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  32,  39,   1,   0,       0xff } */ {    0x12720, (void*)(g_uintptr_t)      0xff }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  40,  47,   1,   8,     0xff00 } */ {  0x8012f28, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  18,  22,   0,  18,   0x7c0000 } */ { 0x12001612, (void*)(g_uintptr_t)  0x7c0000 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable51 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  48,  51,   1,  16,    0xf0000 } */ { 0x10013330, (void*)(g_uintptr_t)   0xf0000 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  23,  25,   0,  23,  0x3800000 } */ { 0x17001917, (void*)(g_uintptr_t) 0x3800000 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  52,  55,   1,  20,   0xf00000 } */ { 0x14013734, (void*)(g_uintptr_t)  0xf00000 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable51 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    12, /* {  52,  63,   1,  20, 0xfff00000 } */ { 0x14013f34, (void*)(g_uintptr_t)0xfff00000 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable650[]
const ged_ins_field_entry_t DecodingTable651[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  24,  25,   0,  24,  0x3000000 } */ { 0x18001918, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable186 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  26,  27,   0,  26,  0xc000000 } */ { 0x1a001b1a, (void*)(g_uintptr_t) 0xc000000 }, RestrictionTable187 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable493 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable494 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  98,  98,   3,   2,        0x4 } */ {  0x2036262, (void*)(g_uintptr_t)       0x4 }, RestrictionTable108 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable370 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable123 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0IsImm,                         DecodingTable428 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable27 }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable147 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  21,  21,   0,  21,   0x200000 } */ { 0x15001515, (void*)(g_uintptr_t)  0x200000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  22,  23,   0,  22,   0xc00000 } */ { 0x16001716, (void*)(g_uintptr_t)  0xc00000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable316 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable27 }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable147 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src2IsImm,                         DecodingTable480 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable384 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  39,  39,   1,   7,       0x80 } */ {  0x7012727, (void*)(g_uintptr_t)      0x80 }, RestrictionTable125 }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src0IsImm,                         DecodingTable490 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src2IsImm,                         DecodingTable491 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable27 }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  46,  46,   1,  14,     0x4000 } */ {  0xe012e2e, (void*)(g_uintptr_t)    0x4000 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  43,  45,   1,  11,     0x3800 } */ {  0xb012d2b, (void*)(g_uintptr_t)    0x3800 }, RestrictionTable165 }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  48,  49,   1,  16,    0x30000 } */ { 0x10013130, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable111 }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable391 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable496 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable396 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable498 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, NULL } // 128
}; // DecodingTable651[]
const ged_ins_field_entry_t DecodingTable652[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable330 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  24,  25,   0,  24,  0x3000000 } */ { 0x18001918, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable186 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  26,  27,   0,  26,  0xc000000 } */ { 0x1a001b1a, (void*)(g_uintptr_t) 0xc000000 }, RestrictionTable187 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_DstAddrMode,                       DecodingTable265 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable331 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0IsImm,                         DecodingTable459 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable332 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable422 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_MathFC,                            DecodingTable503 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable425 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  48,  49,   1,  16,    0x30000 } */ { 0x10013130, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src0IsImm,                         DecodingTable505 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0IsImm,                         DecodingTable428 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0IsImm,                         DecodingTable429 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0IsImm,                         DecodingTable453 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0IsImm,                         DecodingTable454 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0IsImm,                         DecodingTable455 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  21,  21,   0,  21,   0x200000 } */ { 0x15001515, (void*)(g_uintptr_t)  0x200000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  22,  23,   0,  22,   0xc00000 } */ { 0x16001716, (void*)(g_uintptr_t)  0xc00000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src0IsImm,                         DecodingTable508 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0IsImm,                         DecodingTable509 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0IsImm,                         DecodingTable510 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0IsImm,                         DecodingTable511 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0IsImm,                         DecodingTable512 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0IsImm,                         DecodingTable513 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_MathFC,                            DecodingTable516 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable175 }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_MathFC,                            DecodingTable343 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_MathFC,                            DecodingTable415 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0IsImm,                         DecodingTable418 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  46,  46,   1,  14,     0x4000 } */ {  0xe012e2e, (void*)(g_uintptr_t)    0x4000 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable652[]
const ged_ins_field_entry_t DecodingTable653[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable653[]
const ged_ins_field_entry_t DecodingTable654[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  24,  25,   0,  24,  0x3000000 } */ { 0x18001918, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable186 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  26,  27,   0,  26,  0xc000000 } */ { 0x1a001b1a, (void*)(g_uintptr_t) 0xc000000 }, RestrictionTable187 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  66,  66,   2,   2,        0x4 } */ {  0x2024242, (void*)(g_uintptr_t)       0x4 }, RestrictionTable108 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  98,  98,   3,   2,        0x4 } */ {  0x2036262, (void*)(g_uintptr_t)       0x4 }, RestrictionTable108 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     6, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  21,  21,   0,  21,   0x200000 } */ { 0x15001515, (void*)(g_uintptr_t)  0x200000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  22,  23,   0,  22,   0xc00000 } */ { 0x16001716, (void*)(g_uintptr_t)  0xc00000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     6, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 104, 111,   3,   8,     0xff00 } */ {  0x8036f68, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable173 }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable109 }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_DescRegFile,                       DecodingTable327 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable409 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable44 }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  49,  49,   1,  17,    0x20000 } */ { 0x11013131, (void*)(g_uintptr_t)   0x20000 }, RestrictionTable109 }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable329 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable402 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable410 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable406 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable654[]
const ged_ins_field_entry_t DecodingTable655[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  24,  25,   0,  24,  0x3000000 } */ { 0x18001918, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable186 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  26,  27,   0,  26,  0xc000000 } */ { 0x1a001b1a, (void*)(g_uintptr_t) 0xc000000 }, RestrictionTable187 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable143 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_SyncFC,                            DecodingTable458 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0IsImm,                         DecodingTable459 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable143 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     6, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     8, { 0 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  48,  49,   1,  16,    0x30000 } */ { 0x10013130, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable77 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src0RegFile,                       DecodingTable461 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable463 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable144 }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable50 }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable194 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  21,  21,   0,  21,   0x200000 } */ { 0x15001515, (void*)(g_uintptr_t)  0x200000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  22,  23,   0,  22,   0xc00000 } */ { 0x16001716, (void*)(g_uintptr_t)  0xc00000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     6, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     8, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable144 }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable50 }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable194 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0IsImm,                         DecodingTable464 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  46,  46,   1,  14,     0x4000 } */ {  0xe012e2e, (void*)(g_uintptr_t)    0x4000 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable193 }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable655[]
const ged_ins_field_entry_t DecodingTable656[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   7,   7,   0,   7,       0x80 } */ {  0x7000707, (void*)(g_uintptr_t)      0x80 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  32,  39,   1,   0,       0xff } */ {    0x12720, (void*)(g_uintptr_t)      0xff }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  40,  47,   1,   8,     0xff00 } */ {  0x8012f28, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  18,  22,   0,  18,   0x7c0000 } */ { 0x12001612, (void*)(g_uintptr_t)  0x7c0000 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable51 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  48,  51,   1,  16,    0xf0000 } */ { 0x10013330, (void*)(g_uintptr_t)   0xf0000 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  23,  25,   0,  23,  0x3800000 } */ { 0x17001917, (void*)(g_uintptr_t) 0x3800000 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  52,  55,   1,  20,   0xf00000 } */ { 0x14013734, (void*)(g_uintptr_t)  0xf00000 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_FRAGMENTED,      5, { 2, FragmentsTable51 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    12, /* {  52,  63,   1,  20, 0xfff00000 } */ { 0x14013f34, (void*)(g_uintptr_t)0xfff00000 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable656[]
const ged_ins_field_entry_t DecodingTable657[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  24,  25,   0,  24,  0x3000000 } */ { 0x18001918, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable186 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  26,  27,   0,  26,  0xc000000 } */ { 0x1a001b1a, (void*)(g_uintptr_t) 0xc000000 }, RestrictionTable187 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0IsImm,                         DecodingTable447 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_DstAddrMode,                       DecodingTable265 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable526 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0IsImm,                         DecodingTable528 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_DstAddrMode,                       DecodingTable423 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     11, { GED_INS_FIELD_DstAddrMode,                       DecodingTable424 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable425 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable271 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  48,  49,   1,  16,    0x30000 } */ { 0x10013130, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src0IsImm,                         DecodingTable449 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     11, { GED_INS_FIELD_Src0IsImm,                         DecodingTable450 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable451 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0IsImm,                         DecodingTable452 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0IsImm,                         DecodingTable429 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0IsImm,                         DecodingTable448 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0IsImm,                         DecodingTable453 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0IsImm,                         DecodingTable454 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0IsImm,                         DecodingTable455 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  21,  21,   0,  21,   0x200000 } */ { 0x15001515, (void*)(g_uintptr_t)  0x200000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  22,  23,   0,  22,   0xc00000 } */ { 0x16001716, (void*)(g_uintptr_t)  0xc00000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     64, { GED_INS_FIELD_Src0IsImm,                         DecodingTable457 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  46,  46,   1,  14,     0x4000 } */ {  0xe012e2e, (void*)(g_uintptr_t)    0x4000 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable657[]
const ged_ins_field_entry_t DecodingTable658[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  24,  25,   0,  24,  0x3000000 } */ { 0x18001918, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable186 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  26,  27,   0,  26,  0xc000000 } */ { 0x1a001b1a, (void*)(g_uintptr_t) 0xc000000 }, RestrictionTable187 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_DstAddrMode,                       DecodingTable265 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable521 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable12 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable522 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable27 }, RestrictionTable139 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable525 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_DstAddrMode,                       DecodingTable423 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     11, { GED_INS_FIELD_DstAddrMode,                       DecodingTable424 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable425 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_DstAddrMode,                       DecodingTable271 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  48,  49,   1,  16,    0x30000 } */ { 0x10013130, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable15 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable426 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     11, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable427 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable428 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0AddrMode,                      DecodingTable277 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0IsImm,                         DecodingTable429 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  80,  80,   2,  16,    0x10000 } */ { 0x10025050, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable15 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  64,  65,   2,   0,        0x3 } */ {    0x24140, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  81,  83,   2,  17,    0xe0000 } */ { 0x11025351, (void*)(g_uintptr_t)   0xe0000 }, RestrictionTable28 }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  84,  86,   2,  20,   0x700000 } */ { 0x14025654, (void*)(g_uintptr_t)  0x700000 }, RestrictionTable188 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  21,  21,   0,  21,   0x200000 } */ { 0x15001515, (void*)(g_uintptr_t)  0x200000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  22,  23,   0,  22,   0xc00000 } */ { 0x16001716, (void*)(g_uintptr_t)  0xc00000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src1IsImm,                         DecodingTable433 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     11, { GED_INS_FIELD_Src1IsImm,                         DecodingTable435 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1IsImm,                         DecodingTable436 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src1IsImm,                         DecodingTable437 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1IsImm,                         DecodingTable430 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src1IsImm,                         DecodingTable431 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src1IsImm,                         DecodingTable438 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1IsImm,                         DecodingTable439 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src1IsImm,                         DecodingTable440 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1IsImm,                         DecodingTable442 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  46,  46,   1,  14,     0x4000 } */ {  0xe012e2e, (void*)(g_uintptr_t)    0x4000 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable658[]
const ged_ins_field_entry_t DecodingTable659[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  24,  25,   0,  24,  0x3000000 } */ { 0x18001918, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable186 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  26,  27,   0,  26,  0xc000000 } */ { 0x1a001b1a, (void*)(g_uintptr_t) 0xc000000 }, RestrictionTable187 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Opcode,                            DecodingTable517 }, NULL }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable521 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable522 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable108 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src0IsImm,                         DecodingTable518 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0IsImm,                         DecodingTable428 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable27 }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable28 }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable188 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  21,  21,   0,  21,   0x200000 } */ { 0x15001515, (void*)(g_uintptr_t)  0x200000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  22,  23,   0,  22,   0xc00000 } */ { 0x16001716, (void*)(g_uintptr_t)  0xc00000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     6, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     8, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable27 }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable28 }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable188 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0IsImm,                         DecodingTable519 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable97 }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  46,  46,   1,  14,     0x4000 } */ {  0xe012e2e, (void*)(g_uintptr_t)    0x4000 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable659[]
const ged_ins_field_entry_t DecodingTable660[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  24,  25,   0,  24,  0x3000000 } */ { 0x18001918, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable186 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  26,  27,   0,  26,  0xc000000 } */ { 0x1a001b1a, (void*)(g_uintptr_t) 0xc000000 }, RestrictionTable187 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable521 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable522 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_Src0IsImm,                         DecodingTable520 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 1 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src0IsImm,                         DecodingTable518 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0IsImm,                         DecodingTable428 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable144 }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable50 }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable194 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  21,  21,   0,  21,   0x200000 } */ { 0x15001515, (void*)(g_uintptr_t)  0x200000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  22,  23,   0,  22,   0xc00000 } */ { 0x16001716, (void*)(g_uintptr_t)  0xc00000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     6, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     8, { 0 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable144 }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable50 }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     3, { 0 }, RestrictionTable194 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src0IsImm,                         DecodingTable519 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_Src1RegFile,                       DecodingTable357 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable97 }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  46,  46,   1,  14,     0x4000 } */ {  0xe012e2e, (void*)(g_uintptr_t)    0x4000 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable660[]
const ged_ins_field_entry_t DecodingTable661[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  24,  25,   0,  24,  0x3000000 } */ { 0x18001918, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable186 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  26,  27,   0,  26,  0xc000000 } */ { 0x1a001b1a, (void*)(g_uintptr_t) 0xc000000 }, RestrictionTable187 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable529 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable530 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  98,  98,   3,   2,        0x4 } */ {  0x2036262, (void*)(g_uintptr_t)       0x4 }, RestrictionTable108 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable531 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  51,  55,   1,  18,   0xf80000 } */ { 0x12013733, (void*)(g_uintptr_t)  0xf80000 }, RestrictionTable103 }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  56,  63,   1,  24, 0xff000000 } */ { 0x18013f38, (void*)(g_uintptr_t)0xff000000 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable123 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  67,  71,   2,   2,       0xf8 } */ {  0x2024743, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0IsImm,                         DecodingTable428 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable27 }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable147 }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  21,  21,   0,  21,   0x200000 } */ { 0x15001515, (void*)(g_uintptr_t)  0x200000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  22,  23,   0,  22,   0xc00000 } */ { 0x16001716, (void*)(g_uintptr_t)  0xc00000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {  99, 103,   3,   2,       0xf8 } */ {  0x2036763, (void*)(g_uintptr_t)      0xf8 }, RestrictionTable103 }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable316 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable27 }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable147 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src2IsImm,                         DecodingTable478 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src2IsImm,                         DecodingTable480 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable532 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  39,  39,   1,   7,       0x80 } */ {  0x7012727, (void*)(g_uintptr_t)      0x80 }, RestrictionTable125 }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src0IsImm,                         DecodingTable490 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src2IsImm,                         DecodingTable491 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 0 }, RestrictionTable27 }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  46,  46,   1,  14,     0x4000 } */ {  0xe012e2e, (void*)(g_uintptr_t)    0x4000 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      4, { 2, FragmentsTable52 }, RestrictionTable208 }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  48,  49,   1,  16,    0x30000 } */ { 0x10013130, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable209 }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable533 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable496 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable534 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable498 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, NULL } // 128
}; // DecodingTable661[]
const ged_ins_field_entry_t DecodingTable662[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable330 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  24,  25,   0,  24,  0x3000000 } */ { 0x18001918, (void*)(g_uintptr_t) 0x3000000 }, RestrictionTable186 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  26,  27,   0,  26,  0xc000000 } */ { 0x1a001b1a, (void*)(g_uintptr_t) 0xc000000 }, RestrictionTable187 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  18,  20,   0,  18,   0x1c0000 } */ { 0x12001412, (void*)(g_uintptr_t)  0x1c0000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_DstAddrMode,                       DecodingTable265 }, NULL }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  36,  39,   1,   4,       0xf0 } */ {  0x4012724, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable138 }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable331 }, NULL }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0IsImm,                         DecodingTable459 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_MathFC,                            DecodingTable332 }, NULL }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable525 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_MathFC,                            DecodingTable503 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstAddrMode,                       DecodingTable425 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  48,  49,   1,  16,    0x30000 } */ { 0x10013130, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable27 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  35,  35,   1,   3,        0x8 } */ {  0x3012323, (void*)(g_uintptr_t)       0x8 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src0IsImm,                         DecodingTable505 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0IsImm,                         DecodingTable428 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0IsImm,                         DecodingTable429 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0IsImm,                         DecodingTable453 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0IsImm,                         DecodingTable454 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0IsImm,                         DecodingTable455 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  21,  21,   0,  21,   0x200000 } */ { 0x15001515, (void*)(g_uintptr_t)  0x200000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  22,  23,   0,  22,   0xc00000 } */ { 0x16001716, (void*)(g_uintptr_t)  0xc00000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      6, { GED_INS_FIELD_Src0IsImm,                         DecodingTable508 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0IsImm,                         DecodingTable509 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0IsImm,                         DecodingTable510 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0IsImm,                         DecodingTable511 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0IsImm,                         DecodingTable512 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_Src0IsImm,                         DecodingTable513 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_MathFC,                            DecodingTable516 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable175 }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_MathFC,                            DecodingTable343 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_MathFC,                            DecodingTable415 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Src0IsImm,                         DecodingTable418 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    10, /* {   8,  17,   0,   8,    0x3ff00 } */ {  0x8001108, (void*)(g_uintptr_t)   0x3ff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  47,  47,   1,  15,     0x8000 } */ {  0xf012f2f, (void*)(g_uintptr_t)    0x8000 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  46,  46,   1,  14,     0x4000 } */ {  0xe012e2e, (void*)(g_uintptr_t)    0x4000 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable662[]
const ged_ins_field_entry_t DecodingTable663[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 2 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  66,  66,   2,   2,        0x4 } */ {  0x2024242, (void*)(g_uintptr_t)       0x4 }, RestrictionTable108 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  98,  98,   3,   2,        0x4 } */ {  0x2036262, (void*)(g_uintptr_t)       0x4 }, RestrictionTable108 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstRegFile,                        DecodingTable269 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {  72,  79,   2,   8,     0xff00 } */ {  0x8024f48, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     5, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* { 104, 111,   3,   8,     0xff00 } */ {  0x8036f68, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable212 }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable109 }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_DescRegFile,                       DecodingTable327 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     32, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable409 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable44 }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  49,  49,   1,  17,    0x20000 } */ { 0x11013131, (void*)(g_uintptr_t)   0x20000 }, RestrictionTable109 }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable329 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable150 }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable402 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      1, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable410 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_ExDescRegFile,                     DecodingTable406 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable663[]
const ged_ins_field_entry_t DecodingTable664[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      4, { GED_INS_FIELD_Opcode,                            DecodingTable365 }, NULL }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable536 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable139 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable538 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  98,  98,   3,   2,        0x4 } */ {  0x2036262, (void*)(g_uintptr_t)       0x4 }, RestrictionTable108 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  51,  55,   1,  19,   0xf80000 } */ { 0x13013733, (void*)(g_uintptr_t)  0xf80000 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstRegFile,                        DecodingTable269 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  67,  71,   2,   3,       0xf8 } */ {  0x3024743, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable275 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable539 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable297 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable315 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  99, 103,   3,   3,       0xf8 } */ {  0x3036763, (void*)(g_uintptr_t)      0xf8 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable316 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable540 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable541 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      5, { GED_INS_FIELD_Src2RegFile,                       DecodingTable317 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src2RegFile,                       DecodingTable318 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable139 }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 0 }, RestrictionTable39 }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  39,  39,   1,   7,       0x80 } */ {  0x7012727, (void*)(g_uintptr_t)      0x80 }, RestrictionTable125 }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src0RegFile,                       DecodingTable322 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,     16, { GED_INS_FIELD_Src2RegFile,                       DecodingTable325 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src2RegFile,                       DecodingTable326 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  48,  49,   1,  16,    0x30000 } */ { 0x10013130, (void*)(g_uintptr_t)   0x30000 }, RestrictionTable111 }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Opcode,                            DecodingTable401 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable664[]
const ged_ins_field_entry_t DecodingTable665[129] =
{
    { GED_INS_FIELD_Opcode,                            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     7, /* {   0,   6,   0,   0,       0x7f } */ {      0x600, (void*)(g_uintptr_t)      0x7f }, NULL }, // 0
    { GED_INS_FIELD_CmptCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  29,  29,   0,  29, 0x20000000 } */ { 0x1d001d1d, (void*)(g_uintptr_t)0x20000000 }, NULL }, // 1
    { GED_INS_FIELD_,                                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_INS_FIELD_Reserved,                          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_INS_FIELD____SrcImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_INS_FIELD_NumOfSourceOperands,               GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     2, { 3 }, NULL }, // 5
    { GED_INS_FIELD_HasDestinationOperand,             GED_TABLE_ENTRY_TYPE_FIXED_VALUE,     1, { 1 }, NULL }, // 6
    { GED_INS_FIELD_AccessMode,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 7
    { GED_INS_FIELD_MaskCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  31,  31,   0,  31, 0x80000000 } */ { 0x1f001f1f, (void*)(g_uintptr_t)0x80000000 }, RestrictionTable2 }, // 8
    { GED_INS_FIELD_DepCtrl,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_INS_FIELD_ExecMaskOffsetCtrl,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 10
    { GED_INS_FIELD_ChannelOffset,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  19,  21,   0,  19,   0x380000 } */ { 0x13001513, (void*)(g_uintptr_t)  0x380000 }, RestrictionTable4 }, // 11
    { GED_INS_FIELD_ThreadCtrl,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  32,  32,   1,   0,        0x1 } */ {    0x12020, (void*)(g_uintptr_t)       0x1 }, RestrictionTable137 }, // 12
    { GED_INS_FIELD_PredCtrl,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  24,  27,   0,  24,  0xf000000 } */ { 0x18001b18, (void*)(g_uintptr_t) 0xf000000 }, RestrictionTable21 }, // 13
    { GED_INS_FIELD_PredInv,                           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  28,  28,   0,  28, 0x10000000 } */ { 0x1c001c1c, (void*)(g_uintptr_t)0x10000000 }, RestrictionTable6 }, // 14
    { GED_INS_FIELD_ExecSize,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  16,  18,   0,  16,    0x70000 } */ { 0x10001210, (void*)(g_uintptr_t)   0x70000 }, RestrictionTable7 }, // 15
    { GED_INS_FIELD_CondModifier,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  92,  95,   2,  28, 0xf0000000 } */ { 0x1c025f5c, (void*)(g_uintptr_t)0xf0000000 }, RestrictionTable8 }, // 16
    { GED_INS_FIELD_AccWrCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  33,  33,   1,   1,        0x2 } */ {  0x1012121, (void*)(g_uintptr_t)       0x2 }, RestrictionTable9 }, // 17
    { GED_INS_FIELD_DebugCtrl,                         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  30,  30,   0,  30, 0x40000000 } */ { 0x1e001e1e, (void*)(g_uintptr_t)0x40000000 }, RestrictionTable10 }, // 18
    { GED_INS_FIELD_Saturate,                          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  34,  34,   1,   2,        0x4 } */ {  0x2012222, (void*)(g_uintptr_t)       0x4 }, RestrictionTable11 }, // 19
    { GED_INS_FIELD_DstRegFile,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  50,  50,   1,  18,    0x40000 } */ { 0x12013232, (void*)(g_uintptr_t)   0x40000 }, RestrictionTable108 }, // 20
    { GED_INS_FIELD_DstDataType,                       GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable536 }, NULL }, // 21
    { GED_INS_FIELD_Src0RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable26 }, RestrictionTable12 }, // 22
    { GED_INS_FIELD_Src0DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable538 }, NULL }, // 23
    { GED_INS_FIELD_Src1RegFile,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  98,  98,   3,   2,        0x4 } */ {  0x2036262, (void*)(g_uintptr_t)       0x4 }, RestrictionTable108 }, // 24
    { GED_INS_FIELD_Src1DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable372 }, NULL }, // 25
    { GED_INS_FIELD_DstChanEn,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 26
    { GED_INS_FIELD_DstSubRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 27
    { GED_INS_FIELD_DstAddrImm,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 28
    { GED_INS_FIELD_DstRegNum,                         GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_DstRegFile,                        DecodingTable269 }, NULL }, // 29
    { GED_INS_FIELD_DstAddrSubRegNum,                  GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 30
    { GED_INS_FIELD_DstHorzStride,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  48,  48,   1,  16,    0x10000 } */ { 0x10013030, (void*)(g_uintptr_t)   0x10000 }, RestrictionTable123 }, // 31
    { GED_INS_FIELD_DstAddrMode,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 32
    { GED_INS_FIELD_Src0ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 33
    { GED_INS_FIELD_Src0SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 34
    { GED_INS_FIELD_Src0AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_INS_FIELD_Src0RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src0RegFile,                       DecodingTable275 }, NULL }, // 36
    { GED_INS_FIELD_Src0AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 37
    { GED_INS_FIELD_Src0SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable539 }, NULL }, // 38
    { GED_INS_FIELD_Src0AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 39
    { GED_INS_FIELD_Src0HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable297 }, NULL }, // 40
    { GED_INS_FIELD_Src0Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 41
    { GED_INS_FIELD_Src0VertStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src0RegFile,                       DecodingTable315 }, NULL }, // 42
    { GED_INS_FIELD_FlagSubRegNum,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  22,  22,   0,  22,   0x400000 } */ { 0x16001616, (void*)(g_uintptr_t)  0x400000 }, NULL }, // 43
    { GED_INS_FIELD_FlagRegNum,                        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  23,  23,   0,  23,   0x800000 } */ { 0x17001717, (void*)(g_uintptr_t)  0x800000 }, NULL }, // 44
    { GED_INS_FIELD_Src1ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 45
    { GED_INS_FIELD_Src1SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 46
    { GED_INS_FIELD_Src1AddrImm,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 47
    { GED_INS_FIELD_Src1RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src1RegFile,                       DecodingTable316 }, NULL }, // 48
    { GED_INS_FIELD_Src1AddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 49
    { GED_INS_FIELD_Src1SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable540 }, NULL }, // 50
    { GED_INS_FIELD_Src1AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 51
    { GED_INS_FIELD_Src1HorzStride,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  96,  97,   3,   0,        0x3 } */ {    0x36160, (void*)(g_uintptr_t)       0x3 }, RestrictionTable27 }, // 52
    { GED_INS_FIELD_Src1Width,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 53
    { GED_INS_FIELD_Src1VertStride,                    GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable32 }, RestrictionTable147 }, // 54
    { GED_INS_FIELD_Imm,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 55
    { GED_INS_FIELD_Src2SrcMod,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Opcode,                            DecodingTable541 }, NULL }, // 56
    { GED_INS_FIELD_SrcDataType,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 57
    { GED_INS_FIELD_Src0RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 58
    { GED_INS_FIELD_Src1RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 59
    { GED_INS_FIELD_Src2RepCtrl,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 60
    { GED_INS_FIELD_Src2ChanSel,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 61
    { GED_INS_FIELD_Src2SubRegNum,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 62
    { GED_INS_FIELD_Src2RegNum,                        GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      8, { GED_INS_FIELD_Src2RegFile,                       DecodingTable318 }, NULL }, // 63
    { GED_INS_FIELD_Src2RegFile,                       GED_TABLE_ENTRY_TYPE_FRAGMENTED,      2, { 2, FragmentsTable31 }, RestrictionTable12 }, // 64
    { GED_INS_FIELD_Src2AddrMode,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 65
    { GED_INS_FIELD_Src2VertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 66
    { GED_INS_FIELD_SFID,                              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 67
    { GED_INS_FIELD_DescRegFile,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 68
    { GED_INS_FIELD_DescDataType,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 69
    { GED_INS_FIELD_DescAddrSubRegNum,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 70
    { GED_INS_FIELD_DescRegNum,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 71
    { GED_INS_FIELD_DescHorzStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 72
    { GED_INS_FIELD_DescWidth,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 73
    { GED_INS_FIELD_DescVertStride,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 74
    { GED_INS_FIELD_MsgDesc,                           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 75
    { GED_INS_FIELD_ExMsgDescImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 76
    { GED_INS_FIELD_EOT,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 77
    { GED_INS_FIELD_MathFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 78
    { GED_INS_FIELD_JIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 79
    { GED_INS_FIELD_UIP,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 80
    { GED_INS_FIELD_ControlIndex,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 81
    { GED_INS_FIELD_DataTypeIndex,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 82
    { GED_INS_FIELD_SubRegIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 83
    { GED_INS_FIELD_Src0Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 84
    { GED_INS_FIELD_Src1Index,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 85
    { GED_INS_FIELD_DescIndex,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 86
    { GED_INS_FIELD_ExDescRegFile,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 87
    { GED_INS_FIELD_DstMathMacroExt,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  51,  54,   1,  19,   0x780000 } */ { 0x13013633, (void*)(g_uintptr_t)  0x780000 }, RestrictionTable90 }, // 88
    { GED_INS_FIELD_Src0MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  67,  70,   2,   3,       0x78 } */ {  0x3024643, (void*)(g_uintptr_t)      0x78 }, RestrictionTable90 }, // 89
    { GED_INS_FIELD_Src1MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  99, 102,   3,   3,       0x78 } */ {  0x3036663, (void*)(g_uintptr_t)      0x78 }, RestrictionTable90 }, // 90
    { GED_INS_FIELD_Src2MathMacroExt,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* { 115, 118,   3,  19,   0x780000 } */ { 0x13037673, (void*)(g_uintptr_t)  0x780000 }, RestrictionTable90 }, // 91
    { GED_INS_FIELD_BranchCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 92
    { GED_INS_FIELD_SourceIndex,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 93
    { GED_INS_FIELD_Src2DataType,                      GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      3, { GED_INS_FIELD_ExecutionDataType,                 DecodingTable386 }, NULL }, // 94
    { GED_INS_FIELD_NoSrcDepSet,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 95
    { GED_INS_FIELD_ExFuncCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 96
    { GED_INS_FIELD_ExMsgLength,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 97
    { GED_INS_FIELD_ExDescAddrSubRegNum,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 98
    { GED_INS_FIELD_ExDescRegNum,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 99
    { GED_INS_FIELD_MsgDescCategory,                   GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 100
    { GED_INS_FIELD_MsgDescScratchAddrOffset,          GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 101
    { GED_INS_FIELD_MsgDescScratchBlockSize,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 102
    { GED_INS_FIELD_MsgDescScratchInvalidateAfterRead, GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 103
    { GED_INS_FIELD_MsgDescScratchChannelMode,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 104
    { GED_INS_FIELD_MsgDescScratchMessageType,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 105
    { GED_INS_FIELD_ExecutionDataType,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  39,  39,   1,   7,       0x80 } */ {  0x7012727, (void*)(g_uintptr_t)      0x80 }, RestrictionTable125 }, // 106
    { GED_INS_FIELD_Src0TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 107
    { GED_INS_FIELD_Src2TernaryImm,                    GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 108
    { GED_INS_FIELD_Src2HorzStride,                    GED_TABLE_ENTRY_TYPE_NEXT_TABLE,      2, { GED_INS_FIELD_Src2RegFile,                       DecodingTable326 }, NULL }, // 109
    { GED_INS_FIELD_SWSB,                              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   8,  15,   0,   8,     0xff00 } */ {  0x8000f08, (void*)(g_uintptr_t)    0xff00 }, NULL }, // 110
    { GED_INS_FIELD_Src1IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 111
    { GED_INS_FIELD_Src0IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 112
    { GED_INS_FIELD_Src0SubRegNumByte,                 GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 113
    { GED_INS_FIELD_SyncFC,                            GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 114
    { GED_INS_FIELD_FusionCtrl,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 115
    { GED_INS_FIELD_DataTypeIndexNoDep,                GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 116
    { GED_INS_FIELD_CompactedImm,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 117
    { GED_INS_FIELD_RepeatCount,                       GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 118
    { GED_INS_FIELD_SystolicDepth,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 119
    { GED_INS_FIELD_Src2Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 120
    { GED_INS_FIELD_Src2SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 121
    { GED_INS_FIELD_Src1Precision,                     GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 122
    { GED_INS_FIELD_Src1SubBytePrecision,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 123
    { GED_INS_FIELD_BfnFC,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 124
    { GED_INS_FIELD_ExBSO,                             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 125
    { GED_INS_FIELD_CPS,                               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 126
    { GED_INS_FIELD_Src1Length,                        GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 127
    { GED_INS_FIELD_Src2IsImm,                         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL } // 128
}; // DecodingTable665[]
const ged_ins_field_entry_t PositionInterpreterTable0[29] =
{
    { GED_PSEUDO_FIELD_ArchReg,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   4,   7,   0,   4,       0xf0 } */ {  0x4000704, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable53 }, // 0
    { GED_PSEUDO_FIELD_ArchRegNum,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   0,   3,   0,   0,        0xf } */ {      0x300, (void*)(g_uintptr_t)       0xf }, NULL }, // 1
    { GED_PSEUDO_FIELD_SwizzleX,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   0,   1,   0,   0,        0x3 } */ {      0x100, (void*)(g_uintptr_t)       0x3 }, RestrictionTable54 }, // 2
    { GED_PSEUDO_FIELD_SwizzleY,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   2,   3,   0,   2,        0xc } */ {  0x2000302, (void*)(g_uintptr_t)       0xc }, RestrictionTable54 }, // 3
    { GED_PSEUDO_FIELD_SwizzleZ,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   4,   5,   0,   4,       0x30 } */ {  0x4000504, (void*)(g_uintptr_t)      0x30 }, RestrictionTable54 }, // 4
    { GED_PSEUDO_FIELD_SwizzleW,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   6,   7,   0,   6,       0xc0 } */ {  0x6000706, (void*)(g_uintptr_t)      0xc0 }, RestrictionTable54 }, // 5
    { GED_PSEUDO_FIELD_MessageLength,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  25,  28,   0,  25, 0x1e000000 } */ { 0x19001c19, (void*)(g_uintptr_t)0x1e000000 }, NULL }, // 6
    { GED_PSEUDO_FIELD_ResponseLength,                GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  20,  24,   0,  20,  0x1f00000 } */ { 0x14001814, (void*)(g_uintptr_t) 0x1f00000 }, NULL }, // 7
    { GED_PSEUDO_FIELD_HeaderPresent,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  19,  19,   0,  19,    0x80000 } */ { 0x13001313, (void*)(g_uintptr_t)   0x80000 }, RestrictionTable55 }, // 8
    { GED_PSEUDO_FIELD_MessageTypeDP_SAMPLER,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable56 }, // 9
    { GED_PSEUDO_FIELD_MessageTypeDP_RC,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable57 }, // 10
    { GED_PSEUDO_FIELD_MessageTypeDP_CC,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable58 }, // 11
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable59 }, // 12
    { GED_PSEUDO_FIELD_TypedSurfaceSlotGroup,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable60 }, // 13
    { GED_PSEUDO_FIELD_TypedAtomicSlotGroup,          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable61 }, // 14
    { GED_PSEUDO_FIELD_UntypedSurfaceSIMDMode,        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable62 }, // 15
    { GED_PSEUDO_FIELD_UntypedAtomicSIMDMode,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable63 }, // 16
    { GED_PSEUDO_FIELD_InvalidateAfterRead,           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  13,  13,   0,  13,     0x2000 } */ {  0xd000d0d, (void*)(g_uintptr_t)    0x2000 }, NULL }, // 17
    { GED_PSEUDO_FIELD_BlockSize,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   8,  10,   0,   8,      0x700 } */ {  0x8000a08, (void*)(g_uintptr_t)     0x700 }, RestrictionTable64 }, // 18
    { GED_PSEUDO_FIELD_RedChannel,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable65 }, // 19
    { GED_PSEUDO_FIELD_GreenChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable66 }, // 20
    { GED_PSEUDO_FIELD_BlueChannel,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  10,  10,   0,  10,      0x400 } */ {  0xa000a0a, (void*)(g_uintptr_t)     0x400 }, RestrictionTable67 }, // 21
    { GED_PSEUDO_FIELD_AlphaChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  11,  11,   0,  11,      0x800 } */ {  0xb000b0b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable68 }, // 22
    { GED_PSEUDO_FIELD_ReturnDataControl,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  13,  13,   0,  13,     0x2000 } */ {  0xd000d0d, (void*)(g_uintptr_t)    0x2000 }, RestrictionTable69 }, // 23
    { GED_PSEUDO_FIELD_AtomicOperationType,           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable70 }, // 24
    { GED_PSEUDO_FIELD_AtomicCounterOperationType,    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable71 }, // 25
    { GED_PSEUDO_FIELD_SubFuncID,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   0,   2,   0,   0,        0x7 } */ {      0x200, (void*)(g_uintptr_t)       0x7 }, RestrictionTable72 }, // 26
    { GED_PSEUDO_FIELD_BindingTableIndex,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   0,   7,   0,   0,       0xff } */ {      0x700, (void*)(g_uintptr_t)      0xff }, NULL }, // 27
    { GED_PSEUDO_FIELD_FuncControl,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    19, /* {   0,  18,   0,   0,    0x7ffff } */ {     0x1200, (void*)(g_uintptr_t)   0x7ffff }, NULL } // 28
}; // PositionInterpreterTable0[]
const ged_ins_field_entry_t PositionInterpreterTable1[30] =
{
    { GED_PSEUDO_FIELD_ArchReg,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   4,   7,   0,   4,       0xf0 } */ {  0x4000704, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable79 }, // 0
    { GED_PSEUDO_FIELD_ArchRegNum,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   0,   3,   0,   0,        0xf } */ {      0x300, (void*)(g_uintptr_t)       0xf }, NULL }, // 1
    { GED_PSEUDO_FIELD_SwizzleX,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   0,   1,   0,   0,        0x3 } */ {      0x100, (void*)(g_uintptr_t)       0x3 }, RestrictionTable54 }, // 2
    { GED_PSEUDO_FIELD_SwizzleY,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   2,   3,   0,   2,        0xc } */ {  0x2000302, (void*)(g_uintptr_t)       0xc }, RestrictionTable54 }, // 3
    { GED_PSEUDO_FIELD_SwizzleZ,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   4,   5,   0,   4,       0x30 } */ {  0x4000504, (void*)(g_uintptr_t)      0x30 }, RestrictionTable54 }, // 4
    { GED_PSEUDO_FIELD_SwizzleW,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   6,   7,   0,   6,       0xc0 } */ {  0x6000706, (void*)(g_uintptr_t)      0xc0 }, RestrictionTable54 }, // 5
    { GED_PSEUDO_FIELD_MessageLength,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  25,  28,   0,  25, 0x1e000000 } */ { 0x19001c19, (void*)(g_uintptr_t)0x1e000000 }, NULL }, // 6
    { GED_PSEUDO_FIELD_ResponseLength,                GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  20,  24,   0,  20,  0x1f00000 } */ { 0x14001814, (void*)(g_uintptr_t) 0x1f00000 }, NULL }, // 7
    { GED_PSEUDO_FIELD_HeaderPresent,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  19,  19,   0,  19,    0x80000 } */ { 0x13001313, (void*)(g_uintptr_t)   0x80000 }, RestrictionTable55 }, // 8
    { GED_PSEUDO_FIELD_MessageTypeDP_SAMPLER,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable80 }, // 9
    { GED_PSEUDO_FIELD_MessageTypeDP_RC,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable81 }, // 10
    { GED_PSEUDO_FIELD_MessageTypeDP_CC,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable58 }, // 11
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable82 }, // 12
    { GED_PSEUDO_FIELD_TypedSurfaceSlotGroup,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable60 }, // 13
    { GED_PSEUDO_FIELD_TypedAtomicSlotGroup,          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable61 }, // 14
    { GED_PSEUDO_FIELD_UntypedSurfaceSIMDMode,        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable62 }, // 15
    { GED_PSEUDO_FIELD_UntypedAtomicSIMDMode,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable63 }, // 16
    { GED_PSEUDO_FIELD_InvalidateAfterRead,           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  13,  13,   0,  13,     0x2000 } */ {  0xd000d0d, (void*)(g_uintptr_t)    0x2000 }, NULL }, // 17
    { GED_PSEUDO_FIELD_BlockSize,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   8,  10,   0,   8,      0x700 } */ {  0x8000a08, (void*)(g_uintptr_t)     0x700 }, RestrictionTable64 }, // 18
    { GED_PSEUDO_FIELD_RedChannel,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable65 }, // 19
    { GED_PSEUDO_FIELD_GreenChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable66 }, // 20
    { GED_PSEUDO_FIELD_BlueChannel,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  10,  10,   0,  10,      0x400 } */ {  0xa000a0a, (void*)(g_uintptr_t)     0x400 }, RestrictionTable67 }, // 21
    { GED_PSEUDO_FIELD_AlphaChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  11,  11,   0,  11,      0x800 } */ {  0xb000b0b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable68 }, // 22
    { GED_PSEUDO_FIELD_ReturnDataControl,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  13,  13,   0,  13,     0x2000 } */ {  0xd000d0d, (void*)(g_uintptr_t)    0x2000 }, RestrictionTable69 }, // 23
    { GED_PSEUDO_FIELD_AtomicOperationType,           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable70 }, // 24
    { GED_PSEUDO_FIELD_AtomicCounterOperationType,    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable71 }, // 25
    { GED_PSEUDO_FIELD_SubFuncID,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   0,   2,   0,   0,        0x7 } */ {      0x200, (void*)(g_uintptr_t)       0x7 }, RestrictionTable72 }, // 26
    { GED_PSEUDO_FIELD_BindingTableIndex,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   0,   7,   0,   0,       0xff } */ {      0x700, (void*)(g_uintptr_t)      0xff }, NULL }, // 27
    { GED_PSEUDO_FIELD_FuncControl,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    19, /* {   0,  18,   0,   0,    0x7ffff } */ {     0x1200, (void*)(g_uintptr_t)   0x7ffff }, NULL }, // 28
    { GED_PSEUDO_FIELD_MessageTypeDP_DC1,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable83 } // 29
}; // PositionInterpreterTable1[]
const ged_ins_field_entry_t PositionInterpreterTable2[33] =
{
    { GED_PSEUDO_FIELD_ArchReg,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   4,   7,   0,   4,       0xf0 } */ {  0x4000704, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable99 }, // 0
    { GED_PSEUDO_FIELD_ArchRegNum,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   0,   3,   0,   0,        0xf } */ {      0x300, (void*)(g_uintptr_t)       0xf }, NULL }, // 1
    { GED_PSEUDO_FIELD_SwizzleX,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   0,   1,   0,   0,        0x3 } */ {      0x100, (void*)(g_uintptr_t)       0x3 }, RestrictionTable54 }, // 2
    { GED_PSEUDO_FIELD_SwizzleY,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   2,   3,   0,   2,        0xc } */ {  0x2000302, (void*)(g_uintptr_t)       0xc }, RestrictionTable54 }, // 3
    { GED_PSEUDO_FIELD_SwizzleZ,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   4,   5,   0,   4,       0x30 } */ {  0x4000504, (void*)(g_uintptr_t)      0x30 }, RestrictionTable54 }, // 4
    { GED_PSEUDO_FIELD_SwizzleW,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   6,   7,   0,   6,       0xc0 } */ {  0x6000706, (void*)(g_uintptr_t)      0xc0 }, RestrictionTable54 }, // 5
    { GED_PSEUDO_FIELD_MessageLength,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  25,  28,   0,  25, 0x1e000000 } */ { 0x19001c19, (void*)(g_uintptr_t)0x1e000000 }, NULL }, // 6
    { GED_PSEUDO_FIELD_ResponseLength,                GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  20,  24,   0,  20,  0x1f00000 } */ { 0x14001814, (void*)(g_uintptr_t) 0x1f00000 }, NULL }, // 7
    { GED_PSEUDO_FIELD_HeaderPresent,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  19,  19,   0,  19,    0x80000 } */ { 0x13001313, (void*)(g_uintptr_t)   0x80000 }, RestrictionTable55 }, // 8
    { GED_PSEUDO_FIELD_MessageTypeDP_SAMPLER,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable80 }, // 9
    { GED_PSEUDO_FIELD_MessageTypeDP_RC,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable100 }, // 10
    { GED_PSEUDO_FIELD_MessageTypeDP_CC,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable58 }, // 11
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_PSEUDO_FIELD_TypedSurfaceSlotGroup,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable60 }, // 13
    { GED_PSEUDO_FIELD_TypedAtomicSlotGroup,          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable61 }, // 14
    { GED_PSEUDO_FIELD_UntypedSurfaceSIMDMode,        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable62 }, // 15
    { GED_PSEUDO_FIELD_UntypedAtomicSIMDMode,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable63 }, // 16
    { GED_PSEUDO_FIELD_InvalidateAfterRead,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_PSEUDO_FIELD_BlockSize,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   8,  10,   0,   8,      0x700 } */ {  0x8000a08, (void*)(g_uintptr_t)     0x700 }, RestrictionTable64 }, // 18
    { GED_PSEUDO_FIELD_RedChannel,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable65 }, // 19
    { GED_PSEUDO_FIELD_GreenChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable66 }, // 20
    { GED_PSEUDO_FIELD_BlueChannel,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  10,  10,   0,  10,      0x400 } */ {  0xa000a0a, (void*)(g_uintptr_t)     0x400 }, RestrictionTable67 }, // 21
    { GED_PSEUDO_FIELD_AlphaChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  11,  11,   0,  11,      0x800 } */ {  0xb000b0b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable68 }, // 22
    { GED_PSEUDO_FIELD_ReturnDataControl,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  13,  13,   0,  13,     0x2000 } */ {  0xd000d0d, (void*)(g_uintptr_t)    0x2000 }, RestrictionTable69 }, // 23
    { GED_PSEUDO_FIELD_AtomicOperationType,           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable70 }, // 24
    { GED_PSEUDO_FIELD_AtomicCounterOperationType,    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable71 }, // 25
    { GED_PSEUDO_FIELD_SubFuncID,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   0,   2,   0,   0,        0x7 } */ {      0x200, (void*)(g_uintptr_t)       0x7 }, RestrictionTable72 }, // 26
    { GED_PSEUDO_FIELD_BindingTableIndex,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   0,   7,   0,   0,       0xff } */ {      0x700, (void*)(g_uintptr_t)      0xff }, NULL }, // 27
    { GED_PSEUDO_FIELD_FuncControl,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    19, /* {   0,  18,   0,   0,    0x7ffff } */ {     0x1200, (void*)(g_uintptr_t)   0x7ffff }, NULL }, // 28
    { GED_PSEUDO_FIELD_MessageTypeDP_DC1,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable102 }, // 29
    { GED_PSEUDO_FIELD_MessageTypeDP0Category,        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  18,  18,   0,  18,    0x40000 } */ { 0x12001212, (void*)(g_uintptr_t)   0x40000 }, NULL }, // 30
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0Legacy,       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable82 }, // 31
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0ScratchBlock, GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  17,  17,   0,  17,    0x20000 } */ { 0x11001111, (void*)(g_uintptr_t)   0x20000 }, RestrictionTable101 } // 32
}; // PositionInterpreterTable2[]
const ged_ins_field_entry_t PositionInterpreterTable3[35] =
{
    { GED_PSEUDO_FIELD_ArchReg,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   4,   7,   0,   4,       0xf0 } */ {  0x4000704, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable99 }, // 0
    { GED_PSEUDO_FIELD_ArchRegNum,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   0,   3,   0,   0,        0xf } */ {      0x300, (void*)(g_uintptr_t)       0xf }, NULL }, // 1
    { GED_PSEUDO_FIELD_SwizzleX,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   0,   1,   0,   0,        0x3 } */ {      0x100, (void*)(g_uintptr_t)       0x3 }, RestrictionTable54 }, // 2
    { GED_PSEUDO_FIELD_SwizzleY,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   2,   3,   0,   2,        0xc } */ {  0x2000302, (void*)(g_uintptr_t)       0xc }, RestrictionTable54 }, // 3
    { GED_PSEUDO_FIELD_SwizzleZ,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   4,   5,   0,   4,       0x30 } */ {  0x4000504, (void*)(g_uintptr_t)      0x30 }, RestrictionTable54 }, // 4
    { GED_PSEUDO_FIELD_SwizzleW,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   6,   7,   0,   6,       0xc0 } */ {  0x6000706, (void*)(g_uintptr_t)      0xc0 }, RestrictionTable54 }, // 5
    { GED_PSEUDO_FIELD_MessageLength,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  25,  28,   0,  25, 0x1e000000 } */ { 0x19001c19, (void*)(g_uintptr_t)0x1e000000 }, NULL }, // 6
    { GED_PSEUDO_FIELD_ResponseLength,                GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  20,  24,   0,  20,  0x1f00000 } */ { 0x14001814, (void*)(g_uintptr_t) 0x1f00000 }, NULL }, // 7
    { GED_PSEUDO_FIELD_HeaderPresent,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  19,  19,   0,  19,    0x80000 } */ { 0x13001313, (void*)(g_uintptr_t)   0x80000 }, RestrictionTable55 }, // 8
    { GED_PSEUDO_FIELD_MessageTypeDP_SAMPLER,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_PSEUDO_FIELD_MessageTypeDP_RC,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable113 }, // 10
    { GED_PSEUDO_FIELD_MessageTypeDP_CC,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_PSEUDO_FIELD_TypedSurfaceSlotGroup,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable60 }, // 13
    { GED_PSEUDO_FIELD_TypedAtomicSlotGroup,          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable61 }, // 14
    { GED_PSEUDO_FIELD_UntypedSurfaceSIMDMode,        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable62 }, // 15
    { GED_PSEUDO_FIELD_UntypedAtomicSIMDMode,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable63 }, // 16
    { GED_PSEUDO_FIELD_InvalidateAfterRead,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_PSEUDO_FIELD_BlockSize,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   8,  10,   0,   8,      0x700 } */ {  0x8000a08, (void*)(g_uintptr_t)     0x700 }, RestrictionTable64 }, // 18
    { GED_PSEUDO_FIELD_RedChannel,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable65 }, // 19
    { GED_PSEUDO_FIELD_GreenChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable66 }, // 20
    { GED_PSEUDO_FIELD_BlueChannel,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  10,  10,   0,  10,      0x400 } */ {  0xa000a0a, (void*)(g_uintptr_t)     0x400 }, RestrictionTable67 }, // 21
    { GED_PSEUDO_FIELD_AlphaChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  11,  11,   0,  11,      0x800 } */ {  0xb000b0b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable68 }, // 22
    { GED_PSEUDO_FIELD_ReturnDataControl,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  13,  13,   0,  13,     0x2000 } */ {  0xd000d0d, (void*)(g_uintptr_t)    0x2000 }, RestrictionTable69 }, // 23
    { GED_PSEUDO_FIELD_AtomicOperationType,           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable70 }, // 24
    { GED_PSEUDO_FIELD_AtomicCounterOperationType,    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable71 }, // 25
    { GED_PSEUDO_FIELD_SubFuncID,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   0,   2,   0,   0,        0x7 } */ {      0x200, (void*)(g_uintptr_t)       0x7 }, RestrictionTable117 }, // 26
    { GED_PSEUDO_FIELD_BindingTableIndex,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   0,   7,   0,   0,       0xff } */ {      0x700, (void*)(g_uintptr_t)      0xff }, NULL }, // 27
    { GED_PSEUDO_FIELD_FuncControl,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    19, /* {   0,  18,   0,   0,    0x7ffff } */ {     0x1200, (void*)(g_uintptr_t)   0x7ffff }, NULL }, // 28
    { GED_PSEUDO_FIELD_MessageTypeDP_DC1,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable114 }, // 29
    { GED_PSEUDO_FIELD_MessageTypeDP0Category,        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  18,  18,   0,  18,    0x40000 } */ { 0x12001212, (void*)(g_uintptr_t)   0x40000 }, NULL }, // 30
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0Legacy,       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable82 }, // 31
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0ScratchBlock, GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  17,  17,   0,  17,    0x20000 } */ { 0x11001111, (void*)(g_uintptr_t)   0x20000 }, RestrictionTable101 }, // 32
    { GED_PSEUDO_FIELD_MessageTypeDP_DC2,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable115 }, // 33
    { GED_PSEUDO_FIELD_MessageTypeDP_DCRO,            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable116 } // 34
}; // PositionInterpreterTable3[]
const ged_ins_field_entry_t PositionInterpreterTable4[35] =
{
    { GED_PSEUDO_FIELD_ArchReg,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   4,   7,   0,   4,       0xf0 } */ {  0x4000704, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable99 }, // 0
    { GED_PSEUDO_FIELD_ArchRegNum,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   0,   3,   0,   0,        0xf } */ {      0x300, (void*)(g_uintptr_t)       0xf }, NULL }, // 1
    { GED_PSEUDO_FIELD_SwizzleX,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   0,   1,   0,   0,        0x3 } */ {      0x100, (void*)(g_uintptr_t)       0x3 }, RestrictionTable54 }, // 2
    { GED_PSEUDO_FIELD_SwizzleY,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   2,   3,   0,   2,        0xc } */ {  0x2000302, (void*)(g_uintptr_t)       0xc }, RestrictionTable54 }, // 3
    { GED_PSEUDO_FIELD_SwizzleZ,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   4,   5,   0,   4,       0x30 } */ {  0x4000504, (void*)(g_uintptr_t)      0x30 }, RestrictionTable54 }, // 4
    { GED_PSEUDO_FIELD_SwizzleW,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   6,   7,   0,   6,       0xc0 } */ {  0x6000706, (void*)(g_uintptr_t)      0xc0 }, RestrictionTable54 }, // 5
    { GED_PSEUDO_FIELD_MessageLength,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  25,  28,   0,  25, 0x1e000000 } */ { 0x19001c19, (void*)(g_uintptr_t)0x1e000000 }, NULL }, // 6
    { GED_PSEUDO_FIELD_ResponseLength,                GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  20,  24,   0,  20,  0x1f00000 } */ { 0x14001814, (void*)(g_uintptr_t) 0x1f00000 }, NULL }, // 7
    { GED_PSEUDO_FIELD_HeaderPresent,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  19,  19,   0,  19,    0x80000 } */ { 0x13001313, (void*)(g_uintptr_t)   0x80000 }, RestrictionTable55 }, // 8
    { GED_PSEUDO_FIELD_MessageTypeDP_SAMPLER,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_PSEUDO_FIELD_MessageTypeDP_RC,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable113 }, // 10
    { GED_PSEUDO_FIELD_MessageTypeDP_CC,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_PSEUDO_FIELD_TypedSurfaceSlotGroup,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable60 }, // 13
    { GED_PSEUDO_FIELD_TypedAtomicSlotGroup,          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable61 }, // 14
    { GED_PSEUDO_FIELD_UntypedSurfaceSIMDMode,        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable62 }, // 15
    { GED_PSEUDO_FIELD_UntypedAtomicSIMDMode,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable63 }, // 16
    { GED_PSEUDO_FIELD_InvalidateAfterRead,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_PSEUDO_FIELD_BlockSize,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   8,  10,   0,   8,      0x700 } */ {  0x8000a08, (void*)(g_uintptr_t)     0x700 }, RestrictionTable64 }, // 18
    { GED_PSEUDO_FIELD_RedChannel,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable65 }, // 19
    { GED_PSEUDO_FIELD_GreenChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable66 }, // 20
    { GED_PSEUDO_FIELD_BlueChannel,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  10,  10,   0,  10,      0x400 } */ {  0xa000a0a, (void*)(g_uintptr_t)     0x400 }, RestrictionTable67 }, // 21
    { GED_PSEUDO_FIELD_AlphaChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  11,  11,   0,  11,      0x800 } */ {  0xb000b0b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable68 }, // 22
    { GED_PSEUDO_FIELD_ReturnDataControl,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  13,  13,   0,  13,     0x2000 } */ {  0xd000d0d, (void*)(g_uintptr_t)    0x2000 }, RestrictionTable69 }, // 23
    { GED_PSEUDO_FIELD_AtomicOperationType,           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable70 }, // 24
    { GED_PSEUDO_FIELD_AtomicCounterOperationType,    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable71 }, // 25
    { GED_PSEUDO_FIELD_SubFuncID,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   0,   2,   0,   0,        0x7 } */ {      0x200, (void*)(g_uintptr_t)       0x7 }, RestrictionTable117 }, // 26
    { GED_PSEUDO_FIELD_BindingTableIndex,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   0,   7,   0,   0,       0xff } */ {      0x700, (void*)(g_uintptr_t)      0xff }, NULL }, // 27
    { GED_PSEUDO_FIELD_FuncControl,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    19, /* {   0,  18,   0,   0,    0x7ffff } */ {     0x1200, (void*)(g_uintptr_t)   0x7ffff }, NULL }, // 28
    { GED_PSEUDO_FIELD_MessageTypeDP_DC1,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable131 }, // 29
    { GED_PSEUDO_FIELD_MessageTypeDP0Category,        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  18,  18,   0,  18,    0x40000 } */ { 0x12001212, (void*)(g_uintptr_t)   0x40000 }, NULL }, // 30
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0Legacy,       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable82 }, // 31
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0ScratchBlock, GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  17,  17,   0,  17,    0x20000 } */ { 0x11001111, (void*)(g_uintptr_t)   0x20000 }, RestrictionTable101 }, // 32
    { GED_PSEUDO_FIELD_MessageTypeDP_DC2,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable115 }, // 33
    { GED_PSEUDO_FIELD_MessageTypeDP_DCRO,            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable116 } // 34
}; // PositionInterpreterTable4[]
const ged_ins_field_entry_t PositionInterpreterTable5[35] =
{
    { GED_PSEUDO_FIELD_ArchReg,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   4,   7,   0,   4,       0xf0 } */ {  0x4000704, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable99 }, // 0
    { GED_PSEUDO_FIELD_ArchRegNum,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   0,   3,   0,   0,        0xf } */ {      0x300, (void*)(g_uintptr_t)       0xf }, NULL }, // 1
    { GED_PSEUDO_FIELD_SwizzleX,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_PSEUDO_FIELD_SwizzleY,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_PSEUDO_FIELD_SwizzleZ,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_PSEUDO_FIELD_SwizzleW,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_PSEUDO_FIELD_MessageLength,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  25,  28,   0,  25, 0x1e000000 } */ { 0x19001c19, (void*)(g_uintptr_t)0x1e000000 }, NULL }, // 6
    { GED_PSEUDO_FIELD_ResponseLength,                GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  20,  24,   0,  20,  0x1f00000 } */ { 0x14001814, (void*)(g_uintptr_t) 0x1f00000 }, NULL }, // 7
    { GED_PSEUDO_FIELD_HeaderPresent,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  19,  19,   0,  19,    0x80000 } */ { 0x13001313, (void*)(g_uintptr_t)   0x80000 }, RestrictionTable55 }, // 8
    { GED_PSEUDO_FIELD_MessageTypeDP_SAMPLER,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_PSEUDO_FIELD_MessageTypeDP_RC,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable113 }, // 10
    { GED_PSEUDO_FIELD_MessageTypeDP_CC,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_PSEUDO_FIELD_TypedSurfaceSlotGroup,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable60 }, // 13
    { GED_PSEUDO_FIELD_TypedAtomicSlotGroup,          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable61 }, // 14
    { GED_PSEUDO_FIELD_UntypedSurfaceSIMDMode,        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable62 }, // 15
    { GED_PSEUDO_FIELD_UntypedAtomicSIMDMode,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable63 }, // 16
    { GED_PSEUDO_FIELD_InvalidateAfterRead,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_PSEUDO_FIELD_BlockSize,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   8,  10,   0,   8,      0x700 } */ {  0x8000a08, (void*)(g_uintptr_t)     0x700 }, RestrictionTable64 }, // 18
    { GED_PSEUDO_FIELD_RedChannel,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable65 }, // 19
    { GED_PSEUDO_FIELD_GreenChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable66 }, // 20
    { GED_PSEUDO_FIELD_BlueChannel,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  10,  10,   0,  10,      0x400 } */ {  0xa000a0a, (void*)(g_uintptr_t)     0x400 }, RestrictionTable67 }, // 21
    { GED_PSEUDO_FIELD_AlphaChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  11,  11,   0,  11,      0x800 } */ {  0xb000b0b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable68 }, // 22
    { GED_PSEUDO_FIELD_ReturnDataControl,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  13,  13,   0,  13,     0x2000 } */ {  0xd000d0d, (void*)(g_uintptr_t)    0x2000 }, RestrictionTable69 }, // 23
    { GED_PSEUDO_FIELD_AtomicOperationType,           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable70 }, // 24
    { GED_PSEUDO_FIELD_AtomicCounterOperationType,    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable71 }, // 25
    { GED_PSEUDO_FIELD_SubFuncID,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   0,   2,   0,   0,        0x7 } */ {      0x200, (void*)(g_uintptr_t)       0x7 }, RestrictionTable117 }, // 26
    { GED_PSEUDO_FIELD_BindingTableIndex,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   0,   7,   0,   0,       0xff } */ {      0x700, (void*)(g_uintptr_t)      0xff }, NULL }, // 27
    { GED_PSEUDO_FIELD_FuncControl,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    19, /* {   0,  18,   0,   0,    0x7ffff } */ {     0x1200, (void*)(g_uintptr_t)   0x7ffff }, NULL }, // 28
    { GED_PSEUDO_FIELD_MessageTypeDP_DC1,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable131 }, // 29
    { GED_PSEUDO_FIELD_MessageTypeDP0Category,        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  18,  18,   0,  18,    0x40000 } */ { 0x12001212, (void*)(g_uintptr_t)   0x40000 }, NULL }, // 30
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0Legacy,       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable82 }, // 31
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0ScratchBlock, GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  17,  17,   0,  17,    0x20000 } */ { 0x11001111, (void*)(g_uintptr_t)   0x20000 }, RestrictionTable101 }, // 32
    { GED_PSEUDO_FIELD_MessageTypeDP_DC2,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable115 }, // 33
    { GED_PSEUDO_FIELD_MessageTypeDP_DCRO,            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable116 } // 34
}; // PositionInterpreterTable5[]
const ged_ins_field_entry_t PositionInterpreterTable6[36] =
{
    { GED_PSEUDO_FIELD_ArchReg,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   4,   7,   0,   4,       0xf0 } */ {  0x4000704, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable99 }, // 0
    { GED_PSEUDO_FIELD_ArchRegNum,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   0,   3,   0,   0,        0xf } */ {      0x300, (void*)(g_uintptr_t)       0xf }, NULL }, // 1
    { GED_PSEUDO_FIELD_SwizzleX,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_PSEUDO_FIELD_SwizzleY,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_PSEUDO_FIELD_SwizzleZ,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_PSEUDO_FIELD_SwizzleW,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_PSEUDO_FIELD_MessageLength,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  25,  28,   0,  25, 0x1e000000 } */ { 0x19001c19, (void*)(g_uintptr_t)0x1e000000 }, NULL }, // 6
    { GED_PSEUDO_FIELD_ResponseLength,                GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  20,  24,   0,  20,  0x1f00000 } */ { 0x14001814, (void*)(g_uintptr_t) 0x1f00000 }, NULL }, // 7
    { GED_PSEUDO_FIELD_HeaderPresent,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  19,  19,   0,  19,    0x80000 } */ { 0x13001313, (void*)(g_uintptr_t)   0x80000 }, RestrictionTable55 }, // 8
    { GED_PSEUDO_FIELD_MessageTypeDP_SAMPLER,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_PSEUDO_FIELD_MessageTypeDP_RC,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable113 }, // 10
    { GED_PSEUDO_FIELD_MessageTypeDP_CC,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_PSEUDO_FIELD_TypedSurfaceSlotGroup,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable60 }, // 13
    { GED_PSEUDO_FIELD_TypedAtomicSlotGroup,          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable61 }, // 14
    { GED_PSEUDO_FIELD_UntypedSurfaceSIMDMode,        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable62 }, // 15
    { GED_PSEUDO_FIELD_UntypedAtomicSIMDMode,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable63 }, // 16
    { GED_PSEUDO_FIELD_InvalidateAfterRead,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_PSEUDO_FIELD_BlockSize,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   8,  10,   0,   8,      0x700 } */ {  0x8000a08, (void*)(g_uintptr_t)     0x700 }, RestrictionTable64 }, // 18
    { GED_PSEUDO_FIELD_RedChannel,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable65 }, // 19
    { GED_PSEUDO_FIELD_GreenChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable66 }, // 20
    { GED_PSEUDO_FIELD_BlueChannel,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  10,  10,   0,  10,      0x400 } */ {  0xa000a0a, (void*)(g_uintptr_t)     0x400 }, RestrictionTable67 }, // 21
    { GED_PSEUDO_FIELD_AlphaChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  11,  11,   0,  11,      0x800 } */ {  0xb000b0b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable68 }, // 22
    { GED_PSEUDO_FIELD_ReturnDataControl,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  13,  13,   0,  13,     0x2000 } */ {  0xd000d0d, (void*)(g_uintptr_t)    0x2000 }, RestrictionTable69 }, // 23
    { GED_PSEUDO_FIELD_AtomicOperationType,           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable70 }, // 24
    { GED_PSEUDO_FIELD_AtomicCounterOperationType,    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable71 }, // 25
    { GED_PSEUDO_FIELD_SubFuncID,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   0,   2,   0,   0,        0x7 } */ {      0x200, (void*)(g_uintptr_t)       0x7 }, RestrictionTable117 }, // 26
    { GED_PSEUDO_FIELD_BindingTableIndex,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   0,   7,   0,   0,       0xff } */ {      0x700, (void*)(g_uintptr_t)      0xff }, NULL }, // 27
    { GED_PSEUDO_FIELD_FuncControl,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    19, /* {   0,  18,   0,   0,    0x7ffff } */ {     0x1200, (void*)(g_uintptr_t)   0x7ffff }, NULL }, // 28
    { GED_PSEUDO_FIELD_MessageTypeDP_DC1,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable156 }, // 29
    { GED_PSEUDO_FIELD_MessageTypeDP0Category,        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  18,  18,   0,  18,    0x40000 } */ { 0x12001212, (void*)(g_uintptr_t)   0x40000 }, NULL }, // 30
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0Legacy,       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable155 }, // 31
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0ScratchBlock, GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  17,  17,   0,  17,    0x20000 } */ { 0x11001111, (void*)(g_uintptr_t)   0x20000 }, RestrictionTable101 }, // 32
    { GED_PSEUDO_FIELD_MessageTypeDP_DC2,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable157 }, // 33
    { GED_PSEUDO_FIELD_MessageTypeDP_DCRO,            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable158 }, // 34
    { GED_PSEUDO_FIELD_ExMessageLength,               GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {   6,  10,   0,   6,      0x7c0 } */ {  0x6000a06, (void*)(g_uintptr_t)     0x7c0 }, NULL } // 35
}; // PositionInterpreterTable6[]
const ged_ins_field_entry_t PositionInterpreterTable7[35] =
{
    { GED_PSEUDO_FIELD_ArchReg,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   4,   7,   0,   4,       0xf0 } */ {  0x4000704, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable171 }, // 0
    { GED_PSEUDO_FIELD_ArchRegNum,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   0,   3,   0,   0,        0xf } */ {      0x300, (void*)(g_uintptr_t)       0xf }, NULL }, // 1
    { GED_PSEUDO_FIELD_SwizzleX,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_PSEUDO_FIELD_SwizzleY,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_PSEUDO_FIELD_SwizzleZ,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_PSEUDO_FIELD_SwizzleW,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_PSEUDO_FIELD_MessageLength,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  25,  28,   0,  25, 0x1e000000 } */ { 0x19001c19, (void*)(g_uintptr_t)0x1e000000 }, NULL }, // 6
    { GED_PSEUDO_FIELD_ResponseLength,                GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  20,  24,   0,  20,  0x1f00000 } */ { 0x14001814, (void*)(g_uintptr_t) 0x1f00000 }, NULL }, // 7
    { GED_PSEUDO_FIELD_HeaderPresent,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  19,  19,   0,  19,    0x80000 } */ { 0x13001313, (void*)(g_uintptr_t)   0x80000 }, RestrictionTable55 }, // 8
    { GED_PSEUDO_FIELD_MessageTypeDP_SAMPLER,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_PSEUDO_FIELD_MessageTypeDP_RC,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable113 }, // 10
    { GED_PSEUDO_FIELD_MessageTypeDP_CC,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_PSEUDO_FIELD_TypedSurfaceSlotGroup,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable60 }, // 13
    { GED_PSEUDO_FIELD_TypedAtomicSlotGroup,          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable61 }, // 14
    { GED_PSEUDO_FIELD_UntypedSurfaceSIMDMode,        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable62 }, // 15
    { GED_PSEUDO_FIELD_UntypedAtomicSIMDMode,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable63 }, // 16
    { GED_PSEUDO_FIELD_InvalidateAfterRead,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_PSEUDO_FIELD_BlockSize,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   8,  10,   0,   8,      0x700 } */ {  0x8000a08, (void*)(g_uintptr_t)     0x700 }, RestrictionTable64 }, // 18
    { GED_PSEUDO_FIELD_RedChannel,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable65 }, // 19
    { GED_PSEUDO_FIELD_GreenChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable66 }, // 20
    { GED_PSEUDO_FIELD_BlueChannel,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  10,  10,   0,  10,      0x400 } */ {  0xa000a0a, (void*)(g_uintptr_t)     0x400 }, RestrictionTable67 }, // 21
    { GED_PSEUDO_FIELD_AlphaChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  11,  11,   0,  11,      0x800 } */ {  0xb000b0b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable68 }, // 22
    { GED_PSEUDO_FIELD_ReturnDataControl,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  13,  13,   0,  13,     0x2000 } */ {  0xd000d0d, (void*)(g_uintptr_t)    0x2000 }, RestrictionTable69 }, // 23
    { GED_PSEUDO_FIELD_AtomicOperationType,           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable70 }, // 24
    { GED_PSEUDO_FIELD_AtomicCounterOperationType,    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable71 }, // 25
    { GED_PSEUDO_FIELD_SubFuncID,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   0,   2,   0,   0,        0x7 } */ {      0x200, (void*)(g_uintptr_t)       0x7 }, RestrictionTable117 }, // 26
    { GED_PSEUDO_FIELD_BindingTableIndex,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   0,   7,   0,   0,       0xff } */ {      0x700, (void*)(g_uintptr_t)      0xff }, NULL }, // 27
    { GED_PSEUDO_FIELD_FuncControl,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    19, /* {   0,  18,   0,   0,    0x7ffff } */ {     0x1200, (void*)(g_uintptr_t)   0x7ffff }, NULL }, // 28
    { GED_PSEUDO_FIELD_MessageTypeDP_DC1,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable156 }, // 29
    { GED_PSEUDO_FIELD_MessageTypeDP0Category,        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  18,  18,   0,  18,    0x40000 } */ { 0x12001212, (void*)(g_uintptr_t)   0x40000 }, NULL }, // 30
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0Legacy,       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable155 }, // 31
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0ScratchBlock, GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  17,  17,   0,  17,    0x20000 } */ { 0x11001111, (void*)(g_uintptr_t)   0x20000 }, RestrictionTable101 }, // 32
    { GED_PSEUDO_FIELD_MessageTypeDP_DC2,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable157 }, // 33
    { GED_PSEUDO_FIELD_MessageTypeDP_DCRO,            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable172 } // 34
}; // PositionInterpreterTable7[]
const ged_ins_field_entry_t PositionInterpreterTable8[45] =
{
    { GED_PSEUDO_FIELD_ArchReg,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   4,   7,   0,   4,       0xf0 } */ {  0x4000704, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable171 }, // 0
    { GED_PSEUDO_FIELD_ArchRegNum,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   0,   3,   0,   0,        0xf } */ {      0x300, (void*)(g_uintptr_t)       0xf }, NULL }, // 1
    { GED_PSEUDO_FIELD_SwizzleX,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_PSEUDO_FIELD_SwizzleY,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_PSEUDO_FIELD_SwizzleZ,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_PSEUDO_FIELD_SwizzleW,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_PSEUDO_FIELD_MessageLength,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  25,  28,   0,  25, 0x1e000000 } */ { 0x19001c19, (void*)(g_uintptr_t)0x1e000000 }, NULL }, // 6
    { GED_PSEUDO_FIELD_ResponseLength,                GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  20,  24,   0,  20,  0x1f00000 } */ { 0x14001814, (void*)(g_uintptr_t) 0x1f00000 }, NULL }, // 7
    { GED_PSEUDO_FIELD_HeaderPresent,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  19,  19,   0,  19,    0x80000 } */ { 0x13001313, (void*)(g_uintptr_t)   0x80000 }, RestrictionTable55 }, // 8
    { GED_PSEUDO_FIELD_MessageTypeDP_SAMPLER,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_PSEUDO_FIELD_MessageTypeDP_RC,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable113 }, // 10
    { GED_PSEUDO_FIELD_MessageTypeDP_CC,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_PSEUDO_FIELD_TypedSurfaceSlotGroup,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable60 }, // 13
    { GED_PSEUDO_FIELD_TypedAtomicSlotGroup,          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable61 }, // 14
    { GED_PSEUDO_FIELD_UntypedSurfaceSIMDMode,        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable62 }, // 15
    { GED_PSEUDO_FIELD_UntypedAtomicSIMDMode,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable63 }, // 16
    { GED_PSEUDO_FIELD_InvalidateAfterRead,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_PSEUDO_FIELD_BlockSize,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   8,  10,   0,   8,      0x700 } */ {  0x8000a08, (void*)(g_uintptr_t)     0x700 }, RestrictionTable64 }, // 18
    { GED_PSEUDO_FIELD_RedChannel,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable65 }, // 19
    { GED_PSEUDO_FIELD_GreenChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable66 }, // 20
    { GED_PSEUDO_FIELD_BlueChannel,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  10,  10,   0,  10,      0x400 } */ {  0xa000a0a, (void*)(g_uintptr_t)     0x400 }, RestrictionTable67 }, // 21
    { GED_PSEUDO_FIELD_AlphaChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  11,  11,   0,  11,      0x800 } */ {  0xb000b0b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable68 }, // 22
    { GED_PSEUDO_FIELD_ReturnDataControl,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  13,  13,   0,  13,     0x2000 } */ {  0xd000d0d, (void*)(g_uintptr_t)    0x2000 }, RestrictionTable69 }, // 23
    { GED_PSEUDO_FIELD_AtomicOperationType,           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable70 }, // 24
    { GED_PSEUDO_FIELD_AtomicCounterOperationType,    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable71 }, // 25
    { GED_PSEUDO_FIELD_SubFuncID,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   0,   2,   0,   0,        0x7 } */ {      0x200, (void*)(g_uintptr_t)       0x7 }, RestrictionTable117 }, // 26
    { GED_PSEUDO_FIELD_BindingTableIndex,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   0,   7,   0,   0,       0xff } */ {      0x700, (void*)(g_uintptr_t)      0xff }, NULL }, // 27
    { GED_PSEUDO_FIELD_FuncControl,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    19, /* {   0,  18,   0,   0,    0x7ffff } */ {     0x1200, (void*)(g_uintptr_t)   0x7ffff }, NULL }, // 28
    { GED_PSEUDO_FIELD_MessageTypeDP_DC1,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable156 }, // 29
    { GED_PSEUDO_FIELD_MessageTypeDP0Category,        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  18,  18,   0,  18,    0x40000 } */ { 0x12001212, (void*)(g_uintptr_t)   0x40000 }, NULL }, // 30
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0Legacy,       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable184 }, // 31
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0ScratchBlock, GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  17,  17,   0,  17,    0x20000 } */ { 0x11001111, (void*)(g_uintptr_t)   0x20000 }, RestrictionTable101 }, // 32
    { GED_PSEUDO_FIELD_MessageTypeDP_DC2,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable157 }, // 33
    { GED_PSEUDO_FIELD_MessageTypeDP_DCRO,            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable185 }, // 34
    { GED_PSEUDO_FIELD_ExMessageLength,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_PSEUDO_FIELD_DPOpcode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {   0,   5,   0,   0,       0x3f } */ {      0x500, (void*)(g_uintptr_t)      0x3f }, RestrictionTable176 }, // 36
    { GED_PSEUDO_FIELD_DPAddrSurfaceType,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  29,  30,   0,  29, 0x60000000 } */ { 0x1d001e1d, (void*)(g_uintptr_t)0x60000000 }, RestrictionTable177 }, // 37
    { GED_PSEUDO_FIELD_DPVectSize,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  12,  14,   0,  12,     0x7000 } */ {  0xc000e0c, (void*)(g_uintptr_t)    0x7000 }, RestrictionTable178 }, // 38
    { GED_PSEUDO_FIELD_DPFlushType,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  12,  14,   0,  12,     0x7000 } */ {  0xc000e0c, (void*)(g_uintptr_t)    0x7000 }, RestrictionTable179 }, // 39
    { GED_PSEUDO_FIELD_DPTranspose,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  15,  15,   0,  15,     0x8000 } */ {  0xf000f0f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable180 }, // 40
    { GED_PSEUDO_FIELD_DPFlushRange,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  15,  15,   0,  15,     0x8000 } */ {  0xf000f0f, (void*)(g_uintptr_t)    0x8000 }, NULL }, // 41
    { GED_PSEUDO_FIELD_DPDataSize,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   9,  11,   0,   9,      0xe00 } */ {  0x9000b09, (void*)(g_uintptr_t)     0xe00 }, RestrictionTable181 }, // 42
    { GED_PSEUDO_FIELD_DPFenceScope,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   9,  11,   0,   9,      0xe00 } */ {  0x9000b09, (void*)(g_uintptr_t)     0xe00 }, RestrictionTable182 }, // 43
    { GED_PSEUDO_FIELD_DPAddrSize,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   7,   8,   0,   7,      0x180 } */ {  0x7000807, (void*)(g_uintptr_t)     0x180 }, RestrictionTable183 } // 44
}; // PositionInterpreterTable8[]
const ged_ins_field_entry_t PositionInterpreterTable9[45] =
{
    { GED_PSEUDO_FIELD_ArchReg,                       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   4,   7,   0,   4,       0xf0 } */ {  0x4000704, (void*)(g_uintptr_t)      0xf0 }, RestrictionTable171 }, // 0
    { GED_PSEUDO_FIELD_ArchRegNum,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   0,   3,   0,   0,        0xf } */ {      0x300, (void*)(g_uintptr_t)       0xf }, NULL }, // 1
    { GED_PSEUDO_FIELD_SwizzleX,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 2
    { GED_PSEUDO_FIELD_SwizzleY,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 3
    { GED_PSEUDO_FIELD_SwizzleZ,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 4
    { GED_PSEUDO_FIELD_SwizzleW,                      GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 5
    { GED_PSEUDO_FIELD_MessageLength,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  25,  28,   0,  25, 0x1e000000 } */ { 0x19001c19, (void*)(g_uintptr_t)0x1e000000 }, NULL }, // 6
    { GED_PSEUDO_FIELD_ResponseLength,                GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  20,  24,   0,  20,  0x1f00000 } */ { 0x14001814, (void*)(g_uintptr_t) 0x1f00000 }, NULL }, // 7
    { GED_PSEUDO_FIELD_HeaderPresent,                 GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  19,  19,   0,  19,    0x80000 } */ { 0x13001313, (void*)(g_uintptr_t)   0x80000 }, RestrictionTable55 }, // 8
    { GED_PSEUDO_FIELD_MessageTypeDP_SAMPLER,         GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 9
    { GED_PSEUDO_FIELD_MessageTypeDP_RC,              GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable113 }, // 10
    { GED_PSEUDO_FIELD_MessageTypeDP_CC,              GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 11
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0,             GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 12
    { GED_PSEUDO_FIELD_TypedSurfaceSlotGroup,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable60 }, // 13
    { GED_PSEUDO_FIELD_TypedAtomicSlotGroup,          GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable61 }, // 14
    { GED_PSEUDO_FIELD_UntypedSurfaceSIMDMode,        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  12,  13,   0,  12,     0x3000 } */ {  0xc000d0c, (void*)(g_uintptr_t)    0x3000 }, RestrictionTable62 }, // 15
    { GED_PSEUDO_FIELD_UntypedAtomicSIMDMode,         GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  12,  12,   0,  12,     0x1000 } */ {  0xc000c0c, (void*)(g_uintptr_t)    0x1000 }, RestrictionTable63 }, // 16
    { GED_PSEUDO_FIELD_InvalidateAfterRead,           GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 17
    { GED_PSEUDO_FIELD_BlockSize,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   8,  10,   0,   8,      0x700 } */ {  0x8000a08, (void*)(g_uintptr_t)     0x700 }, RestrictionTable64 }, // 18
    { GED_PSEUDO_FIELD_RedChannel,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   8,   8,   0,   8,      0x100 } */ {  0x8000808, (void*)(g_uintptr_t)     0x100 }, RestrictionTable65 }, // 19
    { GED_PSEUDO_FIELD_GreenChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {   9,   9,   0,   9,      0x200 } */ {  0x9000909, (void*)(g_uintptr_t)     0x200 }, RestrictionTable66 }, // 20
    { GED_PSEUDO_FIELD_BlueChannel,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  10,  10,   0,  10,      0x400 } */ {  0xa000a0a, (void*)(g_uintptr_t)     0x400 }, RestrictionTable67 }, // 21
    { GED_PSEUDO_FIELD_AlphaChannel,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  11,  11,   0,  11,      0x800 } */ {  0xb000b0b, (void*)(g_uintptr_t)     0x800 }, RestrictionTable68 }, // 22
    { GED_PSEUDO_FIELD_ReturnDataControl,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  13,  13,   0,  13,     0x2000 } */ {  0xd000d0d, (void*)(g_uintptr_t)    0x2000 }, RestrictionTable69 }, // 23
    { GED_PSEUDO_FIELD_AtomicOperationType,           GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable70 }, // 24
    { GED_PSEUDO_FIELD_AtomicCounterOperationType,    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {   8,  11,   0,   8,      0xf00 } */ {  0x8000b08, (void*)(g_uintptr_t)     0xf00 }, RestrictionTable71 }, // 25
    { GED_PSEUDO_FIELD_SubFuncID,                     GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   0,   2,   0,   0,        0x7 } */ {      0x200, (void*)(g_uintptr_t)       0x7 }, RestrictionTable117 }, // 26
    { GED_PSEUDO_FIELD_BindingTableIndex,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     8, /* {   0,   7,   0,   0,       0xff } */ {      0x700, (void*)(g_uintptr_t)      0xff }, NULL }, // 27
    { GED_PSEUDO_FIELD_FuncControl,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,    19, /* {   0,  18,   0,   0,    0x7ffff } */ {     0x1200, (void*)(g_uintptr_t)   0x7ffff }, NULL }, // 28
    { GED_PSEUDO_FIELD_MessageTypeDP_DC1,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable156 }, // 29
    { GED_PSEUDO_FIELD_MessageTypeDP0Category,        GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  18,  18,   0,  18,    0x40000 } */ { 0x12001212, (void*)(g_uintptr_t)   0x40000 }, NULL }, // 30
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0Legacy,       GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     4, /* {  14,  17,   0,  14,    0x3c000 } */ {  0xe00110e, (void*)(g_uintptr_t)   0x3c000 }, RestrictionTable184 }, // 31
    { GED_PSEUDO_FIELD_MessageTypeDP_DC0ScratchBlock, GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  17,  17,   0,  17,    0x20000 } */ { 0x11001111, (void*)(g_uintptr_t)   0x20000 }, RestrictionTable101 }, // 32
    { GED_PSEUDO_FIELD_MessageTypeDP_DC2,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable157 }, // 33
    { GED_PSEUDO_FIELD_MessageTypeDP_DCRO,            GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     5, /* {  14,  18,   0,  14,    0x7c000 } */ {  0xe00120e, (void*)(g_uintptr_t)   0x7c000 }, RestrictionTable185 }, // 34
    { GED_PSEUDO_FIELD_ExMessageLength,               GED_TABLE_ENTRY_TYPE_NOT_SUPPORTED,   0, { 0 }, NULL }, // 35
    { GED_PSEUDO_FIELD_DPOpcode,                      GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     6, /* {   0,   5,   0,   0,       0x3f } */ {      0x500, (void*)(g_uintptr_t)      0x3f }, RestrictionTable202 }, // 36
    { GED_PSEUDO_FIELD_DPAddrSurfaceType,             GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {  29,  30,   0,  29, 0x60000000 } */ { 0x1d001e1d, (void*)(g_uintptr_t)0x60000000 }, RestrictionTable177 }, // 37
    { GED_PSEUDO_FIELD_DPVectSize,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  12,  14,   0,  12,     0x7000 } */ {  0xc000e0c, (void*)(g_uintptr_t)    0x7000 }, RestrictionTable178 }, // 38
    { GED_PSEUDO_FIELD_DPFlushType,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {  12,  14,   0,  12,     0x7000 } */ {  0xc000e0c, (void*)(g_uintptr_t)    0x7000 }, RestrictionTable203 }, // 39
    { GED_PSEUDO_FIELD_DPTranspose,                   GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  15,  15,   0,  15,     0x8000 } */ {  0xf000f0f, (void*)(g_uintptr_t)    0x8000 }, RestrictionTable180 }, // 40
    { GED_PSEUDO_FIELD_DPFlushRange,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     1, /* {  15,  15,   0,  15,     0x8000 } */ {  0xf000f0f, (void*)(g_uintptr_t)    0x8000 }, NULL }, // 41
    { GED_PSEUDO_FIELD_DPDataSize,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   9,  11,   0,   9,      0xe00 } */ {  0x9000b09, (void*)(g_uintptr_t)     0xe00 }, RestrictionTable181 }, // 42
    { GED_PSEUDO_FIELD_DPFenceScope,                  GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     3, /* {   9,  11,   0,   9,      0xe00 } */ {  0x9000b09, (void*)(g_uintptr_t)     0xe00 }, RestrictionTable204 }, // 43
    { GED_PSEUDO_FIELD_DPAddrSize,                    GED_TABLE_ENTRY_TYPE_CONSECUTIVE,     2, /* {   7,   8,   0,   7,      0x180 } */ {  0x7000807, (void*)(g_uintptr_t)     0x180 }, RestrictionTable183 } // 44
}; // PositionInterpreterTable9[]
